/* Copyright (C) 2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     ArchMgrDmn
// @module  ArchMgrDmn | This is the main module for the OpenSTA Architecture
//          Manager daemon.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// OpenSTA includes.
#include <stdafx.h>
#include <malloc.h>
#include "ArchMgrRun.h"
#include "ArchMgrDmn.h"
#include "OstaTraceDef.h"
#include "CECInetInfo.h"
#include "mutex.hpp"
#include "ctime.hpp"

#pragma warning (disable:4786)
#pragma warning (disable:4503)

///////////////////////////////////////////////////////////////////////////////
// Standard includes.
#include <iostream>
#include <signal.h>
#include <string>
using namespace std;

///////////////////////////////////////////////////////////////////////////////
// Function declarations
void Usage();
void StartLogging();
void StopLogging();
int CreateOSTADaemon(HANDLE* hOSTADaemon ,DWORD* dwOSTADaemonPid);
int StartProcess (CString& Command ,HANDLE* pReturnHandle ,DWORD* pReturnPid);
void AbortDaemon(int);
void Shutdown();

///////////////////////////////////////////////////////////////////////////////
// Global variables
// @globalv ArchMgrRun* | g_pArchMgrRun | Pointer to Architecture Manager object.
// @globalv CORBAServer* | g_pCorba | Pointer to CORBA server object.
// @globalv CString | g_strOSTAHome | Pointer to OPENSTA_HOME path.
// @globalv int | g_nTraceLevel | Trace level.
ArchMgrRun* g_pArchMgrRun = NULL;
CORBAServer* g_pCorba = NULL;
CString g_strOSTAHome;
int g_nTraceLevel = 0;
std::ofstream* g_pLogFile;

int main(int argc, char** argv)
{
    CyrtoolsInitialize();

    // Indicate what to do if this image is aborted.
    signal(SIGINT, AbortDaemon);
    signal(SIGABRT, AbortDaemon);

    // Parse the command line and act on any options.
    int count = 0;      // Argument count
    while (count < argc)
    {
        if (strcmp( argv[count], "--help" ) == 0)
        {
            // Display usage message and terminate application
            Usage();
            return 0;
        }
        else if (strcmp( argv[count], "--trace" ) == 0)
        {
            // Next parameter is the requested trace level
            count++;
            g_nTraceLevel = atoi(argv[count]);
        }
        count++;
    }

    try
    {
        StartLogging();
        // Start the OpenSTA Daemon (cyrdmn.exe) for the localhost.
        HANDLE hOSTADaemon;
        DWORD dwOSTADaemonPid;
        if (CreateOSTADaemon(&hOSTADaemon, &dwOSTADaemonPid) != 0) return 1;

        // Create CORBA Server object.
        g_pCorba = new CORBAServer(argc, argv);

        // Instantiate new Architecture Manager Daemon object.
        g_pArchMgrRun = new ArchMgrRun(g_pCorba, hOSTADaemon, dwOSTADaemonPid, g_strOSTAHome);

        // Invoke the main Architecture Manager Daemon runtime.
        g_pArchMgrRun->Run();
    }
    catch(CyrDaemonException ex)
    {
        std::string strMsg = "Architecture Manager Daemon has thrown exception: " +
            ex.what() + " in " + ex.where();
        WriteLogMsg (TRCLEV_NONE, strMsg);
    }
    catch(...)
    {
        WriteLogMsg(TRCLEV_NONE, "Unknown exception caught by Architecture Manager Daemon.");
    }

    // Call shutdown procedure.
    Shutdown();
    StopLogging();
    CyrtoolsUninitialize();

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @func   This function displays the help and usage information for the
//         Architecture Manager daemon.
//
void Usage()
{
    cout << "OpenSTA Architecture Manager daemon. This application runs in" << endl;
    cout << "the background and allows the Architecture Manager within Commander" << endl;
    cout << "to communicate with remote OpenSTA machines.\n" << endl;
    cout << "archmgrdmn [--help] [--trace <level>]" << endl;
    cout << "\n" << endl;
    cout << "   --help      Displays this message.\n" << endl;
    cout << "   --trace     Sets initial trace level for the daemon.\n" << endl;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function starts Architecture Manager Daemon logging. It
//          creates and opens the Architecture Manager Daemon Log file
//          OPENSTA_HOME\Server\archmgrdmn_IPADR.log, where IPADR is the IP
//          address (with dots replaced by underscores) of the daemon host.
//
void StartLogging()
{
	char* pOpenStaHome = getenv("OPENSTA_HOME") ;
    if (!pOpenStaHome) return;
    string strLocalPath = pOpenStaHome;
    CECInetInfo inet;
    string strRegNam = inet.GetIPAddressAsString();
    string strLogFilNam = "archmgrdmn_" + strRegNam + ".log";
#ifdef _WIN32
    strLocalPath += "Server\\Logs\\" + strLogFilNam;
#else
    strLocalPath += "Server/Logs/" + strLogFilNam;
#endif
    g_pLogFile = new ofstream( strLocalPath.c_str(), std::ios::binary|std::ios::trunc|std::ios::out);

    WriteLogMsg(TRCLEV_NONE, string("Architecture Manager Daemon started."));
    char szTrcLev [12];
    itoa(g_nTraceLevel, szTrcLev, 10);
    WriteLogMsg(TRCLEV_NONE, string("Trace Level=") + szTrcLev);
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Stop logging to local trace file.
//
void StopLogging()
{
    WriteLogMsg(TRCLEV_NONE, string("Architecture Manager Daemon stopped."));
    g_pLogFile->close();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method is responsible for starting the OpenSTA Daemon
//          (cyrdmn.exe) process.
//
int CreateOSTADaemon(HANDLE* phOSTADaemon ,DWORD* pdwOSTADaemonPid)
{
    // Translate OPENSTA_HOME and format the executable name
    std::string strLocalPath;
	char *pOSTAHome = getenv("OPENSTA_HOME");
	if (pOSTAHome != NULL) strLocalPath = pOSTAHome ;
#ifdef _WIN32
	strLocalPath += "Server\\";
#else
	strLocalPath += "Server/";
#endif
    GetShortPathName(pOSTAHome, g_strOSTAHome.GetBuffer(MAX_PATH), MAX_PATH);
    g_strOSTAHome.ReleaseBuffer();
    if (g_strOSTAHome.Right(1) != '\\') g_strOSTAHome += "\\";

    // Establish the OpenSTA Daemon command line.
    CString strCommand;
    strCommand.Format ("%sServer\\cyrdmn.exe --trace %d" ,g_strOSTAHome, g_nTraceLevel);
    // ...and run it.
    int retsts = StartProcess (strCommand ,phOSTADaemon ,pdwOSTADaemonPid);
    if (retsts != 0)
    {
        WriteLogMsg(TRCLEV_NONE, string("Failed to start OpenSTA Daemon."));
    }
    else
    {
        WriteLogMsg(TRCLEV_LOW, string("OpenSTA Daemon started successfully."));
    }
    return retsts;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method creates a process to execute a specified command. It
//          returns its pid and handle.
//
int StartProcess (CString& Command ,HANDLE* pReturnHandle ,DWORD* pReturnPid)
{
    CString     ErrorBuffer;
    STARTUPINFO StartInfo = {sizeof(StartInfo)};

    STARTUPINFO si;
    ZeroMemory(&si,sizeof(STARTUPINFO));
    si.cb = sizeof(STARTUPINFO);
    si.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
    si.hStdOutput = NULL;
    si.hStdInput  = NULL;
    si.hStdError  = NULL;
    si.wShowWindow = SW_HIDE;

    // Change the current directory to the server directory
    CString strServerDir;
    strServerDir.Format ("%sServer", g_strOSTAHome);

    ::SetCurrentDirectory(strServerDir.GetBuffer(0));

    // Copy the command line, because CreateProcess may modify it directly!
    char* zCmdLine = (char*)_alloca(Command.GetLength() + 1);
    lstrcpy(zCmdLine, Command);

    // Create the process to execute the command
    PROCESS_INFORMATION pi;
    if (!CreateProcess(NULL,
                       zCmdLine,            // command line
                       NULL,                // process security attributes
                       NULL,                // primary thread security attributes
                       TRUE,                // handles are inherited
                       CREATE_NEW_CONSOLE,  // creation flags
                       NULL,                // use parent's environment
                       strServerDir.GetBuffer(0),       // Current directory
                       &si,                 // STARTUPINFO pointer
                       &pi))                // receives PROCESS_INFORMATION
    {
        ErrorBuffer.Format ("Error executing command: %s" ,Command);
        LPTSTR lpMsgBuf;

        FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                       NULL,
                       GetLastError(),
                       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                       (LPTSTR) &lpMsgBuf,
                       0,
                       NULL);
        //**AfxMessageBox (lpMsgBuf);
        WriteLogMsg (TRCLEV_NONE, lpMsgBuf);

        // Free the buffer.
        LocalFree(lpMsgBuf);
        return -1;
    }

    // Return the process details.
    *pReturnHandle = pi.hProcess;
    *pReturnPid = pi.dwProcessId;

    // Release the unused handle:
    CloseHandle(pi.hThread);

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @func    This function is invoked if the Architecture Manager Daemon is
//          aborted. It performs any required tidying up and exits.
//
void AbortDaemon(int)
{
    WriteLogMsg (TRCLEV_LOW, string("Abort Architecture Manager Daemon requested."));

    Shutdown();
    CyrtoolsUninitialize();

    // Log message and exit.
    WriteLogMsg (TRCLEV_NONE, string("Architecture Manager Daemon aborted."));
    exit(0);
}

///////////////////////////////////////////////////////////////////////////////
// @func    This function deallocates memory and shuts down the CORBA server.
//
void Shutdown()
{
    try
    {
        // Tidy memory.
        if (g_pArchMgrRun != NULL)
        {
            delete g_pArchMgrRun;
            g_pArchMgrRun = NULL;
        }

        // Shutdown CORBA Server.
        if (g_pCorba != NULL)
        {
            delete g_pCorba;
            g_pCorba = NULL;
        }
    }
    catch(...)
    {}
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method writes a time-stamped message to the local
//          Architecture Manager Daemon log file.
//
// @parm    Severity of message to be logged (TRCLEV_* symbols 1 to 1000).
// @parm    Message to be logged.
//
void WriteLogMsg(int severity, const string& msg)
{
    static CTMutex mutex;

    CTLockGuard<CTMutex> lock(mutex);

    // If g_pLogFile is NULL, we cannot log an error, so don't try.
    if (g_pLogFile == NULL) return;

	// Check if this severity is to be logged, just return if not.
	if(severity > g_nTraceLevel) return;

    if(g_pLogFile->good())
    {
        CTTime t;
        t.now();
        string strTim(t.Format("%Y-%m-%d %H:%M:%S "));

        g_pLogFile->write(strTim.c_str(), strTim.length());
        g_pLogFile->write(msg.c_str(), msg.length());
        g_pLogFile->write("\r\n", 2);
        g_pLogFile->flush();
    }
}