/* Copyright (C) 2000-2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
#include "InjectorStatus_i.h"
#include <CECInetInfo.h>
#include "mutex.hpp"

using namespace std;

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Constructor. Create local IS log file (ISLog_IPADR.txt) and
//          register the InjectorStatus interface with the name service.
//
// @parm    Pointer to CORBA server object.
// @parm    Injector Control object reference.
//
InjectorStatusImpl::InjectorStatusImpl(CORBAServer* corba, InjectorControlImpl& ic)
    : m_ic(ic), m_pRepository(NULL)
{
    // Create and open the IS Log file OPENSTA_HOME\Engines\ISLog_IPADR.txt, where
    // IPADR is the IP address (with dots replaced by underscores) of the daemon host.
   	m_bTraceLevel=10;

	char * pOpenStaHome = getenv("OPENSTA_HOME") ;
    if (!pOpenStaHome) return ;
    string strLocalPath = pOpenStaHome;
    CECInetInfo inet;
    m_registered_name=inet.GetIPAddressAsString();
    string strISLogFilNam = "ISLog_" + m_registered_name + ".txt";
#ifdef _WIN32
    strLocalPath += "Engines\\Temp\\" + strISLogFilNam;
#else
    strLocalPath += "Engines/Temp/" + strISLogFilNam;
#endif
    m_pLogFile = new ofstream( strLocalPath.c_str(), std::ios::binary|std::ios::trunc|std::ios::out);

	// Register the InjectorStatus interface with the name service.
    m_corba = corba;
    WriteLogMsg(TRCLEV_NONE,string("InjectorStatusImpl::InjectorStatusImpl - Register IS with name " + m_registered_name));
    m_corba->NameObject(this->_this(), InjectorStatus::CLASS_NAME, m_registered_name.c_str());
    WriteLogMsg(TRCLEV_NONE,string("InjectorStatusImpl::InjectorStatusImpl - Register IS with name complete"));
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Destructor. Deregister the InjectorStatus object from the name
//          service.
//
InjectorStatusImpl::~InjectorStatusImpl()
{
	// Deregister the InjectorStatus interface.
    m_corba->UnNameObject(InjectorStatus::CLASS_NAME, m_registered_name.c_str());
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method gets a 'snapshot' of the status data for each Task
//          Group Executer running on the local Injector.
//
// @rdesc   Return status. true=method successful, false=method failed.
//
// @parm    List of status 'snapshots' for all the Task Group Executers running
//          on the local Injector.
//
CORBA::Boolean InjectorStatusImpl::GetStatsSnapshot(TESnapList& InjectorSnapList)
{
    TestExecuterSnapshot TeSnap;            // Executer status snapshot
    TaskGroupSnap TgSnap;                   // Task Group Snapshot
    string tenam = "";

    CTLockGuard<CTMutex> lock(m_ic.m_TEMutex);

    try
	{

        // Get the status snapshot for each Executer on the local Injector
        // and append it to the list of snapshots for the Injector.
        InjectorSnapList.length(0);
        for(InjectorControlImpl::TestExecuters_t::iterator ppTE = m_ic.m_TEs.begin(); ppTE != m_ic.m_TEs.end(); ++ppTE)
		{
            tenam = (*ppTE)->m_strRegName;
			if ((*ppTE)->m_pTE->GetStatsSnapshot(TeSnap))
            {
                InjectorSnapList.length(InjectorSnapList.length() + 1);
                InjectorSnapList[ppTE - m_ic.m_TEs.begin()] = TeSnap;
            }
            else
            {
                string strmsg = "InjectorStatusImpl::GetStatsSnapshot: Failed to get snapshot from TG Executer '"
                    + tenam + "'";
#ifdef _WIN32
                LogError (TRCLEV_LOW,strmsg);
#endif
            }
            tenam = "";
		}
	}
	catch(...)
	{
        string strmsg = "InjectorStatusImpl::GetStatsSnapshot: Exception getting snapshot from TG Executer '"
                + tenam + "'; Could be TG stopped or completed";
        LogError (TRCLEV_LOW,strmsg);
		return false;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method determines the number of active Virtual Users on each
//          Executer running on the local Injector. It adds these numbers
//          together to get the total number of active Virtual Users on the
//          local Injector.
//
//          It is quite legitimate for one or more of the Test Executers in the
//          list of Executers not to be running (e.g. the Executer has just
//          been stopped) when we come to traverse the list of Executers. In
//          order to cater for this, we trap all exceptions when traversing
//          the list.
//
// @rdesc   Total number of active Virtual Users on the local Injector.
//
long InjectorStatusImpl::GetInjectorVUs(long lPreviousTestVUCount)
{
    long InjectorVUs = 0;               // Total number of active VUs on local Injector.
    string tenam;                       // Test/Task Group Executer Name

    // Get the number of active Virtual Users on each Executer on the local
    // Injector and add to the running total.
    CTLockGuard<CTMutex> lock(m_ic.m_TEMutex);

    try
	{
        for (InjectorControlImpl::TestExecuters_t::iterator ppTE = m_ic.m_TEs.begin();
             ppTE != m_ic.m_TEs.end();
             ++ppTE)
        {
            tenam = (*ppTE)->m_strRegName;
			if ((*ppTE)->m_tgtType == tgt_script)
			{
	            Script_TestExecuter_var pSTE(Script_TestExecuter::_narrow((*ppTE)->m_pTE));
	            if(!CORBA::is_nil(pSTE))
	                InjectorVUs += pSTE->GetTEVUs(lPreviousTestVUCount);
			}
		}
    }
    catch(...)
	{
        string strmsg =
            "InjectorStatusImpl::GetInjectorVUs: Exception getting VU count from TG Executer '"
            + tenam + "'; Could be TG stopped or completed";
        LogError (TRCLEV_LOW,strmsg);
    }

    // Return total number of active Virtual Users on the local Injector.
	return InjectorVUs;
}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method determines the number of active  and total Virtual Users 
//          on each
//          Executer running on the local Injector. It adds these numbers
//          together to get the total number of active and total Virtual Users 
//	        on the
//          local Injector.
//
//          It is quite legitimate for one or more of the Test Executers in the
//          list of Executers not to be running (e.g. the Executer has just
//          been stopped) when we come to traverse the list of Executers. In
//          order to cater for this, we trap all exceptions when traversing
//          the list.
//
//
void InjectorStatusImpl::GetInjectorTotalAndActiveVUs(CORBA::Long lPreviousTestActiveVUCount, CORBA::Long& lActiveVUCount, CORBA::Long lPreviousTestTotalVUCount, CORBA::Long& lTotalVUCount)
{
    long lActiveInjectorVUs = 0; // Total number of active VUs on local Injector.
    long lTotalInjectorVUs = 0;  // Total number of started VUs on local Injector.
    string tenam;                // Test/Task Group Executer Name

    // Get the number of active Virtual Users on each Executer on the local
    // Injector and add to the running total.
    CTLockGuard<CTMutex> lock(m_ic.m_TEMutex);

    try
	{
	    WriteLogMsg(TRCLEV_NONE,"InjectorStatusImpl::GetInjectorTotalAndActiveVUs");
		long lActiveVUs = 0 ;
		long lTotalVUs = 0 ;
        for (InjectorControlImpl::TestExecuters_t::iterator ppTE = m_ic.m_TEs.begin();
             ppTE != m_ic.m_TEs.end();
             ++ppTE)
        {
            tenam = (*ppTE)->m_strRegName;
		    WriteLogMsg(TRCLEV_NONE,string("InjectorStatusImpl::GetInjectorTotalAndActiveVUs, TExecuter name : " + tenam));
			if ((*ppTE)->m_tgtType == tgt_script)
			{
	            Script_TestExecuter_var pSTE(Script_TestExecuter::_narrow((*ppTE)->m_pTE));
	            if(!CORBA::is_nil(pSTE))
				{
					lActiveVUs = 0 ;
					lTotalVUs = 0 ;

				    WriteLogMsg(TRCLEV_NONE,string("InjectorStatusImpl::GetInjectorTotalAndActiveVUs, calling pSTE->GetTETotalAndActiveVUs..."));
	                pSTE->GetTETotalAndActiveVUs(lPreviousTestActiveVUCount, lActiveVUs, lPreviousTestTotalVUCount, lTotalVUs);
					lActiveInjectorVUs += lActiveVUs ;
					lTotalInjectorVUs += lTotalVUs ;
				}
			}
		}
		lActiveVUCount = lActiveInjectorVUs ;
		lTotalVUCount = lTotalInjectorVUs ;
    }
    catch(...)
	{
        string strmsg =
            "InjectorStatusImpl::GetInjectorTotalAndActiveVUs: Exception getting VU count from TG Executer '"
            + tenam + "'; Could be TG stopped or completed";
        LogError (TRCLEV_LOW,strmsg);
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method writes a message to the local Injector Status log file
//          AND the OpenSTA Error Log.
//
// @parm    Message to be logged.
//
void InjectorStatusImpl::LogError (int severity, const string& strmsg)
{
	if(m_ic.ShowErrors())
	{
		WriteLogMsg(severity,strmsg);
		CECErrorLog errmsg(m_ic.GetTimeOffset(),m_ic.GetTestName().c_str(),InjectorStatus::CLASS_NAME, strmsg.c_str());
		m_ic.AddTestOutput(CECErrorLog::OUTPUT_TYPE, errmsg.ToString().c_str());
	}
	else
	{
		WriteLogMsg(severity,string("TEST STOPPING ERROR SUPRESSED - " + strmsg));
	}

	return;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method writes a time-stamped message to the local Injector
//          Status log file.
//
// @parm    Message to be logged.
//
void InjectorStatusImpl::WriteLogMsg(int severity,const string& msg)
{
	/*
	TRCLEV_NONE          0              // Fatal Errors only
	TRCLEV_LOW          10              // Important Trace error
	TRCLEV_MEDIUM       20              // Normal Trace Level
	TRCLEV_HIGH         50              // High Trace level
    */

    static CTMutex mutex;

    CTLockGuard<CTMutex> lock(mutex);

    // If m_pLogFile is NULL, we cannot log an error, so don't try.
    if (m_pLogFile == NULL) return;

	//check if severity of error is greater than m_bTraceLevel of IC
	if(severity>m_ic.GetTraceLevel())
	{
		return;
	}

    if(m_pLogFile->good())
    {
        CTTime t;
        t.now();
        string strTim(t.Format("%Y-%m-%d %H:%M:%S "));

        m_pLogFile->write(strTim.c_str(), strTim.length());
        m_pLogFile->write(msg.c_str(), msg.length());
        m_pLogFile->write("\r\n", 2);
        m_pLogFile->flush();
    }
}

CORBA::Boolean InjectorStatusImpl::GetTaskSnapshot(const TaskIdList& taskIDs, TaskSnap2List& taskSnapshots)
{
	bool bRet = false;
    CTLockGuard<CTMutex> lock(m_ic.m_TEMutex);

    try
	{
        for (InjectorControlImpl::TestExecuters_t::iterator ppTE = m_ic.m_TEs.begin();
             ppTE != m_ic.m_TEs.end();
             ++ppTE)
        {
			if ((*ppTE)->m_tgtType == tgt_script)
			{
	            Script_TestExecuter_var pSTE(Script_TestExecuter::_narrow((*ppTE)->m_pTE));
	            if(!CORBA::is_nil(pSTE))
	                bRet = pSTE->GetTaskSnapshot(taskIDs, taskSnapshots);
			}
		}
    }
    catch(...)
	{
        string strmsg ="Exception from: InjectorStatusImpl::GetTaskSnapshot";
        LogError (TRCLEV_NONE ,strmsg);
    }

	return bRet;
}
