/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc CyrStApi CyrWrapLib
///////////////////////////////////////////////////////////////////////////////

#ifndef _CECTestOutput_h
#define _CECTestOutput_h

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <ctime.hpp>

#include <string>
#include <list>

using std::list;
using std::string;
///////////////////////////////////////////////////////////////////////////////
// @class   CECTestOutput | This class is the base class for all Test Output.
//          It provides only a single method.  This method is a pure virtual
//          method that transforms the object into a string version of itself.
//          This string will be stored as the test output.
class CECTestOutput
{
    // @access Public Members
public:
    CECTestOutput(long t_offsets);            // @cmember Constructor
    virtual ~CECTestOutput();   // @cmember Destructor

    virtual string GetType() = 0;    // @cmember Returns type of output
    virtual string ToString() = 0;   // @cmember Serialize data to a string
    CTTime GetCreatedTime() const;      // @cmember Get time of timer creation

    // @access Protected Members
protected:
    CTTime    m_time;    // @cmember Creation time field
}
;

///////////////////////////////////////////////////////////////////////////////
// @class   CECTimerOutput | This class is used to create Timer Output records.
// @base    public | CECTestOutput
class CECTimerOutput : public CECTestOutput
{
    // @access Public Members
public:
    // @cmember Constructor to create Test Output
    CECTimerOutput( long t_offsets, const char* desc, long value, long active );
    // @cmember Constructor to retrieve Test Output
    CECTimerOutput( const char* str );
    virtual ~CECTimerOutput();          // @cmember Destructor

    static const char OUTPUT_TYPE[];     // @cmember Type of output
    virtual string GetType();        // @cmember Returns type of output
    virtual string ToString();       // @cmember Serialize data to a string

    string GetDescription() const;   // @cmember Get timer description
    long GetValue() const;              // @cmember Get timer value
    long GetActiveThreads() const;      // @cmember Get number of active threads

    // @access Protected Members
protected:
    string m_desc;    // @cmember Description field
    long      m_value;   // @cmember Value field
    long      m_active;  // @cmember Number of active threads
}
;

///////////////////////////////////////////////////////////////////////////////
// @class   CECTestLog | This is used to create test log records.
// @base    public | CECTestOutput
class CECTestLog : public CECTestOutput
{
    // @access Public Members
public:
    CECTestLog( long t_offsets, const char* text, bool create = true );  // @cmember Construtor
    virtual ~CECTestLog();          // @cmember Destructor

    static const char OUTPUT_TYPE[]; // @cmember Type of output
    virtual string GetType();    // @cmember Returns type of output
    virtual string ToString();   // @cmember Serialize data to a string

    string GetText() const;      // @cmember Returns test log text

    // @access Protected Members
protected:
    string m_text;  // @cmember Text entry
}
;

class CECTestReport : public CECTestLog
{
    // @access Public Members
public:
    CECTestReport( long t_offsets, const char* text, bool create = true );  // @cmember Constructor

    static const char OUTPUT_TYPE[];     // @cmember Type of output
    virtual string GetType();        // @cmember Returns type of output
}
;

class CECTestDiagnostic : public CECTestLog
{
    // @access Public Members
public:
    CECTestDiagnostic( const char* text, bool create = true );  // @cmember Constructor

    static const char OUTPUT_TYPE[];     // @cmember Type of output
    virtual string GetType();        // @cmember Returns type of output
}
;

class CECTestHistory : public CECTestLog
{
    // @access Public Members
public:
    CECTestHistory( long t_offsets, const char* text, bool create = true );  // @cmember Constructor

    static const char OUTPUT_TYPE[];     // @cmember Type of output
    virtual string GetType();        // @cmember Returns type of output
}
;

///////////////////////////////////////////////////////////////////////////////
// OTA Test Output
///////////////////////////////////////////////////////////////////////////////
typedef struct tagsRESULT
{
    int start;
    int stop;
    int result;
}
RESULT_S, *PRESULT_S;

typedef list<PRESULT_S> ResultData;

///////////////////////////////////////////////////////////////////////////////
// @class   CECOtaSimpleQuery | This class represents a simple set of query
//          results generated by an OTA query.
// @base    public | CECTestOutput
class CECOtaSimpleQuery : public CECTestOutput
{
    // @access Public Members
public:
    // @cmember Constructor for retrieval
    CECOtaSimpleQuery( const char* str );
    // @cmember Constructor for creation of data entry
    CECOtaSimpleQuery( long t_offsets, const char* node, int index, int start, int stop, int result );
    virtual ~CECOtaSimpleQuery();   // @cmember Destructor

    static const char OUTPUT_TYPE[]; // @cmember Type of output
    virtual string GetType();    // @cmember Returns type of output
    virtual string ToString();   // @cmember Serialize data to a string

    string GetNodeName();        // @cmember Returns name of node
    string GetDescription();     // @cmember Returns description
    int GetStartTime();
    int GetStopTime();
    int GetResult();
    int GetIndex();

    // @access Private Members
private:
    int         m_index;        // @cmember Query index
    bool        m_header;       // @cmember True if header
    string   m_node;         // @cmember Name of node where query executed
    string   m_desc;         // @cmember Description for query
    int         m_start;        // @cmember Start time
    int         m_stop;         // @cmember Start time
    int         m_result;       // @cmember Result
}
;


///////////////////////////////////////////////////////////////////////////////
// @class   CECOtaSimpleQuery | This class represents a simple set of query
//          results generated by an OTA query.
// @base    public | CECTestOutput
class CECOtaHeader : public CECTestOutput
{
    // @access Public Members
public:
    // @cmember Constructor for retrieval
    CECOtaHeader( const char* str );
    // @cmember Constructor for creation of header
    CECOtaHeader( long t_offsets, const char* node, int index, const char* desc );
    virtual ~CECOtaHeader();   // @cmember Destructor

    static const char OUTPUT_TYPE[]; // @cmember Type of output
    virtual string GetType();    // @cmember Returns type of output
    virtual string ToString();   // @cmember Serialize data to a string

    string GetNodeName();        // @cmember Returns name of node
    string GetDescription();     // @cmember Returns description
    int GetIndex();                 // @cmember Returns index

    // @access Private Members
private:
    string   m_node;         // @cmember Name of node where query executed
    string   m_desc;         // @cmember Description for query
    int m_index;
};

///////////////////////////////////////////////////////////////////////////////
// @class   CECNTPHeader | This class represents query header generated by and
//			NT Performance Query
// @base    public | CECTestOutput
class CECNTPHeader : public CECTestOutput
{
    // @access Public Members
public:
    // @cmember Constructor for retrieval
    CECNTPHeader( const char* str );
    // @cmember Constructor for creation of header
    CECNTPHeader( long t_offsets, int nTaskGroupID, int index, const char* desc );
    virtual ~CECNTPHeader();   // @cmember Destructor

    static const char OUTPUT_TYPE[]; // @cmember Type of output
    virtual string GetType();    // @cmember Returns type of output
    virtual string ToString();   // @cmember Serialize data to a string

    int GetTaskGroupID();        // @cmember Returns name of node
    string GetDescription();     // @cmember Returns description
    int GetIndex();                 // @cmember Returns index

    // @access Private Members
private:
    int   m_nTG;         // @cmember Name of node where query executed
    string   m_desc;         // @cmember Description for query
    int m_index;
};

///////////////////////////////////////////////////////////////////////////////
// @class   CECNTPQuery | This class represents a query result the NT Performance monitor
// @base    public | CECTestOutput
class CECNTPQuery : public CECTestOutput
{
    // @access Public Members
public:
    // @cmember Constructor for retrieval
    CECNTPQuery( const char* str );
    // @cmember Constructor for creation of data entry
    CECNTPQuery( long t_offsets, int nTaskGroupID, int index, int start, int stop, int result );
    virtual ~CECNTPQuery();   // @cmember Destructor

    static const char OUTPUT_TYPE[]; // @cmember Type of output
    virtual string GetType();    // @cmember Returns type of output
    virtual string ToString();   // @cmember Serialize data to a string

    int GetTaskGroupID();        // @cmember Returns name of node
    string GetDescription();     // @cmember Returns description
    int GetStartTime();
    int GetStopTime();
    int GetResult();
    int GetIndex();

    // @access Private Members
private:
    int         m_index;        // @cmember Query index
    int   m_nTG;         // @cmember Name of node where query executed
    string   m_desc;         // @cmember Description for query
    int         m_start;        // @cmember Start time
    int         m_stop;         // @cmember Start time
    int         m_result;       // @cmember Result
}
;

// @class   CECSMPHeader | This class represents query header generated by and
//			SNMP Performance Query
// @base    public | CECTestOutput
class CECSNMPHeader : public CECTestOutput
{
    // @access Public Members
public:
    // @cmember Constructor for retrieval
    CECSNMPHeader( const char* str );
    // @cmember Constructor for creation of header
    CECSNMPHeader( long t_offsets, int nTaskGroupID, int index, const char* desc );
    virtual ~CECSNMPHeader();   // @cmember Destructor

    static const char OUTPUT_TYPE[]; // @cmember Type of output
    virtual string GetType();    // @cmember Returns type of output
    virtual string ToString();   // @cmember Serialize data to a string

    int GetTaskGroupID();        // @cmember Returns name of node
    string GetDescription();     // @cmember Returns description
    int GetIndex();                 // @cmember Returns index

    // @access Private Members
private:
    int   m_nTG;         // @cmember Name of node where query executed
    string   m_desc;         // @cmember Description for query
    int m_index;
};

///////////////////////////////////////////////////////////////////////////////
// @class   CECSNMPQuery | This class represents a query result the SNMP monitor
//          results generated by an OTA query.
// @base    public | CECTestOutput
class CECSNMPQuery : public CECTestOutput
{
    // @access Public Members
public:
    // @cmember Constructor for retrieval
    CECSNMPQuery( const char* str );
    // @cmember Constructor for creation of data entry
    CECSNMPQuery( long t_offsets, int nTaskGroupID, int index, int start, int stop, int result );
    virtual ~CECSNMPQuery();   // @cmember Destructor

    static const char OUTPUT_TYPE[]; // @cmember Type of output
    virtual string GetType();    // @cmember Returns type of output
    virtual string ToString();   // @cmember Serialize data to a string

    int GetTaskGroupID();        // @cmember Returns name of node
    string GetDescription();     // @cmember Returns description
    int GetStartTime();
    int GetStopTime();
    int GetResult();
    int GetIndex();

    // @access Private Members
private:
    int         m_index;        // @cmember Query index
    bool        m_header;       // @cmember True if header
    int   m_nTG;         // @cmember Name of node where query executed
    string   m_desc;         // @cmember Description for query
    int         m_start;        // @cmember Start time
    int         m_stop;         // @cmember Start time
    int         m_result;       // @cmember Result
}
;

///////////////////////////////////////////////////////////////////////////////
// @class   CECCollector1Header | This class represents an index for data collected
//          by Data Collector1.
//
// @base    public | CECTestOutput
class CECCollector1Header : public CECTestOutput
{
    // @access Public Members
public:
    // @cmember Constructor for retrieval
    CECCollector1Header( const char* str );
    // @cmember Constructor for creation of header
    CECCollector1Header( long t_offset, int timestamp, int index, const char* prfnam);
    virtual ~CECCollector1Header();   // @cmember Destructor

    static const char OUTPUT_TYPE[]; // @cmember Type of output
    virtual string GetType();    // @cmember Returns type of output
    virtual string ToString();   // @cmember Serialize data to a string

	int		GetTimestamp();
	int		GetIndex();
    string	GetProfileName();

    // @access Private Members
private:
	int		m_timestamp;
	int		m_index;
    string  m_prfnam;           // @cmember Anomaly type
};

///////////////////////////////////////////////////////////////////////////////
// @class   CECCollector1Data | This class represents a set of data collected
//          by Data Collector1.
//
// @base    public | CECTestOutput
class CECCollector1Data : public CECTestOutput
{
    // @access Public Members
public:
    // @cmember Constructor for retrieval
    CECCollector1Data(const char* str);
    // @cmember Constructor for creation of data entry
    CECCollector1Data(long t_offsets, const char* anotyp, int anosev, int anoval, int pid, int anoid,
        const char* dbnam, const char* dbobjnam, const char* prfnam,
        int begtim, int modtim, int endtim, int anotim);
    virtual ~CECCollector1Data();   // @cmember Destructor

    static const char OUTPUT_TYPE[]; // @cmember Type of output
    virtual string GetType();    // @cmember Returns type of output
    virtual string ToString();   // @cmember Serialize data to a string

    int GetBeginTime();
    int GetModTime();
    int GetEndTime();
    int GetAnomalyTime();
    string GetAnomalyType();
	int GetAnomalySeverity();
    int GetAnomalyValue();
    int GetPid();
    int GetAnomalyId();
    string GetDBName();
    string GetDBObjName();
    string GetProfileName();

    // @access Private Members
private:
    int     m_begtim;           // @cmember Anomaly begin time
    int     m_modtim;           // @cmember Anomaly last modified time
    int     m_endtim;           // @cmember Anomaly end time
    int     m_anotim;           // @cmember Anomaly duration
    string  m_anotyp;           // @cmember Anomaly type
	int		m_anosev;			// @cmember Anomaly severity
    int     m_anoval;           // @cmember Anomaly value
    int     m_pid;              // @cmember Anomaly process ID
    int     m_anoid;            // @cmember Anomaly ID
    string  m_dbnam;            // @cmember Anomaly DB name
    string  m_dbobjnam;         // @cmember Anomaly DB object name
    string  m_prfnam;           // @cmember Profile name
};

class CECVULogOutput : public CECTestOutput
{
    // @access Public Members
public:
    CECVULogOutput(long t_offsets, int nActiveVUsers, int nTotalVUsers);
	CECVULogOutput(const char* strEntry);

    virtual ~CECVULogOutput();   // @cmember Destructor
    static const char OUTPUT_TYPE[]; // @cmember Type of output

    CTTime GetTime();      // @cmember Get time of timer creation
    int GetActiveVUsers();    // @cmember Returns type of output
    int GetTotalVUsers();    // @cmember Returns type of output

    string GetType();    // @cmember Returns type of output
    string ToString();   // @cmember Serialize data to a string

    // @access Protected Members
protected:
	int m_nActiveVUsers;
	int m_nTotalVUsers;
}
;

class CECErrorLog : public CECTestOutput
{
    // @access Public Members
public:
    CECErrorLog(long t_offsets, const char* strTestName, const char* strLocation, const char* strMsg);
	CECErrorLog(const char* strEntry);

    virtual ~CECErrorLog();   // @cmember Destructor
    static const char OUTPUT_TYPE[]; // @cmember Type of output

    CTTime GetTime();      // @cmember Get time of timer creation
    string GetTestName();    // @cmember Returns type of output
    string GetLocation();    // @cmember Returns type of output
    string GetMessage();    // @cmember Returns type of output

    string GetType();    // @cmember Returns type of output
    string ToString();   // @cmember Serialize data to a string

    // @access Protected Members
protected:
	string m_strTestName;
	string m_strLocation;
	string m_strMsg;
}
;
#endif
