/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrFFApi
// @module  Impact Project Wrapper | This module contains the implementation of
//          the CECFFImpactProject wrapper class.  This class extends the
//          CECFFProject wrapper class and creates all the appropriate child
//          folders.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include "CECFFFile.h"
#include "CECFFImpactProject.h"
#include "CECFFImpactFolders.h"

///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Constants
const char* BUILD_FILE = "BuildSettings";

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor retrieves an existing Impact project from the
//          OpenSTARep database.
// @parm    Pointer to an instance of the parent Server wrapper class.
// @parm    The name of the Project to retrieve.
CECFFImpactProject::CECFFImpactProject( CECFFServer* server, const char* name )
        : CECFFProject( server, string( string("Cyr.Impact.")+ name  ).c_str() )
{
    m_debug = false;
    m_list = false;
    m_checkSteps = false;
    m_steps = 1;

    ReadBuildSettings();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor allows the creation of a new Impact Project within
//          the OpenSTARep database.
// @parm    Pointer to an instance of the parent Server wrapper class.
// @parm    The name of the new Project.
// @parm    The description of the new Project.
CECFFImpactProject::CECFFImpactProject( CECFFServer* server, const char* name,
                                        const char* desc )
        : CECFFProject( server, string( string("Cyr.Impact.")+ name ).c_str() , desc )
{
    // Get root folder
    CECFFFolder* rootFolder = new CECFFFolder( this );

    // Create all child folders
    CECFFSourceFolder* pSource = new CECFFSourceFolder( rootFolder, desc );
    CECFFTestFolder* pTests = new CECFFTestFolder( rootFolder, desc );
    CECFFDataFolder* pData = new CECFFDataFolder( rootFolder, desc );
    CECFFObjectCodeFolder* pObjCode = new CECFFObjectCodeFolder( rootFolder, desc );
    CECFFDependancyFolder* pDependancies = new CECFFDependancyFolder( rootFolder, desc );
    CECFFParameterFolder* pParameters = new CECFFParameterFolder( rootFolder, desc );
    CECFFCaptureFolder* pCapture = new CECFFCaptureFolder( rootFolder, desc );

    // Tidy memory until such time as actual folders are requested
    delete pObjCode;
    delete pData;
    delete pTests;
    delete pSource;
    delete pDependancies;
    delete pParameters;
    delete pCapture;
    delete rootFolder;

    m_debug = false;
    m_list = false;
    m_checkSteps = false;
    m_steps = 1;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor of this class writes out the current buidl settings.
CECFFImpactProject::~CECFFImpactProject()
{
    WriteBuildSettings();
}

bool CECFFImpactProject::GetBuildDebug()
{
    return m_debug;
}

void CECFFImpactProject::SetBuildDebug( bool flag )
{
    m_debug = flag;
}

bool CECFFImpactProject::GetBuildList()
{
    return m_list;
}

void CECFFImpactProject::SetBuildList( bool flag )
{
    m_list = flag;
}

bool CECFFImpactProject::GetBuildCheckSteps()
{
    return m_checkSteps;
}

void CECFFImpactProject::SetBuildCheckSteps( bool flag )
{
    m_checkSteps = flag;
}

long CECFFImpactProject::GetBuildMaxSteps()
{
    return m_steps;
}

void CECFFImpactProject::SetBuildMaxSteps( long steps )
{
    m_steps = steps;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method reads the build settings from a remote file.
void CECFFImpactProject::ReadBuildSettings()
{
    // Get root folder
    CECFFFolder* rootFolder = new CECFFFolder( this );
    CECFFFile* settings = NULL;

    try
    {
        // Get latest build settings file from root folder
        settings = new CECFFFile( rootFolder, BUILD_FILE );
        settings->CheckOut( false );

        // Open file and read bytes of file as appropriate
        std::ifstream* f = new std::ifstream( settings->GetFullPath().c_str());
        if ( f->good() )
        {
            *f>>m_debug>>m_list>>m_checkSteps>>m_steps;
            /*f->read( m_debug );
            f->read( m_list );
            f->read( m_checkSteps );
            f->read( m_steps );*/
        }
        delete f;
    }
    catch( ... )
    {}

    // Tidy memory

    if ( settings != NULL )
        delete settings;
    if ( rootFolder != NULL )
        delete rootFolder;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method writes the build settings to a file and then checks
//          that file in.
void CECFFImpactProject::WriteBuildSettings()
{
    // Error information
    const char* where = "CECFFImpactProject::WriteBuildSettings";

    // Get root folder object
    CECFFFolder* rootFolder = new CECFFFolder( this );

    // Create full path name of settings file
    string fullpath = rootFolder->GetFullPath();
    fullpath += BUILD_FILE;

    try
    {
        // Open file and write contents to it
        std::ofstream* f = new std::ofstream( fullpath.c_str());
        if ( f->good() )
        {
            f->write( (const char*)m_debug ,sizeof(m_debug));
            f->write( (const char*)m_list ,sizeof(m_list));
            f->write( (const char*)m_checkSteps,sizeof(m_checkSteps) );
            f->write( (const char*)m_steps ,sizeof(m_steps ));
        }
        delete f;

        try
        {
            // Try to update existing entry
            CECFFFile* settings = new CECFFFile( rootFolder, BUILD_FILE );
            settings->Lock();
            settings->CheckIn( true );
            delete settings;
        }
        catch( CyrRepositoryException ex )
        {
            if ( ex.status() != CYR_FILE_NOT_FOUND )
                throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
            else
            {
                // Entry does not exist so Create it
                CECFFFile* settings = new CECFFFile( rootFolder, BUILD_FILE, BUILD_FILE );
                delete settings;
            }
        }
    }
    catch( CyrRepositoryException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), ex.where().c_str(), ex.status() );
    }

    // Tidy memory
    delete rootFolder;
}

