/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyranoDaemon
// @module  Cyrano Daemon | This module contains the implementation of the
//          Cyrano daemon application.  This application is intended for use as
//          a daemon process.  It creates and maintains an instance of the
//          <c CyrDaemon> class.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Daemon/CyrDmn.h>
#pragma warning (disable:4786)
#pragma warning (disable:4503)
#include <Logging/CECLog.h>


///////////////////////////////////////////////////////////////////////////////
// Standard includes
#include <iostream>
#include <signal.h>
#include <string>
using std::cerr;
using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Function declarations
void Usage();
void AbortDaemon( int );
void Shutdown();

///////////////////////////////////////////////////////////////////////////////
// Global variables
// @globalv CyrDaemon* | daemon | Pointer to daemon functionality.
// @globalv CORBAServer* | server | Pointer to CORBA helper functionality
// @globalv CECLog* | flogger | Pointer to global logging functionality
CyrDaemon* daemon = NULL;
CORBAServer* server = NULL;
CECLog* flogger = NULL;

///////////////////////////////////////////////////////////////////////////////
// @func    This function is the main function for the Cyrano Daemon
//          executable.  It creates a <c CyrDaemon> class and sets up various
//          parameters for running the daemon.  This executable is a command-
//          line application.
// @parm    The number of command-line arguments.
// @parm    An array of command-line arguments.
#include "process.hpp"
int main( int argc, char** argv )
{
    CyrtoolsInitialize();

    int period = -1;
    int trace = 0;

    // Tell application what to do in case daemon is aborted.
    signal( SIGINT, AbortDaemon );
    signal( SIGABRT, AbortDaemon );

    // Parse command line and act on any daemon options
    int count = 0;
    while ( count < argc )
    {
        if ( strcmp( argv[count], "--help" ) == 0 )
        {
            // Display usage message and terminate application
            Usage();
            return 0;
        }
        else if ( strcmp( argv[count], "--trace" ) == 0 )
        {
            // Next parameter is the requested trace level
            count++;
            trace = atoi( argv[count] );
        }
        else if ( strcmp( argv[count], "--period" ) == 0 )
        {
            // Next parameter is the period in seconds
            count++;
            period = atoi( argv[count] );
        }

        count++;
    }

    // Calculate name for this daemon.  This will be the host address.
    CECInetInfo inet;
    string name=inet.GetIPAddressAsString();
    try
    {
        // Start ORB and BOA
        server = new CORBAServer( argc, argv );

        // Start logging object
        string logName = name;
        logName += "_";
        logName += CYR_DAEMON;
        flogger = new CECLog( server, logName.c_str(), trace );

        // Construct a new Daemon object
        daemon = new CyrDaemon( server, name.c_str() );

        // If the period argument has been set then run daemon for a
        // specified period of time (time in seconds).  Otherwise run
        // the daemon forever.
        if ( period != -1 )
        {
            daemon->Run( true );
            CyrSleep( period * 1000 );
        }
        else
            daemon->Run( false );
    }
    catch( CyrDaemonException ex )
    {
        std::cerr << "Cyrano daemon has thrown exception " << ex.what();
        std::cerr << " in " << ex.where() << std::endl;
    }
    catch( ... )
    {
        std::cerr << "Unknown exception caught by daemon." << std::endl;
    }

    // Call shutdown procedure
    Shutdown();

    CyrtoolsUninitialize();

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @func   This function tidies up any allocated memory if the application is
//         aborted for any reason.
void AbortDaemon( int )
{
    Shutdown();
    CyrtoolsUninitialize();

    // Output message and quit application
    cerr << "Cyrano daemon application aborted." << endl;
    exit( 0 );
}

///////////////////////////////////////////////////////////////////////////////
// @func   This function displays the help and usage information for the daemon
//         application.
void Usage()
{
    cout << "Cyrano daemon application.  This application runs in" << endl;
    cout << "the background and allows remote Cyrano applications" << endl;
    cout << "to talk to this machine.\n" << endl;

    cout << "cyrdmn [--help] [--trace <level>] [--period <secs>]" << endl;
    cout << "\n" << endl;
    cout << "   --help      Displays this message." << endl;
    cout << "   --trace     Sets initial trace level for the daemon.\n";
    cout << "   --period    Daemon will only run for the specified length\n";
    cout << "               of time (value in seconds)." << endl;
}

///////////////////////////////////////////////////////////////////////////////
// @func    This function shuts down any active CORBA objects and attempts to
//          clear up any memory
void Shutdown()
{
    try
    {
        // Tidy memory
        if ( daemon != NULL )
        {
            delete daemon;
            daemon = NULL;
        }

        // Get rid of logging object
        if ( flogger != NULL )
        {
            delete flogger;
            flogger = NULL;
        }

        // Shut down ORB and BOA
        if ( server != NULL )
        {
            delete server;
            server = NULL;
        }
    }
    catch(...)
    {}

}
