/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrStRepository
// @module  Byte Iterators | This module contains an implementation of
//          the CyrStByteIt CORBA object.  This CORBA object is a mechanism
//          for transferring file data in small and easy to handle chunks.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Repository/CyrStByteIt_i.h>

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the current size of the byte array.
long CyrStByteItImpl::GetSize()
{
    return m_byteArray.size();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a pointer to a sequence of bytes from the
//          array we are currently iterating through.  The maximum size of the
//          sequence is defined and iteration is done by repeated requests
//          to this method.  A sequence that has no elements indicates that the
//          iteration is complete.
// @rdesc   Pointer to the sequence of bytes.
ByteSeq* CyrStByteItImpl::ReadBytes()
{
    // Determine size of array to read - limiting it to the maximum we
    // have specified.  If we have read to the end of the array then we
    // return an empy sequence so that the iteration can be stopped.
    int count = GetSize();
    if ( count <= 0 )
    {
        // Create and return empty sequence
        ByteSeq_var bySeq = new ByteSeq();
        return bySeq._retn();
    }

    if ( count > BYTESEQ_SIZE )
        count = BYTESEQ_SIZE;

    // Allocate new buffer for bytes
    CORBA::Octet* buf = ByteSeq::allocbuf( count );

    // Read characters from array and destroy them
    for ( size_t i = 0; i < (size_t)count; i++ )
    {
        buf[i] = m_byteArray.front();
        m_byteArray.pop_front();
    }

    // Create sequence of bytes for transmission
    ByteSeq_var bySeq = new ByteSeq( count, count, buf, 1 );

    // Return reference and inform client that it has responsiblity for
    // deallocating memory.  Code copied from Henning and Vinoski pg 423.
    return bySeq._retn();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method inserts a collection of bytes into the internal array.
// @parm    Reference to a sequence of bytes.
void CyrStByteItImpl::WriteBytes( const ByteSeq& bytes )
{
    // Loop through all bytes in sequence and add them to array
    for ( size_t i = 0; i < bytes.length(); i++ )
        m_byteArray.push_back( bytes[i] );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc  This method reads a specified number of bytes into a byte array.  By
//         convincing the method that a specific type is in fact a byte array
//         we can rebuild longs, shorts, etc.
// @parm   The array of bytes to copy into.
void CyrStByteItImpl::ReadData( CORBA::Octet* bytes, size_t size )
{
    // Read specified number of bytes from array
    for ( size_t i = 0; i < size; i++ )
    {
        bytes[i] = m_byteArray.front();
        m_byteArray.pop_front();
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method writes a specified number of bytes into the byte array.
//          By casting another type to be a byte array we can parse longs,
//          shorts, etc. into individual bytes.
void CyrStByteItImpl::WriteData( CORBA::Octet* bytes, size_t size )
{
    // Write specified number of bytes into array
    for ( size_t i = 0; i < size; i++ )
        m_byteArray.push_back( bytes[i] );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method allows a long value to be read from the byte array.
// @rdesc   The long value.
long CyrStByteItImpl::ReadLong()
{
    long data;
    ReadData( (CORBA::Octet*)&data, sizeof(long) );
    return data;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds the bytes from a long to the byte array.
// @parm    The long to be added.
void CyrStByteItImpl::WriteLong( long data )
{
    WriteData( (CORBA::Octet*)&data, sizeof(long) );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method allows a short value to be read from the byte array.
// @rdesc   The short value.
short CyrStByteItImpl::ReadShort()
{
    short data;
    ReadData( (CORBA::Octet*)&data, sizeof(short) );
    return data;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds the bytes from a short to the byte array.
// @parm    The short to be added.
void CyrStByteItImpl::WriteShort( short data )
{
    WriteData( (CORBA::Octet*)&data, sizeof(short) );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method allows a Boolean value to be read from the byte array.
// @rdesc   The Boolean value.
CORBA::Boolean CyrStByteItImpl::ReadBool()
{
    CORBA::Boolean data;
    ReadData( (CORBA::Octet*)&data, sizeof(CORBA::Boolean) );
    return data;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds the bytes from a Boolean to the byte array.
// @parm    The Boolean to be added.
void CyrStByteItImpl::WriteBool( CORBA::Boolean data )
{
    WriteData( (CORBA::Octet*)&data, sizeof(CORBA::Boolean) );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method allows a single byte to be read from the byte array.
// @rdesc   The byte value.
CORBA::Octet CyrStByteItImpl::ReadByte()
{
    CORBA::Octet  oct=m_byteArray.front();
    m_byteArray.pop_front();
    return oct;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method adds a single byte to the byte array.
// @parm    The byte to be added.
void CyrStByteItImpl::WriteByte( CORBA::Octet data )
{
    m_byteArray.push_back( data );
}

