/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __IPMUTEX__
#  include "ipMutex.hpp"
#endif  // __IPMUTEX__


///////////////////////// CTIPMutex /////////////////////////

#ifdef _WIN32
CTIPMutex::CTIPMutex(const char* zName, unsigned long ulTimeOutMs)
    : bAutoName_(zName == NULL), hMutex_(NULL), dwMilliseconds_(ulTimeOutMs)
{
    hMutex_ = CreateMutex(NULL, FALSE, zName);
    if(hMutex_)
    {
        if(zName)
            strName_ = zName;
        else
        {
            // Mark this handle as inheritable and
            // generate the name as the handle.
            if(!SetHandleInformation(hMutex_, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT))
            {
                // Failure.
                CloseHandle(hMutex_);
                hMutex_ = NULL;
            }
            else
            {
                char zVar[32];
                wsprintf(zVar, "%ld", (long)hMutex_);
                strName_ = zVar;
            }
        }
    }
}

CTIPMutex::CTIPMutex(unsigned long ulTimeOutMs)
    : bAutoName_(true), hMutex_(NULL), dwMilliseconds_(ulTimeOutMs)
{
}

CTIPMutex::~CTIPMutex()
{
    if(hMutex_)
        CloseHandle(hMutex_);
}

bool CTIPMutex::attach(const char* zAutoName)
{
    bool bRet = false;
    if(bAutoName_ && zAutoName)
    {
        hMutex_ = (HANDLE)atol(zAutoName);
        bRet = true;
    }
    return bRet;
}

CtWaitResult CTIPMutex::lock()
{
    CtWaitResult wr;
    DWORD dwRet = WaitForSingleObject(hMutex_, dwMilliseconds_);
    switch(dwRet)
    {
        case WAIT_OBJECT_0:
        case WAIT_ABANDONED: wr = CT_WAIT_OK;      break;
        case WAIT_TIMEOUT:   wr = CT_WAIT_TIMEOUT; break;
        default:             wr = CT_WAIT_ERROR;   break;
    }
    return wr;
}

bool CTIPMutex::unlock()
{
    return ReleaseMutex(hMutex_) != FALSE;
}

bool CTIPMutex::isValid(void)
{
    return hMutex_ != NULL;
}


///////////////////////// CIPSignal /////////////////////////

CIPSignal::CIPSignal(const char* zName, unsigned long ulTimeOutMs)
    : bAutoName_(zName == NULL), hEvent_(NULL), dwMilliseconds_(ulTimeOutMs)
{
    hEvent_ = CreateEvent(NULL, TRUE, FALSE, zName);
    if(hEvent_)
    {
        if(zName)
            strName_ = zName;
        else
        {
            // Mark this handle as inheritable and
            // generate the name as the handle.
            if(!SetHandleInformation(hEvent_, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT))
            {
                // Failure.
                CloseHandle(hEvent_);
                hEvent_ = NULL;
            }
            else
            {
                char zVar[32];
                wsprintf(zVar, "%ld", (long)hEvent_);
                strName_ = zVar;
            }
        }
    }
}

CIPSignal::CIPSignal(unsigned long ulTimeOutMs)
    : bAutoName_(true), hEvent_(NULL), dwMilliseconds_(ulTimeOutMs)
{
}

CIPSignal::~CIPSignal()
{
    if(hEvent_)
        CloseHandle(hEvent_);
}

bool CIPSignal::attach(const char* zAutoName)
{
    bool bRet = false;
    if(bAutoName_ && zAutoName)
    {
        hEvent_ = (HANDLE)atol(zAutoName);
        bRet = true;
    }
    return bRet;
}


CtWaitResult CIPSignal::wait()
{
    CtWaitResult wr;
    DWORD dwRet = WaitForSingleObject(hEvent_, dwMilliseconds_);
    switch(dwRet)
    {
        case WAIT_OBJECT_0: wr = CT_WAIT_OK;      break;
        case WAIT_TIMEOUT:  wr = CT_WAIT_TIMEOUT; break;
        default:            wr = CT_WAIT_ERROR;   break;
    }
    return wr;
}

bool CIPSignal::signal()
{
    return SetEvent(hEvent_) != FALSE;
}

bool CIPSignal::isValid(void)
{
    return hEvent_ != NULL;
}

#elif defined( SUN )

///////////////////////// CTIPMutex /////////////////////////
CTIPMutex::CTIPMutex(const char* zName, unsigned long ulTimeOutMs)
{
}

CTIPMutex::CTIPMutex(unsigned long ulTimeOutMs)
{
}

CTIPMutex::~CTIPMutex()
{
}

bool CTIPMutex::attach(const char* zAutoName)
{
    return true;
}

CtWaitResult CTIPMutex::lock()
{
    return CT_WAIT_OK;
}

bool CTIPMutex::unlock()
{
    return true ;
}

bool CTIPMutex::isValid(void)
{
    return true;
}


///////////////////////// CIPSignal /////////////////////////

int CIPSignal::iShareMemoryId_ = 0 ;
CTMutex CIPSignal::signalMutex_ ;

CIPSignal::CIPSignal(const char* zName, unsigned long ulTimeOutMs)
:pCondition_(NULL), ipcFd_(0), iMilliseconds_(ulTimeOutMs)
{
	if (zName == NULL)
	{
		// Create a unique name for the memory map file
		char mapFileName[100] ;
		{
			CTLockGuard<CTMutex> guard(signalMutex_) ;
			iShareMemoryId_ ++ ;
			sprintf(mapFileName, "/tmp/%s%d", SHAREMEMORYNAME, iShareMemoryId_) ;
		}
		strName_ = mapFileName ;
		ipcFd_ = creat(mapFileName, O_CREAT|O_RDWR );
		pthread_cond_t condTemp ;
		memset(&condTemp, 0, sizeof(pthread_cond_t));
		write(ipcFd_, &condTemp, 2);
		close(ipcFd_);
		chmod(mapFileName, S_IRWXU|S_IRWXG|S_IRWXO);

		ipcFd_ = open(mapFileName, O_RDWR);
		pCondition_ = (pthread_cond_t *)mmap(NULL, sizeof(pthread_cond_t), PROT_READ|PROT_WRITE, MAP_SHARED, ipcFd_, 0);
		pthread_condattr_t    cv_attr;
		pthread_condattr_init(&cv_attr);     // initialize cv_attr with defaults
		pthread_condattr_setpshared(&cv_attr, PTHREAD_PROCESS_SHARED);
		if (pthread_cond_init(pCondition_, &cv_attr) != 0)
		{
			pCondition_ = NULL ;
		}
	}
	else
	{
		strName_ = zName ;
		while((ipcFd_ = open(zName, O_RDWR)) == -1)
			sleep(1);
		pCondition_ = (pthread_cond_t *)mmap(NULL, sizeof(pthread_cond_t), PROT_READ|PROT_WRITE, MAP_SHARED, ipcFd_, 0);

	}
}

CIPSignal::CIPSignal(unsigned long ulTimeOutMs)
:pCondition_(NULL), ipcFd_(0), iMilliseconds_(ulTimeOutMs)
{
}

CIPSignal::~CIPSignal()
{
	if (pCondition_ != NULL)
		pthread_cond_destroy (pCondition_) ;
	close(ipcFd_) ;
}

bool CIPSignal::attach(const char* zName)
{
	strName_ = zName ;
	while((ipcFd_ = open(zName, O_RDWR)) == -1)
		sleep(1);
	pCondition_ = (pthread_cond_t *)mmap(NULL, sizeof(pthread_cond_t), PROT_READ|PROT_WRITE, MAP_SHARED, ipcFd_, 0);
	if (pCondition_ != NULL)
		return true ;
	return false ;
}

CtWaitResult CIPSignal::wait()
{
	timespec tsp;
	timeval tv ;
	gettimeofday(&tv, NULL);
	tsp.tv_sec =  tv.tv_sec + iMilliseconds_ / 1000 ;
	tsp.tv_nsec = 0 ;
	CTMutexInfo *pMutexInfo = (CTMutexInfo *)condMutex_.getSyncObjectInfo() ;
	int iRet = pthread_cond_timedwait(pCondition_, pMutexInfo->hMutex_, &tsp);
	switch(iRet)
	{
		case 0		: return CT_WAIT_OK ;
		case ETIME 	: return CT_WAIT_TIMEOUT ;
	}
	printf("ERROR : %s\n", strerror(errno));
	return CT_WAIT_ERROR ;
}

bool CIPSignal::signal()
{
	return  (pthread_cond_signal(pCondition_) == 0) ;
}

bool CIPSignal::isValid(void)
{
    return pCondition_ != NULL;
}

#else

#error *** IPMutex classes not defined for this platform! ***

#endif
