/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_FILE_RTNS               */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains all the PSL routines associated with        */
/*  opening files.                               */
/*                                       */
/*===========================================================================*/

#include <psl_macros.h>

/* Standard ANSI headers */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef UNIX
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#endif  /* UNIX */

#ifdef VMS
#include <unixio.h>
#endif  /* VMS */

#ifdef WIN32
#include <errno.h>
#include <direct.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif  /* WIN32 */

#ifdef VMS
#pragma nostandard
#include <lib$routines.h>       /* Library Routines */
#include <starlet.h>            /* System services */
#pragma standard
#endif  /* VMS */

#include <psl_rtndef.h>         /* PSL Routines */


/*
 *  Local structures.
 *  structure for a single element in a filespec, typically a directory.
 */
#ifdef UNIX
struct elem
{
    int len;
    char *elp;
};
#endif /*UNIX*/

/* Function Prototypes for internal routines */

#if defined(UNIX) || defined(WIN32)
static void      BUILD_FILENAME (char *outbuf_ptr,
                                 const char *inpnam_ptr);
static Boolean       IS_DIR         (const char *filnam_ptr);

static unsigned long PARSE_FILESPEC (const char *input_ptr,
                                     char *path_ptr,
                                     char *file_ptr,
                                     char *extn_ptr);

#endif  /* UNIX & WIN32 */

#ifdef  UNIX
static int TOTAL_LEN    (struct elem *elp, int nelems);
static void FREE_ELEMS  (struct elem *elp, int nelems);
#endif  /* UNIX */

/*   Structure definitions and parameters  */

#include <pslfil_prmdef.h>      /* PSL_FILE_PARSE Symbols */
#include <shrmsg.h>         /* PSL Messages */

#ifdef  VMS
#pragma nostandard
#include <fab.h>            /* RMS FAB */
#include <nam.h>            /* RMS NAM */
#include <rmsdef.h>         /* RMS messages */
#include <descrip.h>            /* String Descriptors */
#pragma standard
#endif  /* VMS */

#include <pslerr_prmdef.h>      /* Error Trapping */
#include <psldiag_prmdef.h>     /* Diagnostics Constants */

/*
 *    Define the character user to separate components of a file path and the 
 *    file path from the file name
 */
#if defined(UNIX)
static char  path_seperator[2] = "/";       /* Path/filename separator  */
#elif defined WIN32
static char  path_seperator[2] = "\\";      /* Path/filename seperator  */
#endif  /* UNIX */

#ifdef  WIN32
#define MAXPATHLEN 256
#endif  /* WIN32 */




/************************************************************************/
/*                                  */
/*  PSL_CHK_FILENAME - Check file specification         */
/*                                  */
/*  This routine takes a file specification and attempts to check   */
/*  to see if it is valid.                      */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_CHK_FILENAME (filename_ptr);              */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret         = Unsigned Long PSL Routine Return Status      */
/*  filename_ptr = Pointer to file specification string     */
/*                                  */
/************************************************************************/
unsigned long   PSL_CHK_FILENAME (const char *filename_ptr)
{
    /*   Local Data */

    unsigned long       iret=PSL_NORMAL;    /* Return status */

#ifdef  VMS

    /*
     *   RMS Control Structures 
     */
    struct FAB      fab;        /* FAB              */
    struct NAM      nam;        /* NAM              */

    /*
     *   Initialize structures.
     */
    (void)memset (&fab ,'\0' ,sizeof (struct FAB));
    (void)memset (&nam ,'\0' ,sizeof (struct NAM));

    /*
     *   Set up RMS structures 
     */
    fab.fab$b_bid = FAB$C_BID;      /* Set up Block Id      */
    fab.fab$b_bln = FAB$C_BLN;      /* and Block length         */
    nam.nam$b_bid = NAM$C_BID;      /* Set up Block Id      */
    nam.nam$b_bln = NAM$C_BLN;      /* and Block length         */
    nam.nam$b_nop |= NAM$M_SYNCHK;  /* Set 'syntax check only' bit  */
    fab.fab$l_nam = &nam;       /* and hook into FAB        */

    /*
     *    Set up filename address and length in FAB 
     */

    fab.fab$l_fna = (char *) filename_ptr;
    /* Set up Filename address */
    fab.fab$b_fns = strlen(filename_ptr);
    /* and length */

    /*  Request PARSE */

    iret = sys$parse (&fab);        /* Perform PARSE */

    /*  END: */

    /*  Routine exit point - return the status to the caller. */

end:
#else   /* VMS */

    /*  Currently anything is valid as a UNIX or WIN32 file specification */

#endif  /* VMS */
    return iret;
}



#ifdef  VMS
/************************************************************************/
/*                                  */
/*  PSL_FILE_PARSE - Parse file specification (VMS Version)     */
/*                                  */
/*  This routine takes a filespec, a default filespec and a     */
/*  related filespec and returns a fully qualified filespec.    */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_FILE_PARSE (inp_file_ptr, def_file_ptr,       */
/*                rel_file_ptr, rel_extn_ptr,       */
/*                exp_file_ptr, max_exp_len, flags);    */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret         = Unsigned Long PSL Routine Return Status      */
/*  inp_file_ptr = Pointer to input file specification string   */
/*  def_file_ptr = Pointer to default file specification string */
/*  rel_file_ptr = Pointer to related file specification string */
/*  rel_extn_ptr = Pointer to expected related file extension   */
/*          (UNIX only)                 */
/*  exp_file_ptr = Pointer to expanded file specification string    */
/*  max_exp_len  = Unsigned long maximum length of expanded file    */
/*             in the range 1-256 (VMS) or 1-MAXPATHLEN (UNIX)  */
/*  flags        = Unsigned Long Flags (as defined in PSLFIL_PRMDEF)*/
/*      FILPRS_RELDIR => Use related device/directory       */
/*      FILPRS_OBSOLETE => do NOT set Syntax Check only flag    */
/*             (VMS only)                   */
/*      FILPRS_INPUT => Perform input file parsing (VMS only)   */
/*                                  */
/*    This version of PSL_FILE_PARSE is for VMS ONLY            */
/*                                  */
/************************************************************************/

unsigned long   PSL_FILE_PARSE (const char *inp_file_ptr,
                                const char *def_file_ptr,
                                const char *rel_file_ptr,
                                const char *rel_extn_ptr,
                                char *exp_file_ptr,
                                unsigned long max_exp_len,
                                unsigned long flags)

{
    /*   Local Data */

    unsigned long       iret=PSL_NORMAL;    /* Return status */

    /*
     *   RMS Control Structures 
     */

    /*
     *   RMS Control Structures 
     */
    struct FAB      fab;        /* FAB              */
    struct NAM      nam;        /* NAM              */
    struct NAM      relnam;     /* Related NAM          */

    /*
     *   Initialize structures.
     */
    (void)memset (&fab ,'\0' ,sizeof (struct FAB));
    (void)memset (&nam ,'\0' ,sizeof (struct NAM));
    (void)memset (&relnam ,'\0' ,sizeof (struct NAM));

    /*
     *   Set up RMS structures 
     */
    fab.fab$b_bid = FAB$C_BID;      /* Set up Block Id      */
    fab.fab$b_bln = FAB$C_BLN;      /* and Block length         */
    nam.nam$b_bid = NAM$C_BID;      /* Set up Block Id      */
    nam.nam$b_bln = NAM$C_BLN;      /* and Block length         */
    nam.nam$b_nop |= NAM$M_PWD;     /* Preserve DECnet Passwords    */
    fab.fab$l_nam = &nam;       /* and hook into FAB        */

    relnam.nam$b_bid = NAM$C_BID;   /* Set up Block Id      */
    relnam.nam$b_bln = NAM$C_BLN;   /* and Block length         */
    nam.nam$l_rlf = &relnam;        /* and hook into NAM block  */

    /*
     *  Set up the Expanded File Pointers 
     */
    if  (max_exp_len > 256)     /* If buffer > 256 characters */
    {
        nam.nam$b_ess = 255;        /* set maximum valid size */
    }
    else                /* Else (buffer <= 256 chars) */
    {
        nam.nam$b_ess = max_exp_len - 1;
        /* set to buffer size */
    }
    nam.nam$l_esa = exp_file_ptr;   /* And address of buffer */

    /*  Set up filename address and length in FAB */

    fab.fab$l_fna = (char *) inp_file_ptr;
    /* Set up Filename address */
    fab.fab$b_fns = strlen(inp_file_ptr);
    /* and length */
    nam.nam$b_esl = 0;      /* and initialize expanded length */

    /*  Set up default filespec address and length in FAB */

    fab.fab$l_dna = (char *) def_file_ptr;
    /* Set up Default Filename address */
    if  (def_file_ptr == NULL)      /* If no default file */
    {
        fab.fab$b_dns = 0;  /* Set length to zero */
    }
    else                /* Else */
    {
        fab.fab$b_dns = strlen(def_file_ptr);
        /* Set to length of string */
    }

    /*  Set up related filespec address and length in RELNAM */

    relnam.nam$l_rsa = (char *) rel_file_ptr;
    /* Set up Default Filename address */
    if  (rel_file_ptr == NULL)      /* If no default file */
    {
        relnam.nam$b_rsl = 0;   /* Set length to zero */
    }
    else                /* Else */
    {
        relnam.nam$b_rsl = strlen(rel_file_ptr);
        /* Set to length of string */
    }



    /*  See if we want to take note of the device/directory on the      */
    /*  related file specification and set/clear the OFP bit        */
    /*  accordingly.                            */

    if  ((flags & FILPRS_RELDIR) != 0)  /* If flag set */
    {
        fab.fab$l_fop &= ~FAB$M_OFP;
        /* clear the bit */
    }
    else                /* Else (flag not set) */
    {
        fab.fab$l_fop |= FAB$M_OFP;
        /* set the bit */
    }



    /*  By default, for output files, we only do a syntax check.        */
    /*  However, for upwards compatibility we must also support a mode  */
    /*  that tries to check that the file could exist.  For input files */
    /*  we must always look for the file itself as that's the only way  */
    /*  to handle search lists.                     */

    if  ((flags & (FILPRS_OBSOLETE | FILPRS_INPUT)) != 0)
        /* If input file or obsolete */
        /* flag set */
    {
        nam.nam$b_nop &= ~NAM$M_SYNCHK;
        /* Clear "syntax only" bit */
    }
    else                /* Else (normal output file) */
    {
        nam.nam$b_nop |= NAM$M_SYNCHK;
        /* Set the bit */
    }


    /*  Request PARSE */

    iret = sys$parse (&fab);        /* Perform PARSE */
    if  (!(psl_test_status(iret))) goto end;
    /* Exit if Error Returned */



    /*  If we are parsing an input file, we need to look for the file   */
    /*  itself to avoid problems with Search Lists.             */

    if  ((flags & FILPRS_INPUT) != 0)   /* If an input file */
    {
        iret = sys$search (&fab);   /* Perform SEARCH */
        if  (!(psl_test_status(iret))) goto end;
        /* Exit if Error Returned */
    }


    /*  END: */

    /*  Routine exit point - Null Terminate the Expanded Filespec, and  */
    /*  return the status to the caller.  Note that, if the error       */
    /*  status is PSL_INSVIRMEM then we never got to RMS at all so we   */
    /*  must sort out the filename by hand.  Note also that $PARSE may  */
    /*  give up before it sets up the expanded filespec.            */

end:
    if  ((iret != PSL_INSVIRMEM)    /* If not memory failure */
            && (nam.nam$b_esl != 0))  /* and we got an expanded filespec */
    {
        *(exp_file_ptr + nam.nam$b_esl) = '\0';
        /* Just null-terminate the string */
    }
    else                /* Else */
    {
        (void) strncpy (exp_file_ptr, inp_file_ptr, max_exp_len - 1);
        /* Copy over input file */
        *(exp_file_ptr + max_exp_len - 1) = '\0';
        /* And ensure null-termination */
    }

    /*
     *    A status of RMS_FNF is half expected here as for input files we may not
     *    have located the file. We should not return an expected status as it is
     *    not really an error. Instead we will return PSL_NORMAL and pretend that
     *    there was no problem. We will still have the expanded file spec anyway
     *    which is what we are really interested in.
     */
    if (iret == RMS$_FNF)       /* if status = File-Not-Found   */
    {
        iret = PSL_NORMAL;      /* ...pretend all o.k. really   */
    }


    return iret;
}
#endif  /* VMS */


#if defined(UNIX) || defined(WIN32)
/************************************************************************/
/*                                  */
/*  PSL_FILE_PARSE - Parse file specification (UNIX/WIN32 version)  */
/*                                  */
/*  This routine takes a filespec, a default filespec and a     */
/*  related filespec and returns a fully qualified filespec.    */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_FILE_PARSE (inp_file_ptr, def_file_ptr,       */
/*                rel_file_ptr, rel_extn_ptr,       */
/*                exp_file_ptr, max_exp_len, flags);    */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret         = Unsigned Long PSL Routine Return Status      */
/*  inp_file_ptr = Pointer to input file specification string   */
/*  def_file_ptr = Pointer to default file specification string */
/*  rel_file_ptr = Pointer to related file specification string */
/*  rel_extn_ptr = Pointer to expected related file extension   */
/*          (UNIX only)                 */
/*  exp_file_ptr = Pointer to expanded file specification string    */
/*  max_exp_len  = Unsigned long maximum length of expanded file    */
/*             in the range 1-256 (VMS) or 1-MAXPATHLEN (UNIX)  */
/*  flags        = Unsigned Long Flags (as defined in PSLFIL_PRMDEF)*/
/*      FILPRS_RELDIR => Use related device/directory       */
/*      FILPRS_OBSOLETE => do NOT set Syntax Check only flag    */
/*             (VMS only)                   */
/*      FILPRS_INPUT => Perform input file parsing (VMS only)   */
/*                                  */
/*    This version of PSL_FILE_PARSE is for UNIX and WIN32 platforms    */
/*                                  */
/************************************************************************/

unsigned long   PSL_FILE_PARSE (const    char *inp_file_ptr,
                                const    char *def_file_ptr,
                                const    char *rel_file_ptr,
                                const    char *rel_extn_ptr,
                                char *exp_file_ptr,
                                unsigned long max_exp_len,
                                unsigned long flags)

{
    /*   Local Data */

    static   char       def_path[MAXPATHLEN]; /* Default file pathname */
    static   char       rel_path[MAXPATHLEN]; /* Related file pathname */
    static   char       cur_path[MAXPATHLEN]; /* Current file pathname */
    static   char       def_file[MAXPATHLEN]; /* Default file name */
    static   char       rel_file[MAXPATHLEN]; /* Related file name */
    static   char       def_extn[MAXPATHLEN]; /* Default file extension */
    static   char       rel_extn[MAXPATHLEN]; /* Related file extension */
    char       *segment_ptr;   /* Pointer to the segment to move */
    unsigned long       out_exp_length; /* Output Expanded length */
    unsigned long       iret;       /* Return status */
    long       comp_status;    /* Comparison status */
#ifdef  WIN32
    char       *chr_ptr;
#endif  /* WIN32 */




    /*  If the expanded filename length is greater than our maximum     */
    /*  then shorten it.  In all cases ensure there is room for a       */
    /*  trailing null (as some of the code below seems to assume it).   */

    if (max_exp_len > MAXPATHLEN)
    {
        out_exp_length = MAXPATHLEN - 1;
    }
    else
    {
        out_exp_length = max_exp_len - 1;
    }


    /*  Validate the input parameters */

    if ((exp_file_ptr == NULL)      /* If no pointer to return file */
            ||  (out_exp_length == 0 ))     /* ...or returned length max=0 */
    {
        return PSL_BADPARAM;        /* Return "Bad parameter" */
    }

    /*  If the input file has been explicitly specified, and is not     */
    /*  just a directory, then we use it without change - however, we   */
    /*  need to expand it to an absolute pathname.              */

    if ((inp_file_ptr != NULL)      /* If input file specified */
            && (strlen (inp_file_ptr) > 0) /* And not null string */
            && (!IS_DIR (inp_file_ptr)))   /* And not a directory */
    {
#ifdef  WIN32
        chr_ptr = _fullpath (exp_file_ptr ,inp_file_ptr ,out_exp_length);
        if (chr_ptr == NULL)
        {
            iret = PSL_FILNOTFOU;
        }
        else
        {
            iret = PSL_NORMAL;
        }
#else   /* WIN32 */
        iret = PSL_REALPATH (       /* Convert to absolute path */
                   (char *) inp_file_ptr,
                   /* Input file */
                   exp_file_ptr,   /* Expanded file */
                   out_exp_length, /* Maximum expanded file length */
                   PFALSE);    /* Do not resolve links */
#endif  /* WIN32 */
        return iret;            /* Return error status */
    }

    /*  We must now chop up the default and related file specifications */
    /*  to see what we have in each.                    */

    iret = PARSE_FILESPEC (def_file_ptr, def_path, def_file, def_extn);
    if (!(psl_test_status(iret))) return iret;
    iret = PARSE_FILESPEC (rel_file_ptr, rel_path, rel_file, rel_extn);
    if (!(psl_test_status(iret))) return iret;

    /*  Get the current path name */

    (void) getcwd (cur_path, sizeof (cur_path));
    (void) strcat (cur_path, path_seperator);

    /*  Now concatenate the bits of these filespecs togethor to give    */
    /*  the expanded file specification, remembering that we must not   */
    /*  exceed the maximum length.                      */

    if ((inp_file_ptr != NULL)      /* If input file specified */
            && (strlen (inp_file_ptr) > 0))    /* And not null string */
    {                   /* Then it must be a directory */
        segment_ptr = (char *) inp_file_ptr;
        /* Segment to move */
    }
    else if (def_path[0] != '\0')   /* If default path specified */
    {
        segment_ptr = def_path;     /* Segment to move */
    }
    else if (((flags & FILPRS_RELDIR) != 0)
             /* If using related pathname */
             && (rel_path[0] != '\0')) /* And related path specified */

    {
        segment_ptr = rel_path;     /* Segment to move */
    }
    else                /* Else use current path */
    {
        segment_ptr = cur_path;     /* Segment to move */
    }

    /*  Convert path segment to absolute path */

    if (segment_ptr != NULL)
    {
#ifdef  WIN32
        chr_ptr = _fullpath (exp_file_ptr ,segment_ptr ,out_exp_length);
        if (chr_ptr == NULL) return PSL_FILNOTFOU;
#else   /* WIN32 */
        iret = PSL_REALPATH (       /* Convert to absolute path */
                   segment_ptr,    /* Input file */
                   exp_file_ptr,   /* Expanded file */
                   out_exp_length, /* Maximum expanded file length */
                   PFALSE);    /* Do not resolve links */
        if (!(psl_test_status(iret))) return iret;
#endif  /* WIN32 */
        segment_ptr = exp_file_ptr + strlen(exp_file_ptr) - 1;
        /* Point to last character */
        if (*segment_ptr != path_seperator[0])  /* If not '/' */
        {
            (void) strcat (exp_file_ptr, path_seperator);
            /* Append '/' */
        }
        segment_ptr = NULL;     /* Reset the segment pointer */
    }
    else
    {

        /*  Reset the current length of the expanded file specification to zero */

        *exp_file_ptr = '\0';       /* Initialize as null string */
    }

    /*  Now do the same for the file segment */

    if  (def_file[0] != '\0')
    {
        segment_ptr = def_file;     /* Segment to move */
    }
    else if (rel_file[0] != '\0')
    {
        segment_ptr = rel_file;     /* Segment to move */

        /*  If we are using the related filename, append the related file   */
        /*  extension unless it is the expected one.                */

        if (rel_extn_ptr != NULL)   /* If expected extn specified */
        {
            comp_status = strcmp (rel_extn, rel_extn_ptr);
            /* Compare actual with expected */
        }
        else
        {
            comp_status = 1;        /* Show actual not = expected */
        }
        if (comp_status != 0)       /* If not the expected extension */
        {
            (void) strcat (rel_file, rel_extn);
            /* Put it back on */
        }
    }

    /*  Check we have enough room for this segment in the expanded file */
    /*  string and if we have move it in.                   */

    if (segment_ptr != NULL)
    {
        if ((strlen (exp_file_ptr) + strlen (segment_ptr)) > out_exp_length)
        {
            return PSL_BADPARAM;
        }
        (void) strcat (exp_file_ptr, segment_ptr);
        segment_ptr = NULL;     /* Reset the segment pointer */
    }

    /*  Now do the file extension */

    if (def_extn[0] != '\0')
    {
        segment_ptr = def_extn;     /* Segment to move */
    }

    /*  Check we have enough room for this segment in the expanded file */
    /*  string and if we have move it in.                   */

    if (segment_ptr != NULL)
    {
        if ((strlen (exp_file_ptr) + strlen (segment_ptr)) > out_exp_length)
        {
            return PSL_BADPARAM;
        }
        (void) strcat (exp_file_ptr, segment_ptr);
    }

    /*  Return with success */

    return PSL_NORMAL;
}
#endif  /* UNIX & WIN32 */

#if defined(UNIX) || defined(WIN32)
/************************************************************************/
/*                                  */
/*  PARSE_FILESPEC - Parse file specification (UNIX and WIN32 only) */
/*                                  */
/*  This routine takes a filespec string and breaks it down     */
/*  into its component parts.                   */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PARSE_FILESPEC (input_ptr, path_ptr, file_ptr,    */
/*                extn_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*  iret      = Unsigned Long PSL Routine Return Status     */
/*  input_ptr = Pointer to input file specification string      */
/*  path_ptr  = Pointer to string to hold returned pathname     */
/*  file_ptr  = Pointer to string to hold returned filename     */
/*  extn_ptr  = Pointer to string to hold returned extension    */
/*                                  */
/*  Note that the buffers pointed to by the last three pointers */
/*  MUST be MAXPATHLEN characters in length.            */
/*                                  */
/************************************************************************/

static unsigned long PARSE_FILESPEC (const char *input_ptr,
                                     char *path_ptr,
                                     char *file_ptr,
                                     char *extn_ptr)
{
    /*    Local data */

    char       *inp_end_ptr;   /* Pointer to end of input name */
    char       *inp_path_ptr;  /* Pointer to the path name */
    char       *inp_extn_ptr;  /* Pointer to the extension */
    unsigned long       token_length;   /* Length of token */

    /*  If we have been passed a NULL pointer then set all of the       */
    /*  components to be NULL pointers and return.              */

    if (input_ptr == NULL)
    {
        *path_ptr = '\0';
        *file_ptr = '\0';
        *extn_ptr = '\0';
        return PSL_NORMAL;
    }

    /*  Find the address of the byte after the end of each filename */

    inp_end_ptr = ((char *) input_ptr + strlen (input_ptr));


    /*  Identify the possible three 3 sections of the input file        */
    /*  specification The device/directory is identified by the last /  */
    /*  in the filename. The filename is between the backslash and the  */
    /*  last '.', and anything after that is the filename.          */

    inp_path_ptr = strrchr (input_ptr, path_seperator[0]);
    /* Locate end of path in input file */

    if (inp_path_ptr == NULL)
    {
        *path_ptr = '\0';       /* Show no path in this file */
        inp_path_ptr = (char *) input_ptr;/* Reset segment pointer to start */

    }
    else
    {
        token_length = inp_path_ptr - input_ptr + 1;
        /* Get length of token */
        if (token_length > (MAXPATHLEN-1))
        {
            return PSL_BADPARAM;
        }
        (void) memcpy (path_ptr, input_ptr, token_length);
        *(path_ptr + token_length) = '\0';
        inp_path_ptr += 1;      /* Consume the backslash from name */
    }

    /*  Now locate a '.' to define that start of the extension. All     */
    /*  between the current segment pointer and the '.' is the file     */
    /*  name. All after is the extension.                   */

    inp_extn_ptr = strrchr (inp_path_ptr, '.');
    /* Locate extension in input file */
    if (inp_extn_ptr == NULL)
    {
        *extn_ptr = '\0';       /* Show no extension in this file */
        inp_extn_ptr = inp_end_ptr;     /* Set pointer to end */
    }
    else
    {
        token_length = inp_end_ptr - inp_extn_ptr;
        /* Calc length of extension */
        if (token_length > (MAXPATHLEN-1))
        {
            return PSL_BADPARAM;
        }
        (void) memcpy (extn_ptr, inp_extn_ptr, token_length);
        *(extn_ptr + token_length) = '\0';
        /* Copy extensiona and null terminate */
    }

    /*  Finally move the file name */

    token_length = inp_extn_ptr - inp_path_ptr;
    /* Get length of token */
    if (token_length > (MAXPATHLEN-1))
    {
        return PSL_BADPARAM;
    }
    (void) memcpy (file_ptr, inp_path_ptr, token_length);
    *(file_ptr + token_length) = '\0'; /* Copy filename and null terminate */

    return PSL_NORMAL;
}
#endif  /* UNIX & WIN32 */

#if defined(UNIX) || defined(WIN32)
/************************************************************************/
/*                                  */
/*    IS_DIR - Is the specified file a directory file           */
/*    ======                                */
/*                                  */
/*    Calling Format:                           */
/*  iret = IS_DIR (filnam_ptr);                 */
/*                                  */
/*    Inputs:                               */
/*  filnam_ptr  - Pointer to null terminated filename       */
/*                                  */
/*    Return Status                             */
/*  PTRUE if file is a directory                    */
/*  PFALSE otherwise.                       */
/*                                  */
/*    This routine is called to determine whether the specified     */
/*    filespec is a directory or not.                   */
/*                                  */
/*    This routine is applicable to UNIX and WIN32 platforms only   */
/*                                  */
/************************************************************************/

static Boolean IS_DIR (const char *filnam_ptr)
{
    /*    Local data */

    struct   stat   file_stat;      /* File status structure */
    int    statsts;        /* Status from stat() */

    /*  We use the stat() function to get information on this file.     */
    /*  stat() will fail if the file does not exist or if the directory */
    /*  this file is in is not searchable - other errors are        */
    /*  unexpected.                             */

    statsts = stat(filnam_ptr, &file_stat);

    /*  Assume file is not a directory if stat fails */

    if (statsts != 0)           /* Not successful */
    {
        return PFALSE;
    }

    /*
     *    Check to see if the current file is in fact a directory and return 
     *    PTRUE if it is.
     */
#ifdef  WIN32
    if ((file_stat.st_mode & S_IFDIR) == 0)
#else
    if (S_ISDIR(file_stat.st_mode) == 0)
#endif
    {
        return PFALSE;
    }
    else
    {
        return PTRUE;
    }

}
#endif      /* UNIX & WIN32 */



/************************************************************************/
/*                                  */
/*  PSL_GET_FDNAME - Return full expanded file specification    */
/*           for a file descriptor.             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_GET_FDNAME (file_dsc, exp_file_ptr, max_exp_len);    */
/*                                  */
/*  Where:                              */
/*                                  */
/*         file_dsc     = Unsigned long file descritor for file     */
/*     exp_file_ptr = Pointer to expanded file specification string */
/*     max_exp_len  = Unsigned long maximum length of expanded file */
/*                in the range 1-256 (VMS) or 1-MAXPATHLEN (UNIX)*/
/*                                  */
/*  This routine will attempt to return the full expanded file  */
/*  specification for a file.  Note that the file must currently    */
/*  be open and that exp_file_ptr must point to the expanded file   */
/*  specification set up by PSL_FILE_PARSE.                 */
/*                                  */
/************************************************************************/
void    PSL_GET_FDNAME (unsigned long file_dsc,
                        char *exp_file_ptr,
                        unsigned long max_exp_len)
{
#ifdef  VMS
    /*  Local data */

    static  char        file_name[256]; /* Local Buffer for file name */

    /*  On VMS just call the GETNAME routine and then copy the result   */
    /*  into the user buffer.                       */

    getname (file_dsc, file_name, 1);   /* Get file name in VMS format */
    (void) strncpy (exp_file_ptr, file_name, max_exp_len - 1);
    /* Copy over file name */
    *(exp_file_ptr + max_exp_len - 1) = '\0';
    /* And ensure null-termination */
#else   /* VMS */

    /*  On UNIX just leave the PSL_FILE_PARSE value unchanged */

#endif  /* VMS */
    return;             /* And return */
}



/************************************************************************/
/*                                  */
/*  PSL_GET_FILENAME - Return full expanded file specification  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_GET_FILENAME (file_ptr, exp_file_ptr, max_exp_len);  */
/*                                  */
/*  Where:                              */
/*                                  */
/*         file_ptr     = Pointer to FILE structure for file        */
/*     exp_file_ptr = Pointer to expanded file specification string */
/*     max_exp_len  = Unsigned long maximum length of expanded file */
/*                in the range 1-256 (VMS) or 1-MAXPATHLEN (UNIX)*/
/*                                  */
/*  This routine will attempt to return the full expanded file  */
/*  specification for a file.  Note that the file must currently    */
/*  be open and that exp_file_ptr must point to the expanded file   */
/*  specification set up by PSL_FILE_PARSE.                 */
/*                                  */
/************************************************************************/
void    PSL_GET_FILENAME (FILE *file_ptr,
                          char *exp_file_ptr,
                          unsigned long max_exp_len)
{
#ifdef  VMS
    /*  Local data */

    static  char        file_name[256]; /* Local Buffer for file name */

    /*  On VMS just call the FGETNAME routine and then copy the result  */
    /*  into the user buffer.                       */

    fgetname (file_ptr, file_name, 1);  /* Get file name in VMS format */
    (void) strncpy (exp_file_ptr, file_name, max_exp_len - 1);
    /* Copy over file name */
    *(exp_file_ptr + max_exp_len - 1) = '\0';
    /* And ensure null-termination */
#else   /* VMS */

    /*  On UNIX just leave the PSL_FILE_PARSE value unchanged */

#endif  /* VMS */
    return;             /* And return */
}


#ifndef VMS

/************************************************************************/
/*                                  */
/*  PSL_RESTOREFILE - Restore original file saved by PSL_SAVEFILE   */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_RESTOREFILE (diag_ptr, filename_ptr)      */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret         = Unsigned Long PSL Routine Return Status   */
/*     diag_ptr     = Pointer to Diagnostics Stack          */
/*     filename_ptr = Pointer to File to be saved               */
/*                                  */
/*  This routine will restore a file that was renamed by a previous */
/*  call to PSL_SAVEFILE.  Note that it is the responsibility of    */
/*  the caller to ensure that such renaming took place.         */
/*                                  */
/*  WINDOWS32 and UNIX version                  */
/*                                  */
/************************************************************************/
unsigned long PSL_RESTOREFILE (struct diagstk *diag_ptr,
                               const char *filename_ptr)
{
    /*  Local Data */

    unsigned long    iret;          /* Routine Return Status */
    unsigned long    start_diag;        /* Start of Diagnostics Stack */
    long    status;        /* Status from UNIX functions */
    char    newfile_name[MAXPATHLEN];
    /* Renamed File Name */

    /*  Maintain the diagnostics stack */

    start_diag = PSL_ADD_STATDIAG (diag_ptr, "During PSL_RESTOREFILE");

    /*  Construct the name of the renamed file and rename it back */

    BUILD_FILENAME (newfile_name, filename_ptr);
    /* Construct the file name */
    do                  /* Loop indefinitely */
    {
        status = rename (newfile_name, filename_ptr);
        /* rename the file back */
    }
    while (status == EINTR);        /* Until it is NOT interrupted */

    /*  Any errors? */

    if (status != 0)
    {
        iret = PSL_RENAMEERR;
        PSL_LOG_CIOERR(diag_ptr, iret, newfile_name);

        goto end;
    }

    /*  Unwind the diagnostics stack */

    PSL_REMOVE_DIAGNOSTICS (diag_ptr, start_diag);

    /*  Return success */

    iret = PSL_NORMAL;

    /*  END: */

    /*  At this point IRET will contain the final return status */

end:
    return iret;
}
#endif  /* NON-VMS */


#ifndef VMS
/************************************************************************/
/*                                  */
/*  PSL_SAVEFILE - Save original file if to be overwritten          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_SAVEFILE (diag_ptr, filename_ptr)         */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret         = Unsigned Long PSL Routine Return Status   */
/*     diag_ptr     = Pointer to Diagnostics Stack          */
/*     filename_ptr = Pointer to File to be saved               */
/*                                  */
/*  This routine will rename any existing file if the specified     */
/*  file already exists.                                        */
/*                                  */
/*  This routine is not applicable to VMS.              */
/*                                  */
/************************************************************************/
unsigned long   PSL_SAVEFILE (struct diagstk *diag_ptr,
                              const char *filename_ptr)
{
    /*  Local Data */

    unsigned long    iret;          /* Routine Return Status */
    unsigned long    start_diag;        /* Start of Diagnostics Stack */
    long    status;        /* Status from UNIX functions */
    struct   stat    stat_buf;      /* stat() returned buffer */
    char    newfile_name[MAXPATHLEN];
    /* New File Name */

    /*  Maintain the diagnostics stack */

    start_diag = PSL_ADD_STATDIAG (diag_ptr, "During PSL_SAVEFILE");

    /*  Determine if the output file already exists... */

    status = stat (filename_ptr, &stat_buf);
    if (status == 0)            /* File exists - rename it */
    {


        /*  Construct the name of the renamed file and rename it */

        BUILD_FILENAME (newfile_name, filename_ptr);
        /* Construct the file name */
        do              /* Loop indefinitely */
        {
            status = rename (filename_ptr, newfile_name);
            /* renaming the file */
        }
        while (status == EINTR);    /* Until it is NOT interrupted */


        /*  Any errors? */

        if (status != 0)
        {
            iret = PSL_RENAMEERR;
            PSL_LOG_CIOERR(diag_ptr, iret, filename_ptr);

            goto end;
        }
    }

    /*  Unwind the diagnostics stack */

    PSL_REMOVE_DIAGNOSTICS (diag_ptr, start_diag);

    /*  Return success */

    iret = PSL_NORMAL;

    /*  END: */

    /*  At this point IRET will contain the final return status */

end:
    return iret;
}
#endif  /* NON-VMS */

#ifndef VMS

/************************************************************************/
/*                                  */
/*  BUILD_FILENAME - Build filename for renamed file            */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     BUILD_FILENAME (outbuf_ptr, inpnam_ptr);         */
/*                                  */
/*  Where:                              */
/*                                  */
/*     inpnam_ptr = Pointer to input file name                  */
/*     outbuf_ptr = Pointer to buffer to hold output file name. */
/*          This must be MAXPATHLEN bytes long.     */
/*                                  */
/*  This routine is called by PSL_SAVEFILE and PSL_RESTOREFILE. */
/*                                  */
/************************************************************************/
static void BUILD_FILENAME (char *outbuf_ptr,
                            const char *inpnam_ptr)
{
    /*  Local Data */

    long    name_len;      /* Length of file name */

    /*  We try to rename the file from "<file>" to "<file>~".  This     */
    /*  could conceivably be impossible if the input file name is too   */
    /*  large so we check for that and truncate the name if so.     */

    name_len = strlen(inpnam_ptr);  /* Get length of file name */
    if  (name_len >= (MAXPATHLEN-1)) name_len = MAXPATHLEN - 2;
    /* Truncate if too large */
    (void) memcpy (outbuf_ptr, inpnam_ptr, name_len);
    /* Copy over file name */
    *(outbuf_ptr + name_len) = '~'; /* And add '~' at end */
    *(outbuf_ptr + name_len + 1) = '\0';

    return;             /* Return to caller */
}

#endif  /* NON-VMS */



#ifdef VMS
/************************************************************************/
/*                                  */
/*  PSL_TRIM_FILESPEC - Trime file specification            */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_TRIM_FILESPEC (inp_file_ptr, out_file_ptr,    */
/*                  max_out_len, flags);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*      iret         = Unsigned long PSL routine return status  */
/*      inp_file_ptr = Pointer to buffer containing input filespec  */
/*      out_file_ptr = Pointer to buffer to hold trimmed filespec   */
/*      max_out_len  = Maximum length thereof           */
/*      flags        = Control flags (currently none - provided */
/*             for possible future use)         */
/*                                  */
/*  This routine trims a file specification in a `sensible' way */
/*  so that it fits into the specified output buffer.       */
/*  On VMS we call the LIB$TRIM_FILESPEC routine.           */
/*                                  */
/*  This version of the routine is for VMS platforms        */
/*                                  */
/************************************************************************/

unsigned long PSL_TRIM_FILESPEC (char *inp_file_ptr,
                                 char *out_file_ptr,
                                 unsigned long max_out_len,
                                 unsigned long flags)
{
    /*
     *  Local Data.
     */
    unsigned long       iret = PSL_NORMAL;  /* Status Return */
    unsigned short      maxlen;     /* Maximum length of output */
    char       *wrk_ptr;       /* Work pointer */
    unsigned short      retlen;     /* Returned length of output */
    struct   dsc$descriptor_s strdsc1;  /* String Descriptor */
    struct   dsc$descriptor_s strdsc2;  /* String Descriptor */

    /*  Set up string descriptor for input filespec.            */

    strdsc1.dsc$b_class   = DSC$K_CLASS_S;
    strdsc1.dsc$b_dtype   = DSC$K_DTYPE_T;
    strdsc1.dsc$a_pointer = inp_file_ptr;
    strdsc1.dsc$w_length  = strlen (strdsc1.dsc$a_pointer);

    /*  Set up string descriptor for output buffer.             */

    maxlen = (unsigned short) max_out_len - 1;
    /* Leave room to null terminate */
    strdsc2.dsc$b_class   = DSC$K_CLASS_S;
    strdsc2.dsc$b_dtype   = DSC$K_DTYPE_T;
    strdsc2.dsc$a_pointer = out_file_ptr;
    strdsc2.dsc$w_length  = maxlen;

    /*  Set up trimmed copy of filespec.                    */

    iret = lib$trim_filespec (      /* Trim filespec */
               &strdsc1,       /* Trim this filespec */
               &strdsc2,       /* Put result here */
               &maxlen,        /* Leave room to null terminate */
               &retlen);       /* Returned length */

    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    *(out_file_ptr + retlen) = '\0';    /* Null terminate */
    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    ;
    return iret;
}
#endif /* VMS */


#ifdef  WIN32
/************************************************************************/
/*                                  */
/*  PSL_TRIM_FILESPEC - Trime file specification            */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_TRIM_FILESPEC (inp_file_ptr, out_file_ptr,    */
/*                  max_out_len, flags);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*      iret         = Unsigned long PSL routine return status  */
/*      inp_file_ptr = Pointer to buffer containing input filespec  */
/*      out_file_ptr = Pointer to buffer to hold trimmed filespec   */
/*      max_out_len  = Maximum length thereof           */
/*      flags        = Control flags (currently none - provided */
/*             for possible future use)         */
/*                                  */
/*  This routine trims a file specification in a `sensible' way */
/*  so that it fits into the specified output buffer.       */
/*                                  */
/*  This version of the routine is for WINDOWS 32 platforms     */
/*                                  */
/************************************************************************/

unsigned long PSL_TRIM_FILESPEC (char *inp_file_ptr,
                                 char *out_file_ptr,
                                 unsigned long max_out_len,
                                 unsigned long flags)
{
    /*
     *  Local Data.
     */
    unsigned long       maxlen;     /* Maximum length of output */
    char       *wrk_ptr;       /* Work pointer */

    /*  Check to see if the full filespec will fit into the output buffer.  */
    /*  If not, strip off the directory.                    */

    wrk_ptr = inp_file_ptr;     /* Point to string */
    maxlen = strlen (wrk_ptr);      /* Get its length */
    if (maxlen > (max_out_len - 1)) /* If it won't fit */
    {

        /*  Check to see if the filename will fit into the output buffer.   */
        /*  If not, truncate it.                        */

        wrk_ptr = strrchr (inp_file_ptr, '\\');
        /* Point to end of directory */
        if (wrk_ptr == NULL)        /* If there isn't one */
        {
            wrk_ptr = inp_file_ptr; /* Reset to full filespec */
        }

        maxlen = strlen (wrk_ptr);  /* Get its length */
        if (maxlen > (max_out_len - 1)) /* If it won't fit */
        {
            maxlen = max_out_len - 1;   /* Truncate so it will */
        }
    }

    (void) memcpy (out_file_ptr, wrk_ptr, maxlen);
    /* Copy string */
    *(out_file_ptr + maxlen) = '\0';    /* Null terminate */

    return PSL_NORMAL;
}
#endif /* WIN32 */


#ifdef UNIX

/************************************************************************/
/*                                  */
/*  PSL_TRIM_FILESPEC - Trim file specification         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_TRIM_FILESPEC (inp_file_ptr, out_file_ptr,    */
/*                  max_out_len, flags);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*      iret         = Unsigned long PSL routine return status  */
/*      ssp          = Pointer to buffer containing input filespec  */
/*      ddp          = Pointer to buffer to hold trimmed filespec   */
/*      dlen         = Maximum length thereof           */
/*      flags        = Control flags (currently none - provided */
/*             for possible future use)         */
/*                                  */
/*  This routine trims a file specification in a `sensible' way */
/*  so that it fits into the specified output buffer.       */
/*      If the filespec is too long the following methods are used to   */
/*      trim it:                                                        */

/*      [1] remove directories from the middle of the filespec until    */
/*          it is small enough or there is only one directory and       */
/*          filename left. The removed directories are replaced with    */
/*          "..."                                                       */
/*      [2] if the resulting string is too long then simply use the     */
/*          filename prefixed with ".../" to show dirs omitted from     */
/*          the start.                                                  */
/*      [3] if the filename itslef is too long then return ".../" +     */
/*          <leading part of filename> + "..."                          */
/*                                  */
/************************************************************************/
unsigned long PSL_TRIM_FILESPEC (char *ssp,
                                 char *ddp,
                                 unsigned long dlen,
                                 unsigned long flags)
{
#define FPREFIX ".../"
#define FSUFFIX "..."
#define PREFIXLEN (strlen(FPREFIX))
#define SUFFIXLEN (strlen(FSUFFIX))
#define PREFSUFFLEN (strlen(FPREFIX)+strlen(FSUFFIX))
#define MAXELEMS 100

    struct elem elems[MAXELEMS];                /* elements of a filspec */
    int nelems = 0;                             /* # of elements in above */
    int xelems;                                 /* temporary elem index */
    int i;                                      /* loop variable */
    char *cp;                                   /* useful chr ptr */
    char sp[512];                               /* copy of source string */
    char dp[512];                               /* copy of destination string */
    int ellipses=0;                             /* ellipses done? */

    /*
     *  Check we have big enough buffers to perform the operation
     */
    if (strlen(ssp) > sizeof(sp))
    {
        *ddp = '\0';
        return PSL_BADPARAM;
    }
    strcpy(sp, ssp);                            /* copy source string */

    /*
     *  If the source string already fits into destination string just copy it over
     */
    if (strlen(ssp) < dlen)
    {
        (void)strcpy(ddp, ssp);
        return PSL_NORMAL;
    }

    /*
     *  break the filespec into directories and filename.
     */
    cp = sp;
    while ((cp = strtok(cp, "/")))
    {
        elems[nelems].elp = (char *)malloc(strlen(cp) + 1);
        (void)strcpy(elems[nelems].elp, cp);
        elems[nelems].len = strlen(elems[nelems].elp);
        nelems++;
        if (nelems >= MAXELEMS) return PSL_BADPARAM;
        cp = NULL;
    };
    /*
     *  if we have <= 2 elements then we can only return the filename.
     */
    if (nelems <= 2)
    {
        cp = strrchr(ssp, '/');                 /* point to filename */
        cp++;                                   /* move past '/' */
        /*
         *      if the filename with the PREFIX fits then return that.
         */
        if (strlen(cp) < (dlen - PREFIXLEN))    /* can't fit all filename */
        {
            (void)strcpy(ddp, FPREFIX);         /* add prefix */
            (void)strcat(ddp, cp);              /* copy to users buffer */
            FREE_ELEMS(&elems[0], nelems);
            return PSL_NORMAL;
        }
        else
        {
            /*
             *          copy as much of the filename that will fit, prefixing ".../" and
             *          adding "..." as a suffix
             */
            (void)strcpy(ddp, FPREFIX);         /* start with prefix */
            (void)memcpy(ddp + PREFIXLEN, cp, dlen - PREFSUFFLEN);
            /* add filename */
            ddp[dlen - SUFFIXLEN] = '\0';       /* terminate */
            (void)strcat(ddp, "...");           /* and add suffix */
            FREE_ELEMS(&elems[0], nelems);
            return PSL_NORMAL;
        }
    }
    /*
     *  we have a filespec with more than one directory in it, so we can strip
     *  directories out the middle until the filespec is small enough, or there
     *  are no more directories to remove.
     */
    xelems = nelems - 2;                        /* use second from last element */
    while((TOTAL_LEN(&elems[0], nelems) > (dlen - SUFFIXLEN - 1)) &&
            /* until small enough */
            (xelems > 0))                         /* or until on first element */
    {
        elems[xelems].len = 0;                  /* remove this element */
        xelems -= 1;                            /* backup one element */
    };
    /*
     *  if we removed all middle directories then we are only left with the
     *  first and last elements of the filespec, first dir and filename.
     */
    if (xelems == 0)
    {
        (void)strcpy(dp, "/");
        (void)strcat(dp, elems[0].elp);
        (void)strcat(dp, "/");
        (void)strcat(dp, elems[nelems - 1].elp);
    }
    else                                        /* build up new filespec */
    {
        (void)strcpy(dp, "/");                  /* start with '/' */
        for (i = 0; i < (nelems - 1); i++)      /* until second last element */
        {
            if (elems[i].len != 0)              /* if to be included */
            {
                (void)strcat(dp, elems[i].elp); /* add it */
                (void)strcat(dp, "/");
            }
            else if (!ellipses)                 /* if not to be included */
                /* and not done ellipses */
            {
                (void)strcat(dp, FPREFIX);      /* add ellipses */
                ellipses = 1;                   /* say we've done it */
            }
        }
        (void)strcat(dp, elems[i].elp);         /* add last element */
    }

    /*
     *  if still too long after we have trimmed all the middle directories out
     *  then pass this string back to ourselves to sort the filename out.
     */
    if (strlen(dp) > (dlen - 1))
    {
        (void)PSL_TRIM_FILESPEC(dp, ddp, dlen, 0);
        FREE_ELEMS(&elems[0], nelems);
        return PSL_NORMAL;
    }
    (void)strcpy(ddp, dp);                      /* copy our file to users */
    FREE_ELEMS(&elems[0], nelems);
    return PSL_NORMAL;
}

#endif /* UNIX */

#ifdef UNIX

/************************************************************************/

/*      TOTAL_LEN - get total length of filespec from element structure */

/*      Find out how long the filespec contained in the element         */
/*      structure would be if we concatenated all the non-zero elements */
/*      together.                                                       */

/************************************************************************/
static int TOTAL_LEN(struct elem *elp,
                     int nelems)
{
    int i;
    int total = 0;

    for (i = 0; i < nelems; i++)
    {
        total += (elp[i].len + 1);
    }
    return total;
}

#endif /* UNIX */

#ifdef UNIX

/************************************************************************/

/*      FREE_ELEMS - Free up elements in an element structure           */

/*      This function free up all the memory allocated in an element    */
/*      structure.                                                      */

/************************************************************************/
static void FREE_ELEMS(struct elem *elp,
                       int nelems)
{
    int i;

    for (i = 0; i < nelems; i++)
    {
        if (elp[i].elp != NULL) free (elp[i].elp);
    }
    return;
}

#endif /* UNIX */
