/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
/*===========================================================================*/
/*									     */
/*			      **** C O M M O N ****   			     */
/*									     */
/*===========================================================================*/
/*									     */
/*			Module Name: PSL_PTYRD_RTNS			     */
/*			============					     */
/*									     */
/* Description:								     */
/* ============								     */
/*									     */
/*  This module contains the interface routines to the PTYRD Deamon.	     */
/*									     */
/*									     */
/*===========================================================================*/

#include  <psl_macros.h>
#ifdef HP

#include <string.h>			/* String/memory functions */
#include <stdlib.h>			/* Standard library */
#include <signal.h>			/* Signal Definition */
#include <errno.h>			/* Error Numbers etc */
#include <sys/ipc.h>			/* Inter Process Communication */
#include <sys/shm.h>			/* Shared Memory */
#include <sys/sem.h>			/* Semaphores */
#include <sys/types.h>			/* System types */
#include <sys/sysmacros.h>		/* System macros */
#include <sys/stat.h>			/* stat() and structure */

/*
 *  Function Prototypes.
 */
#include  psl_common_file (psl_rtndef.h)	/* PSL Routines */

/*
 *  Structure Definitions.
 */
#include psl_common_file(psl_ptyrddef.h)
					/* PTYRD structures */
/*
 *  Parameters
 */
#include psl_common_file(psl_ipcdef.h)	    /* IPC definitions */
#include psl_common_file(shrmsg.h)	    /* PSL Messages */
#include psl_common_file(psldiag_prmdef.h)  /* Diagnostic Routine Parameters */
#include psl_common_file(pslerr_prmdef.h)   /* Erro parameters */
#include psl_common_file(psl_ptyrd_configdef.h)
					    /* Configuration parameters */


/************************************************************************/
/*									*/
/*    PSL_PTYRD_OPEN              					*/
/*    ==============							*/
/*									*/
/*    Calling Format:							*/
/*	desc_ptr = PSL_PTYRD_OPEN (diag_ptr)				*/
/*									*/
/*    Inputs								*/
/*      diag_ptr	- Ptr to a diagnostic stack.			*/
/*									*/
/*    Return Value							*/
/*	desc_ptr	- Ptr to a PTYRD shared memory structure.	*/
/*									*/
/*	This routine is called to map the PTYRD shared memory segment	*/
/*	and return a PTYRD shared memory structure which may be		*/
/*	passed to the other PSL_PTYRD_XXX functions.			*/
/*									*/
/************************************************************************/

PTYMEM *PSL_PTYRD_OPEN(struct diagstk *diag_ptr)
{
	 PTYMEM		*descriptor;	/* shm descriptor */
	 char		key_file[512];	/* Name of key file */
unsigned long		iret;		/* PSL return status */
	 int		status;		/* Status from ipc function */
	 int		saved_errno;	/* Saved errno */
	 key_t		key;		/* Key for semaphores/shared memory */
	 int		shm_id;		/* Shared memory id */
	 int		sem_id;		/* Semaphore id */
	 long		major;		/* Major device number */
	 long		nptys;		/* Max number of ptys */
	 pty_entry_t	*ptyentryp;	/* Ptr to base of shared memory */
struct	 stat		statbuf;	/* stat structure */
struct	 shmid_ds	shmbuf;		/* Shared memory structure */

/*
 *  Define the semaphore operations for the semop() calls.
 */
static struct sembuf op_attach[1] =
    {
	CLIENTATTACH, +1, SEM_UNDO	/* Increment attach count */
    };

/*
 *  Obtain the name of the key file.
 */
    PSL_GET_NODEDIR(key_file, sizeof(key_file));
    (void)strcat(key_file, "/ptyrd_pid");

/*
 *  Obtain the key for the shared memory /semaphores.
 */
    key = ftok(key_file, PSL_PROJECT_ID);
    if (key == (key_t)-1)
    {
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr, "Getting file id for %s", key_file);
	iret = PSL_DIAG_ERRNO(diag_ptr, "ftok()");
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	return NULL;
    }

/*
 *  Get the shared memory id.
 */
    shm_id = shmget(key, 0, 0);
    if(shm_id < 0)
    {
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr,
			         "Getting shared memory id for %s with key %d",
				 key_file, key);
	iret = PSL_DIAG_ERRNO(diag_ptr, "shmget()");
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	return NULL;
    }

/*
 *  Obtain the id of the sempahore from the key.
 */
    sem_id = semget(key, PTYRD_SEMS, 0);
    if(sem_id < 0)
    {
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr,
				 "Getting semaphore id for %s with key %d",
				 key_file, key);
	iret = PSL_DIAG_ERRNO(diag_ptr, "semget()");
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	return NULL;
    }

/*
 *  Obtain the major device number for ptys. We first try the device /dev/ttyp0
 *  and if this does not exist we then try /dev/pty/ttyp0. If neither exist
 *  we can only assume this machine has no ptys.
 */
    status = stat("/dev/ttyp0", &statbuf);
    if (status != 0)
    {
	status = stat("/dev/pty/ttyp0", &statbuf);
	if (status != 0)
	{
	    (void)PSL_ADD_DIAGNOSTIC(diag_ptr,
		    "No ptys found on this machine");
	    iret = PSL_DIAG_ERRNO(diag_ptr,
		    "stat(/dev/ttyp0 | /dev/pty/ttyp0)");
	    PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	    return NULL;
	}
    }
    major = major(statbuf.st_rdev);

/*
 *  Map the shared memory.
 */
    ptyentryp = (pty_entry_t *)
	       shmat(shm_id, (char *)0, SHM_RDONLY);
    if(ptyentryp == (pty_entry_t *) -1)
    {
	saved_errno = errno;
        (void)shmdt((char *)ptyentryp);
        errno = saved_errno;
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr,
				 "mapping shared memory %s with id %d",
				 key_file, shm_id);
	iret = PSL_DIAG_ERRNO(diag_ptr, "shmat()");
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
        return NULL;
    }

/*
 *  Register this shared memory attach in the CLIENTATTACH semaphore.
 */
    status = -1;
    errno = EINTR;
    while ((status == -1) && (errno == EINTR))
    {
	status = semop(sem_id, &op_attach[0], 1);
    }
    if (status == -1)
    {
	saved_errno= errno;
        (void)shmdt((char *)ptyentryp);
	errno = saved_errno;
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr, "Registering PTYRD client attach");
	iret = PSL_DIAG_ERRNO(diag_ptr, "semop()");    
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	return NULL;
    }

/*
 *  Find out the size of the shared memory so we can work out the maximum
 *  possible ptys on this machine.
 */
    status = shmctl(shm_id, IPC_STAT, &shmbuf);
    if (status != 0)
    {
	saved_errno= errno;
        (void)shmdt((char *)ptyentryp);
	errno = saved_errno;
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr,
	    "Reading shared memory control structure");
	iret = PSL_DIAG_ERRNO(diag_ptr, "shmctl()");    
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	return NULL;
    }
    nptys = shmbuf.shm_segsz / sizeof(struct pty_status_struct);

/*
 *  Allocate memory for a PTYMEM structure.
 */
    descriptor = (PTYMEM *)malloc(sizeof(PTYMEM));
    if (descriptor == NULL)
    {
        (void)shmdt((char *)ptyentryp);
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr, "Allocating memory for PTYMEM");
	PSL_PUTDIAG(diag_ptr, 1, PSL_INSVIRMEM);
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, PSL_INSVIRMEM);
	return NULL;
    }

/*
 *  Now that we have been successful, populate the returned structure.
 */
    descriptor->key = key;
    descriptor->sem_id = sem_id;
    descriptor->shm_id = shm_id;
    descriptor->major = major;
    descriptor->nptys = nptys;
    descriptor->shm_base = ptyentryp;

/*
 *  Return the shared memory descriptor.
 */
    return descriptor;
}


/************************************************************************/
/*									*/
/*    PSL_PTYRD_CLOSE              					*/
/*    ==============							*/
/*									*/
/*    Calling Format:							*/
/*	 iret = PSL_PTYRD_CLOSE (desc_ptr, diag_ptr)			*/
/*									*/
/*    Inputs								*/
/*	desc_ptr	- Ptr to a PTYRD shared memory descriptor	*/
/*      diag_ptr	- Ptr to a diagnostic stack.			*/
/*									*/
/*    Return status							*/
/*	iret		- PSL return status				*/
/*									*/
/*	This routine is called to unm,ap the PTYRD shared memory	*/
/*	and deallocate all memory associated with the PSL_PTYRD_OPEN.	*/
/*									*/
/************************************************************************/
unsigned long PSL_PTYRD_CLOSE(PTYMEM *desc_ptr,
			      struct diagstk *diag_ptr)
{
	 int	status;			/* shmdt() return status */
	 int	saved_errno;		/* Saved error number */
unsigned long	iret;			/* PSL return status */


/*
 *  Define the semaphore operation arrays for the semop() calls.
 */
static struct sembuf	op_detach[1] =
    {
	CLIENTATTACH, -1, SEM_UNDO    /* Decrement attach count */
    };

/*
 *  Check the descriptor is OK.
 */
    if ((desc_ptr->sem_id == -1) ||
	(desc_ptr->shm_id == -1) ||
	(desc_ptr->key == (key_t)-1))
    {
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr, "Invalid descriptor");
	iret = PSL_BADPARAM;
	PSL_PUTDIAG(diag_ptr, 1, iret);
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	return iret;
    }

/*
 *  Detach the shared memory.
 */
    status = shmdt((char *)desc_ptr->shm_base);
    if (status != 0)
    {
	saved_errno= errno;
	free(desc_ptr);
        (void)free(desc_ptr);		/* Free the PTYMEM */
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr, "Detaching PTYRD shared memory");
	iret = PSL_DIAG_ERRNO(diag_ptr, "shmdt()");
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	return iret;
    }

/*
 *  Register this shared memory dettach in the CLIENTATTACH semaphore.
 */
    status = -1;
    errno = EINTR;
    while ((status == -1) && (errno == EINTR))
    {
	status = semop(desc_ptr->sem_id, &op_detach[0], 1);
    }
    if (status == -1)
    {
	saved_errno= errno;
        (void)free(desc_ptr);		/* Free the PTYMEM */
	errno = saved_errno;
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr, "Registering PTYRD client detach");
	iret = PSL_DIAG_ERRNO(diag_ptr, "semop()");
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	return iret;
    }

/*
 *  Trash the descriptor.
 */
    desc_ptr->shm_id = -1;
    desc_ptr->sem_id = -1;
    desc_ptr->key = (key_t)-1;
    desc_ptr->shm_base = 0;

/*
 *  Free up the descriptor.
 */
    free(desc_ptr);

    return PSL_NORMAL;
}


/************************************************************************/
/*									*/
/*    PSL_PTYRD_STATUS              					*/
/*    ================							*/
/*									*/
/*    Calling Format:							*/
/*	pty_stat_t = PSL_PTYRD_OPEN (desc_ptr, pty, stamp_ptr		*/
/*				     pid_ptr, update_ptr, diag_ptr)	*/
/*									*/
/*    Inputs								*/
/*	desc_ptr    - Ptr to PTYRD descriptor				*/
/*	pty	    - pty to get the status for				*/
/*      diag_ptr    - Ptr to a diagnostic stack.			*/
/*									*/
/*    Outputs								*/
/*	stamp_ptr   - Returned time stamp				*/
/*	pid_ptr	    - Returned pid					*/
/*	update_ptr  - Returned daemon update number			*/
/*									*/
/*    Return Value							*/
/*	pty_stat_t  - Enumerated type					*/
/*		      Currently:					*/
/*			0 - PTY_ERROR					*/
/*			1 - PTY_ACTIVE					*/
/*			2 - PTY_READ_WAIT				*/
/*									*/
/*	This routine is called to determine the state of a particular	*/
/*	pty.								*/
/*									*/
/************************************************************************/
pty_stat_t PSL_PTYRD_STATUS(PTYMEM *desc_ptr,
			    dev_t pty,
			    clock_t *stamp_ptr,
			    pid_t *pid_ptr,
			    unsigned short *update_ptr,
			    struct diagstk *diag_ptr)
{
         int	    minor;		/* Minor device number */
         int	    major;		/* Major device number */
	 int	    status;		/* Status from ipc function */
	 pty_stat_t pty_status;		/* Status of pty */
unsigned long	    iret;		/* PSL return status */
	 Boolean    paramerr;		/* Parameter error */

/*
 *  Define the semaphore operation arrays for the semop() calls.
 */
static struct sembuf	op_read[2] =
    {
	DWRTLCK, 0, 0,		/* Wait for Deamon write lock to reset */
	CLIENTREAD, 1, SEM_UNDO /* Increment client read count */
    };

static struct sembuf	op_end_read[1] =
    {
	CLIENTREAD, -1, SEM_UNDO/* Decrement client read count */
    };

/*
 *  Initialize local variables.
 */
    paramerr = PFALSE;

/*
 *  Check the descriptor is OK.
 */
    if ((desc_ptr->sem_id == -1) ||
	(desc_ptr->shm_id == -1) ||
	(desc_ptr->key == (key_t)-1))
    {
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr, "Invalid descriptor");
	paramerr = PTRUE;
    }

/*
 *  Get the minor device number of the pty.
 */
    minor = minor(pty);
    major = major(pty);

/*
 *  Check the specified device is OK.
 */
    if ((minor < 0) || (minor >= desc_ptr->nptys))
    {
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr, "Invalid device, minor=%d", minor);
	paramerr = PTRUE;
    }
    if (major != desc_ptr->major)
    {
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr, "Invalid device, major=%d", major);
	paramerr = PTRUE;
    }

/*
 *  If we had a parameter problem then report it.
 */
    if (paramerr)
    {
	iret = PSL_BADPARAM;
	PSL_PUTDIAG(diag_ptr, 1, iret);
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	return PTY_ERROR;
    }

/*
 *  Wait for the Deamon to relinquish the write lock on the shared memory.
 */
    status = -1;
    errno = EINTR;
    while ((status == -1) && (errno == EINTR))
    {
	status = semop(desc_ptr->sem_id, &op_read[0], 2);
    }
    if (status == -1)
    {
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr, "Waiting to read semaphore");
	iret = PSL_DIAG_ERRNO(diag_ptr, "semop()");    
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	return PTY_ERROR;
    }

/*
 *  Return the time stamp.
 */
    *stamp_ptr = desc_ptr->shm_base[minor].pty_change;

/*
 *  Return the pid.
 */
    *pid_ptr = desc_ptr->shm_base[minor].pty_wt_pid;

/*
 *  Get the status of this pty.
 */
    pty_status = desc_ptr->shm_base[minor].pty_status;

/*
 *  Return the daemon update count.
 */
    *update_ptr = (unsigned short)
		  semctl(desc_ptr->sem_id, UPDATECOUNT, GETVAL, 0);

/*
 *  Register we are finished reading the shared memory.
 */
    status = -1;
    errno = EINTR;
    while ((status == -1) && (errno == EINTR))
    {
	status = semop(desc_ptr->sem_id, &op_end_read[0], 1);
    }
    if (status == -1)
    {
	(void)PSL_ADD_DIAGNOSTIC(diag_ptr, "Releasing read count");
	iret = PSL_DIAG_ERRNO(diag_ptr, "semop()");
	PSL_SIGNAL_ERROR(diag_ptr, PSLERRCLS_USER1, iret);
	return PTY_ERROR;
    }

    return pty_status;

}
#endif
