/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*            **** C O M M O N ****                  */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_ROOT_WIN32              */
/*          ============                         */
/*                                       */
/*  Description:                                 */
/*  ============                                 */
/*                                       */
/*  These routines are used, under WIN32, to return or concatenate       */
/*  the translation of the "psl_root" locations to the user.         */
/*                                       */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  This module is WIN32-Specific...
 */
#ifdef WIN32                    /* On WIN32 */

/*
 *  Standard ANSI header files.
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>    /* PSL Routines */

/*
 *  Structure Definitions.
 */
#include <shrmsg.h>    /* PSL Messages */


/************************************************************************/
/*                                  */
/*  PSL_ADDROOT - Prefix Specified String with psl_root (location)  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_ADDROOT (string_ptr, outbuf_ptr, outbuf_size)            */
/*                                  */
/*  Where:                              */
/*                                  */
/*     string_ptr  = Pointer to string to be appended to /psl_root  */
/*     outbuf_ptr  = Pointer to output string for result        */
/*           May be NULL if just psl_root to be returned    */
/*     outbuf_size = Size of output string                          */
/*                                  */
/*  This routine will prefix the (location of) psl_root to the      */
/*  specified string.  It will translate the "PSL_ROOT"     */
/*  environment variable.  If this fails, no prefix will be added.  */
/*                                  */
/************************************************************************/

void    PSL_ADDROOT (const char *string_ptr,
                     char *outbuf_ptr,
                     size_t outbuf_size)
{
    /*
     *  Local Data
     */
    char    *wrk_ptr = NULL;            /* Work Pointer */
    char    module_path[_MAX_PATH];         /* Work buffer */
    char    full_path[_MAX_PATH];           /* Work buffer */
    unsigned long    wrk_len;           /* Length of Work String */

    /*
     *  Get pathname of current executable
     */
    wrk_len = GetModuleFileName (NULL, module_path, sizeof(module_path));
    /*
     *  Did we succeed?
     */
    if (wrk_len != 0)
    {
        /*
         *  Convert to full pathname (i.e. not 8.3)
         */
        wrk_len = GetFullPathName (module_path, sizeof(full_path), full_path, &wrk_ptr);
        if (wrk_len <= 0)
        {
            *outbuf_ptr = '\0';
        }
        else
        {
            /*
             *  Truncate at filename
             */
            *(wrk_ptr-1) = '\0';
            wrk_len = strlen (full_path);

            /*
             *  Copy as many characters as we can.
             */
            if (wrk_len < outbuf_size)
            {
                (void) memcpy (outbuf_ptr, full_path, wrk_len);
                *(outbuf_ptr + wrk_len) = '\0';
            }
            else
            {
                (void) memcpy (outbuf_ptr, full_path, outbuf_size - 1);
                *(outbuf_ptr + outbuf_size - 1) = '\0';
            }
        }
    }
    else
        /*
         *  If we couldn't get pathname, store null prefix
         */
    {
        *outbuf_ptr = '\0';
    }
    /*
     *  How much have we left to copy.
     */
    if ((string_ptr != NULL) &&
            (strlen (outbuf_ptr) < (outbuf_size - 2)))
    {

        /*
         *  Append the string.
         */
        wrk_ptr = strchr (outbuf_ptr, '\0');
        wrk_len = wrk_ptr - outbuf_ptr; /* Space used */
        wrk_len = outbuf_size - wrk_len;
        /* Space available */
        if (wrk_len > strlen (string_ptr))
        {
            wrk_len = strlen (string_ptr) + 1;
            /* Size of string */
        }

        (void) memcpy (wrk_ptr, string_ptr, wrk_len - 1);
        *(wrk_ptr + wrk_len - 1) = '\0';
    }

    return;
}


/************************************************************************/
/*                                  */
/*  PSL_GET_NODEDIR - Return node-specific directory                */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_GET_NODEDIR (outbuf_ptr, outbuf_size)                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     outbuf_ptr  = Pointer to output string for result        */
/*     outbuf_size = Size of output string                          */
/*                                  */
/*  This routine will return the current psl_root/nodename location */
/*                                  */
/************************************************************************/

void    PSL_GET_NODEDIR (char *outbuf_ptr,
                         unsigned long outbuf_size)
{
    /*
     *  Local Data
     */
    char    nodename[256];     /* Nodename */

    /*
     *  First get the current nodename - prefixed with a '\'.
     */
    nodename[0] = '\\';
    (void) PSL_GET_NODENAME (NULL, &nodename[1], sizeof (nodename) -1);

    /*
     *  Prefix this name with the psl_root location.
     */
    PSL_ADDROOT (nodename, outbuf_ptr, outbuf_size);

    return;
}


/************************************************************************/
/*                                  */
/*  PSL_GET_PSUSER - Return PSL username and UID                */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_GET_PSUSER (diag_ptr, username_ptr,       */
/*                username_size, uid_ptr, gid_ptr)  */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret             = Unsigned Long PSL Routine Return Status   */
/*     diag_ptr         = Pointer to Diagnostics Stack      */
/*     username_ptr     = Pointer for username return       */
/*                        May be NULL if not required       */
/*     username_size    = Size of username string           */
/*     uid_ptr          = Pointer for UID return                    */
/*     gid_ptr          = Pointer for GID return                    */
/*                                  */
/*  This routine will return the User information selected during   */
/*  the installation. It currently returns "WINDOWS" on WIN32.  */
/*                                  */
/************************************************************************/

unsigned long   PSL_GET_PSUSER (struct diagstk *diag_ptr,
                                char *username_ptr,
                                unsigned long username_size,
                                void *uid_ptr,
                                void *gid_ptr)
{
    /*
     *  Local Data
     */

    /*
     *  Return "WINDOWS"
     */
    (void) strcpy (username_ptr, "WINDOWS");

    /*
     *  Return success.
     */
    return PSL_NORMAL;

}

#endif      /* WIN32-Specific */
