/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*            **** C O M M O N ****                  */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_SLEEP                   */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*    This routine allows a process to hibernate for a number of 10ms units  */
/*                                       */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

#ifdef  VMS
#pragma nostandard
#include <starlet.h>            /* VMS System Services      */
#pragma standard

#include <itemlist.h>           /* Itemlists            */
#endif  /* VMS */

#ifdef  UNIX
#include <unistd.h>         /* Unix functions       */
#include <signal.h>         /* Signals          */
#include <setjmp.h>         /* SETJMP definitions       */
#include <time.h>           /* TIME definitions     */
#if defined(IBM) || defined(OSF)
#include <sys/time.h>           /* For itimerval */
#endif
#endif  /* UNIX */

#ifdef  WIN32
#include <windows.h>
#endif  /* WIN32 */

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines         */

/*
 *    UNIX requires some static data to hold the setjmp offset and a function
 *    prototype for the alarm signal handler.
 */
#ifdef  UNIX
static void     ALRM_HANDLER (int signo);

#include <pslerr_prmdef.h>      /* Error Routine parameters */

#endif

/*
 *    EFN for VMS routine.
 */
#ifdef  VMS
#define TIMER_EFN   17
#endif

#ifdef  VMS
/************************************************************************/
/*                                  */
/*    PSL_SLEEP - Sleep for a number of 10ms units          */

/*    Calling Convention:                       */
/*                                  */
/*  PSL_SLEEP (centiseconds);                   */
/*                                  */
/*    Where:                                */
/*  centiseconds - UL number of 10ms ticks to sleep for.        */
/*                                  */
/*    This routine is called to force the process to sleep for the      */
/*    defined number of centiseconds (1/100th second). This routine may */
/*    be used in place of sleep when a period of less than 1 second is  */
/*    required.                             */
/*                                  */
/************************************************************************/

void    PSL_SLEEP       (unsigned long centiseconds)

{
    /*
     *    Local data
     */
    long wake_time[2]={0,-1};
    unsigned long iret;

    /*
     *    To prevent conflicts with other AST's which may be running we implement
     *    the PSL_SLEEP by initiating a timer for the required period and then
     *    waiting for the event flag associated with the timer to be set.
     *
     *    This should ONLY occur when the timer expires.
     */

    /*
     *    Set the base longword of the sleep time to be a number of 100 nanosecond
     *    units.
     */
    wake_time[0] = -100000 * centiseconds;

    /*
     *    ...and initiate the timer to fire after the specified period, specifying
     *    an event flag of 17 which will be set when the timer fires.
     */
    iret = sys$setimr (TIMER_EFN ,wake_time ,NULL ,0 ,0);
    if (!(psl_test_status (iret))) goto end;

    /*
     *    ...and then wait for the specified event flag to be set.
     */
    (void) sys$waitfr (TIMER_EFN);

    /*
     *    ...and return
     */
end:
    ;
    return;
}
#endif  /* VMS */

#ifdef  UNIX
/*
 *    Signal handler for the SIGALRM signal.
 */
static void     ALRM_HANDLER    (int signo)
{
    return;
}

/************************************************************************/
/*                                  */
/*    PSL_SLEEP - Sleep for a number of centiseconds            */

/*    Calling Convention:                       */
/*                                  */
/*  PSL_SLEEP (centiseconds);                   */
/*                                  */
/*    Where:                                */
/*  centiseconds - UL number of centiseconds to sleep for.      */
/*                                  */
/*    This routine is called to force the process to sleep for the      */
/*    defined number of centiseconds (1/100th second). This routine may */
/*    be used in place of sleep when period less than 1 second are  */
/*    required                              */
/*                                  */
/*    This routine is implemented using the POSIX compliant SIGNAL  */
/*    routines. It must not be called by routines/utilities which make  */
/*    use of signals from non-compatible signal sets such as signal or  */
/*    sigset.                               */
/*                                  */
/************************************************************************/

void    PSL_SLEEP       (unsigned long centiseconds)

{
    /*
     *    Local data
     */
    struct   itimerval  itm;        /* Timer structure      */
    int        status;     /* Signed status        */
    long       secs;       /* # of seconds to wait     */
    long       usecs;      /* # of microseconds to wait    */
    unsigned long       iret;       /* PSL return status        */
    struct   sigaction  sig_action;         /* Signal action        */
    struct   sigaction  old_action;     /* Old Signal action structure  */
    sigset_t   newmask;            /* New signal mask      */
    sigset_t   oldmask;            /* Old signal mask      */
    sigset_t   zeromask;           /* Empty signal mask        */

    /*
     *    Setup a signal handler for SIGALRM, saving the old signal action so that
     *    we can restore it at the end.
     */
#if defined(IBM) && !defined(AIXV41)
    sig_action.sa_handler = (void *)ALRM_HANDLER;
#else /*IBM */
sig_action.sa_handler = ALRM_HANDLER;
#endif
#ifdef SVR4
    sig_action.sa_flags = SA_RESTART;           /* Restart interrupted f()s */
#else /* SVR4 */
sig_action.sa_flags = 0;
#endif /* SVR4 */
    (void) sigemptyset(&sig_action.sa_mask);

    (void) sigaction (SIGALRM, &sig_action, &old_action);

    /*
     *    Setup the timer structure with the required time. We shall have to convert
     *    the number of 100ms ticks into seconds and micro-seconds as this is what
     *    is expected in the time structure
     */
    secs = centiseconds / 100;      /* Calculate # of seconds   */
    usecs = (centiseconds % 100) * 10000;
    /* ...and usecs         */

    itm.it_interval.tv_sec = 0;
    itm.it_interval.tv_usec = 0;
    itm.it_value.tv_sec = secs;
    itm.it_value.tv_usec = usecs;

    /*
     *    Hold off SIGALRM. We do this by obtaining a clear signal mask, adding
     *    the SIGALRM signal to it and then blocking all signals in the mask.
     */
    (void) sigemptyset (&zeromask); /* Get zero mask        */
    (void) sigemptyset (&newmask);  /* ... and another      */
    (void) sigaddset (&newmask, SIGALRM);
    /* Add SIGALRM          */
    status = sigprocmask (SIG_BLOCK, &newmask, &oldmask);
    /* Block SIGALRM    */
    /*
     *    If we have an error then signal it remembering that we must restore the
     *    old signal action before we return.
     */
    if (status != 0)
    {
        (void) sigaction (SIGALRM, &old_action, (struct sigaction *) NULL);
        iret = PSL_DIAG_ERRNO (NULL ,"sigprocmask(SIGBLOCK)");
        PSL_SIGNAL_ERROR (NULL ,PSLERRCLS_USER1, iret);
        goto end;
    }

    /*
     *  Start the timer.
     */
    status = setitimer (ITIMER_REAL, &itm, NULL);

    /*
     *    If we have an error then signal it remebering that we must restore the
     *    old signal action and the signal mask before we return.
     */
    if (status != 0)
    {
        (void) sigprocmask (SIG_SETMASK, &oldmask, NULL);
        (void) sigaction (SIGALRM, &old_action, (struct sigaction *) NULL);
        iret = PSL_DIAG_ERRNO (NULL, "setitimer");
        PSL_SIGNAL_ERROR (NULL ,PSLERRCLS_USER1, iret);
        goto end;
    }

    /*
     *  Allow all signals through and wait for one to arrive.
     */
    (void) sigsuspend (&zeromask);

    /*
     *  Restore old signal mask
     */
    status = sigprocmask(SIG_SETMASK, &oldmask, NULL);
    if (status != 0)
    {
        (void) sigaction (SIGALRM, &old_action, (struct sigaction *) NULL);
        iret = PSL_DIAG_ERRNO(NULL, "setprocmask");
        PSL_SIGNAL_ERROR(NULL ,PSLERRCLS_USER1, iret);
    }

    /*
     *    ...and restore the old signal action.
     */
    (void) sigaction (SIGALRM, &old_action, (struct sigaction *) NULL);
end:
    ;
    return;
}
#endif  /* UNIX */


#ifdef  WIN32
/************************************************************************/
/*                                  */
/*    PSL_SLEEP - Sleep for a number of 100ms units         */

/*    Calling Convention:                       */
/*                                  */
/*  PSL_SLEEP (centiseconds);                   */
/*                                  */
/*    Where:                                */
/*  centiseconds - UL number of centiseconds to sleep for.      */
/*                                  */
/*    This routine is called to force the process to sleep for the      */
/*    defined number of 10ms ticks (1/100th second). This routine may   */
/*    be used in place of sleep when period less than 1 second are  */
/*    required                              */
/*                                  */
/************************************************************************/

void    PSL_SLEEP       (unsigned long centiseconds)

{
    /*
     *  Sleep for required number of milliseconds
     */
    Sleep (centiseconds * 10);

    return;
}
#endif  /* WIN32 */

