/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: TPA_ROUTINES                */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  Generic Tparse routines.                         */
/*                                       */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <ctype.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines */
#include <tpa_rtndef.h>         /* Tparse Routines */
#include <trc_rtndef.h>         /* Trace Routines */

static Boolean CHK_CONDITIONAL (struct tdb *tdb_ptr);

static Boolean DO_CONDITIONAL (struct tdb *tdb_ptr,
                               char *srclin_ptr,
                               char *prslin_ptr);

static void DO_ELIF (struct tdb *tdb_ptr,
                     char *srclin_ptr,
                     char *prslin_ptr);

static void DO_ELSE (struct tdb *tdb_ptr,
                     char *srclin_ptr,
                     char *prslin_ptr);

static void DO_ENDIF (struct tdb *tdb_ptr,
                      char *srclin_ptr,
                      char *prslin_ptr);


static void DO_ERROR (struct tdb *tdb_ptr,
                      char *srclin_ptr,
                      char *prslin_ptr);


static void DO_IFDEF (struct tdb *tdb_ptr,
                      char *srclin_ptr,
                      char *prslin_ptr);

static void DO_IFNDEF (struct tdb *tdb_ptr,
                       char *srclin_ptr,
                       char *prslin_ptr);

static void DO_INCLUDE (struct tdb *tdb_ptr,
                        char *srclin_ptr,
                        char *prslin_ptr);

static void EXTEND_TOKEN_BLANK (struct tpaarg *argblk_ptr);

static void EXTEND_TOKEN_LINE (struct tpaarg *argblk_ptr);

static void EXTEND_TOKEN_STD (struct tpaarg *argblk_ptr);

static Boolean GET_VARIANT (struct tdb *tdb_ptr,
                            char *srclin_ptr,
                            char *buf_ptr,
                            char *varnam_ptr);

static Boolean OPEN_INCLUDE (struct tdb *tdb_ptr,
                             char *prslin_ptr,
                             char *srclin_ptr,
                             FILE **file_ptr_ptr,
                             char *name_ptr);

static void READ_NEXT_LINE (struct tdb *tdb_ptr,
                            long *srcsiz_ptr,
                            char *srclin_ptr,
                            long *prssiz_ptr,
                            char *prslin_ptr,
                            char *errlin_ptr,
                            Boolean *eof_flag_ptr);

static void RESET_STRUCT (struct tdb *tdb_ptr);

/*
 *  Structure Definitions.
 */
#include <shrmsg.h>         /* PSL Messages */
#include <psldiag_prmdef.h>     /* Diagnostic Routine Symbols */
#include <pslerr_prmdef.h>      /* Error Routine Symbols */
#include <pslfil_prmdef.h>      /* File Handling Symbols */
#include <psl_trc_prmdef.h>     /* Tracing Symbols */
#include <tdbdef.h>         /* TDB Structure */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block Structure */


/************************************************************************/
/*                                  */
/*  TPA_INIT - Initialize TPA routines              */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_INIT (tdb_ptr);                   */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret    = Unsigned Long PSL Routine Return Status        */
/*     tdb_ptr = Pointer to TPA structure               */
/*                                  */
/*  This routine initializes a TPA structure fopr use by the    */
/*  TPA routines.                           */
/*                                  */
/************************************************************************/

unsigned long TPA_INIT (struct tdb *tdb_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret = PSL_NORMAL;  /* Return status */
    unsigned long       diag_save;      /* Saved diagnostic context */
    struct   diagstk    *diag_ptr;      /* Pointer to Diagnostic Stack */

    /*
     *  Copy pointers for ease of reference.
     */
    diag_ptr = (*tdb_ptr).tdb_diag_ptrps;

    /*
     *  First add a diagnostic to the stack.
     */
    diag_save = PSL_ADD_STATDIAG (diag_ptr, "During TPA_INIT");

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_GEN_TPARSE))
    {
        PSL_OUTPUT_TRACE ("Initializing TPA structure\n");
    }

    /*
     *  Set up structure type.
     */
    (*tdb_ptr).tdb_strtypl = TDB_STRNAM;

    /*
     *  Allocate line buffers.
     *
     *  Note that we allocate 4 bytes extra to allow for a terminating null
     *  and to allow use to read more characters than the maximum allowed so
     *  that we can check for overflow.
     */
    (*tdb_ptr).tdb_srclin_ptrpc =
        (char *) malloc ((*tdb_ptr).tdb_max_linsizl + 4);
    /* Allocate buffer */
    if ((*tdb_ptr).tdb_srclin_ptrpc == NULL)
        /* If an error */
    {
        iret = PSL_SYSERVERR;       /* Set up error */
        PSL_PUTDIAG (           /* Put Diagnostic */
            diag_ptr,       /* Use this stack */
            2,          /* Number of message segments */
            iret,           /* First message */
            "malloc for Source Line",
            /* Parameter */
            PSL_INSVIRMEM);     /* Secondary message */
        PSL_SIGNAL_ERROR (      /* Signal Error */
            diag_ptr,       /* Use this stack */
            PSLERRCLS_USER5,    /* Error Class */
            iret);          /* Message */
        goto end;           /* And exit */
    }

    (*tdb_ptr).tdb_prslin_ptrpc =
        (char *) malloc ((*tdb_ptr).tdb_max_linsizl + 4);
    /* Allocate buffer */
    if ((*tdb_ptr).tdb_prslin_ptrpc == NULL)
        /* If an error */
    {
        iret = PSL_SYSERVERR;       /* Set up error */
        PSL_PUTDIAG (           /* Put Diagnostic */
            diag_ptr,       /* Use this stack */
            2,          /* Number of message segments */
            iret,           /* First message */
            "malloc for Parsable Line",
            /* Parameter */
            PSL_INSVIRMEM);     /* Secondary message */
        PSL_SIGNAL_ERROR (      /* Signal Error */
            diag_ptr,       /* Use this stack */
            PSLERRCLS_USER5,    /* Error Class */
            iret);          /* Message */
        goto end;           /* And exit */
    }

    (*tdb_ptr).tdb_errlin_ptrpc =
        (char *) malloc ((*tdb_ptr).tdb_max_linsizl + 4);
    /* Allocate buffer */
    if ((*tdb_ptr).tdb_errlin_ptrpc == NULL)
        /* If an error */
    {
        iret = PSL_SYSERVERR;       /* Set up error */
        PSL_PUTDIAG (           /* Put Diagnostic */
            diag_ptr,       /* Use this stack */
            2,          /* Number of message segments */
            iret,           /* First message */
            "malloc for Error Line",
            /* Parameter */
            PSL_INSVIRMEM);     /* Secondary message */
        PSL_SIGNAL_ERROR (      /* Signal Error */
            diag_ptr,       /* Use this stack */
            PSLERRCLS_USER5,    /* Error Class */
            iret);          /* Message */
        goto end;           /* And exit */
    }

    /*
     *  Initialize other fields.
     */
    RESET_STRUCT (tdb_ptr);

    /*
     *  Reset the diagnostics.
     */
    PSL_REMOVE_DIAGNOSTICS (diag_ptr, diag_save);

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return iret;
}


/************************************************************************/
/*                                  */
/*  TPA_END - Terminate TPA routines                */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     TPA_END (tdb_ptr);                       */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr = Pointer to TDB structure               */
/*                                  */
/*  This routine clears up a TDB structure.             */
/*                                  */
/************************************************************************/

void TPA_END (struct tdb *tdb_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       diag_save;      /* Saved diagnostic context */
    struct   diagstk    *diag_ptr;      /* Pointer to Diagnostic Stack */

    /*
     *  Copy pointers for ease of reference.
     */
    diag_ptr = (*tdb_ptr).tdb_diag_ptrps;

    /*
     *  First add a diagnostic to the stack.
     */
    diag_save = PSL_ADD_STATDIAG (diag_ptr, "During TPA_END");

    /*
     *  Free line buffers.
     */
    free ((*tdb_ptr).tdb_srclin_ptrpc); /* Free source line */
    (*tdb_ptr).tdb_srclin_ptrpc = NULL; /* Trash pointer */

    free ((*tdb_ptr).tdb_prslin_ptrpc); /* Free parsable line */
    (*tdb_ptr).tdb_prslin_ptrpc = NULL; /* Trash pointer */

    free ((*tdb_ptr).tdb_errlin_ptrpc); /* Free error line */
    (*tdb_ptr).tdb_errlin_ptrpc = NULL; /* Trash pointer */

    /*
     *  Reset the diagnostics.
     */
    PSL_REMOVE_DIAGNOSTICS (diag_ptr, diag_save);

    return;
}


/************************************************************************/
/*                                  */
/*  TPA_CHK_CONDEND - Check for missing #ENDIFs         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     TPA_CHK_CONDEND (tdb_ptr);                   */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr = Pointer to TPA Structure               */
/*                                  */
/*  This routine checks to see if there are any missing #ENDIF  */
/*  directives, and outputs a message if there are.         */
/*                                  */
/************************************************************************/

void TPA_CHK_CONDEND (struct tdb *tdb_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       diag_save;      /* Saved diagnostic context */
    struct   diagstk    *diag_ptr;      /* Pointer to Diagnostic Stack */

    /*
     *  Copy pointers for ease of reference.
     */
    diag_ptr = (*tdb_ptr).tdb_diag_ptrps;

    /*
     *  First add a diagnostic to the stack.
     */
    diag_save = PSL_ADD_STATDIAG (diag_ptr, "During TPA_CHK_CONDEND");

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_GEN_TPARSE))
    {
        PSL_OUTPUT_TRACE ("Checking for missing #ENDIFs\n");
    }

    /*
     *  If there are any missing #ENDIFs report an error.
     */
    if ((*tdb_ptr).tdb_if_levell != -1) /* If missing #ENDIFs */
    {
        PSL_PUTDIAG (       /* Add Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_MISCONDEND,     /* Message */
            (*tdb_ptr).tdb_if_levell+1,
            /* First parameter */
            (*tdb_ptr).tdb_expsrcnamc256);
        /* Second parameter */
        PSL_OUTPUT_DIAG (       /* Output Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* From this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Output Last Error */

        (*tdb_ptr).tdb_err_countl += 1; /* Count it */
    }

    /*
     *  Reset the diagnostics.
     */
    PSL_REMOVE_DIAGNOSTICS (diag_ptr, diag_save);

    return;
}


/************************************************************************/
/*                                  */
/*  TPA_CHK_EOF - Check for End of Source File          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_CHK_EOF (argblk_ptr);             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine checks the End of File flag and returns success    */
/*  if it is set.                           */
/*                                  */
/************************************************************************/

unsigned long TPA_CHK_EOF (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   tdb        *tdb_ptr;       /* Pointer to TDB structure */

    /*
     *  Copy pointers for ease of reference.
     */
    tdb_ptr = (struct tdb *) (*argblk_ptr).user_ptrp;

    if ((*tdb_ptr).tdb_eof_flagn)   /* If end of file */
    {
        return 1;           /* Signal success */
    }
    else                /* Otherwise */
    {
        return 0;           /* Signal failure */
    }
}


/************************************************************************/
/*                                  */
/*  TPA_CHK_EOL - Check for End of Source Line          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_CHK_EOL (argblk_ptr);             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine checks that there is no more text after the    */
/*  current point in the line.                  */
/*                                  */
/************************************************************************/

unsigned long TPA_CHK_EOL (struct tpaarg *argblk_ptr)
{

    if ((*argblk_ptr).stringcntl == 0)  /* If no characters left */
    {
        return 1;           /* Signal success */
    }
    else                /* Some characters left */
    {
        return 0;           /* Signal failure */
    }

}


/************************************************************************/
/*                                  */
/*  TPA_END_SOURCE - Terminate source file              */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_END_SOURCE (tdb_ptr);             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret    = Unsigned Long Action Routine Return Status     */
/*     tdb_ptr = Pointer to TPA Structure               */
/*                                  */
/*  This routine terminates the handling of a source file.      */
/*  It performs the following:                  */
/*                                  */
/*      1. Closes the source file                   */
/*                                  */
/************************************************************************/

unsigned long TPA_END_SOURCE (struct tdb *tdb_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret = PSL_NORMAL;  /* Return status */
    unsigned long       diag_save;      /* Saved diagnostic context */
    struct   diagstk    *diag_ptr;      /* Pointer to Diagnostic Stack */
    long       status;     /* File I/O status */

    /*
     *  Copy pointers for ease of reference.
     */
    diag_ptr = (*tdb_ptr).tdb_diag_ptrps;

    /*
     *  First add a diagnostic to the stack.
     */
    diag_save = PSL_ADD_STATDIAG (diag_ptr, "During TPA_END_SOURCE");

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_GEN_TPARSE))
    {
        PSL_OUTPUT_TRACE ("Terminating source file \"%s\"\n",
                          (*tdb_ptr).tdb_expsrcnamc256);
    }

    /*
     *  Close the file.
     */
    status = fclose ((*tdb_ptr).tdb_srcfil_ptrpf);
    if (status != 0)
    {
        iret = PSL_CLOSEIN;     /* Set up error */
        PSL_LOG_CIOERR (        /* Log file error */
            diag_ptr,       /* Use this stack */
            iret,           /* Error to report */
            (*tdb_ptr).tdb_expsrcnamc256);
        /* Filespec */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            diag_ptr,       /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        goto end;           /* And exit */
    }

    (*tdb_ptr).tdb_srcfil_ptrpf = NULL; /* Trash pointer */

    /*
     *  Reset the diagnostics.
     */
    PSL_REMOVE_DIAGNOSTICS (diag_ptr, diag_save);

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return iret;
}


/************************************************************************/
/*                                  */
/*  TPA_ERROR_BLANK - Extend Token to Next Blank and Log Error  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_ERROR_BLANK (argblk_ptr);             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine adds the error message specified in the Argument   */
/*  Block to the error array (if there is room), extends the    */
/*  current token to the next 'blank' and underlines the current    */
/*  token.                              */
/*                                  */
/*  The routine return status is always set to success.     */
/*                                  */
/************************************************************************/

unsigned long TPA_ERROR_BLANK (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   tdb        *tdb_ptr;       /* Pointer to TDB structure */

    /*
     *  Copy pointers for ease of reference.
     */
    tdb_ptr = (struct tdb *) (*argblk_ptr).user_ptrp;

    /*
     *  Store the error in the array, if possible.
     */
    TPA_STORE_ERROR (tdb_ptr, (*argblk_ptr).paraml);

    /*
     *  Perform standard token extension.
     */
    EXTEND_TOKEN_BLANK (argblk_ptr);

    /*
     *  Underline the Token in Error.
     */
    TPA_UNDERLINE_ERROR (argblk_ptr);

    /*
     *  Set routine return status to success.
     */
    return 1;
}



/************************************************************************/
/*                                  */
/*  TPA_ERROR_LINE - Extend Token to End of Line and Log Error  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_ERROR_LINE (argblk_ptr);              */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine adds the error message specified in the Argument   */
/*  Block to the error array (if there is room), extends the    */
/*  current token to the end of the current line and underlines */
/*  the current token.                      */
/*                                  */
/*  The routine return status is always set to success.     */
/*                                  */
/************************************************************************/

unsigned long TPA_ERROR_LINE (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   tdb        *tdb_ptr;       /* Pointer to TDB structure */

    /*
     *  Copy pointers for ease of reference.
     */
    tdb_ptr = (struct tdb *) (*argblk_ptr).user_ptrp;

    /*
     *  Store the error in the array, if possible.
     */
    TPA_STORE_ERROR (tdb_ptr, (*argblk_ptr).paraml);

    /*
     *  Extend the current token to the End of Line.
     */
    EXTEND_TOKEN_LINE (argblk_ptr);

    /*
     *  Underline the Token in Error.
     */
    TPA_UNDERLINE_ERROR (argblk_ptr);

    /*
     *  Set routine return status to success.
     */
    return 1;
}



/************************************************************************/
/*                                  */
/*  TPA_ERROR_MSG - Log Error with No Token Extension/Underline */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_ERROR_MSG (argblk_ptr);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine adds the error message specified in the Argument   */
/*  Block to the error array (if there is room). No token       */
/*  extension or token underlining is performed.            */
/*                                  */
/*  The routine return status is always set to success.     */
/*                                  */
/************************************************************************/

unsigned long TPA_ERROR_MSG (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   tdb        *tdb_ptr;       /* Pointer to TDB structure */

    /*
     *  Copy pointers for ease of reference.
     */
    tdb_ptr = (struct tdb *) (*argblk_ptr).user_ptrp;

    /*
     *  Store the error in the array, if possible.
     */
    TPA_STORE_ERROR (tdb_ptr, (*argblk_ptr).paraml);

    /*
     *  Set routine return status to success.
     */
    return 1;
}



/************************************************************************/
/*                                  */
/*  TPA_ERROR_STD - Perform Standard Token Extension and Log Error  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_ERROR_STD (argblk_ptr);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine adds the error message specified in the Argument   */
/*  Block to the error array (if there is room), extends the    */
/*  current token in the standard way (using the user defined   */
/*  token separators) and underlines the current token.     */
/*                                  */
/*  The routine return status is always set to success.     */
/*                                  */
/************************************************************************/

unsigned long TPA_ERROR_STD (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   tdb        *tdb_ptr;       /* Pointer to TDB structure */

    /*
     *  Copy pointers for ease of reference.
     */
    tdb_ptr = (struct tdb *) (*argblk_ptr).user_ptrp;

    /*
     *  Store the error in the array, if possible.
     */
    TPA_STORE_ERROR (tdb_ptr, (*argblk_ptr).paraml);

    /*
     *  Perform standard token extension.
     */
    EXTEND_TOKEN_STD (argblk_ptr);

    /*
     *  Underline the Token in Error.
     */
    TPA_UNDERLINE_ERROR (argblk_ptr);

    /*
     *  Set routine return status to success.
     */
    return 1;
}



/************************************************************************/
/*                                  */
/*  TPA_ERROR_TOKEN - Log Error with No Token Extension     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_ERROR_TOKEN (argblk_ptr);             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine adds the error message specified in the Argument   */
/*  Block to the error array (if there is room) and underlines  */
/*  the current token. No token extension is performed.     */
/*                                  */
/*  The routine return status is always set to success.     */
/*                                  */
/************************************************************************/

unsigned long TPA_ERROR_TOKEN (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   tdb        *tdb_ptr;       /* Pointer to TDB structure */

    /*
     *  Copy pointers for ease of reference.
     */
    tdb_ptr = (struct tdb *) (*argblk_ptr).user_ptrp;

    /*
     *  Store the error in the array, if possible.
     */
    TPA_STORE_ERROR (tdb_ptr, (*argblk_ptr).paraml);

    /*
     *  Underline the Token in Error.
     */
    TPA_UNDERLINE_ERROR (argblk_ptr);

    /*
     *  Set routine return status to success.
     */
    return 1;
}


/************************************************************************/
/*                                  */
/*  TPA_STORE_ERROR - Store an Error                */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     TPA_STORE_ERROR (tdb_ptr, errnum);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr = Pointer to TPA Structure               */
/*     errnum  = Error Number                   */
/*                                  */
/*  This routine is responsible for storing an Error in the array   */
/*  of errors for the current source line and incrementing the  */
/*  appropriate counters.                       */
/*                                  */
/************************************************************************/

void TPA_STORE_ERROR (struct tdb *tdb_ptr,
                      unsigned long errnum)
{
    /*
     *  Local Data.
     */
    unsigned long       jsub;       /* Loop Control */

    /*
     *  Check to see if we have already filled the error array.
     */
    if ((*tdb_ptr).tdb_errs_in_linel == TDB_MAX_ERRS_IN_LINE) goto end;

    /*
     *  Loop through the error array to make sure that the new error
     *  is not already in there. If it is, we just underline it.
     */
    for (jsub = 0; jsub < (*tdb_ptr).tdb_errs_in_linel; jsub++)
    {
        if ((*tdb_ptr).tdb_err_arrayl5 [jsub] == errnum) goto end;
    }

    /*
     *  If we get here the error is not already in the array so we insert it
     *  in the first free slot and count it.
     */
    (*tdb_ptr).tdb_err_arrayl5 [(*tdb_ptr).tdb_errs_in_linel] = errnum;

    (*tdb_ptr).tdb_errs_in_linel += 1;

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    (*tdb_ptr).tdb_err_countl += 1; /* Count error */

    return;
}


/************************************************************************/
/*                                  */
/*  TPA_UNDERLINE_ERROR - Underline the current Token       */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     UNDERLINE_ERROR (argblk_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine is responsible for underlining the current Token.  */
/*  The ERROR_LINE is set up with asterisks in the correct places.  */
/*                                  */
/************************************************************************/

void TPA_UNDERLINE_ERROR (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    long       jsub;       /* Loop control */
    unsigned long       offset;     /* Character string offset */
    char       *wrk_ptr;       /* Work pointer */
    struct   tdb        *tdb_ptr;       /* Pointer to TDB structure */

    /*
     *  Copy pointers for ease of reference.
     */
    tdb_ptr = (struct tdb *) (*argblk_ptr).user_ptrp;

    /*
     *  Calculate the offset of the character pointed to by the Token Pointer
     *  in the parsable line buffer.
     */
    offset = (long) ((*argblk_ptr).token_ptrpc - (*tdb_ptr).tdb_prslin_ptrpc);

    /*
     *  Set the corresponding characters in the error line buffer to asterisks,
     *  as long as they are not tabs.
     */
    for (jsub = 0, wrk_ptr = (*tdb_ptr).tdb_errlin_ptrpc+offset;
            jsub < (*argblk_ptr).tokencntl;
            jsub++, wrk_ptr++)
    {
        if (*wrk_ptr != '\t')       /* If it's not a tab */
        {
            *wrk_ptr = '*';     /* Replace with asterisk */
        }
    }

    /*
     *  Set the flag that indicates that an Underline Line is required.
     */
    (*tdb_ptr).tdb_undlin_flagn = PTRUE;

    return;
}


/************************************************************************/
/*                                  */
/*  TPA_GET_ERR_COUNT - Get error count             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     err_count = TPA_GET_ERR_COUNT (tdb_ptr);         */
/*                                  */
/*  Where:                              */
/*                                  */
/*     err_count = Unsigned Long returned error count       */
/*     tdb_ptr   = Pointer to TPA Structure             */
/*                                  */
/*  This routine returns the error count from the TDB structure.    */
/*                                  */
/************************************************************************/

unsigned long TPA_GET_ERR_COUNT (struct tdb *tdb_ptr)
{
    return (*tdb_ptr).tdb_err_countl;
}


/************************************************************************/
/*                                  */
/*  TPA_GET_EXP_SRCNAM - Get expanded source file name      */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     TPA_GET_EXP_SRCNAM (tdb_ptr, buf_ptr, maxlen);       */
/*                                  */
/*  Where:                              */
/*                                  */
/*      tdb_ptr = Pointer to TPA Structure              */
/*      buf_ptr = Pointer to buffer to hold returned filename   */
/*      maxlen  = Maximum length of buffer              */
/*                                  */
/*  This routine returns the expanded source file name from the */
/*  TDB structure.                          */
/*                                  */
/************************************************************************/

void TPA_GET_EXP_SRCNAM (struct tdb *tdb_ptr,
                         char *buf_ptr,
                         long maxlen)
{
    /*
     *  Local data.
     */
    long       length;     /* String length */

    /*
     *  Work out how much of the filename we can return to the caller.
     */
    length = strlen ((*tdb_ptr).tdb_expsrcnamc256);

    if (length > (maxlen - 1))
    {
        length = maxlen - 1;
    }

    (void) memcpy (buf_ptr, (*tdb_ptr).tdb_expsrcnamc256, length);
    /* Copy string */
    *(buf_ptr + length) = '\0';     /* Null terminate */

    return;
}


/************************************************************************/
/*                                  */
/*  TPA_GET_LINE_COUNT - Get line count             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     line_count = TPA_GET_LINE_COUNT (tdb_ptr);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*     err_count = Unsigned Long returned error count       */
/*     tdb_ptr   = Pointer to TPA Structure             */
/*                                  */
/*  This routine returns the line count from the TDB structure. */
/*                                  */
/************************************************************************/

unsigned long TPA_GET_LINE_COUNT (struct tdb *tdb_ptr)
{
    return (*tdb_ptr).tdb_line_countl;
}


/************************************************************************/
/*  TPA_GET_PRSLIN_PTR - Get pointer to parsable line       */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     prslin_ptr = TPA_GET_PRSLIN_PTR (tdb_ptr);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*     prslin_ptr = Pointer to parsable line            */
/*     tdb_ptr    = Pointer to TPA Structure            */
/*                                  */
/*  This routine returns a pointer to the parsable line buffer  */
/*  from the TDB structure.                     */
/*                                  */
/************************************************************************/

char *TPA_GET_PRSLIN_PTR (struct tdb *tdb_ptr)
{
    return (*tdb_ptr).tdb_prslin_ptrpc;
}


/************************************************************************/
/*                                  */
/*  TPA_GET_SRCLIN_PTR - Get pointer to source line         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     srclin_ptr = TPA_GET_SRCLIN_PTR (tdb_ptr);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*     srclin_ptr = Pointer to source line              */
/*     tdb_ptr    = Pointer to TPA Structure            */
/*                                  */
/*  This routine returns a pointer to the source line buffer from   */
/*  the TDB structure.                      */
/*                                  */
/************************************************************************/

char *TPA_GET_SRCLIN_PTR (struct tdb *tdb_ptr)
{
    return (*tdb_ptr).tdb_srclin_ptrpc;
}


/************************************************************************/
/*                                  */
/*  TPA_GET_USRDAT_PTR - Get user data pointer          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     usr_ptr = TPA_GET_USRDAT_PTR (tdb_ptr);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*     usr_ptr = Pointer to void returned user data pointer     */
/*     tdb_ptr = Pointer to TPA Structure               */
/*                                  */
/*  This routine returns the user data pointer from the TPA     */
/*  structure.                          */
/*                                  */
/************************************************************************/

void *TPA_GET_USRDAT_PTR (struct tdb *tdb_ptr)
{
    return (*tdb_ptr).tdb_usrdat_ptrp;
}


/************************************************************************/
/*                                  */
/*  TPA_GET_NEXT_LINE - Get the next line from the Source File  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_GET_NEXT_LINE (argblk_ptr);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine gets the next line from the source file.       */
/*                                  */
/*  It first checks to see if there are any errors in the current   */
/*  source line and, if so, outputs the underline and reports the   */
/*  errors.                             */
/*                                  */
/************************************************************************/

unsigned long TPA_GET_NEXT_LINE (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Status Return */
    unsigned long       jsub;       /* Loop Control */
    struct   tdb        *tdb_ptr;       /* Pointer to TDB structure */

    /*
     *  Copy pointers for ease of reference.
     */
    tdb_ptr = (struct tdb *) (*argblk_ptr).user_ptrp;

    /*
     *  Are there any errors to report?
     */
    if ((*tdb_ptr).tdb_errs_in_linel != 0)
    {

        /*
         *  Output the underline buffer, if required.
         */
        if ((*tdb_ptr).tdb_undlin_flagn)
        {

            /*
             *  Output to the listing file. If it fails a message will already have
             *  been output so we signal the ENDABORT message (which causes us to exit).
             */
            (void) fprintf (stderr, "\n%s\n", (*tdb_ptr).tdb_srclin_ptrpc);

            (void) fprintf (stderr, "%s\n", (*tdb_ptr).tdb_errlin_ptrpc);

            if ((*tdb_ptr).tdb_listn)
            {
                iret = LIST_LINE ((*tdb_ptr).tdb_list_ctx_ptrps,
                                  (*tdb_ptr).tdb_errlin_ptrpc);
                if (!(psl_test_status (iret)))
                    /* If it failed */
                {
                    PSL_SIGDIAG (   /* Signal Diagnostic */
                        (*tdb_ptr).tdb_diag_ptrps,
                        /* Use this stack */
                        1,      /* Number of message segments */
                        PSL_ENDABORT,   /* Message */
                        (*tdb_ptr).tdb_expsrcnamc256);
                    /* Parameter */
                    goto end;
                }

            }

        }

        /*
         *  Loop through the array of errors and report them.
         */
        for (jsub = 0; jsub < (*tdb_ptr).tdb_errs_in_linel; jsub++)
        {
            PSL_PUTDIAG (       /* Add Diagnostic */
                (*tdb_ptr).tdb_diag_ptrps,
                /* Use this stack */
                1,          /* Number of message segments */
                (*tdb_ptr).tdb_err_arrayl5[jsub]);
            /* Message */
            PSL_OUTPUT_DIAG (       /* Output Diagnostic */
                (*tdb_ptr).tdb_diag_ptrps,
                /* From this stack */
                0,          /* No Context */
                DIOUFL_LSTERR);     /* Output Last Error */
        }

        /*
         *  Reset counter and underline flag.
         */
        (*tdb_ptr).tdb_errs_in_linel = 0;
        (*tdb_ptr).tdb_undlin_flagn = PFALSE;

    }

    /*
     *  Read the next line.
     */
    READ_NEXT_LINE (tdb_ptr,
                    &(*tdb_ptr).tdb_srclin_lenl,
                    (*tdb_ptr).tdb_srclin_ptrpc,
                    &(*tdb_ptr).tdb_prslin_lenl,
                    (*tdb_ptr).tdb_prslin_ptrpc,
                    (*tdb_ptr).tdb_errlin_ptrpc,
                    &(*tdb_ptr).tdb_eof_flagn);

    /*
     *  If we got end of file return failure.
     */
    if ((*tdb_ptr).tdb_eof_flagn)
    {
        return 0;
    }

    /*
     *  Set up to parse the new line.
     */
    (*argblk_ptr).string_ptrpc = (*tdb_ptr).tdb_prslin_ptrpc;
    (*argblk_ptr).stringcntl = (*tdb_ptr).tdb_prslin_lenl;

    /*
     *  List the new source line.
     */
    TPA_LIST_LINE (tdb_ptr, (*tdb_ptr).tdb_srclin_ptrpc);

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return 1;
}


/************************************************************************/
/*                                  */
/*  TPA_GET_CONT_LINE - Get continued line from the Source File */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_GET_CONT_LINE (argblk_ptr);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine must be called from a transition specifying    */
/*  a continuation character as the transition type. It checks  */
/*  that the current token is the last token on the line before */
/*  reading the next source line. It is implemented as a shell  */
/*  around the TPA_GET_NEXT_LINE routine. If the current token  */
/*  is NOT the last token on the line the routine returns a     */
/*  failure status and does NOT read the next source line,      */
/*  otherwise it calls the TPA_GET_NEXT_LINE routine to read    */
/*  the next source line and returns with the Action Routine    */
/*  status from TPA_GET_NEXT_LINE.                  */
/*                                  */
/*  Note that any trailing comments on a source line are removed    */
/*  when the line is read. This allows us to do a simple 'end of    */
/*  line' check here to determine whether the continuation      */
/*  character is positioned correctly (the continuation character   */
/*  may come before or after a trailing comment).           */
/*                                  */
/************************************************************************/

unsigned long TPA_GET_CONT_LINE (struct tpaarg *argblk_ptr)
{

    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Action Routine return status */

    /*
     *  If the current token is the last on the current line, get the next
     *  source line using GET_NEXT_LINE, otherwise return a failure status.
     */
    if ((*argblk_ptr).stringcntl == 0)  /* If last token on line */
    {
        iret = TPA_GET_NEXT_LINE (argblk_ptr);
    }
    else
    {
        iret = 0;           /* Return failure status */
    }

    /*
     *  Return with Action Routine return status in 'iret'.
     */
    return iret;
}



/************************************************************************/
/*                                  */
/*  TPA_INC_ERR_COUNT - Increment error count           */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     TPA_INC_ERR_COUNT (tdb_ptr);                 */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr = Pointer to TPA Structure               */
/*                                  */
/*  This routine increments the error count in the TDB structure.   */
/*                                  */
/************************************************************************/

void TPA_INC_ERR_COUNT (struct tdb *tdb_ptr)
{
    (*tdb_ptr).tdb_err_countl += 1;
    return;
}


/************************************************************************/
/*                                  */
/*  TPA_INIT_SOURCE - Initialize source file            */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_INIT_SOURCE (tdb_ptr, src_name_ptr);      */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret         = Unsigned Long Action Routine Return Status    */
/*     tdb_ptr      = Pointer to TPA Structure          */
/*     src_name_ptr = Pointer to name of source file        */
/*                                  */
/*  This routine initializes the handline of a source file.     */
/*  It performs the following:                  */
/*                                  */
/*      1. Opens the source file                    */
/*      2. Resets the TDB structures                */
/*                                  */
/************************************************************************/

unsigned long TPA_INIT_SOURCE (struct tdb *tdb_ptr,
                               char *src_name_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret = PSL_NORMAL;  /* Return status */
    unsigned long       diag_save;      /* Saved diagnostic context */
    struct   diagstk    *diag_ptr;      /* Pointer to Diagnostic Stack */

    /*
     *  Copy pointers for ease of reference.
     */
    diag_ptr = (*tdb_ptr).tdb_diag_ptrps;

    /*
     *  First add a diagnostic to the stack.
     */
    diag_save = PSL_ADD_STATDIAG (diag_ptr, "During TPA_INIT_SOURCE");

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_GEN_TPARSE))
    {
        PSL_OUTPUT_TRACE ("Initializing source file \"%s\"\n", src_name_ptr);
    }

    /*
     *  Clear any C file errors hanging around.
     */
    PSL_CLEARERR ();

    /*
     *  Get the name of the file.
     */
    iret = PSL_FILE_PARSE (     /* Parse filespec */
               src_name_ptr,       /* This File */
               NULL,           /* No Default File */
               NULL,           /* No Related File */
               NULL,           /* No Expected Related Extension */
               (*tdb_ptr).tdb_expsrcnamc256,
               /* Returned File */
               sizeof ((*tdb_ptr).tdb_expsrcnamc256),
               /* Maximum length thereof */
               FILPRS_INPUT);      /* Flags */

    if (!(psl_test_status (iret)))  /* If an error */
    {
        PSL_PUTDIAG (           /* Signal Diagnostic */
            diag_ptr,       /* Use this diagnostic stack */
            2,          /* Number of message segments */
            PSL_OPENIN,     /* First message */
            (*tdb_ptr).tdb_expsrcnamc256,
            /* Parameter */
            iret);          /* Secondary message */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            diag_ptr,       /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        iret = PSL_OPENIN;      /* Set up error return */
        goto end;           /* And exit */
    }

    /*
     *  Open the file.
     */
    (*tdb_ptr).tdb_srcfil_ptrpf = fopen ((*tdb_ptr).tdb_expsrcnamc256, "r");
    if ((*tdb_ptr).tdb_srcfil_ptrpf == NULL)
    {
        iret = PSL_OPENIN;      /* Set up error */
        PSL_LOG_CIOERR (        /* Log file error */
            diag_ptr,       /* Use this stack */
            iret,           /* Error to report */
            (*tdb_ptr).tdb_expsrcnamc256);
        /* Filespec */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        goto end;           /* And exit */
    }

    /*
     *  Get the full name.
     */
    PSL_GET_FILENAME ((*tdb_ptr).tdb_srcfil_ptrpf,
                      (*tdb_ptr).tdb_expsrcnamc256,
                      sizeof ((*tdb_ptr).tdb_expsrcnamc256));

    /*
     *  Reset TDB structure.
     */
    RESET_STRUCT (tdb_ptr);

    /*
     *  Reset the diagnostics.
     */
    PSL_REMOVE_DIAGNOSTICS (diag_ptr, diag_save);

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return iret;
}


/************************************************************************/
/*                                  */
/*  TPA_LIST_LINE - Output Source Line to Listing File      */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     TPA_LIST_LINE (tdb_ptr, buf_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr = Pointer to TPA Structure               */
/*     buf_ptr = Pointer to Buffer to write             */
/*                                  */
/*  This routine outputs a line from the source file to the     */
/*  listing file.                           */
/*                                  */
/************************************************************************/

void TPA_LIST_LINE (struct tdb *tdb_ptr,
                    char *buf_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Status Return */

    /*
     *  Print the line if the Listing option is active.
     */
    if ((*tdb_ptr).tdb_listn)
    {
        iret = LIST_TESTPAGE ((*tdb_ptr).tdb_list_ctx_ptrps, 3);
        if (!(psl_test_status (iret))) goto err_exit;

        iret = LIST_LINE_REFERENCED ((*tdb_ptr).tdb_list_ctx_ptrps, buf_ptr);
        if (!(psl_test_status (iret))) goto err_exit;
    }

    goto end;

    /*
     *  ERR_EXIT:
     *
     *  An error from a Listing routine will already have been reported
     *  so we just signal the ENDABORT message (which causes us to exit).
     */
err_exit:
    PSL_SIGDIAG (           /* Signal Diagnostic */
        (*tdb_ptr).tdb_diag_ptrps,  /* Use this stack */
        1,              /* Number of message segments */
        PSL_ENDABORT,           /* Message */
        (*tdb_ptr).tdb_expsrcnamc256);  /* Parameter */
    /****    goto end;***/

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return;
}


/************************************************************************/
/*                                  */
/*  TPA_OUTPUT_SUM - Output summary message             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     TPA_OUTPUT_SUM (tdb_ptr);                    */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr = Pointer to TPA Structure               */
/*                                  */
/*  This routine outputs a summary message after compilation.   */
/*                                  */
/*  The message it may output is defined as follows:        */
/*                                  */
/*      ENDNOOBJ    if errors and /OBJECT specified         */
/*      ENDDIAGS    if errors and /OBJECT not specified     */
/*      COMPILED    if no errors and /LOG specified but /OBJECT */
/*          not specified                   */
/*                                  */
/************************************************************************/

void TPA_OUTPUT_SUM (struct tdb *tdb_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Status Return */

    if ((*tdb_ptr).tdb_err_countl != 0) /* If errors detected */
    {
        if ((*tdb_ptr).tdb_objectn) /* If /OBJECT specified */
        {
            iret = PSL_ENDNOOBJ;    /* No object file produced */
        }
        else                /* Otherwise */
        {
            iret = PSL_ENDDIAGS;    /* Completed with Diagnostics */
        }
        PSL_PUTDIAG (           /* Add Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            iret,           /* Message */
            (*tdb_ptr).tdb_expsrcnamc256,
            /* First Parameter */
            (*tdb_ptr).tdb_err_countl);
        /* Second Parameter */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Output Last Error */
    }
    else                /* No Errors */
    {
        if (!(*tdb_ptr).tdb_objectn /* If /OBJECT not specified */
                && (*tdb_ptr).tdb_logn)    /* ... and /LOG specified */
        {
            PSL_PUTDIAG (       /* Add Diagnostic */
                (*tdb_ptr).tdb_diag_ptrps,
                /* Use this stack */
                1,          /* Number of message segments */
                PSL_COMPILED,       /* Message */
                (*tdb_ptr).tdb_expsrcnamc256);
            /* Parameter */
            PSL_OUTPUT_DIAG (       /* Output Diagnostics */
                (*tdb_ptr).tdb_diag_ptrps,
                /* Use this stack */
                0,          /* No Context */
                DIOUFL_LSTERR);     /* Output Last Error */
        }
    }

    return;
}


/************************************************************************/
/*                                  */
/*  TPA_PREMATURE_EOF - Check for premature End of File     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = TPA_PREMATRURE_EOF (argblk_ptr);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long Action Routine Return Status  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine checks the End of File flag and, if set,       */
/*  reports a 'Premature End of File' error.            */
/*                                  */
/************************************************************************/

unsigned long TPA_PREMATURE_EOF (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    struct   tdb        *tdb_ptr;       /* Pointer to TDB structure */

    /*
     *  Copy pointers for ease of reference.
     */
    tdb_ptr = (struct tdb *) (*argblk_ptr).user_ptrp;

    if ((*tdb_ptr).tdb_eof_flagn)   /* If End of File */
    {
        PSL_PUTDIAG (           /* Add Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_PREMEOF);       /* This message */
        PSL_OUTPUT_DIAG (       /* Output Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* From this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Output Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment Error Count */
        return 1;           /* Signal success */
    }
    else                /* Otherwise */
    {
        return 0;           /* Signal failure */
    }
}





/************************************************************************/
/*                                  */
/*  TPA_SET_USRDAT_PTR - Set user data pointer          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     TPA_SET_USRDAT_PTR (tdb_ptr, usr_ptr);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr = Pointer to TPA Structure               */
/*     usr_ptr = Pointer to void user data pointer          */
/*                                  */
/*  This routine sets the user data pointer in the TPA structure.   */
/*                                  */
/************************************************************************/

void TPA_SET_USRDAT_PTR (struct tdb *tdb_ptr,
                         void *usr_ptr)
{
    (*tdb_ptr).tdb_usrdat_ptrp = usr_ptr;
    return;
}




/************************************************************************/
/*                                  */
/*  CHK_CONDITIONAL - Check conditional state           */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     retval = CHK_CONDITIONAL (tdb_ptr);              */
/*                                  */
/*  Where:                              */
/*                                  */
/*     retval  = Boolean returned conditional flag          */
/*           PTRUE if line is to be ignored         */
/*           PFALSE otherwise                   */
/*     tdb_ptr = Pointer to TPA Structure               */
/*                                  */
/*  This routine checks to see if the current line should be    */
/*  included in the output or not.                  */
/*                                  */
/*  The routine scans through the 'if' stack and tries to match */
/*  the specified variant with those defined by the user.       */
/*                                  */
/*  If there is no match and the variant action is 'include'    */
/*  then the line is rejected.                  */
/*                                  */
/*  If there is a match and the variant action is 'exclude'     */
/*  then the line is rejected.                  */
/*                                  */
/************************************************************************/

static Boolean CHK_CONDITIONAL (struct tdb *tdb_ptr)
{
    /*
     *  Local Data.
     */
    Boolean         retval;     /* Return value */
    long       match_status;   /* Status from strcmp */
    long       jsub;       /* Loop control */
    unsigned long       ksub;       /* Loop control */
    char       *if_name_ptr;   /* Pointer to string */
    char       **var_ptr_ptr;  /* Pointer to pointer to string */
    struct   ifs        *ifs_ptr;       /* Pointer to Conditional Compilation */
    /* ... stack entry */

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Checking Conditional state...\n");
    }

    /*
     *  Assume line is to be INCLUDED.
     */
    retval = PFALSE;

    /*
     *  Loop through the 'if' stack.
     */
    for (jsub = 0; jsub <= (*tdb_ptr).tdb_if_levell; jsub++)
    {

        /*
         *  Copy pointers for ease of reference.
         */
        ifs_ptr = &(*tdb_ptr).tdb_if_stkr10 [jsub];
        if_name_ptr = (*ifs_ptr).ifs_namec20;

        /*
         *  Loop through the list of variants and try to get a match.
         */
        for (ksub = 0, var_ptr_ptr = (*tdb_ptr).tdb_variant_listppc;
                ksub < (*tdb_ptr).tdb_variant_countl;
                ksub++, var_ptr_ptr++)
        {
            match_status = strcmp (if_name_ptr, *var_ptr_ptr);

            if (match_status == 0)  /* If we have a match */
            {

                /*
                 *  Output trace information.
                 */
                if (PSL_TEST_TRACE (PTRC_USR_USER2))
                {
                    PSL_OUTPUT_TRACE ("Match found for Variant : \"%s\"\n",
                                      if_name_ptr);
                }

                /*
                 *  If the action for this variant is 'exclude' then return
                 *  an appropriate value to the caller.
                 */
                if ((*ifs_ptr).ifs_actionl == TDB_IFAC_EXC)
                {
                    retval = PTRUE; /* Ignore line */
                    goto end;       /* And exit */
                }
                else
                {
                    goto next_if;   /* Check next variant */
                }
            }               /* End of if match */

        }               /* End of loop through variants */

        /*
         *  Output trace information.
         */
        if (PSL_TEST_TRACE (PTRC_USR_USER2))
        {
            PSL_OUTPUT_TRACE ("No Match for Variant : \"%s\"\n", if_name_ptr);
        }

        /*
         *  We couldn't find a match for the variant.
         *  If the action for this variant is 'include' then return
         *  an appropriate value to the caller.
         */
        if ((*ifs_ptr).ifs_actionl == TDB_IFAC_INC)
        {
            retval = PTRUE;     /* Ignore line */
            goto end;           /* And exit */
        }

        /*
         *  NEXT_IF:
         *
         *  We branch here if we have a match for the current variant.
         */
next_if:
        ;
    }                   /* End of loop through 'if' stack */

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    /*
     *    Check if OS or PLATFORM specific code was parsed.
     *    Relevant flags in the TDB structure are set.
     *
     *    N.B. It is the responsibilty of the user of these routines
     *    to clear the TDB flags as necessary.
     */

    /*
     *  First the Operating System.
     */
#ifdef UNIX
    if (strncmp((*ifs_ptr).ifs_namec20, "%UNIX",5) ==0)
    {
        (*tdb_ptr).tdb_opsys_spec_coden = PTRUE;
    }
#elif VMS
if (strncmp((*ifs_ptr).ifs_namec20, "%VMS",4) == 0)
    {
        (*tdb_ptr).tdb_opsys_spec_coden = PTRUE;
    }

#elif WIN32
    if (strncmp((*ifs_ptr).ifs_namec20, "%WIN32",6) == 0)
    {
        (*tdb_ptr).tdb_opsys_spec_coden = PTRUE;

    }

#else
#error Platform unsupported
#endif

    /*
     *  Now the Platform.
     */
#ifdef VAX
    if (strncmp((*ifs_ptr).ifs_namec20, "%VAX",4)==0)
    {
        (*tdb_ptr).tdb_plat_spec_coden = PTRUE;
    }
#elif ALPHA
if (strncmp((*ifs_ptr).ifs_namec20, "%ALPHA",6)==0)
    {
        (*tdb_ptr).tdb_plat_spec_coden = PTRUE;
    }
#elif HP
    if (strncmp((*ifs_ptr).ifs_namec20, "%HP",3)==0)
    {
        (*tdb_ptr).tdb_plat_spec_coden = PTRUE;
    }
#elif IBM
    if (strncmp((*ifs_ptr).ifs_namec20, "%IBM",4)==0)
    {
        (*tdb_ptr).tdb_plat_spec_coden = PTRUE;
    }
#elif NCR
    if (strncmp((*ifs_ptr).ifs_namec20, "%NCR",4)==0)
    {
        (*tdb_ptr).tdb_plat_spec_coden = PTRUE;
    }
#elif SUN
    if (strncmp((*ifs_ptr).ifs_namec20, "%SUN",4)==0)
    {
        (*tdb_ptr).tdb_plat_spec_coden = PTRUE;
    }

#elif NTI
    if (strncmp((*ifs_ptr).ifs_namec20, "%NTI",4)==0)
    {
        (*tdb_ptr).tdb_plat_spec_coden = PTRUE;
    }

#elif OSF
    if (strncmp((*ifs_ptr).ifs_namec20, "%OSF",4)==0)
    {
        (*tdb_ptr).tdb_plat_spec_coden = PTRUE;
    }

#else
#error Platform unsupported
#endif

    return retval;
}


/************************************************************************/
/*                                  */
/*  DO_CONDITIONAL - Process a conditional compilation directive    */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     retval = DO_CONDITIONAL (tdb_ptr, srclin_ptr, prslin_ptr);   */
/*                                  */
/*  Where:                              */
/*                                  */
/*     retval     = Boolean return value indicating whether or not  */
/*          the line contained a conditional compilation    */
/*          directive                   */
/*          PTRUE if conditional directive found        */
/*          PFALSE otherwise                */
/*     tdb_ptr    = Pointer to TPA Structure            */
/*     srclin_ptr = Pointer to source line              */
/*     prslin_ptr = Pointer to parsable line            */
/*                                  */
/*  This routine checks for and processes the conditional       */
/*  compilation directives.                     */
/*                                  */
/************************************************************************/

static Boolean DO_CONDITIONAL (struct tdb *tdb_ptr,
                               char *srclin_ptr,
                               char *prslin_ptr)
{
    /*
     *  Local Data.
     */
    Boolean         retval;     /* Return value */

    /*
     *  Assume a conditional compilation directive.
     */
    retval = PTRUE;

    /*
     *  Check for each one in turn.
     */
    if (memcmp (prslin_ptr, "#IFDEF", 6) == 0)
        /* If #IFDEF */
    {
        TPA_LIST_LINE (tdb_ptr, srclin_ptr);
        /* List the line */
        DO_IFDEF (tdb_ptr, srclin_ptr, prslin_ptr);
        /* Handle it */
    }
    else if (memcmp (prslin_ptr, "#IFNDEF", 7) == 0)
        /* Else if #IFNDEF */
    {
        TPA_LIST_LINE (tdb_ptr, srclin_ptr);
        /* List the line */
        DO_IFNDEF (tdb_ptr, srclin_ptr, prslin_ptr);
        /* Handle it */
    }
    else if (memcmp (prslin_ptr, "#ELIF", 5) == 0)
        /* Else if #ELIF */
    {
        TPA_LIST_LINE (tdb_ptr, srclin_ptr);
        /* List the line */
        DO_ELIF (tdb_ptr, srclin_ptr, prslin_ptr);
        /* Handle it */
    }
    else if (memcmp (prslin_ptr, "#ELSE", 5) == 0)
        /* Else if #ELSE */
    {
        TPA_LIST_LINE (tdb_ptr, srclin_ptr);
        /* List the line */
        DO_ELSE (tdb_ptr, srclin_ptr, prslin_ptr);
        /* Handle it */
    }
    else if (memcmp (prslin_ptr, "#ENDIF", 6) == 0)
        /* Else if #ENDIF */
    {
        TPA_LIST_LINE (tdb_ptr, srclin_ptr);
        /* List the line */
        DO_ENDIF (tdb_ptr, srclin_ptr, prslin_ptr);
        /* Handle it */
    }
    else                /* Not a conditional directive */
    {
        retval = PFALSE;        /* Signal the fact */
    }

    return retval;
}


/************************************************************************/
/*                                  */
/*  DO_IFDEF - Process a #IFDEF directive               */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     DO_IFDEF (tdb_ptr, srclin_ptr, prslin_ptr);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr    = Pointer to TPA Structure            */
/*     srclin_ptr = Pointer to source line              */
/*     prslin_ptr = Pointer to parsable line            */
/*                                  */
/*  This routine processes a #IFDEF directive.          */
/*                                  */
/************************************************************************/

static void DO_IFDEF (struct tdb *tdb_ptr,
                      char *srclin_ptr,
                      char *prslin_ptr)
{
    /*
     *  Local Data.
     */
    Boolean         retval;     /* Return value */
    char       var_name [256]; /* Work Buffer */
    struct   ifs        *ifs_ptr;       /* Pointer to Conditional Compilation */
    /* ... stack entry */

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Processing #IFDEF...\n");
    }

    /*
     *  Check that we have not exceeded the maximum "IF" nesting.
     */
    if ((*tdb_ptr).tdb_if_levell == (TDB_MAX_IF_LEVEL - 1))
    {
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_EXCONDDEPTH,    /* Message */
            (*tdb_ptr).tdb_line_countl,
            /* First Parameter */
            (*tdb_ptr).tdb_expsrcnamc256);
        /* Second Parameter */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
        goto end;           /* And exit */
    }

    /*
     *  Parse the input line buffer starting after the directive to extract
     *  the name of the variant.
     */
    retval = GET_VARIANT (tdb_ptr, srclin_ptr, prslin_ptr+6, var_name);
    if (!retval) goto end;      /* Exit if some error */

    /*
     *  Increment the nesting level and save the specified value and
     *  initialize the action to 'include'.
     */
    (*tdb_ptr).tdb_if_levell += 1;  /* Increment nesting level */

    ifs_ptr = &(*tdb_ptr).tdb_if_stkr10 [(*tdb_ptr).tdb_if_levell];
    /* Set up ptr for ease of reference */

    (void) strcpy ((*ifs_ptr).ifs_namec20, var_name);
    /* Save name */
    (*ifs_ptr).ifs_actionl = TDB_IFAC_INC;
    /* Set action = INCLUDE */
    (*ifs_ptr).ifs_flagsl = 0;      /* Clear all flags */

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("...new level = %d\n", (*tdb_ptr).tdb_if_levell);
        PSL_OUTPUT_TRACE ("...variant = \"%s\"\n", (*ifs_ptr).ifs_namec20);
    }

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return;
}


/************************************************************************/
/*                                  */
/*  DO_IFNDEF - Process a #IFNDEF directive             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     DO_IFNDEF (tdb_ptr, srclin_ptr, prslin_ptr);         */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr    = Pointer to TPA Structure            */
/*     srclin_ptr = Pointer to source line              */
/*     prslin_ptr = Pointer to parsable line            */
/*                                  */
/*  This routine processes a #IFNDEF directive.         */
/*                                  */
/************************************************************************/

static void DO_IFNDEF (struct tdb *tdb_ptr,
                       char *srclin_ptr,
                       char *prslin_ptr)
{
    /*
     *  Local Data.
     */
    Boolean         retval;     /* Return value */
    char       var_name [256]; /* Work Buffer */
    struct   ifs        *ifs_ptr;       /* Pointer to Conditional Compilation */
    /* ... stack entry */

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Processing #IFNDEF...\n");
    }

    /*
     *  Check that we have not exceeded the maximum "IF" nesting.
     */
    if ((*tdb_ptr).tdb_if_levell == (TDB_MAX_IF_LEVEL - 1))
    {
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_EXCONDDEPTH,    /* Message */
            (*tdb_ptr).tdb_line_countl,
            /* First Parameter */
            (*tdb_ptr).tdb_expsrcnamc256);
        /* Second Parameter */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
        goto end;           /* And exit */
    }

    /*
     *  Parse the input line buffer starting after the directive to extract
     *  the name of the variant.
     */
    retval = GET_VARIANT (tdb_ptr, srclin_ptr, prslin_ptr+7, var_name);
    if (!retval) goto end;      /* Exit if some error */

    /*
     *  Increment the nesting level and save the specified value and
     *  initialize the action to 'exclude'.
     */
    (*tdb_ptr).tdb_if_levell += 1;  /* Increment nesting level */

    ifs_ptr = &(*tdb_ptr).tdb_if_stkr10 [(*tdb_ptr).tdb_if_levell];
    /* Set up ptr for ease of reference */

    (void) strcpy ((*ifs_ptr).ifs_namec20, var_name);
    /* Save name */
    (*ifs_ptr).ifs_actionl = TDB_IFAC_EXC;
    /* Set action = EXCLUDE */
    (*ifs_ptr).ifs_flagsl = 0;      /* Clear all flags */

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("...new level = %d\n", (*tdb_ptr).tdb_if_levell);
        PSL_OUTPUT_TRACE ("...variant = \"%s\"\n", (*ifs_ptr).ifs_namec20);
    }

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return;
}


/************************************************************************/
/*                                  */
/*  DO_ELIF - Process a #ELIF directive             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     DO_ELIF (tdb_ptr, srclin_ptr, prslin_ptr);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr    = Pointer to TPA Structure            */
/*     srclin_ptr = Pointer to source line              */
/*     prslin_ptr = Pointer to parsable line            */
/*                                  */
/*  This routine processes a #ELIF directive.           */
/*                                  */
/************************************************************************/

static void DO_ELIF (struct tdb *tdb_ptr,
                     char *srclin_ptr,
                     char *prslin_ptr)
{
    /*
     *  Local Data.
     */
    Boolean         retval;     /* Return value */
    char       var_name [256]; /* Work Buffer */
    struct   ifs        *ifs_ptr;       /* Pointer to Conditional Compilation */
    /* ... stack entry */

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Processing #ELIF...\n");
    }

    /*
     *  Copy pointers for ease of reference.
     */
    ifs_ptr = &(*tdb_ptr).tdb_if_stkr10 [(*tdb_ptr).tdb_if_levell];

    /*
     *  Check that we are inside a conditional block and that we haven't
     *  already had a #ELSE directive.
     */
    if (((*tdb_ptr).tdb_if_levell == -1)
            /* If not in conditional block */
            || (((*ifs_ptr).ifs_flagsl & TDB_IFFL_ELSE) != 0))
        /* or #ELSE already specified */
    {
        (void) fprintf (stderr, "\n%s\n", srclin_ptr);
        /* Output line in question */
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_BADCOND,        /* Message */
            (*tdb_ptr).tdb_line_countl,
            /* First Parameter */
            (*tdb_ptr).tdb_expsrcnamc256);
        /* Second Parameter */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
        goto end;           /* And exit */
    }

    /*
     *  Parse the input line buffer starting after the directive to extract
     *  the name of the variant.
     */
    retval = GET_VARIANT (tdb_ptr, srclin_ptr, prslin_ptr+5, var_name);
    if (!retval) goto end;      /* Exit if some error */

    /*
     *  Save the specified value and initialize the action to 'include'.
     */
    (void) strcpy ((*ifs_ptr).ifs_namec20, var_name);
    /* Save name */
    (*ifs_ptr).ifs_actionl = TDB_IFAC_INC;
    /* Set action = INCLUDE */
    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("...current level = %d\n", (*tdb_ptr).tdb_if_levell);
        PSL_OUTPUT_TRACE ("...new variant = \"%s\"\n", (*ifs_ptr).ifs_namec20);
    }

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return;
}


/************************************************************************/
/*                                  */
/*  DO_ELSE - Process a #ELSE directive             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     DO_ELSE (tdb_ptr, srclin_ptr, prslin_ptr);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr    = Pointer to TPA Structure            */
/*     srclin_ptr = Pointer to source line              */
/*     prslin_ptr = Pointer to parsable line            */
/*                                  */
/*  This routine processes a #ELSE directive.           */
/*                                  */
/************************************************************************/

static void DO_ELSE (struct tdb *tdb_ptr,
                     char *srclin_ptr,
                     char *prslin_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       length;     /* Buffer Length */
    struct   ifs        *ifs_ptr;       /* Pointer to Conditional Compilation */
    /* ... stack entry */

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Processing #ELSE\n");
    }

    /*
     *  Copy pointers for ease of reference.
     */
    ifs_ptr = &(*tdb_ptr).tdb_if_stkr10 [(*tdb_ptr).tdb_if_levell];

    /*
     *  Check that we are inside a conditional block and that we haven't
     *  already had a #ELSE directive.
     */
    if (((*tdb_ptr).tdb_if_levell == -1)
            /* If not in conditional block */
            || (((*ifs_ptr).ifs_flagsl & TDB_IFFL_ELSE) != 0))
        /* or #ELSE already specified */
    {
        (void) fprintf (stderr, "\n%s\n", srclin_ptr);
        /* Output line in question */
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_BADCOND,        /* Message */
            (*tdb_ptr).tdb_line_countl,
            /* First Parameter */
            (*tdb_ptr).tdb_expsrcnamc256);
        /* Second Parameter */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
        goto end;           /* And exit */
    }

    /*
     *  Report an error if there is anything else on the line.
     */
    length = strlen (prslin_ptr+5); /* Get length of any extra text */
    if (length != 0)            /* If there is some */
    {
        (void) fprintf (stderr, "\n%s\n", srclin_ptr);
        /* Output line in question */
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_EXTTXT);        /* Message */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
        goto end;           /* And exit */
    }

    /*
     *  Reset the action to the opposite of whatever it currently is.
     */
    if ((*ifs_ptr).ifs_actionl == TDB_IFAC_INC)
        /* If action = INCLUDE */
    {
        (*ifs_ptr).ifs_actionl = TDB_IFAC_EXC;
        /* Reset to EXCLUDE */
    }
    else                /* Else action = EXCLUDE */
    {
        (*ifs_ptr).ifs_actionl = TDB_IFAC_INC;
        /* Reset to INCLUDE */
    }

    /*
     *  Set the flag that indicates we have had a #ELSE directive.
     */
    (*ifs_ptr).ifs_flagsl = TDB_IFFL_ELSE;

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("...current level = %d\n", (*tdb_ptr).tdb_if_levell);
        PSL_OUTPUT_TRACE ("...variant = \"%s\"\n", (*ifs_ptr).ifs_namec20);
    }

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return;
}


/************************************************************************/
/*                                  */
/*  DO_ENDIF - Process a #ENDIF directive               */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     DO_ENDIF (tdb_ptr, srclin_ptr, prslin_ptr);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr    = Pointer to TPA Structure            */
/*     srclin_ptr = Pointer to source line              */
/*     prslin_ptr = Pointer to parsable line            */
/*                                  */
/*  This routine processes a #ENDIF directive.          */
/*                                  */
/************************************************************************/

static void DO_ENDIF (struct tdb *tdb_ptr,
                      char *srclin_ptr,
                      char *prslin_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       length;     /* Buffer Length */

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("Processing #ENDIF\n");
    }

    /*
     *  Check that we are inside a conditional block.
     */
    if ((*tdb_ptr).tdb_if_levell == -1)
    {
        (void) fprintf (stderr, "\n%s\n", srclin_ptr);
        /* Output line in question */
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_BADCOND,        /* Message */
            (*tdb_ptr).tdb_line_countl,
            /* First Parameter */
            (*tdb_ptr).tdb_expsrcnamc256);
        /* Second Parameter */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
        goto end;           /* And exit */
    }

    /*
     *  Report an error if there is anything else on the line.
     */
    length = strlen (prslin_ptr+6); /* Get length of any extra text */
    if (length != 0)            /* If there is some */
    {
        (void) fprintf (stderr, "\n%s\n", srclin_ptr);
        /* Output line in question */
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_EXTTXT);        /* Message */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
        goto end;           /* And exit */
    }

    (*tdb_ptr).tdb_if_levell -= 1;  /* Decrement nesting level */

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_USR_USER2))
    {
        PSL_OUTPUT_TRACE ("...new level = %d\n", (*tdb_ptr).tdb_if_levell);
    }

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return;
}





/************************************************************************/
/*                                  */
/*  DO_ERROR - Process a #ERROR directive               */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     DO_ERROR (tdb_ptr, srclin_ptr, prslin_ptr);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr    = Pointer to TPA Structure            */
/*     srclin_ptr = Pointer to line as input            */
/*     prslin_ptr = Pointer to line minus comments and uppercased   */
/*                                  */
/*  This routine processes a #ERROR directive.          */
/*                                  */
/************************************************************************/

static void DO_ERROR (struct tdb *tdb_ptr,
                      char *srclin_ptr,
                      char *prslin_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       diag_save;      /* Saved diagnostic context */
    struct   diagstk    *diag_ptr;      /* Pointer to Diagnostic Stack */
    unsigned long       length;     /* Buffer Length */
    char       *wrk_ptr;       /* Work Pointer */
    char       wrkbuf [256];   /* Work Buffer */

    /*
     *  Copy pointers for ease of reference.
     */
    diag_ptr = (*tdb_ptr).tdb_diag_ptrps;

    /*
     *  First add a diagnostic to the stack.
     */
    diag_save = PSL_ADD_STATDIAG (diag_ptr, "During DO_ERROR");

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_GEN_TPARSE))
    {
        PSL_OUTPUT_TRACE ("Handling #ERROR directive\n");
    }

    /*
     *  Parse the input line buffer to extract the message text.
     *  First, locate the start of the text.
     */
    wrk_ptr = prslin_ptr + 6;       /* Start after "#error" */
    length = strspn (wrk_ptr, " \t");   /* Get length of leading whitespace */

    /*
     *  Check that the first character is a double quote.
     */
    wrk_ptr += length;          /* Point to first char of message */

    if (*wrk_ptr != '\"') goto bad_message;

    (void) strcpy (wrkbuf, wrk_ptr+1);  /* Copy string to work buffer */

    /*
     *  Now locate the end of the message.
     */
    length = strcspn (wrkbuf, " \t\0"); /* Get length of message */

    wrk_ptr = wrkbuf + length - 1;  /* Point to last char of message */

    if (*wrk_ptr != '\"') goto bad_message;
    /* Check it's a double quote */
    *wrk_ptr = '\0';            /* Null terminate */

    /*
     *  Output the text as part of a message.
     */
    PSL_PUTDIAG (           /* Signal Diagnostic */
        diag_ptr,       /* Use this diagnostic stack */
        1,          /* Number of message segments */
        PSL_USERERROR,      /* Message */
        wrkbuf);        /* Parameter */

    PSL_OUTPUT_DIAG (           /* Output Diagnostics */
        diag_ptr,       /* Use this stack */
        0,          /* No Context */
        DIOUFL_LSTERR);     /* Last Error */

    (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */

    /*
     *  Reset the diagnostics.
     */
    PSL_REMOVE_DIAGNOSTICS (diag_ptr, diag_save);

    goto end;

    /*
     *  BAD_MESSAGE:
     *
     *  Report 'Invalid Message' error.
     */
bad_message:
    (void) fprintf (stderr, "\n%s\n", srclin_ptr);
    /* Output line in question */
    PSL_PUTDIAG (           /* Put Diagnostic */
        diag_ptr,       /* Use this stack */
        1,          /* Number of message segments */
        PSL_BADERRMSG,      /* Message */
        (*tdb_ptr).tdb_line_countl,
        /* First Parameter */
        (*tdb_ptr).tdb_expsrcnamc256);
    /* Second Parameter */
    PSL_OUTPUT_DIAG (           /* Output Diagnostics */
        diag_ptr,       /* Use this stack */
        0,          /* No Context */
        DIOUFL_LSTERR);     /* Last Error */
    (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return;
}




/************************************************************************/
/*                                  */
/*  DO_INCLUDE - Process a #INCLUDE directive           */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     DO_INCLUDE (tdb_ptr, srclin_ptr, prslin_ptr);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr    = Pointer to TPA Structure            */
/*     srclin_ptr = Pointer to line as input            */
/*     prslin_ptr = Pointer to line minus comments and uppercased   */
/*                                  */
/*  This routine processes a #INCLUDE directive.            */
/*                                  */
/************************************************************************/

static void DO_INCLUDE (struct tdb *tdb_ptr,
                        char *srclin_ptr,
                        char *prslin_ptr)
{
    /*
     *  Local Data.
     */
    Boolean         retval;     /* Return value */
    char       tmp_name [256]; /* Temporary storage buffer */

    /*
     *  Check that we have not exceeded the maximum include file nesting.
     */
    if ((*tdb_ptr).tdb_inc_levell == TDB_MAX_INC_LEVEL)
    {
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_EXINCDEPTH,     /* Message */
            (*tdb_ptr).tdb_line_countl,
            /* First Parameter */
            (*tdb_ptr).tdb_expsrcnamc256);
        /* Second Parameter */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
        retval = PFALSE;        /* Signal failure */
        goto end;           /* And exit */
    }

    /*
     *  Open the include file.
     */
    retval = OPEN_INCLUDE (     /* Open Include File */
                 tdb_ptr,
                 prslin_ptr, /* Get name from this */
                 srclin_ptr, /* Report errors from this */
                 &(*tdb_ptr).tdb_file_ptr_stkpf8
                 [(*tdb_ptr).tdb_inc_levell],
                 /* Put pointer here */
                 tmp_name);  /* Put name here */

    /*
     *  If it worked we store the current information on the appropriate
     *  stack and reset for the new source file.
     */
    if (retval)             /* If it succeeded */
    {
        (*tdb_ptr).tdb_line_count_stkl8 [(*tdb_ptr).tdb_inc_levell] =
            (*tdb_ptr).tdb_line_countl;
        /* Save line count */

        (void) strcpy ((*tdb_ptr).tdb_file_name_stkc256_8
                       [(*tdb_ptr).tdb_inc_levell],
                       (*tdb_ptr).tdb_expsrcnamc256);
        /* Save file name */

        (*tdb_ptr).tdb_line_countl = 0; /* Reset Line Count */

        (void) strcpy ((*tdb_ptr).tdb_expsrcnamc256, tmp_name);
        /* Reset file name */

        (*tdb_ptr).tdb_inc_levell += 1; /* Increment Include nesting level */
    }

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return;
}





/************************************************************************/
/*                                  */
/*  EXTEND_TOKEN_BLANK - Extend Token to Next Blank         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     EXTEND_TOKEN_BLANK (argblk_ptr);             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine extends the current token to the next space,   */
/*  non-printable character or end of line, whichever comes     */
/*  first.                              */
/*                                  */
/************************************************************************/

static void EXTEND_TOKEN_BLANK (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       tokencnt;       /* Local copy of tokencntl */
    char       *wrk_ptr;       /* Work Pointer */

    /*
     *  Copy Token Count for manipulation.
     */
    tokencnt = (*argblk_ptr).tokencntl;
    wrk_ptr  = (*argblk_ptr).token_ptrpc + 1;

    /*
     *  Test if this is at the end of the input String.
     */
    while (((*argblk_ptr).stringcntl > 0) &&
            (isgraph (*wrk_ptr) != 0))
    {
        tokencnt += 1;
        wrk_ptr += 1;
        (*argblk_ptr).string_ptrpc += 1;
        (*argblk_ptr).stringcntl -= 1;
    }

    (*argblk_ptr).tokencntl = tokencnt; /* Update Token Count */

    return;
}






/************************************************************************/
/*                                  */
/*  EXTEND_TOKEN_LINE - Extend Token to End of Line         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     EXTEND_TOKEN_LINE (argblk_ptr);              */
/*                                  */
/*  Where:                              */
/*                                  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine extends the current token to the end of the    */
/*  current line.                           */
/*                                  */
/************************************************************************/

static void EXTEND_TOKEN_LINE (struct tpaarg *argblk_ptr)
{
    (*argblk_ptr).tokencntl += (*argblk_ptr).stringcntl;
    (*argblk_ptr).string_ptrpc += (*argblk_ptr).tokencntl;
    (*argblk_ptr).stringcntl = 0;
    return;
}






/************************************************************************/
/*                                  */
/*  EXTEND_TOKEN_STD - Perform Standard Token Extension     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     EXTEND_TOKEN_STD (argblk_ptr);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine performs standard token extension. This extends    */
/*  the current token to the next non-printing character, space,    */
/*  user defined token separator or end of line, whichever comes    */
/*  first.                              */
/*                                  */
/************************************************************************/

static void EXTEND_TOKEN_STD (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       tokencnt;       /* Local copy of tokencntl */
    char       *wrk_ptr;       /* Work Pointer */
    struct   tdb        *tdb_ptr;       /* Pointer to TDB structure */

    /*
     *  Copy pointers for ease of reference.
     */
    tdb_ptr = (struct tdb *) (*argblk_ptr).user_ptrp;

    /*
     *  Copy Token Count for manipulation.
     */
    tokencnt = (*argblk_ptr).tokencntl;
    wrk_ptr  = (*argblk_ptr).token_ptrpc + 1;

    /*
     *  Test if this is at the end of the input String.
     */
    while (((*argblk_ptr).stringcntl > 0) &&
            (isgraph (*wrk_ptr) != 0) &&
            (*wrk_ptr != (*tdb_ptr).tdb_sep_charb4 [0]) &&
            (*wrk_ptr != (*tdb_ptr).tdb_sep_charb4 [1]) &&
            (*wrk_ptr != (*tdb_ptr).tdb_sep_charb4 [2]) &&
            (*wrk_ptr != (*tdb_ptr).tdb_sep_charb4 [3]))
    {
        tokencnt += 1;
        wrk_ptr += 1;
        (*argblk_ptr).string_ptrpc += 1;
        (*argblk_ptr).stringcntl -= 1;
    }

    (*argblk_ptr).tokencntl = tokencnt; /* Update Token Count */

    return;
}




/************************************************************************/
/*                                  */
/*  GET_VARIANT - Get variant name                  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     retval = GET_VARIANT (tdb_ptr, buf_ptr, srclin_ptr,      */
/*               varnam_ptr);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*     retval     = Boolean indicating whether or not a variant */
/*          name was extracted              */
/*          PTRUE if success                */
/*          PFALSE otherwise                */
/*     tdb_ptr    = Pointer to TPA Structure            */
/*     srclin_ptr = Pointer to source line              */
/*     buf_ptr    = Pointer to search position in parsable line */
/*     varnam_ptr = Pointer to buffer to hold returned variant  */
/*          name                        */
/*                                  */
/*  This routine is called when a Conditional Compilation directive */
/*  that requires a variant is encountered. It parses the specified */
/*  line and extracts the name of the variant.          */
/*  It reports an error if there is no variant name or if there */
/*  is anything after the variant name.             */
/*                                  */
/************************************************************************/

static Boolean GET_VARIANT (struct tdb *tdb_ptr,
                            char *srclin_ptr,
                            char *buf_ptr,
                            char *varnam_ptr)
{
    /*
     *  Local Data.
     */
    Boolean         retval;     /* Return value */
    unsigned long       length;     /* Buffer Length */
    char       var_name [256]; /* Work Buffer */
    char       *wrk_ptr;       /* Work Pointer */

    /*
     *  Assume failure so that we can drop out at any time after an error.
     */
    retval = PFALSE;

    /*
     *  Locate the start of the variant by stepping over any whitespace.
     */
    (void) memset (var_name, '\0', sizeof (var_name));
    /* Clear local buffer */
    length = strspn (buf_ptr, " \t");   /* Get length of leading whitespace */
    wrk_ptr = buf_ptr + length;     /* Point to first char of var. name */
    (void) strcpy (var_name, wrk_ptr);  /* Copy string to work buffer */

    length = strcspn (var_name, " \t\0");
    /* Get length of var. name */
    /*
     *  Report an error if we haven't got a variant name.
     */
    if (length == 0)
    {
        (void) fprintf (stderr, "\n%s\n", srclin_ptr);
        /* Output line in question */
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_PREMEOL);       /* Message */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
        goto end;           /* And exit */
    }

    wrk_ptr = var_name + length;    /* Point to char after var. name */
    *wrk_ptr = '\0';            /* Null terminate */

    /*
     *  Report an error if there is anything else on the line.
     */
    wrk_ptr += 1;           /* Point to next character */
    length = strlen (wrk_ptr);      /* Get length of any extra text */
    if (length != 0)            /* If there is some */
    {
        (void) fprintf (stderr, "\n%s\n", srclin_ptr);
        /* Output line in question */
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_EXTTXT);        /* Message */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
        goto end;           /* And exit */
    }

    /*
     *  Return variant name to caller.
     */
    (void) strcpy (varnam_ptr, var_name);

    retval = PTRUE;         /* Signal success */

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return retval;
}


/************************************************************************/
/*                                  */
/*  OPEN_INCLUDE - Open include file                */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     retval = OPEN_INCLUDE (tdb_ptr, prslin_ptr, srclin_ptr,  */
/*                file_ptr_ptr, name_ptr);      */
/*                                  */
/*  Where:                              */
/*                                  */
/*     retval       = Boolean indicating whether or not the file    */
/*            was opened                    */
/*            PTRUE if success              */
/*            PFALSE otherwise              */
/*     tdb_ptr      = Pointer to TPA Structure          */
/*     prslin_ptr   = Pointer to parsable line          */
/*     srclin_ptr   = Pointer to source line            */
/*     file_ptr_ptr = Pointer to variable to hold returned file */
/*            pointer                   */
/*     name_ptr     = Pointer to variable to hold returned file name*/
/*                                  */
/*  This routine is called when a "#INCLUDE" command is     */
/*  encountered. It parses the specified line read from the     */
/*  current input file and opens the specified file.        */
/*                                  */
/************************************************************************/

static Boolean OPEN_INCLUDE (struct tdb *tdb_ptr,
                             char *prslin_ptr,
                             char *srclin_ptr,
                             FILE **file_ptr_ptr,
                             char *name_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Return Status */
    unsigned long       diag_save;      /* Saved diagnostic context */
    struct   diagstk    *diag_ptr;      /* Pointer to Diagnostic Stack */
    unsigned long       length;     /* Buffer Length */
    char       *wrk_ptr;       /* Work Pointer */
    char       wrkbuf [256];   /* Work Buffer */
    Boolean         retval;     /* Return value */

    /*
     *  Copy pointers for ease of reference.
     */
    diag_ptr = (*tdb_ptr).tdb_diag_ptrps;

    /*
     *  First add a diagnostic to the stack.
     */
    diag_save = PSL_ADD_STATDIAG (diag_ptr, "During OPEN_INCLUDE");

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_GEN_TPARSE))
    {
        PSL_OUTPUT_TRACE ("Opening Include File\n");
    }

    /*
     *  Parse the input line buffer to extract the name of the file to include.
     *  First, locate the start of the file name.
     */
    wrk_ptr = prslin_ptr + 8;       /* Start after "#include" */
    length = strspn (wrk_ptr, " \t");   /* Get length of leading whitespace */

    /*
     *  Check that the first character is a double quote.
     */
    wrk_ptr += length;          /* Point to first char of file name */

    if (*wrk_ptr != '\"') goto bad_name;

    (void) strcpy (wrkbuf, wrk_ptr+1);  /* Copy string to work buffer */

    /*
     *  Now locate the end of the filename.
     */
    length = strcspn (wrkbuf, " \t\0"); /* Get length of file name */

    wrk_ptr = wrkbuf + length - 1;  /* Point to last char of file name */

    if (*wrk_ptr != '\"') goto bad_name;
    /* Check it's a double quote */
    *wrk_ptr = '\0';            /* Null terminate */

    /*
     *  Open the Include File.
     */
    PSL_CLEARERR ();            /* Clear any C file errors */

    /*
     *  Get the name of the include file.
     */
    iret = PSL_FILE_PARSE (     /* Parse filespec */
               wrkbuf,         /* This File */
               (*tdb_ptr).tdb_srcextc16,
               /* Default File */
               NULL,           /* No Related File */
               NULL,           /* No Expected Related Extension */
               name_ptr,       /* Returned File */
               256,            /* Maximum length thereof */
               FILPRS_INPUT);      /* Flags */

    if (!(psl_test_status (iret)))  /* If an error */
    {
        PSL_PUTDIAG (           /* Signal Diagnostic */
            diag_ptr,       /* Use this diagnostic stack */
            2,          /* Number of message segments */
            PSL_OPENIN,     /* First message */
            name_ptr,       /* Parameter */
            iret);          /* Secondary message */

        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            diag_ptr,       /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */

        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */

        retval = PFALSE;        /* Signal failure */

        goto end;           /* And exit */
    }

    /*
     *  Open the include file.
     */
    *file_ptr_ptr = fopen (name_ptr, "r");
    /* Open the file */
    if (*file_ptr_ptr == NULL)      /* If it failed */
    {
        PSL_LOG_CIOERR (        /* Log file error */
            diag_ptr,       /* Use this stack */
            PSL_OPENIN,     /* Error to report */
            name_ptr);      /* Filespec */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            diag_ptr,       /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
        retval = PFALSE;        /* Signal failure */
        goto end;           /* And exit */
    }

    /*
     *  Get the full name.
     */
    PSL_GET_FILENAME (*file_ptr_ptr, name_ptr, 256);

    retval = PTRUE;         /* Signal success */

    /*
     *  Reset the diagnostics.
     */
    PSL_REMOVE_DIAGNOSTICS (diag_ptr, diag_save);

    goto end;

    /*
     *  BAD_NAME:
     *
     *  Report 'Invalid Filename' error.
     */
bad_name:
    (void) fprintf (stderr, "\n%s\n", srclin_ptr);
    /* Output line in question */
    PSL_PUTDIAG (           /* Put Diagnostic */
        diag_ptr,       /* Use this stack */
        1,          /* Number of message segments */
        PSL_BADINCNAME,     /* Message */
        (*tdb_ptr).tdb_line_countl,
        /* First Parameter */
        (*tdb_ptr).tdb_expsrcnamc256);
    /* Second Parameter */
    PSL_OUTPUT_DIAG (           /* Output Diagnostics */
        diag_ptr,       /* Use this stack */
        0,          /* No Context */
        DIOUFL_LSTERR);     /* Last Error */
    (*tdb_ptr).tdb_err_countl += 1; /* Increment error count */
    retval = PFALSE;            /* Signal failure */

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return retval;
}


/************************************************************************/
/*                                  */
/*  READ_NEXT_LINE - Read next line from Source File        */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     READ_NEXT_LINE (tdb_ptr,                 */
/*             srcsiz_ptr, srclin_ptr,          */
/*             prssiz_ptr, prslin_ptr,          */
/*             errlin_ptr, eof_flag_ptr);           */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr      = Pointer to TPA Structure          */
/*     srcsiz_ptr   = Pointer to returned no. chars in input line   */
/*     srclin_ptr   = Pointer to line as input          */
/*     prssiz_ptr   = Pointer to returned no. chars in parsable line*/
/*     prslin_ptr   = Pointer to line minus comments and uppercased */
/*     errlin_ptr   = Pointer to error line             */
/*     eof_flag_ptr = Pointer to End of File flag           */
/*                                  */
/*  This routine reads the next line from the the source file,  */
/*  and outputs that line to the Listing file, if required.     */
/*                                  */
/*  The routine also initializes the Error underline to the same    */
/*  value as the original source text line, to ease the process of  */
/*  underlining errors.                     */
/*                                  */
/*  The routine converts the input text line (excluding comments    */
/*  and quoted strings) to Uppercase, strips off Comments (unless   */
/*  the 'keep comments' flag TDFL_KEPCMT is specified), leaving */
/*  only the parsable text.                     */
/*                                  */
/************************************************************************/

static void READ_NEXT_LINE (struct tdb *tdb_ptr,
                            long *srcsiz_ptr,
                            char *srclin_ptr,
                            long *prssiz_ptr,
                            char *prslin_ptr,
                            char *errlin_ptr,
                            Boolean *eof_flag_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Return Status */
    long       status;     /* Return status */
    long       jsub;       /* Loop Control */
    long       ksub;       /* Loop Control */
    unsigned long       length;     /* Length */
    long       match_status;   /* Status from memcmp */
    Boolean         cont_found;     /* Continuation Found flag */
    Boolean         retval;     /* Return value */
    char       quote;      /* Quote Character */
    char       *src_ptr;       /* Pointer to srclin */
    char       *prs_ptr;       /* Pointer to prslin */
    char       *err_ptr;       /* Pointer to errlin */
    Boolean     comment;       /* Parsing comment flag */

    /*
     *  The following data is used when reading from the 'current' source
     *  file, which may be the primary source file or an include file.
     */
    FILE       *cur_file_ptr;  /* Pointer to current file */

    /*
     *  If we have already found the End of File on a previous parsing
     *  operation, return straight away. This will enable this routine to be
     *  called at lots of places in the TPARSE state table, and will always
     *  produce the same result.
     */
    if (*eof_flag_ptr) goto end;

    /*
     *  NEXT_LINE:
     *
     *  Read next line into the work area line buffer.
     *
     *  Pay attention, 'cos this is a bit tricky. If you request 'n'
     *  characters, fgets reads at most 'n-1' characters. Therefore,
     *  if we want to check whether the line is too long for us to
     *  handle (which we do!) we must read 'n+1' characters and so
     *  must ask fgets for 'n+2' characters. Clear?
     *
     *  Also, if we have a non-zero Include file nesting level we read
     *  from the current include file rather than the primary source file.
     */
next_line:
    ;
    if ((*tdb_ptr).tdb_inc_levell != 0)
    {
        cur_file_ptr = (*tdb_ptr).tdb_file_ptr_stkpf8
                       [(*tdb_ptr).tdb_inc_levell - 1];
    }
    else
    {
        cur_file_ptr = (*tdb_ptr).tdb_srcfil_ptrpf;
    }

    (void) fgets (srclin_ptr,       /* Read into this buffer */
                  (*tdb_ptr).tdb_max_linsizl+2,
                  /* This number of characters */
                  cur_file_ptr);    /* From this file */

    if (ferror (cur_file_ptr) != 0) /* If an error */
    {
        iret = PSL_READERR;     /* Set up Error */
        PSL_LOG_CIOERR (        /* Log file error */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            iret,           /* Error to report */
            (*tdb_ptr).tdb_expsrcnamc256);
        /* Filespec */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        goto end_abort;
    }

    /*
     *  If we are at the end of the primary file then set the flag and exit.
     *  If we are at the end of an include file we close the file,
     *  decrement the include level and go back and try again.
     */
    if (feof (cur_file_ptr) != 0)
    {
        if ((*tdb_ptr).tdb_inc_levell == 0)
            /* If reading primary file */
        {
            *eof_flag_ptr = PTRUE;  /* Set 'End of File' flag */
            *srcsiz_ptr = 0;        /* Return zero line size */
            goto end;           /* And exit */
        }
        else                /* Reading an include file */
        {

            /*
             *  Close the Include File.
             */
            status = fclose (cur_file_ptr);
            /* Close the file */
            if (status != 0)        /* If an error */
            {
                PSL_SIG_CIOERR (    /* Log file error */
                    (*tdb_ptr).tdb_diag_ptrps,
                    /* Use this stack */
                    PSL_CLOSEIN,    /* Error to report */
                    (*tdb_ptr).tdb_expsrcnamc256);
                /* Filespec */
                goto end;       /* And exit */
            }
            (*tdb_ptr).tdb_file_ptr_stkpf8
            [(*tdb_ptr).tdb_inc_levell] = NULL;
            /* Trash pointer */

            (*tdb_ptr).tdb_inc_levell -= 1;
            /* Decrement include file level */

            (*tdb_ptr).tdb_line_countl =
                (*tdb_ptr).tdb_line_count_stkl8 [(*tdb_ptr).tdb_inc_levell];
            /* Restore line count */

            (void) strcpy ((*tdb_ptr).tdb_expsrcnamc256,
                           (*tdb_ptr).tdb_file_name_stkc256_8 [(*tdb_ptr).tdb_inc_levell]);
            /* Restore file name */

            goto next_line;     /* And go back to the previous file */
        }
    }

    /*
     *  Count the line.
     */
    (*tdb_ptr).tdb_line_countl += 1;

    /*
     *  Tidy the line up, measure and count it, reporting errors as we go.
     */
    *srcsiz_ptr = strlen (srclin_ptr);  /* Set length of record read */

    src_ptr = srclin_ptr + *srcsiz_ptr - 1;
    /* Point to last character */

    if (*src_ptr == '\n')       /* If last character is newline */
    {
        *src_ptr = '\0';        /* Remove it */
        *srcsiz_ptr -= 1;       /* And decrement length */
    }

    if (*srcsiz_ptr > (*tdb_ptr).tdb_max_linsizl)
        /* Error if line is too big */
    {
        PSL_PUTDIAG (           /* Add Diagnostic */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            2,          /* Number of message segments */
            PSL_READERR,        /* First message */
            (*tdb_ptr).tdb_expsrcnamc256,
            /* Parameter */
            PSL_BADRECSIZ);     /* Secondary message */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tdb_ptr).tdb_diag_ptrps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        (*tdb_ptr).tdb_err_countl += 1; /* Count error */
        *srcsiz_ptr = (*tdb_ptr).tdb_max_linsizl;
        /* And truncate record */
    }

    /*
     *  We allow continuation characters to be specified either before or
     *  after a trailing line comment so we check here if the last character
     *  in the line is a continuation character. If so, we set a flag to
     *  remind ourselves of this later.
     */
    src_ptr = srclin_ptr + *srcsiz_ptr - 1;
    /* Point to last character */
    cont_found = PFALSE;        /* Assume its not a continuation */

    if (((*tdb_ptr).tdb_flagsl & TDFL_KEPCMT) == 0)
    {
        for (ksub = 0; ksub < (*tdb_ptr).tdb_cont_numl; ksub++)
        {
            if (*src_ptr == (*tdb_ptr).tdb_cont_charb4 [ksub])
            {
                cont_found = PTRUE;
            }
        }
    }

    /*
     *  Set up the Parsable line to contain the same as the input line,
     *  without the Comments, and converted to Uppercase.
     */
    *prssiz_ptr = 0;

    /*
     *  No Quotes yet.
     */
    quote = '\0';
    comment = PFALSE;           /* Not parsing comment */

    for (jsub = 0,
            src_ptr = srclin_ptr, prs_ptr = prslin_ptr, err_ptr = errlin_ptr;
            jsub < *srcsiz_ptr;
            jsub++, src_ptr++, prs_ptr++, err_ptr++)
    {

        /*
         *  If we have a Quote it either starts or ends a quoted string (we hope).
         */
        if ((*src_ptr == '\"') ||
                (*src_ptr == '\''))
        {
            if (quote == '\0')
            {
                quote = *src_ptr;
            }
            else if (quote == *src_ptr)
            {
                quote = '\0';
            }
        }

        /*
         *  If we have a comment that is NOT inside a quoted string, get out.
         */
        for (ksub = 0; ksub < (*tdb_ptr).tdb_comment_numl; ksub++)
        {
            match_status = memcmp (src_ptr,
                                   (*tdb_ptr).tdb_comment_strc4_4 [ksub],
                                   (*tdb_ptr).tdb_comment_lenl4 [ksub]);
            if ((match_status == 0) && (quote == '\0'))
            {
                if (((*tdb_ptr).tdb_flagsl & TDFL_KEPCMT) != 0)
                {
                    comment = PTRUE;    /* Parsing comment */
                }
                else
                {
                    goto end_copy_loop;
                }
            }
        }

        /*
         *  Transfer the character to the parsable line buffer.
         *  If it alphabetic and not inside a quoted string we convert it
         *  to uppercase. Otherwise we copy it unchanged.
         */
        if ((isalpha (*src_ptr)) && /* If it is alphabetic... */
                (quote == '\0') &&      /* ... and outside quoted string */
                (!comment))         /* ... and outside comment */
        {
            *prs_ptr = toupper (*src_ptr);
            /* Convert to Uppercase */
        }
        else                /* Anything else */
        {
            *prs_ptr = *src_ptr;    /* Gets copied as is */
        }
        *prssiz_ptr += 1;

        /*
         *  Set up Error Line as a mask of the parsable line with Tabs in the
         *  same place.
         */
        if (*prs_ptr == '\t')       /* If it was a TAB character */
        {
            *err_ptr = '\t';        /* mask errror line similarly */
        }
        else                /* Anything else */
        {
            *err_ptr = ' ';     /* is masked with a SPACE */
        }

    }

    /*
     *  END_COPY_LOOP:
     *
     *  We come here when we have finished copying characters to the
     *  parsable line buffer.
     */
end_copy_loop:
    ;

    /*
     *  If we have anything vaguely parsable, trim off any trailing
     *  non-printing characters.
     */
    if (*prssiz_ptr > 0)
    {
        for (length = *prssiz_ptr, prs_ptr = prslin_ptr + *prssiz_ptr - 1;
                length != 0;
                length--, prs_ptr--)
        {
            if (isgraph (*prs_ptr) != 0) goto end_trim_loop;
        }

        /*
         *  END_TRIM_LOOP:
         *
         *  We come here if we locate a printing character in the parsable line.
         */
end_trim_loop:
        ;

        *prssiz_ptr = length;       /* Reset parsable size */

    }

    /*
     *  If we detected that the last character on the line was a continuation
     *  and the last character we have now is NOT a continuation then we
     *  append a continuation character by hand.
     */
    if (*prssiz_ptr > 0)
    {
        if (cont_found)
        {
            prs_ptr = prslin_ptr + *prssiz_ptr - 1;
            /* Point to last character */

            for (ksub = 0; ksub < (*tdb_ptr).tdb_cont_numl; ksub++)
            {
                if (*prs_ptr == (*tdb_ptr).tdb_cont_charb4 [ksub])
                {
                    goto end_cont_check;
                }
            }

            /*
             *  The last character is not a continuation character - append one.
             */
            prs_ptr += 1;
            *prs_ptr = ' ';
            prs_ptr += 1;
            *prs_ptr = (*tdb_ptr).tdb_cont_charb4 [0];
            *prssiz_ptr += 2;
        }
    }

    /*
     *  END_CONT_CHECK:
     *
     *  We come here when we have ensured that a continuation line has
     *  a continuation character at the end.
     */
end_cont_check:
    ;

    /*
     *  List the line.
     *
     *  Test if the whole parsable line was blank. If so, list the line
     *  read, and read the next line.
     *
     *  n.b. If line is not blank, we do not list it here since it may be
     *  necessary to output an error message relating to previous line(s)
     *  before listing this one.
     */
    if (*prssiz_ptr == 0)
    {
        if (*srcsiz_ptr > 0)
        {
            TPA_LIST_LINE (tdb_ptr, srclin_ptr);
        }
        else
        {
            TPA_LIST_LINE (tdb_ptr, " ");
        }
        goto next_line;
    }

    /*
     *  Null terminate the parsable and error lines.
     */
    prs_ptr = prslin_ptr + *prssiz_ptr;
    *prs_ptr = '\0';

    err_ptr = errlin_ptr + *prssiz_ptr;
    *err_ptr = '\0';

    /*
     *  We now handle the various compiler directives, which start with
     *  a "#" and must begin in the fisrt column of the parsable line.
     *
     *  Note that each of thes is switchable so that the user can disable
     *  our handling of them if required.
     *
     *  Note also that we must list any such lines here since they won't
     *  get done anywhere else.
     *  Also, we must list any lines that we reject because the necessary
     *  conditions are not met. We attempt to make the user aware of them
     *  by outputting them without a line number.
     */

    /*
     *  #IFDEF and similar
     *  ==================
     */
    if (((*tdb_ptr).tdb_flagsl & TDFL_ENBCND) != 0)
        /* If conditional handling enabled */
    {

        /*
         *  Handle conditional compilation directives.
         */
        retval = DO_CONDITIONAL (tdb_ptr, srclin_ptr, prslin_ptr);
        if (retval)             /* If it was a conditional directive */
        {
            goto next_line;     /* Loop back to read next record */
        }

        /*
         *  Check to see if we want to just ignore this line.
         */
        if ((*tdb_ptr).tdb_if_levell != -1)
            /* If conditions active */
        {
            retval = CHK_CONDITIONAL (tdb_ptr);
            if (retval)
            {
                if ((*tdb_ptr).tdb_listn)
                {
                    (void) LIST_LINE ((*tdb_ptr).tdb_list_ctx_ptrps,
                                      srclin_ptr);
                    /* List the line */
                }
                goto next_line;     /* Loop back to read next record */
            }
        }
    }                   /* End (if conditional handling enabled) */

    /*
     *  #INCLUDE
     *  ========
     *
     *  This directive specifies the name of an include file from which
     *  subsequent lines are to be read.
     */
    if (((*tdb_ptr).tdb_flagsl & TDFL_ENBINC) != 0)
        /* If INCLUDE handling enabled */
    {
        if (memcmp (prslin_ptr, "#INCLUDE", 8) == 0)
            /* If #INCLUDE */
        {
            DO_INCLUDE (tdb_ptr, srclin_ptr, prslin_ptr);
            /* Handle it */
            goto next_line;     /* Loop back to read next record */
        }
    }                   /* End (if include handling enabled) */


    /*
     *  #ERROR
     *  ======
     *
     *  This directive specifies the text that is to be output as part
     *  of an error message.
     */
    if (((*tdb_ptr).tdb_flagsl & TDFL_ENBERR) != 0)
        /* If ERROR handling enabled */
    {
        if (memcmp (prslin_ptr, "#ERROR", 6) == 0)
            /* If #ERROR */
        {
            DO_ERROR (tdb_ptr, srclin_ptr, prslin_ptr);
            /* Handle it */
            goto next_line;     /* Loop back to read next record */
        }
    }                   /* End (if error handling enabled) */


    /*
     *  Here if there is a new line for parsing.
     */
    goto end;

    /*
     *  END_ABORT:
     *
     *  Report the Abort diagnostic.
     */
end_abort:
    ;
    PSL_SIGDIAG (           /* Add Diagnostic */
        (*tdb_ptr).tdb_diag_ptrps,  /* Use this stack */
        1,              /* Number of message segments */
        PSL_ENDABORT,           /* Message */
        (*tdb_ptr).tdb_expsrcnamc256);  /* Parameter */

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return;
}


/************************************************************************/
/*                                  */
/*  RESET_STRUCT - Reset TPA routines               */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     RESET_STRUCT (tdb_ptr);                  */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tdb_ptr = Pointer to TDB structure               */
/*                                  */
/*  This routine resets a TDB structure for re-use by the       */
/*  TPA routines.                           */
/*                                  */
/************************************************************************/

static void RESET_STRUCT (struct tdb *tdb_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       jsub;       /* Loop Control */
    unsigned long       diag_save;      /* Saved diagnostic context */
    struct   diagstk    *diag_ptr;      /* Pointer to Diagnostic Stack */

    /*
     *  Copy pointers for ease of reference.
     */
    diag_ptr = (*tdb_ptr).tdb_diag_ptrps;

    /*
     *  First add a diagnostic to the stack.
     */
    diag_save = PSL_ADD_STATDIAG (diag_ptr, "During RESET_STRUCT");

    /*
     *  Output trace information.
     */
    if (PSL_TEST_TRACE (PTRC_GEN_TPARSE))
    {
        PSL_OUTPUT_TRACE ("Resetting TDB structure\n");
    }

    /*
     *  Initialize fields.
     */
    (*tdb_ptr).tdb_line_countl = 0; /* No lines read */
    (*tdb_ptr).tdb_err_countl = 0;  /* No errors encountered */
    (*tdb_ptr).tdb_eof_flagn = PFALSE;  /* Not at end of file */
    (*tdb_ptr).tdb_undlin_flagn = PFALSE;
    /* No error underline required */
    (*tdb_ptr).tdb_errs_in_linel = 0;   /* No errors in line */
    (*tdb_ptr).tdb_inc_levell = 0;  /* Include level */
    (*tdb_ptr).tdb_if_levell = -1;  /* If level */

    for (jsub = 0; jsub < TDB_MAX_ERRS_IN_LINE; jsub++)
    {
        (*tdb_ptr).tdb_err_arrayl5 [jsub] = 0;
        /* No errors */
    }

    /*
     *  Reset the diagnostics.
     */
    PSL_REMOVE_DIAGNOSTICS (diag_ptr, diag_save);

    return;
}
