/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TScriptRule.cpp
/**
 * Method definitions for TScriptRule class.
 * This class holds all the rules that describe the form and function
 * of a transaction script (TScript), including the description of the
 * file or serialization format, known as tof, which is a form of atvf
 **/

// Local includes
#include "TScriptRule.h"
#include "TRecMake.h"
#include "AtvfRealRecRules.h"
#include "mdbexcept.h"

// Construct ODL rule components
AtvfRule            TScriptRule::_odl("odl.");
const AtvfSimplRule TScriptRule::_prmRule("parameter");
const AtvfSimplRule TScriptRule::_recRule("record");
const AtvfSimplRule TScriptRule::_optRule("option");
const AtvfDelimRule TScriptRule::_lopRule("loption");
const AtvfSimplRule TScriptRule::_defRule("default");
const AtvfSimplRule TScriptRule::_atrRule("attribute");
const AtvfSimplRule TScriptRule::_rowRule("row");
const AtvfSimplRule TScriptRule::_colRule("column");
const AtvfSimplRule TScriptRule::_incRule("#include");

// Create static rule
const bool TScriptRule::_createOdlRule()
{
    _odl.add(SharedPtr<AtvfRecRule>(new AtvfSimplRule(_prmRule)));
    _odl.add(SharedPtr<AtvfRecRule>(new AtvfSimplRule(_recRule)));
    _odl.add(SharedPtr<AtvfRecRule>(new AtvfSimplRule(_optRule)));
    _odl.add(SharedPtr<AtvfRecRule>(new AtvfDelimRule(_lopRule)));
    _odl.add(SharedPtr<AtvfRecRule>(new AtvfSimplRule(_defRule)));
    _odl.add(SharedPtr<AtvfRecRule>(new AtvfSimplRule(_atrRule)));
    _odl.add(SharedPtr<AtvfRecRule>(new AtvfSimplRule(_rowRule)));
    _odl.add(SharedPtr<AtvfRecRule>(new AtvfSimplRule(_colRule)));
    _odl.add(SharedPtr<AtvfRecRule>(new AtvfSimplRule(_incRule)));

    return true;
}

const bool TScriptRule::_odlRuleCreated = _createOdlRule();

/**
 * Default Constructor
 **/
TScriptRule::TScriptRule(const string& odlname)
{

    // Call recursive function to read required odl file, and
    // construct the script rule.
    read_odl(odlname);
}

/**
 * Private recursive helper function used to read an odl, and any included odl's
 * The parameter include_odl is set when called to read an included odl file.
 **/
void TScriptRule::read_odl(const string& odlname, bool include_odl)
{
    // Open and validate the odl file
    // set AtvfInput to ignore comments
    AtvfInput odlin(odlname.c_str(),true);
    string* widgetP = _odl.extractHead(odlin);
    if(widgetP == NULL) throw mdbParseException("Invalid odl file format");

    if ( !include_odl )
        // If main odl then create a new token rule to hold Atvf rule associated with
        // all record token parameter rules. If an included odl this rule should already
        // exist.
        _atvfToknRuleP = new AtvfRule(*widgetP);

    // Create local atvf rule for current ODL to hold all the atvf rules associated with
    // ONLY parameter token rules. ( Not the record token parameter rules )
    AtvfRule* _atvfRuleP = new AtvfRule(*widgetP);
    // Store in vector
    _atvfRulesVect.push_back(_atvfRuleP);
    delete widgetP;
    widgetP = NULL;

    // Create a local container to store an exclusive set of parameter rules based
    // on the content of the current odl. Stores the first occurrence of each parameter
    // and is used in the creation of duplicate parameter.
    prulep_cont_t params;
    // Read in the odl file and use the information to create file format

    AtvfRecord* recP;
    SharedPtr<TRecCreator> creatorP;

    if ( !include_odl )
    {
        // If main odl create a TSriptHead creator use this as the script rule's script
        // head creator component
        _headCreateP = SharedPtr<TRecCreator>(new TRecMake<TScriptHead>());
        creatorP = _headCreateP;
    }
    else
    {
        // Otherwise for included odl use the temporary script head creator, discard
        // when finished.
        creatorP = SharedPtr<TRecCreator>(new TRecMake<TScriptHead>());
    }

    TParamRule* prP = NULL;
    CConfigTable* constTablePtr;
    bool startRowFlag = false;
    string tableRowString;

    while((recP = _odl.extractRecord(odlin)) != NULL)
    {
        if(recP->isKnown())
        { // Ignore unkown records
            AtvfKey rKey = recP->key();
            if ( rKey == _incRule.key() )
            {
                // #include other odl
                /*                string szFileName;
                #ifdef CYRANO_TP_STANDALONE
                                string::size_type endPath = 0;
                                if((endPath = odlname.find_last_of("/\\")) != string::npos)
                                    szFileName = odlname.substr(0, endPath + 1) + recP->value();
                                else
                #endif
                                    szFileName = recP->value();
                                read_odl(szFileName, true);
                            } else if((rKey == _prmRule.key()) || // 'parameter' odl record or
                               (rKey == _recRule.key())) { // 'record' odl record
                                if (prP != NULL) { // terminate table entry
                                    startRowFlag = false;
                                    tableRowString.resize(0);
                                    prP->addConfigs(constTablePtr);
                                }
                */
#ifdef CYRANO_TP_STANDALONE
                int endPath = 0;
                if ( (endPath = odlname.find_last_of("/\\")) != string::npos )
                    read_odl(odlname.substr(0, endPath + 1) + recP->value(), true);
                else
#endif
                    read_odl(recP->value(), true);
            }
            else if((rKey == _prmRule.key()) || // 'parameter' odl record or
                    (rKey == _recRule.key()))
            { // 'record' odl record
                if (prP != NULL)
                { // terminate table entry
                    startRowFlag = false;
                    tableRowString.resize(0);
                    prP->addConfigs(constTablePtr);
                }





                prP = makePrmRule(recP->value(), params );

                if(prP == NULL)
                    throw mdbParseException(recP->value() + " -  odl invalid");
                // In case of duplicate TParamRule recover table pointer from TParamRule
                constTablePtr = const_cast<CConfigTable*>(prP->config());
                if(rKey == _prmRule.key())
                { // is a paramater
                    // Add rule to current AtvfRule
                    _atvfRuleP->add(prP->innerRule()); // Add atvf rule to the set
                    creatorP->addPrmRule(prP); // add to existing record rule
                }

                else
                { // must be a new record rule
                    // Add rule to Token AtvfRule
                    _atvfToknRuleP->add(prP->innerRule()); // Add atvf rule to the set
                    creatorP = getTRecCreator(prP->key()); // make a new one
                    trec_crtrs_t::iterator prevRecIt;
                    if (  (prevRecIt = _creators.find(prP->key()) )
                            != _creators.end())
                    {
                        // If already defined remove current entry, and create new one
                        if ( (*prevRecIt).second != creatorP )
                        {
                            // Must be a TRecord creator
                            (*prevRecIt).second = creatorP;
                        }
                        else
                            // Otherwise a Trcd"type" creator
                            creatorP->clearDefinition();
                    }
                    else
                    {
                        _creators[prP->key()] = creatorP; // and put into map
                        _creatorKeyIdx.push_back(prP->key());
                    }
                    // Move to after
                    creatorP->addTypRule(prP); // add this first param
                    // Map record to current rule.
                    _objRuleMap[prP->key()]=_atvfRuleP;
                    prP = NULL; // So any following odl stuff goes to rrec
                }

            }
            else if((rKey == _optRule.key()) || // 'option' odl record
                    (rKey == _lopRule.key()))
            { // 'loption' odl record
                if(prP != NULL)
                { // description belongs to last parameter
                    prP->addOption(recP->value());
                }
                else
                { // description belongs to last attribute
                    creatorP->addOption(recP->value());
                }
            }
            else if(rKey == _defRule.key())
            {
                if(prP != NULL)
                { // default value belongs to last param
                    prP->defStrVal(recP->value());
                }
                else
                { // default value belongs to last record
                    creatorP->addDefVal(recP->value());
                }
            }
            else if(rKey == _atrRule.key())
            {
                if(prP != NULL)
                { // attribute value belongs to last param
                    prP->setAttr(recP->value());
                }
                else
                { // attribute value belongs to last record
                    creatorP->setAttr(recP->value());
                }
            }
            else if(rKey == _rowRule.key())
            { // new row or table
                tableRowString.resize(0);
                tableRowString = recP->value();
                startRowFlag = true;
            }
            else if (rKey == _colRule.key() && startRowFlag)
            {// new value
                if (constTablePtr==NULL)
                { // New table
                    constTablePtr = new CConfigTable;
                    if ((constTablePtr->AddValue(tableRowString, recP->value()))
                            == BAD_ROW_ID)
                    {
                        // If invalid row
                        delete constTablePtr;
                        constTablePtr = NULL;
                        startRowFlag = false;
                        tableRowString.resize(0);
                    }
                }
                else
                { // Existing table
                    if ((constTablePtr->AddValue(tableRowString, recP->value()))
                            == BAD_ROW_ID)
                    {
                        // If invalid row
                        tableRowString.resize(0);
                        startRowFlag = false;
                    }
                }
            }
        }
        delete recP;
        recP = NULL;
    }

    if (prP != NULL && constTablePtr != NULL)
        // terminate table entry for last parameter
        prP->addConfigs(constTablePtr);

}

/**
 * Private function to help in constructing the TScript Rule
 * Parses the contents of transaction and parameter odl records.
 * This probably wants to go along with above.
 **/
TParamRule* TScriptRule::makePrmRule(const string& text, prulep_cont_t& params)
{
    // Parse the text to extract the useful bits
    string::size_type pos = text.find_first_of(" \t");
    string type = text.substr(0, pos);
    if(pos == string::npos) return NULL;
    pos = text.find_first_not_of(" \t", pos);
    if(pos == string::npos) return NULL;
    string::size_type num = text.find_first_of(" \t", pos);
    if(num != string::npos) num -= pos;
    string name = text.substr(pos, num);

    // If we already have this parameter type simply duplicate the
    // first one so as to pick up its already defined attributes.
    AtvfKey pKey = AtvfKey(name);

    prulep_cont_t::const_iterator i = params.find(pKey);
    if(i != params.end())
    {
        return params[pKey]->duplicate(name, type);
    }

    // Create a parameter rule of the correct type.
    TParamRule* prP = NULL;
    if(type == TPrmToknRule::TYPE)
    { // Token
        prP = new TPrmToknRule(name);
    }
    else if(type == TPrmINumRule::TYPE)
    { // Integer
        prP = new TPrmINumRule(name);
    }
    else if(type == TPrmSTxtRule::TYPE)
    { // Short text
        prP = new TPrmSTxtRule(name);
    }
    else if(type == TPrmDateRule::TYPE)
    { // Date
        prP = new TPrmDateRule(name);
    }
    else if(type == TPrmLTxtRule::TYPE)
    { // Long text
        prP = new TPrmLTxtRule(name);
    }

    // If have produced a valid parameter add it to hash of these
    if(prP != NULL) params[pKey] = prP;

    return prP;
}

/**
 * Destructor
 **/
TScriptRule::~TScriptRule()
{
    for (incrulevect_t::iterator ii = _atvfRulesVect.begin();
            ii != _atvfRulesVect.end(); ii++)
    {
        // clearRule() causes the destruction of the dynamically allocated rule components.
        (*ii)->clearRule();
        delete (*ii);
    }
    _atvfRulesVect.erase(_atvfRulesVect.begin(),_atvfRulesVect.end());

    // Destruct rule components
    _atvfToknRuleP->clearRule();
    delete _atvfToknRuleP;

    // Destruct creators
    _creators.erase(_creators.begin(), _creators.end());
}

/**
 * Use the rule to get the tof header record and its
 * parameters from AtvfInput
 **/
TScriptHead* TScriptRule::extractHead(AtvfInput& in) const
{
    // Note, for script AtvfRule should always be in _atvfRulesVect[0]
    string* verP = _atvfRulesVect[0]->extractHead(in);
    if(verP == NULL) return NULL; // Invalid tof file
#ifndef __VAX
    TScriptHead* headP = static_cast<TScriptHead*>(_headCreateP->create());
#else
    TScriptHead* headP = (TScriptHead*)_headCreateP->create();
#endif
    headP->setVer(verP);

    // Read in the paramaters associated with header block
    AtvfRecord* atvfRecP;
    while((atvfRecP =  _atvfRulesVect[0]->extractRecord(in)) != NULL)
    {
        if(!headP->addData(atvfRecP)) break;
    }

    return headP;
}

/**
 * Use the rule to get the tof header record and its
 * parameters from AtvfInput
 **/
TScriptHead* TScriptRule::newHead() const
{
#ifndef __VAX
    return static_cast<TScriptHead*>(_headCreateP->create());
#else
    return (TScriptHead*)_headCreateP->create();
#endif
}

/**
 * Use the rule to get a tof record from the AtvfInput given.
 * Extracts record token using _atvfToknRuleP, then extracts
 * remaining parameters using either, the AtvfRule 
 * indexed from _objRuleMap by token key, or if key this does
 * not exist (unknown record) then uses _atvfRulesVect[0].
 **/
TRecord* TScriptRule::extractRecord(AtvfInput& in) const
{
    AtvfRecord* atvfRecP = _atvfToknRuleP->extractRecord(in);
    if(atvfRecP == NULL) return NULL; // At end of file

    // Read in any extra blank records,
    while(atvfRecP->isBlank())
    {
        delete atvfRecP;
        atvfRecP = _atvfToknRuleP->extractRecord(in);
        if(atvfRecP == NULL) return NULL; // At end of file
    }

    // Local pointer to point to appropriate rule to use for input.

    AtvfRule* _atvfRuleP;
    // Create the Transaction record that is id'd by this Atvf record
    TRecord* recP = NULL;
    trec_crtrs_t::const_iterator i = _creators.find(atvfRecP->key());
    if(i != _creators.end())
    {
        recP = ((*i).second)->create();
        recP->addHead(atvfRecP);
        // Get appropriate AtvfRecRule
        _atvfRuleP = (*(_objRuleMap.find(atvfRecP->key()))).second;
    }

    // Create an Unknown record to hold data if not one we know about
    if(recP == NULL)
    {
        recP = new TRecord();
        recP->addHead(atvfRecP);
        // If unknown record use _atvfRulesVect[0]
        _atvfRuleP = _atvfRulesVect[0];
    }

    // Read in all the parameters associated with this record
    while((atvfRecP =  _atvfRuleP->extractRecord(in)) != NULL)
    {
        if(!recP->addData(atvfRecP)) break;
    }

    return recP;
}

/**
 * Method to find out the keys of the records that can be
 * validly part of a script described by this rule.
 * May eventually take some sort of option to control subset
 * of returned values, at moment returns all but the header
 * and declarations.
 **/
keyvect_t* TScriptRule::getRecordsKeys() const
{
    keyvect_t* kVectP = new keyvect_t;
    for(keyvect_t::const_iterator k = _creatorKeyIdx.begin();
            k != _creatorKeyIdx.end();
            k++)
    {
        trec_crtrs_t::const_iterator i = _creators.find(*k);
        if((i != _creators.end()) &&
                !(((*i).second)->hasAttr("declaration")))
        {
            kVectP->push_back(*k);
        }
    }

    return kVectP;
}

/**
 * Method to find out information of the records that can be
 * validly part of a script described by this rule
 * May eventually take some sort of option to control subset
 * of returned values, at moment returns all but the header
 * and declarations.
 **/
strvect_t* TScriptRule::getRecordsInfo(const string& optkey,
                                       const keyvect_t* kVectP) const
{
    keyvect_t* intKVP = NULL;
    strvect_t* sVectP = new strvect_t;
    if(kVectP == NULL) kVectP = intKVP = getRecordsKeys();
    for(keyvect_t::const_iterator k = kVectP->begin();
            k != kVectP->end();
            k++)
    {
        trec_crtrs_t::const_iterator i = _creators.find(*k);
        if(i != _creators.end())
        {
            sVectP->push_back(((*i).second)->option(optkey));
        }
        else
        {
            sVectP->push_back(string());
        }
    }
    delete intKVP;
    return sVectP;
}

/**
 * Method to find specific parameter rule.
 **/
const TParamRule* TScriptRule::getParamRule
(const string& ruleName, const string& creatName) const
{
    if (creatName == "head")
        return _headCreateP->getParamRule(ruleName);

    trec_crtrs_t::const_iterator creator = _creators.find(AtvfKey(creatName));
    return (creator != _creators.end()) ?
           (*creator).second->getParamRule(ruleName) :
#ifndef __VAX
           static_cast<TParamRule*>(NULL);
#else
           (TParamRule*)NULL;
#endif
}

/**
 * Method used to create records if the key is known
 **/
TRecord* TScriptRule::makeNewRecord(const AtvfKey& key) const
{
    trec_crtrs_t::const_iterator i = _creators.find(key);
    if(i != _creators.end())
    {
        TRecord* recP = ((*i).second)->create();
        recP->addHead();
        return recP;
    }
    else
    {
        return NULL;
    }
}

/**
 * Accessor to save the header record of the file this describes
 **/
void TScriptRule::saveHead(const string& version, ostream& out) const
{
    _atvfRulesVect[0]->saveHead(version, out);
}

/**
 * Friend operator <<
 **/
ostream& operator << (ostream& os, TScriptRule& tS)
{
    // Deal with header TParamCreator
    os << "HEADER" << endl << "-------------------------------------------------------"
    << endl << *tS._headCreateP << endl << endl;

    os << "RECORDS" << endl << "-------------------------------------------------------";
    for (trec_crtrs_t::const_iterator _ruleIt = tS._creators.begin();
            _ruleIt != tS._creators.end(); ++_ruleIt)
        os << endl << *((*_ruleIt).second) << endl
        << "-------------------------------------------------------";
    return os;
}


