/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrException
// @module  Cyrano Exceptions | This module contains the implementation of any
//          Cyrano-defined exceptions.  Every exception contains three pieces
//          of information.  The first is a string detailing the error that
//          occurred.  The second is a string detailing where the exception
//          occured.  The final piece of information is a status value, which
//          will allow an application to determine whether processing should
//          continue or not.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef _CyrException_h
#define _CyrException_h

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CyrError.h>

#include <string>
using std::string;

///////////////////////////////////////////////////////////////////////////////
// @class   This class is the base class for all Cyrano-defined exceptions.
class CyrException
{
    // @access Public Members
public:
    // @cmember,mfunc This constructor is responsible for assigning the what,
    // where and status member variables for this class.
    CyrException( string what, string where, int status )
    {
        if ( !what.empty() )
            m_what = what;
        else
            m_what = "Unknown error";
        if ( !where.empty())
            m_where = where;
        else
            m_where = "Unknown location";
        m_status = status;
    }

    // @cmember,mfunc This method returns the string explaining what exception
    // occurred.
    string what() const
    {
        return m_what;
    }

    // @cmember,mfunc This method returns the string explaining where the
    // exception occurred.
    string where() const
    {
        return m_where;
    }

    // @cmember,mfunc This method returns a status value that determines
    // which error occured.  This is will be more useful than simply a
    // text message.  Error codes are stored in the file "CyrError.h"
    int status() const
    {
        return m_status;
    }

    // @access Protected Members
protected:
    string  m_what;     // @cmember What happened.
    string  m_where;    // @cmember Where it happened.
    int       m_status;   // @cmember Status of exception.
}
;

///////////////////////////////////////////////////////////////////////////////
// This #define allows simple declarations of normal exceptions.  Using the
// macro will fully define the exception with the specified name.
#define CYRANO_NORMAL_EXCEPTION( name )                         \
class name : public CyrException                            \
{                                                           \
public:                                                     \
name( const char* what, const char* where, int status ) \
: CyrException( what, where, status )   {};         \
}

///////////////////////////////////////////////////////////////////////////////
// @class   CyrCORBAException | This class is the only exception thrown by the
//          CORBA wrappers. It contains a single string which describes the
//          exception and which method caused it.  All other exceptions should
//          be trapped by these classes and a developer need only trap
//          exceptions of this particular type. It should throw no exceptions
//          of its own.
// @base    public | CyrException
CYRANO_NORMAL_EXCEPTION( CyrCORBAException );

///////////////////////////////////////////////////////////////////////////////
// @class   CyrDaemonException | This class is the only exception thrown by the
//          CyrDaemon class.  It contains a single string which describes the
//          exception and which method caused it.  All other exceptions should
//          be trapped by this classes and a developer need only trap
//          exceptions of this particular type. It should throw no exceptions
//          of its own.
// @base    public | CyrException
CYRANO_NORMAL_EXCEPTION( CyrDaemonException );

///////////////////////////////////////////////////////////////////////////////
// @class   CyrRepositoryException | This class is the only exception thrown by
//          the Repository classes.  It contains a single string which
//          describes the exception and which method caused it.  All other
//          exceptions should be trapped by this classes and a developer need
//          only trap exceptions of this particular type. It should throw no
//          exceptions of its own.
// @base    public | CyrException
CYRANO_NORMAL_EXCEPTION( CyrRepositoryException );

///////////////////////////////////////////////////////////////////////////////
// @class   CyrQueryException | This class is the only exception thrown by any
//          of the OTA query classes.  It contains a single string which
//          describes the exception and which method caused it.  All other
//          exceptions should be trapped by this classes and a developer need
//          only trap exceptions of this particular type. It should throw no
//          exceptions of its own.
// @base    public | CyrException
CYRANO_NORMAL_EXCEPTION( CyrQueryException );

#endif
