/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _XMLPATH_HPP
#define _XMLPATH_HPP

//// XPath support.

#include "xmlparser.hpp"
#include <set>
#include <algorithm>
#include <sstream>
#include <iomanip>
#include <malloc.h>

template <typename U>
class XPathExpression
{
public:
    // Definitions:
    typedef const XmlParse<U>::XMLTree::Node* NodePTR;
    typedef const U* CUPTR;

    // Axis type. Even values are forward axes, odd values are reverse axes.
    enum Axis
    {
        ancestor            =  1,
        ancestor_or_self    =  3,
        attribute           =  4,
        child               =  6,
        descendant          =  8,
        descendant_or_self  = 10,
        following           = 12,
        following_sibing    = 14,
        namespace_axis      = 16,
        parent              = 18,
        preceding           = 21,
        preceding_sibling   = 23,
        self                = 24    // Special case: forward AND reverse axis.
    };
    BOOL IsForwardAxis(Axis a) const {return (a & 0x1) == 0;}
    BOOL IsReverseAxis(Axis a) const {return (a & 0x1) != 0 || a == self;}

    // Node test. Odd values have an accompanying string.
    enum NodeTestKind
    {
        nt_star                           = 0,  // *
        nt_ncname_star                    = 1,  // NCName ':' '*'
        nt_qname                          = 3,  // QName
        nt_comment                        = 4,  // 'comment' '(' ')'
        nt_text                           = 6,  // 'text' '(' ')'
        nt_processing_instruction         = 8,  // 'processing-instruction' '(' ')'
        nt_processing_instruction_literal = 9,  // 'processing-instruction' '(' Literal ')'
        nt_node                           = 10, // 'node' '(' ')'
    };
    BOOL HasString(NodeTestKind nt) const {return (nt & 0x1) != 0;}

    // Predicate for document-order sorts:
    struct DocumentOrder
    {
        enum Order {forward, reverse};
        DocumentOrder(Order eOrder = forward) : m_eOrder(eOrder) {}
        BOOL operator() (NodePTR pNode1, NodePTR pNode2) const
        {
            return m_eOrder == forward ?
                pNode1->m_dwID <  pNode2->m_dwID :
                pNode1->m_dwID >  pNode2->m_dwID;
        }
        BOOL operator==(const DocumentOrder& o) const {return m_eOrder == o.m_eOrder;}
        Order m_eOrder;
    };
    typedef std::set<NodePTR, DocumentOrder> OrderedNodeSet;

    static BOOL IsDocumentOrder(const OrderedNodeSet& nodeSet)
    {
        return nodeSet.key_comp().m_eOrder == DocumentOrder::forward;
    }

    // XPath Expression type.
    enum Type
    {
        TYPE_UNKNOWN,   // An undefined XPath expression.
        TYPE_NODESET,   // A node set.
        TYPE_BOOLEAN,   // A boolean.
        TYPE_NUMBER,    // A floating point number.
        TYPE_STRING     // A constant string.
    };

    // XPath expression value.
    struct XPathExprValue
    {
        Type            m_eType;    // The type of the XPath expression.
        OrderedNodeSet  m_nodeSet;  // Valid if m_eType is TYPE_NODESET.
        BOOL            m_bBoolean; // Valid if m_eType is TYPE_BOOLEAN.
        double          m_dNumber;  // Valid if m_eType is TYPE_NUMBER.
        U*              m_string;   // Valid if m_eType is TYPE_STRING.

        XPathExprValue() : m_eType(TYPE_UNKNOWN), m_bBoolean(FALSE),
            m_dNumber(0.0), m_string(NULL) {}
        XPathExprValue(const XPathExprValue& val) : m_eType(val.m_eType),
            m_bBoolean(val.m_bBoolean), m_dNumber(val.m_dNumber),
            m_nodeSet(val.m_nodeSet),
            m_string(DuplicateString(val.m_string)) {}
        ~XPathExprValue() {if(m_string) delete [] m_string;}
        XPathExprValue& operator=(const XPathExprValue& val)
        {
            Clear();
            m_nodeSet  = val.m_nodeSet;
            m_bBoolean = val.m_bBoolean;
            m_dNumber  = val.m_dNumber;
            if(val.m_eType == TYPE_STRING)
                m_string = DuplicateString(val.m_string);
            m_eType    = val.m_eType;
            return *this;
        }
        XPathExprValue& operator=(BOOL b)
        {
            Clear();
            m_eType    = TYPE_BOOLEAN;
            m_bBoolean = b;
            return *this;
        }
        XPathExprValue& operator=(double d)
        {
            Clear();
            m_eType   = TYPE_NUMBER;
            m_dNumber = d;
            return *this;
        }
        XPathExprValue& operator=(CUPTR pStr)
        {
            Clear();
            m_eType  = TYPE_STRING;
            m_string = DuplicateString(pStr);
            return *this;
        }
        XPathExprValue& operator+=(CUPTR pStr)
        {
            U* pOldString = m_string;
            m_string = ConcatString(m_string, pStr);
            delete [] pOldString;
            return *this;
        }
        XPathExprValue& operator=(const OrderedNodeSet& nodeSet)
        {
            Clear();
            m_eType   = TYPE_NODESET;
            m_nodeSet = nodeSet;
            return *this;
        }
        void Clear()
        {
            if(m_string)
                delete [] m_string;
            m_nodeSet.clear();
            m_bBoolean = FALSE;
            m_dNumber  = 0.0;
            m_string   = NULL;
            m_eType    = TYPE_UNKNOWN;
        }
        BOOL IsValid() const {return m_eType != TYPE_UNKNOWN;}
    };

    // Conversions.
    // No types of object can be converted to a node set.
    static BOOL ToBOOL(XPathExprValue& val);
    static BOOL ToNumber(XPathExprValue& val);
    static BOOL ToString(XPathExprValue& val);

    // Exception thrown when encoutering a parsing error.
    struct ParsingError
    {
        DWORD m_dwLineBeg, m_dwColBeg;  // Coordinates of the first character of the token
                                        // when the error occurred.
    protected:
        ParsingError(DWORD dwLineBeg, DWORD dwColBeg) : m_dwLineBeg(dwLineBeg), m_dwColBeg(dwColBeg) {}
    };
    struct LexicalError : public ParsingError
    {
        LexicalError(DWORD dwLineBeg, DWORD dwColBeg, DWORD dwLineCur, DWORD dwColCur)
            : ParsingError(dwLineBeg, dwColBeg), m_dwLineCur(dwLineCur), m_dwColCur(dwColCur) {}
        DWORD m_dwLineCur, m_dwColCur;  // Coordinates of the character at which
                                        // the error occurred.
    };
    struct SyntaxError  : public ParsingError
    {
        SyntaxError(DWORD dwLineBeg, DWORD dwColBeg) : ParsingError(dwLineBeg, dwColBeg) {}
    };

    // Construct an XPath expression.
    XPathExpression() : m_pXMLTreeRoot(NULL), m_pExpressionString(NULL) {}
    ~XPathExpression();

    // Parse an XPath expression and build the equivalent expression object.
    BOOL Parse(CUPTR zXPathExpr);

    // Retrieve the XPath expression string. Call after a parse has been successfully performed.
    CUPTR GetExpressionString() const {return m_pExpressionString;}

    // Evaluate the XPath expression for the XML document tree having pRootNode as root node.
    void Evaluate(const XmlParse<U>::XMLTree::NodeRoot* pRootNode, XPathExprValue& outVal);

#ifdef XML_DUMP
    // Dumping the syntax tree:
    void DumpSyntaxTree(std::basic_ostream<U>& os) const;
    void DumpSyntaxTreeAsArray(std::basic_ostream<U>& os) const;
#endif

    //// Public helpers:
    static double round(double d)
    {
        return ceil(d) - d <= d - floor(d) ?
                ceil(d) : floor(d);
    }
    static char* DuplicateString(const char* pcString)
    {
        return lstrcpyA(new char[lstrlenA(pcString) + 1], pcString);
    }
    static wchar_t* DuplicateString(const wchar_t* pcString)
    {
        return lstrcpyW(new wchar_t[lstrlenW(pcString) + 1], pcString);
    }
    static char* ConcatString(const char* pcString1, const char* pcString2)
    {
        return lstrcatA(
            lstrcpyA(new char[lstrlenA(pcString1) + lstrlenA(pcString2) + 1], pcString1),
            pcString2);
    }
    static wchar_t* ConcatString(const wchar_t* pcString1, const wchar_t* pcString2)
    {
        return lstrcatW(
            lstrcpyW(new wchar_t[lstrlenW(pcString1) + lstrlenW(pcString2) + 1], pcString1),
            pcString2);
    }
    static BOOL StartsWith(const char* pcString1, const char* pcString2)
    {
        return strstr(pcString1, pcString2) == pcString1;
    }
    static BOOL StartsWith(const wchar_t* pcString1, const wchar_t* pcString2)
    {
        return wcsstr(pcString1, pcString2) == pcString1;
    }
    static BOOL Contains(const char* pcString1, const char* pcString2)
    {
        return strstr(pcString1, pcString2) != NULL;
    }
    static BOOL Contains(const wchar_t* pcString1, const wchar_t* pcString2)
    {
        return wcsstr(pcString1, pcString2) != NULL;
    }
    static char* SubString(const char* pcString, long lStart, long lLength = -1)
    {
        long lLen = lstrlenA(pcString);
        if(lStart > lLen || lLength == 0)
            return DuplicateString("");
        else
        {
            if(lStart < 1)
                lLength = lLength + lStart - 1; // Actual length (IEEE 754).
            if(lStart < 1)
                lStart = 1;
            if(lLength == -1 || lLength > lLen - lStart + 1)
                lLength = lLen - lStart + 1;
            return lstrcpynA(new char[lLength + 1], pcString + lStart - 1, lLength + 1);
        }
    }
    static wchar_t* SubString(const wchar_t* pcString, long lStart, long lLength = -1)
    {
        long lLen = lstrlenW(pcString);
        if(lStart > lLen || lLength == 0)
            return DuplicateString(L"");
        else
        {
            if(lStart < 1)
                lLength = lLength + lStart - 1; // Actual length (IEEE 754).
            if(lStart < 1)
                lStart = 1;
            if(lLength == -1 || lLength > lLen - lStart + 1)
                lLength = lLen - lStart + 1;
            return lstrcpynW(new wchar_t[lLength + 1], pcString + lStart - 1, lLength + 1);
        }
    }
    static char* SubStringBefore(const char* pcString, const char* pcDelimiter)
    {
        char *pc = strstr(pcString, pcDelimiter);
        return pc ? SubString(pcString, 1, pc - pcString) : DuplicateString("");
    }
    static wchar_t* SubStringBefore(const wchar_t* pcString, const wchar_t* pcDelimiter)
    {
        wchar_t *pc = wcsstr(pcString, pcDelimiter);
        return pc ? SubString(pcString, 1, pc - pcString) : DuplicateString(L"");
    }
    static char* SubStringAfter(const char* pcString, const char* pcDelimiter)
    {
        char *pc = strstr(pcString, pcDelimiter);
        DWORD dwLen = lstrlenA(pcDelimiter);
        return pc ? SubString(pcString, pc + dwLen - pcString + 1) : DuplicateString("");
    }
    static wchar_t* SubStringAfter(const wchar_t* pcString, const wchar_t* pcDelimiter)
    {
        wchar_t *pc = wcsstr(pcString, pcDelimiter);
        DWORD dwLen = lstrlenW(pcDelimiter);
        return pc ? SubString(pcString, pc + dwLen - pcString + 1) : DuplicateString(L"");
    }

private:
    //// Forbidden:
    XPathExpression(const XPathExpression<U>& expr);
    XPathExpression& operator=(const XPathExpression& expr);

    //// For parsing:
    enum XPATH_TOKEN                            // The token type.
    {
        TOKEN_XPATH_ERROR,                      // Lexical error.
        TOKEN_XPATH_LEFT_PARENTHESIS,           // (
        TOKEN_XPATH_RIGHT_PARENTHESIS,          // )
        TOKEN_XPATH_LEFT_SQUAREBRACKET,         // [
        TOKEN_XPATH_RIGHT_SQUAREBRACKET,        // ]
        TOKEN_XPATH_DOT,                        // .
        TOKEN_XPATH_DOTDOT,                     // ..
        TOKEN_XPATH_AT,                         // @
        TOKEN_XPATH_COMMA,                      // ,
        TOKEN_XPATH_AXIS_DELIMITER,             // ::
        TOKEN_XPATH_NAMETEST,                   // * or NCName:* or QName
        TOKEN_XPATH_NODETYPE,                   // 'comment' or 'text' or 'processing-instruction' or 'node'
        TOKEN_XPATH_OPERATOR,                   // as in [32]
        TOKEN_XPATH_FUNCTIONNAME,               // QName not matching a NodeType
        TOKEN_XPATH_AXISNAME,                   // as in [6]
        TOKEN_XPATH_LITERAL,                    // single or double-quoted string, as in [29]
        TOKEN_XPATH_NUMBER,                     // as in [30]
        TOKEN_XPATH_VARIABLE_REFERENCE,         // $ QName, as in [36]
        TOKEN_XPATH_END                         // End of input XPath string.
    };
    enum XPATH_TOKEN_SUBTYPE                    // The token subtype.
    {
        XPATH_TOKEN_SUBTYPE_NONE,               // No subtype.
        NAMETEST_STAR,                          // *
        NAMETEST_PREFIX_STAR,                   // NCName:*
        NAMETEST_QNAME,                         // QName
        OPERATOR_NAME_OR,                       // or
        OPERATOR_NAME_AND,                      // and
        OPERATOR_NAME_MOD,                      // mod
        OPERATOR_NAME_DIV,                      // div
        OPERATOR_EQUALS,                        // =
        OPERATOR_NOT_EQUALS,                    // !=
        OPERATOR_GREATER,                       // >
        OPERATOR_LESS,                          // <
        OPERATOR_GREATER_OR_EQUAL,              // >=
        OPERATOR_LESS_OR_EQUAL,                 // <=
        OPERATOR_PLUS,                          // +
        OPERATOR_MINUS,                         // -
        OPERATOR_MULTIPLY,                      // *
        OPERATOR_PIPE,                          // |
        OPERATOR_SLASH,                         // /
        OPERATOR_DOUBLE_SLASH,                  // //
        NODETYPE_PROCESSING_INSTRUCTION,        // processing-instruction
    };
    struct Token                                // A token.
    {
        XPATH_TOKEN         m_eToken;           // The token type.
        XPATH_TOKEN_SUBTYPE m_eSubtype;         // The token subtype.
        DWORD               m_dwLineBeg, m_dwColBeg;  // Coordinates of the first character of the lexeme.
        DWORD               m_dwLineCur, m_dwColCur;  // Coordinates of the next character after the lexeme.
        CUPTR               m_pcCur;            // Pointer to the current character in the XPath expression.
                                                // Points 1 character after the lexeme.
        CUPTR               m_pcBeg;            // Points to the first character of the lexeme.

        Token(CUPTR pcBeg)
                : m_pcBeg(pcBeg), m_pcCur(pcBeg), m_eToken(TOKEN_XPATH_ERROR), m_eSubtype(XPATH_TOKEN_SUBTYPE_NONE),
                  m_dwLineBeg(1), m_dwColBeg(1),
                  m_dwLineCur(1), m_dwColCur(1) {}
        Token(CUPTR pcBeg, CUPTR pcCur, XPATH_TOKEN eToken, XPATH_TOKEN_SUBTYPE eSubtype,
              DWORD dwLineBeg, DWORD dwColBeg, DWORD dwLineCur, DWORD dwColCur)
                : m_pcBeg(pcBeg), m_pcCur(pcCur), m_eToken(eToken), m_eSubtype(eSubtype),
                  m_dwLineBeg(dwLineBeg), m_dwColBeg(dwColBeg),
                  m_dwLineCur(dwLineCur), m_dwColCur(dwColCur) {}
        DWORD LexemeLength()    const {return m_pcCur - m_pcBeg;}
        operator LexicalError() const {return LexicalError(m_dwLineBeg, m_dwColBeg, m_dwLineCur, m_dwColCur);}
        operator SyntaxError()  const {return SyntaxError(m_dwLineBeg, m_dwColBeg);}
        double   ToNumber()     const
        {
            DWORD dwLen = LexemeLength();
            U* pU = (U*)_alloca((dwLen + 1) * sizeof(U));
            CopyMemory(pU, m_pcBeg, dwLen);
            pU[dwLen] = 0;
            return StringToDouble(pU);
        }

    private:
        Token();    // Forbidden.
    };

    //// Internal representation of an XPath expression.
    //// The XPath expression parser builds a syntax tree.
    //// The nodes of the tree are held in an array. Nodes may occupy several
    //// entries in the array, depending on their number of parameters.
    //// A node can have a fixed number of parameters, possibly representing
    //// additional information and/or indexes into the array to the node's children.
    //// A node can point to a list of other nodes by keeping an index to the
    //// first node, which itself will point to the next.
    enum SYNTAX_TREE_NODE           // Syntax tree node types.
                                    // All operators are left-associative, and
                                    // listed by order of increasing precedence.
    {
        STREE_OR,                   // 2 parameters: indexes to the child nodes
        STREE_AND,                  // 2 parameters: indexes to the child nodes
        STREE_EQUALITY,             // 3 parameters: subtype, indexes to the child nodes
        STREE_RELATIONAL,           // 3 parameters: subtype, indexes to the child nodes
        STREE_ADDITIVE,             // 3 parameters: subtype, indexes to the child nodes
        STREE_MULTIPLICATIVE,       // 3 parameters: subtype, indexes to the child nodes
        STREE_UNARY_MINUS,          // 1 parameter:  index to the child node
        STREE_UNION,                // 2 parameters: indexes to the child nodes
        STREE_ROOT_SLASH,           // 1 parameter:  index to the child node
        STREE_SLASH,                // 2 parameters: indexes to the child nodes
        STREE_FILTER_EXPRESSION,    // 2 parameters: index to the child node, index to the first predicate
        STREE_STEP,                 // 3 parameters: indexes to axis, nodetest, first predicate
        STREE_AXIS,                 // 1 parameter:  axis type
        STREE_NODETEST,             // 1 or 2 parameters: nodetest type, nothing or NCName or QName or literal depending on the nodetest type
        STREE_PREDICATE,            // 2 parameters: index to expression, index to next predicate
        STREE_LITERAL,              // 1 parameter:  string value
        STREE_VARIABLE_REF,         // 1 parameter:  variable name (string value)
        STREE_NUMBER,               // 1 parameter:  value (size of a double)
        STREE_FUNCTION_CALL,        // 2 parameters: function name (string value), index to first argument
        STREE_FUNCTION_ARGUMENT,    // 2 parameters: index to expression, index to next argument
    };
    enum SYNTAX_TREE_NODE_SUBTYPE   // Detail of the node subtypes.
    {
        EQUALITY_EQUALS,
        EQUALITY_NOT_EQUALS,
        RELATIONAL_GREATER,
        RELATIONAL_GREATER_OR_EQUAL,
        RELATIONAL_LESS,
        RELATIONAL_LESS_OR_EQUAL,
        ADDITIVE_PLUS,
        ADDITIVE_MINUS,
        MULTIPLICATIVE_MUTIPLY,
        MULTIPLICATIVE_DIV,
        MULTIPLICATIVE_MOD,
    };
    U*    m_pExpressionString;      // Character string for the XPath expression.
    std::vector<DWORD> m_syntaxTree;// The syntax tree array. The first entry (0) is unused!
    DWORD m_dwRootNode;             // Index into the array of the root node.
    const XmlParse<U>::XMLTree::NodeRoot* m_pXMLTreeRoot;   // Root of the XML document tree.

    void GetNextXPathToken(Token& token);
    void Match(XPATH_TOKEN eToken, Token& lookahead);

    // Syntax rules helpers. The return value of each function
    // represents the index into the syntax tree array of the
    // root node of the subtree built by the function call.
    DWORD AbbreviatedAbsoluteLocationPath(Token& lookahead);
    DWORD AbbreviatedAxisSpecifier       (Token& lookahead);
    DWORD AbbreviatedStep                (Token& lookahead);
    DWORD AbsoluteLocationPath           (Token& lookahead);
    DWORD AdditiveExpr                   (Token& lookahead);
    DWORD AndExpr                        (Token& lookahead);
    DWORD Argument                       (Token& lookahead);
    DWORD AxisSpecifier                  (Token& lookahead);
    DWORD EqualityExpr                   (Token& lookahead);
    DWORD Expr                           (Token& lookahead);
    DWORD FilterExpr                     (Token& lookahead);
    DWORD FunctionCall                   (Token& lookahead);
    DWORD LocationPath                   (Token& lookahead);
    DWORD MultiplicativeExpr             (Token& lookahead);
    DWORD NodeTest                       (Token& lookahead);
    DWORD OrExpr                         (Token& lookahead);
    DWORD PathExpr                       (Token& lookahead);
    DWORD Predicate                      (Token& lookahead);
    DWORD PredicateExpr                  (Token& lookahead);
    DWORD PrimaryExpr                    (Token& lookahead);
    DWORD RelationalExpr                 (Token& lookahead);
    DWORD RelativeLocationPath           (Token& lookahead, DWORD dwLeft = 0);
    DWORD Step                           (Token& lookahead);
    DWORD UnaryExpr                      (Token& lookahead);
    DWORD UnionExpr                      (Token& lookahead);

    friend struct Token;

    //// Semantic helpers:

    // Creates a new node, and also reserves space for its parameters.
    // One parameter is assumed to occupy a DWORD.
    DWORD MakeNode(SYNTAX_TREE_NODE eNodeType, DWORD dwNumParam)
    {
        DWORD dwNewIndex = m_syntaxTree.size();
        m_syntaxTree.push_back(eNodeType);
        m_syntaxTree.resize(dwNewIndex + 1 + dwNumParam, 0);
        return dwNewIndex;
    }

    // Creates a new node with 1 parameter.
    DWORD Make1Node(SYNTAX_TREE_NODE eNodeType, DWORD dwParam1)
    {
        DWORD dwRet = MakeNode(eNodeType, 1);
        SetNodeParam(dwRet, 1, dwParam1);
        return dwRet;
    }

    // Creates a new node with 1 parameter of type double, which will be stored as
    // 2 DWORD parameters.
    DWORD Make1Node(SYNTAX_TREE_NODE eNodeType, double d)
    {
        DWORD dwRet = MakeNode(eNodeType, 2);
        __int64 *phi = (__int64*)&d;
        SetNodeParam(dwRet, 1, (DWORD)((*phi) >> 32));
        SetNodeParam(dwRet, 2, (DWORD)((*phi) & 0xFFFFFFFF));
        return dwRet;
    }

    // Retrieve a double stored as 2 DWORD parameters.
    // dwIndex is the index of the first parameter.
    double GetDoubleParam(DWORD dwIndex) const
    {
        double d;
        __int64 *phi = (__int64*)&d;
        *phi = m_syntaxTree[dwIndex];
        *phi <<= 32;
        *phi |= m_syntaxTree[dwIndex + 1];
        return d;
    }

    // Creates a new node with 2 parameters.
    DWORD Make2Node(SYNTAX_TREE_NODE eNodeType, DWORD dwParam1, DWORD dwParam2)
    {
        DWORD dwRet = MakeNode(eNodeType, 2);
        SetNodeParam(dwRet, 1, dwParam1);
        SetNodeParam(dwRet, 2, dwParam2);
        return dwRet;
    }

    // Creates a new node with 3 parameters.
    DWORD Make3Node(SYNTAX_TREE_NODE eNodeType, DWORD dwParam1, DWORD dwParam2, DWORD dwParam3)
    {
        DWORD dwRet = MakeNode(eNodeType, 3);
        SetNodeParam(dwRet, 1, dwParam1);
        SetNodeParam(dwRet, 2, dwParam2);
        SetNodeParam(dwRet, 3, dwParam3);
        return dwRet;
    }

    // Sets the value of a parameter. dwParam is the parameter number
    // (starting from 1).
    void SetNodeParam(DWORD dwIndex, DWORD dwParam, DWORD dwValue)
    {
        m_syntaxTree[dwIndex + dwParam] = dwValue;
    }

    // Extract a string from a lexeme. Calls new U[...].
    // Caller must later deallocate.
    // dwBeg: index of 1st character to extract.
    // dwLen: number of characters to extract (excluding NULL terminator). If 0xFFFFFFFF,
    //        the length of the whole lexeme is computed.
    U* CopyLexeme(const Token& token, DWORD dwBeg = 0, DWORD dwLen = 0xFFFFFFFF)
    {
        if(dwLen == 0xFFFFFFFF)
            dwLen = token.LexemeLength();
        return StringCopy(new U[dwLen + 1], token.m_pcBeg + dwBeg, token.m_pcBeg + dwBeg + dwLen);
    }


    //// Expression evaluation helpers:

    // Composing a sequence of steps with a following step S.
    // dwIndex: the index of S in the parse tree.
    // After the call, outNodeSet is sorted as dictated by the axis of S.
    void ComposeWithStep(const OrderedNodeSet& initialNodeSet,
        DWORD dwIndex, OrderedNodeSet& outNodeSet);

    // Evaluating a location step for a given context node. Builds initial node set,
    // which is the set of nodes having the relationship to the context node specified
    // by the axis, and having the node type and expanded name dpecified by the node
    // test.
    void EvaluateStep(NodePTR pContextNode, DWORD dwIndex, Axis eAxis, OrderedNodeSet& outNodeSet);

    // Filter a node set through a predicate.
    void Filter(DWORD dwIndexPredicate, Axis eAxis, OrderedNodeSet& outNodeSet);

    // Filter a node a through predicate. Returns TRUE if the node was
    // filtered out by the predicate.
    BOOL FilterNode(DWORD dwIndexPredicate, Axis eAxis, NodePTR pContextNode, DWORD dwContextSize, DWORD dwContextPosition);

    // Evaluate a node test for a single node. If the evaluation succeeds, the node
    // is added to outNodeSet. bAttributeAxis is TRUE when the current axis is 'attribute'.
    void EvaluateNodeTestForNode(NodePTR pContextNode, NodeTestKind nt, CUPTR pString, OrderedNodeSet& outNodeSet, BOOL bAttributeAxis = FALSE);

    // Evaluate a node test for a node's children, and union the
    // resulting nodes into the output node set.
    void EvaluateNodeTestForChildren(NodePTR pNode, NodeTestKind nt, CUPTR pString, OrderedNodeSet& outNodeSet);

    // Evaluate a node test for a node's descendants, and union the
    // resulting nodes into the output node set.
    void EvaluateNodeTestForDescendants(NodePTR pNode, NodeTestKind nt, CUPTR pString, OrderedNodeSet& outNodeSet);

    // Evaluate an XPath expression.
    void EvaluateExpression(DWORD dwIndexExpression, Axis eAxis, NodePTR pContextNode,
        DWORD dwContextSize, DWORD dwContextPosition, XPathExprValue& val);

    // Boolean test between a node set and a value that is not a node set.
    // These functions are used to test equality or relational expressions.
    BOOL NodeSetCompare(const OrderedNodeSet& nodeSet, const XPathExprValue& val, SYNTAX_TREE_NODE_SUBTYPE eOperatorType);
    BOOL NodeSetCompare(const XPathExprValue& val, const OrderedNodeSet& nodeSet, SYNTAX_TREE_NODE_SUBTYPE eOperatorType);

    // Boolean test between 2 node sets.
    // These functions are used to test equality or relational expressions.
    BOOL NodeSetCompare(const OrderedNodeSet& nodeSet1, const OrderedNodeSet& nodeSet2, SYNTAX_TREE_NODE_SUBTYPE eOperatorType);

    // Other heplers:
    static U* StringCopy(U* pDest, CUPTR pSource, CUPTR pEnd)
    {
        CopyMemory(pDest, pSource, (pEnd-pSource) * sizeof(U));
        pDest[pEnd-pSource] = 0;
        return pDest;
    }
    static BOOL ToNumber(CUPTR pString, double& dRet)
    {
        BOOL bRet = FALSE;
        dRet = 0.0;
        BOOL bMinus = FALSE;
        CUPTR pStr = pString;
        while(*pStr == ' ' || *pStr == '\t' || *pStr == '\r' || *pStr == '\n')
            ++pStr;
        if(*pStr == '-')
        {
            // Optional minus sign present.
            bMinus = TRUE;
            ++pStr;
            while(*pStr == ' ' || *pStr == '\t' || *pStr == '\r' || *pStr == '\n')
                ++pStr;
        }
        CUPTR pBegin = pStr;
        if(*pStr >= '0' && *pStr <= '9')
        {
            // No integral part.
            while(*pStr >= '0' && *pStr <= '9')
                ++pStr;
        }
        if(*pStr == '.')
        {
            // Pickup decimal part.
            ++pStr;
            if(*pStr >= '0' && *pStr <= '9')
            {
                while(*pStr >= '0' && *pStr <= '9')
                    ++pStr;
            }
        }
        if(pBegin != pStr)
        {
            DWORD dwLen = pStr - pBegin;
            U* pU = (U*)_alloca((dwLen + 1) * sizeof(U));
            CopyMemory(pU, pBegin, dwLen);
            pU[dwLen] = 0;
            dRet = StringToDouble(pU);
            if(bMinus)
                dRet = -dRet;
            bRet = TRUE;
        }
        return bRet;
    }
    static double StringToDouble(const char* pString)
    {
        return strtod(pString, NULL);
    }
    static double StringToDouble(const wchar_t* pString)
    {
        return wcstod(pString, NULL);
    }
    static int compare(const char *string1, const char *string2)
    {
        return lstrcmpA(string1, string2);
    }
    static int compare(const wchar_t *string1, const wchar_t *string2)
    {
        return lstrcmpW(string1, string2);
    }

    // String selector:
    static const char*    string_selector(const char*, const char* s1, const wchar_t*)
    {
        return s1;
    }
    static const wchar_t* string_selector(const wchar_t*, const char*, const wchar_t* s2)
    {
        return s2;
    }
    #define STR(s)        string_selector((const U*)0, s, L##s)


#ifdef XML_DUMP
    // Dumping helpers:
    static void Margin(std::basic_ostream<U>& os, DWORD dwIndent)
    {
        U* pcMargin = (U*)_alloca((dwIndent + 1) * sizeof(U));
        for(DWORD dw = 0; dw < dwIndent; ++dw)
            pcMargin[dw] = 32;
        pcMargin[dwIndent] = 0;
        os << pcMargin;
    }
    DWORD DumpNode(std::basic_ostream<U>& os, DWORD dwIndex, DWORD dwIndent = 0, BOOL bRecurse = TRUE) const;
    DWORD DumpNodeIndirect(std::basic_ostream<U>& os, DWORD dwIndex, DWORD dwIndent = 0, BOOL bRecurse = TRUE) const
    {return DumpNode(os, m_syntaxTree[dwIndex], dwIndent, bRecurse);}
#endif
};

#include "XPath.inl"

#endif // _XMLPATH_HPP
