/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include <math.h>

template<typename U>
BOOL XPathExpression<U>::ToBOOL(XPathExprValue& val)
{
    BOOL bRet = FALSE;
    switch(val.m_eType)
    {
        case TYPE_NODESET:
            val = bRet = !val.m_nodeSet.empty();
            break;

        case TYPE_STRING:
            val = bRet = *val.m_string != 0;
            break;

        case TYPE_NUMBER:
            val = bRet = val.m_dNumber != 0.0;
            break;

        case TYPE_BOOLEAN:
            bRet = TRUE;
            break;

        case TYPE_UNKNOWN:
            break;
    }

    return bRet;
}

template<typename U>
BOOL XPathExpression<U>::ToNumber(XPathExprValue& val)
{
    BOOL bRet = FALSE;
    switch(val.m_eType)
    {
        case TYPE_NODESET:
            {
                if(val.m_nodeSet.empty())
                {
                    val.Clear();    // Convertion failed.
                    break;
                }
                else
                {
                    // First, convert to a string:
                    NodePTR pNode = IsDocumentOrder(val.m_nodeSet) ?
                        *val.m_nodeSet.begin() : *val.m_nodeSet.rbegin();
                    val = pNode->StringValue().c_str();
                }
            }
            // Fall through:

        case TYPE_STRING:
            {
                double d;
                bRet = ToNumber(val.m_string, d);
                if(bRet)
                    val = d;
                else
                    val.Clear();
            }
            break;

        case TYPE_BOOLEAN:
            bRet = TRUE;
            val = val.m_bBoolean ? 1.0 : 0.0;
            break;

        case TYPE_NUMBER:
            bRet = TRUE;
            break;

        case TYPE_UNKNOWN:
            break;
    }

    return bRet;
}

template<typename U>
BOOL XPathExpression<U>::ToString(XPathExprValue& val)
{
    BOOL bRet = FALSE;
    switch(val.m_eType)
    {
        case TYPE_NODESET:
            if(val.m_nodeSet.empty())
                val = STR("empty");
            else
            {
                NodePTR pNode = IsDocumentOrder(val.m_nodeSet) ?
                    *val.m_nodeSet.begin() : *val.m_nodeSet.rbegin();
                val = pNode->StringValue().c_str();
            }
            bRet = TRUE;
            break;

        case TYPE_BOOLEAN:
            val = val.m_bBoolean ? STR("true") : STR("false");
            bRet = TRUE;
            break;

        case TYPE_NUMBER:
            {
                std::basic_ostringstream<U> os;
                os << std::setprecision(18) << val.m_dNumber;
                val = os.str().c_str();
            }
            break;

        case TYPE_STRING:
            bRet = TRUE;
            break;

        case TYPE_UNKNOWN:
            break;
    }

    return bRet;
}



            /////////////////// Lexical rules ///////////////////

template<typename U>
void XPathExpression<U>::GetNextXPathToken(Token& token)
{
    if(token.m_eToken == TOKEN_XPATH_END)
        return;

    // Delare some useful shortcuts.
    XPATH_TOKEN& eToken           = token.m_eToken;
    XPATH_TOKEN_SUBTYPE& eSubtype = token.m_eSubtype;
    CUPTR& pcCur  = token.m_pcCur;
    DWORD& dwLine = token.m_dwLineCur;
    DWORD& dwCol  = token.m_dwColCur;

    // Save previous token type.
    XPATH_TOKEN ePreviousTokenType = eToken;

    // Reinitialise the current token.
    eToken            = TOKEN_XPATH_ERROR;
    eSubtype          = XPATH_TOKEN_SUBTYPE_NONE;
    token.m_pcBeg     = pcCur;
    token.m_dwLineBeg = dwLine;
    token.m_dwColBeg  = dwCol;

    BOOL bContinue;
    do
    {
        bContinue = FALSE;

        switch(*pcCur)
        {
            case 0:
                eToken = TOKEN_XPATH_END;
                break;

            case '\n':
                    ++dwLine;
                    dwCol = 1;  // New line.
                    // Fall through.
            case ' ':
            case '\t':
            case '\r':
                // Whitespace is transparent.
                {
                    if(*pcCur == ' ' || *pcCur == '\t')
                        ++dwCol;
                    ++pcCur;
                    U c;
                    while((c = *pcCur) == ' ' || c == '\t' || c == '\r' || c == '\n')
                    {
                        if(c == '\n')
                        {
                            ++dwLine;
                            dwCol = 1;  // New line.
                        }
                        else if(c != '\r')
                            ++dwCol;
                        ++pcCur;
                    }
                    token.m_pcBeg = pcCur;
                    bContinue = TRUE;
                }
                break;

            case '(':
                eToken = TOKEN_XPATH_LEFT_PARENTHESIS;
                ++pcCur;
                ++dwCol;
                break;

            case ')':
                eToken = TOKEN_XPATH_RIGHT_PARENTHESIS;
                ++pcCur;
                ++dwCol;
                break;

            case '[':
                eToken = TOKEN_XPATH_LEFT_SQUAREBRACKET;
                ++pcCur;
                ++dwCol;
                break;

            case ']':
                eToken = TOKEN_XPATH_RIGHT_SQUAREBRACKET;
                ++pcCur;
                ++dwCol;
                break;

            case '.':
                ++pcCur;
                ++dwCol;
                if(*pcCur == '.')
                {
                    eToken = TOKEN_XPATH_DOTDOT;
                    ++pcCur;
                    ++dwCol;
                }
                else if(*pcCur >= '0' && *pcCur <= '9')
                {
                    // Number.
                    eToken = TOKEN_XPATH_NUMBER;
                    ++pcCur;
                    ++dwCol;
                    while(*pcCur >= '0' && *pcCur <= '9')
                    {
                        ++pcCur;
                        ++dwCol;
                    }
                }
                else
                    eToken = TOKEN_XPATH_DOT;
                break;

            case '@':
                eToken = TOKEN_XPATH_AT;
                ++pcCur;
                ++dwCol;
                break;

            case ',':
                eToken = TOKEN_XPATH_COMMA;
                ++pcCur;
                ++dwCol;
                break;

            case ':':
                if(pcCur[1] == ':')
                {
                    eToken = TOKEN_XPATH_AXIS_DELIMITER;
                    pcCur += 2;
                    dwCol += 2;
                }
                break;

            case '"':
            case '\'':
                {
                    U cDelim = *pcCur;
                    ++pcCur;
                    ++dwCol;
                    while(*pcCur && *pcCur != cDelim)
                    {
                        ++pcCur;
                        if(*pcCur == '\n')
                        {
                            ++dwLine;
                            dwCol = 1;   // New line.
                        }
                        else if(*pcCur != '\r')
                            ++dwCol;
                    }
                    if(*pcCur == cDelim)
                    {
                        eToken = TOKEN_XPATH_LITERAL;
                        ++pcCur;
                        ++dwCol;
                    }
                    else
                        throw LexicalError(token);
                }
                break;

            case '$':
                ++pcCur;
                ++dwCol;
                if(IS_NCNAME_FIRSTCHAR(*pcCur))
                {
                    eToken = TOKEN_XPATH_VARIABLE_REFERENCE;
                    ++pcCur;
                    ++dwCol;
                    while(IS_NCNAME_OTHERCHAR(*pcCur))
                    {
                        ++pcCur;
                        ++dwCol;
                    }
                    if(*pcCur == ':')
                    {
                        if(IS_NCNAME_OTHERCHAR(pcCur[1]))
                        {
                            // Match the LocalPart of the QName:
                            pcCur += 2;
                            dwCol += 2;
                            while(IS_NCNAME_OTHERCHAR(*pcCur))
                            {
                                ++pcCur;
                                ++dwCol;
                            }
                        }
                    }
                }
                else
                    throw LexicalError(token);
                break;

            case '/':
                eToken = TOKEN_XPATH_OPERATOR;
                ++pcCur;
                ++dwCol;
                if(*pcCur == '/')
                {
                    eSubtype = OPERATOR_DOUBLE_SLASH;
                    ++pcCur;
                    ++dwCol;
                }
                else
                    eSubtype = OPERATOR_SLASH;
                break;

            case '|':
                eToken = TOKEN_XPATH_OPERATOR;
                eSubtype = OPERATOR_PIPE;
                ++pcCur;
                ++dwCol;
                break;

            case '+':
                eToken = TOKEN_XPATH_OPERATOR;
                eSubtype = OPERATOR_PLUS;
                ++pcCur;
                ++dwCol;
                break;

            case '-':
                eToken = TOKEN_XPATH_OPERATOR;
                eSubtype = OPERATOR_MINUS;
                ++pcCur;
                ++dwCol;
                break;

            case '*':
                ++pcCur;
                ++dwCol;

                // Disambiguating rule 1 (section 3.7).
                if(ePreviousTokenType == TOKEN_XPATH_ERROR              ||  // No preceding token.
                   ePreviousTokenType == TOKEN_XPATH_AT                 ||
                   ePreviousTokenType == TOKEN_XPATH_AXIS_DELIMITER     ||
                   ePreviousTokenType == TOKEN_XPATH_LEFT_PARENTHESIS   ||
                   ePreviousTokenType == TOKEN_XPATH_LEFT_SQUAREBRACKET ||
                   ePreviousTokenType == TOKEN_XPATH_COMMA              ||
                   ePreviousTokenType == TOKEN_XPATH_OPERATOR)
                {
                   eToken = TOKEN_XPATH_NAMETEST;
                   eSubtype = NAMETEST_STAR;
                }
                else
                {
                    eToken = TOKEN_XPATH_OPERATOR;
                    eSubtype = OPERATOR_MULTIPLY;
                }
                break;

            case '=':
                eToken = TOKEN_XPATH_OPERATOR;
                eSubtype = OPERATOR_EQUALS;
                ++pcCur;
                ++dwCol;
                break;

            case '<':
                eToken = TOKEN_XPATH_OPERATOR;
                ++pcCur;
                ++dwCol;
                if(*pcCur == '=')
                {
                    eSubtype = OPERATOR_LESS_OR_EQUAL;
                    ++pcCur;
                    ++dwCol;
                }
                else
                    eSubtype = OPERATOR_LESS;
                break;

            case '>':
                eToken = TOKEN_XPATH_OPERATOR;
                ++pcCur;
                ++dwCol;
                if(*pcCur == '=')
                {
                    eSubtype = OPERATOR_GREATER_OR_EQUAL;
                    ++pcCur;
                    ++dwCol;
                }
                else
                    eSubtype = OPERATOR_GREATER;
                break;

            case '!':
                if(pcCur[1] == '=')
                {
                    eToken = TOKEN_XPATH_OPERATOR;
                    eSubtype = OPERATOR_NOT_EQUALS;
                    pcCur += 2;
                    dwCol += 2;
                }
                break;

            default:
                // Number.
                if(*pcCur >= '0' && *pcCur <= '9')
                {
                    eToken = TOKEN_XPATH_NUMBER;
                    ++pcCur;
                    ++dwCol;
                    while(*pcCur >= '0' && *pcCur <= '9')
                    {
                        ++pcCur;
                        ++dwCol;
                    }
                    if(*pcCur == '.')
                    {
                        ++pcCur;
                        ++dwCol;
                        while(*pcCur >= '0' && *pcCur <= '9')
                        {
                            ++pcCur;
                            ++dwCol;
                        }
                    }
                }
                else
                {
                    // Node type or operator or axis name keywords.
                    switch(*pcCur)
                    {
                        case 'a':
                            if(pcCur[1] == 'n')
                            {
                                if(pcCur[2] == 'd')
                                {
                                    eToken = TOKEN_XPATH_OPERATOR;
                                    eSubtype = OPERATOR_NAME_AND;
                                    pcCur += 3;
                                    dwCol += 3;
                                }
                                else
                                {
                                    const U* pc = pcCur + 1;
                                    if(    *++pc == 'c'
                                        && *++pc == 'e'
                                        && *++pc == 's'
                                        && *++pc == 't'
                                        && *++pc == 'o'
                                        && *++pc == 'r'
                                      )
                                    {
                                        eToken = TOKEN_XPATH_AXISNAME;
                                        pcCur += 8;
                                        dwCol += 8;
                                        if(    *++pc == '-'
                                            && *++pc == 'o'
                                            && *++pc == 'r'
                                            && *++pc == '-'
                                            && *++pc == 's'
                                            && *++pc == 'e'
                                            && *++pc == 'l'
                                            && *++pc == 'f'
                                          )
                                        {
                                            pcCur += 8;
                                            dwCol += 8;
                                        }
                                    }
                                }
                            }
                            else if(pcCur[1] == 't')
                            {
                                const U* pc = pcCur + 1;
                                if(    *++pc == 't'
                                    && *++pc == 'r'
                                    && *++pc == 'i'
                                    && *++pc == 'b'
                                    && *++pc == 'u'
                                    && *++pc == 't'
                                    && *++pc == 'e'
                                  )
                                {
                                    eToken = TOKEN_XPATH_AXISNAME;
                                    pcCur += 9;
                                    dwCol += 9;
                                }
                            }
                            break;

                        case 'c':
                            if(pcCur[1] == 'h')
                            {
                                const U* pc = pcCur + 1;
                                if(    *++pc == 'i'
                                    && *++pc == 'l'
                                    && *++pc == 'd'
                                  )
                                {
                                    eToken = TOKEN_XPATH_AXISNAME;
                                    pcCur += 5;
                                    dwCol += 5;
                                }
                            }
                            else
                            {
                                const U* pc = pcCur;
                                if(    *++pc == 'o'
                                    && *++pc == 'm'
                                    && *++pc == 'm'
                                    && *++pc == 'e'
                                    && *++pc == 'n'
                                    && *++pc == 't'
                                  )
                                {
                                    eToken = TOKEN_XPATH_NODETYPE;
                                    pcCur += 7;
                                    dwCol += 7;
                                }
                            }
                            break;

                        case 'd':
                            if(pcCur[1] == 'i' && pcCur[2] == 'v')
                            {
                                eToken = TOKEN_XPATH_OPERATOR;
                                eSubtype = OPERATOR_NAME_DIV;
                                pcCur += 3;
                                dwCol += 3;
                            }
                            else
                            {
                                const U* pc = pcCur;
                                if(    *++pc == 'e'
                                    && *++pc == 's'
                                    && *++pc == 'c'
                                    && *++pc == 'e'
                                    && *++pc == 'n'
                                    && *++pc == 'd'
                                    && *++pc == 'a'
                                    && *++pc == 'n'
                                    && *++pc == 't'
                                  )
                                {
                                    eToken = TOKEN_XPATH_AXISNAME;
                                    pcCur += 10;
                                    dwCol += 10;
                                    if(    *++pc == '-'
                                        && *++pc == 'o'
                                        && *++pc == 'r'
                                        && *++pc == '-'
                                        && *++pc == 's'
                                        && *++pc == 'e'
                                        && *++pc == 'l'
                                        && *++pc == 'f'
                                      )
                                    {
                                        pcCur += 8;
                                        dwCol += 8;
                                    }
                                }
                            }
                            break;

                        case 'f':
                            {
                                const U* pc = pcCur;
                                if(    *++pc == 'o'
                                    && *++pc == 'l'
                                    && *++pc == 'l'
                                    && *++pc == 'o'
                                    && *++pc == 'w'
                                    && *++pc == 'i'
                                    && *++pc == 'n'
                                    && *++pc == 'g'
                                  )
                                {
                                    eToken = TOKEN_XPATH_AXISNAME;
                                    pcCur += 9;
                                    dwCol += 9;
                                    if(    *++pc == '-'
                                        && *++pc == 's'
                                        && *++pc == 'i'
                                        && *++pc == 'b'
                                        && *++pc == 'l'
                                        && *++pc == 'i'
                                        && *++pc == 'n'
                                        && *++pc == 'g'
                                      )
                                    {
                                        pcCur += 8;
                                        dwCol += 8;
                                    }
                                }
                            }
                            break;

                        case 'm':
                            {
                                const U* pc = pcCur;
                                if(    *++pc == 'o'
                                    && *++pc == 'd'
                                  )
                                {
                                    eToken = TOKEN_XPATH_OPERATOR;
                                    eSubtype = OPERATOR_NAME_MOD;
                                    pcCur += 3;
                                    dwCol += 3;
                                }
                            }
                            break;

                        case 'n':
                            if(pcCur[1] == 'o' && pcCur[2] == 'd' && pcCur[3] == 'e')
                            {
                                eToken = TOKEN_XPATH_NODETYPE;
                                pcCur += 4;
                                dwCol += 4;
                            }
                            else
                            {
                                const U* pc = pcCur;
                                if(    *++pc == 'a'
                                    && *++pc == 'm'
                                    && *++pc == 'e'
                                    && *++pc == 's'
                                    && *++pc == 'p'
                                    && *++pc == 'a'
                                    && *++pc == 'c'
                                    && *++pc == 'e'
                                  )
                                {
                                    eToken = TOKEN_XPATH_AXISNAME;
                                    pcCur += 9;
                                    dwCol += 9;
                                }
                            }
                            break;

                        case 'o':
                            if(pcCur[1] == 'r')
                            {
                                eToken = TOKEN_XPATH_OPERATOR;
                                eSubtype = OPERATOR_NAME_OR;
                                pcCur += 2;
                                dwCol += 2;
                            }
                            break;

                        case 'p':
                            if(pcCur[1] == 'a')
                            {
                                const U* pc = pcCur + 1;
                                if(    *++pc == 'r'
                                    && *++pc == 'e'
                                    && *++pc == 'n'
                                    && *++pc == 't'
                                  )
                                {
                                    eToken = TOKEN_XPATH_AXISNAME;
                                    pcCur += 6;
                                    dwCol += 6;
                                }
                            }
                            else if(pcCur[1] == 'r')
                            {
                                if(pcCur[2] == 'o')
                                {
                                    const U* pc = pcCur + 2;
                                    if(    *++pc == 'c'
                                        && *++pc == 'e'
                                        && *++pc == 's'
                                        && *++pc == 's'
                                        && *++pc == 'i'
                                        && *++pc == 'n'
                                        && *++pc == 'g'
                                        && *++pc == '-'
                                        && *++pc == 'i'
                                        && *++pc == 'n'
                                        && *++pc == 's'
                                        && *++pc == 't'
                                        && *++pc == 'r'
                                        && *++pc == 'u'
                                        && *++pc == 'c'
                                        && *++pc == 't'
                                        && *++pc == 'i'
                                        && *++pc == 'o'
                                        && *++pc == 'n'
                                      )
                                    {
                                        eToken = TOKEN_XPATH_NODETYPE;
                                        eSubtype = NODETYPE_PROCESSING_INSTRUCTION;
                                        pcCur += 22;
                                        dwCol += 22;
                                    }
                                }
                                else
                                {
                                    const U* pc = pcCur + 1;
                                    if(    *++pc == 'e'
                                        && *++pc == 'c'
                                        && *++pc == 'e'
                                        && *++pc == 'd'
                                        && *++pc == 'i'
                                        && *++pc == 'n'
                                        && *++pc == 'g'
                                      )
                                    {
                                        eToken = TOKEN_XPATH_AXISNAME;
                                        pcCur += 9;
                                        dwCol += 9;
                                        if(    *++pc == '-'
                                            && *++pc == 's'
                                            && *++pc == 'i'
                                            && *++pc == 'b'
                                            && *++pc == 'l'
                                            && *++pc == 'i'
                                            && *++pc == 'n'
                                            && *++pc == 'g'
                                          )
                                        {
                                            pcCur += 8;
                                            dwCol += 8;
                                        }
                                    }
                                }
                            }
                            break;

                        case 's':
                            {
                                const U* pc = pcCur;
                                if(    *++pc == 'e'
                                    && *++pc == 'l'
                                    && *++pc == 'f'
                                  )
                                {
                                    eToken = TOKEN_XPATH_AXISNAME;
                                    pcCur += 4;
                                    dwCol += 4;
                                }
                            }
                            break;

                        case 't':
                            {
                                const U* pc = pcCur;
                                if(    *++pc == 'e'
                                    && *++pc == 'x'
                                    && *++pc == 't'
                                  )
                                {
                                    eToken = TOKEN_XPATH_NODETYPE;
                                    pcCur += 4;
                                    dwCol += 4;
                                }
                            }
                            break;
                    }

                    // If a NodeType or AxisName or Operator literal was collected,
                    // it could still be the prefix of an NCName!
                    if(eToken == TOKEN_XPATH_NODETYPE ||
                       eToken == TOKEN_XPATH_OPERATOR ||
                       eToken == TOKEN_XPATH_AXISNAME)
                    {
                        if(*pcCur == ':')
                        {
                            if(pcCur[1] == '*')
                            {
                                eToken = TOKEN_XPATH_NAMETEST;
                                eSubtype = NAMETEST_PREFIX_STAR;
                                pcCur += 2;
                                dwCol += 2;
                            }
                            else if(IS_NCNAME_OTHERCHAR(pcCur[1]))
                            {
                                // Match the LocalPart of the QName:
                                eToken = TOKEN_XPATH_NAMETEST;
                                eSubtype = NAMETEST_QNAME;
                                pcCur += 2;
                                dwCol += 2;
                                while(IS_NCNAME_OTHERCHAR(*pcCur))
                                {
                                    ++pcCur;
                                    ++dwCol;
                                }
                            }
                        }
                        else if(IS_NCNAME_OTHERCHAR(*pcCur))
                        {
                            // Longer match as an NCName or QName is possible.
                            eToken = TOKEN_XPATH_NAMETEST;
                            ++pcCur;
                            ++dwCol;
                            while(IS_NCNAME_OTHERCHAR(*pcCur))
                            {
                                ++pcCur;
                                ++dwCol;
                            }
                            if(*pcCur == ':')
                            {
                                if(pcCur[1] == '*')
                                {
                                    eSubtype = NAMETEST_PREFIX_STAR;
                                    pcCur += 2;
                                    dwCol += 2;
                                }
                                else if(IS_NCNAME_OTHERCHAR(pcCur[1]))
                                {
                                    // Match the LocalPart of the QName:
                                    eSubtype = NAMETEST_QNAME;
                                    pcCur += 2;
                                    dwCol += 2;
                                    while(IS_NCNAME_OTHERCHAR(*pcCur))
                                    {
                                        ++pcCur;
                                        ++dwCol;
                                    }
                                }
                            }
                            else
                                eSubtype = NAMETEST_QNAME;
                        }
                    }
                    else
                    {
                        // NCName or QName.
                        if(IS_NCNAME_FIRSTCHAR(*pcCur))
                        {
                            eToken = TOKEN_XPATH_NAMETEST;
                            ++pcCur;
                            ++dwCol;
                            while(IS_NCNAME_OTHERCHAR(*pcCur))
                            {
                                ++pcCur;
                                ++dwCol;
                            }
                            if(*pcCur == ':')
                            {
                                if(pcCur[1] == '*')
                                {
                                    eSubtype = NAMETEST_PREFIX_STAR;
                                    pcCur += 2;
                                    dwCol += 2;
                                }
                                else if(IS_NCNAME_OTHERCHAR(pcCur[1]))
                                {
                                    // Match the LocalPart of the QName:
                                    eSubtype = NAMETEST_QNAME;
                                    pcCur += 2;
                                    dwCol += 2;
                                    while(IS_NCNAME_OTHERCHAR(*pcCur))
                                    {
                                        ++pcCur;
                                        ++dwCol;
                                    }
                                }
                            }
                            else
                                eSubtype = NAMETEST_QNAME;
                        }
                    }

                    // Disambiguating rule 1 (section 3.7).
                    if(eToken == TOKEN_XPATH_OPERATOR)
                    {
                        if(ePreviousTokenType == TOKEN_XPATH_AT                 ||
                           ePreviousTokenType == TOKEN_XPATH_AXIS_DELIMITER     ||
                           ePreviousTokenType == TOKEN_XPATH_LEFT_PARENTHESIS   ||
                           ePreviousTokenType == TOKEN_XPATH_LEFT_SQUAREBRACKET ||
                           ePreviousTokenType == TOKEN_XPATH_COMMA              ||
                           ePreviousTokenType == TOKEN_XPATH_OPERATOR)
                        {
                            eToken = TOKEN_XPATH_NAMETEST;
                            eSubtype = NAMETEST_QNAME;
                        }
                    }

                    // Disambiguating rule 2 (section 3.7) for FunctionName.
                    if(eToken == TOKEN_XPATH_NAMETEST && eSubtype == NAMETEST_QNAME)
                    {
                        // Examine first character after possible whitespace.
                        const U* pc = pcCur;
                        while(*pc == ' ' || *pc == '\t' || *pc == '\r' || *pc == '\n')
                            ++pc;
                        if(*pc == '(')
                        {
                            // Switch to FunctionName.
                            eToken = TOKEN_XPATH_FUNCTIONNAME;
                            eSubtype = XPATH_TOKEN_SUBTYPE_NONE;
                        }
                    }
                    else if(eToken == TOKEN_XPATH_NODETYPE) // Disambiguating rule 2 (section 3.7) for NodeType.
                    {
                        // Examine first character after possible whitespace.
                        const U* pc = pcCur;
                        while(*pc == ' ' || *pc == '\t' || *pc == '\r' || *pc == '\n')
                            ++pc;
                        if(*pc != '(')
                        {
                            // Switch to NameTest.
                            eToken = TOKEN_XPATH_NAMETEST;
                            eSubtype = NAMETEST_QNAME;
                        }
                    }

                    // Disambiguating rule 3 (section 3.7).
                    if(eToken == TOKEN_XPATH_AXISNAME)
                    {
                        // Examine first 2 characters after possible whitespace.
                        const U* pc = pcCur;
                        while(*pc == ' ' || *pc == '\t' || *pc == '\r' || *pc == '\n')
                            ++pc;
                        if(*pc != ':' || pc[1] != ':')
                        {
                            // Switch to NameTest.
                            eToken = TOKEN_XPATH_NAMETEST;
                            eSubtype = NAMETEST_QNAME;
                        }
                    }
                }
                break;
        }
    } while(bContinue);
}

template<typename U>
void XPathExpression<U>::Match(XPATH_TOKEN eToken, Token& lookahead)
{
    if(lookahead.m_eToken != eToken)
        throw SyntaxError(lookahead);
    else
        GetNextXPathToken(lookahead);
}

template<typename U>
BOOL XPathExpression<U>::Parse(CUPTR zXPathExpr)
{
    m_syntaxTree.clear();
    m_syntaxTree.reserve(128);
    m_syntaxTree.push_back(0xFFFFFFFF);  // The first entry is unused!
    Token lookahead(zXPathExpr);
    GetNextXPathToken(lookahead);
    m_dwRootNode = Expr(lookahead);
    if(m_dwRootNode == 0 || lookahead.m_eToken != TOKEN_XPATH_END)
        throw SyntaxError(lookahead);

    m_pExpressionString = DuplicateString(zXPathExpr);
    return TRUE;
}


            /////////////////// Syntax rules ///////////////////


// Rule [1]:
template<typename U>
DWORD XPathExpression<U>::LocationPath(Token& lookahead)
{
    DWORD dwRet;
    if(lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
       (lookahead.m_eSubtype == OPERATOR_SLASH ||
        lookahead.m_eSubtype == OPERATOR_DOUBLE_SLASH)
      )
        dwRet = AbsoluteLocationPath(lookahead);
    else
        dwRet = RelativeLocationPath(lookahead);

    return dwRet;
}

// Rule [2]:
template<typename U>
DWORD XPathExpression<U>::AbsoluteLocationPath(Token& lookahead)
{
    DWORD dwRet;
    if(lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
       lookahead.m_eSubtype == OPERATOR_SLASH)
    {
        GetNextXPathToken(lookahead);
        switch(lookahead.m_eToken)
        {
            case TOKEN_XPATH_AXISNAME:
            case TOKEN_XPATH_DOT:
            case TOKEN_XPATH_DOTDOT:
            case TOKEN_XPATH_AT:
            case TOKEN_XPATH_NAMETEST:
            case TOKEN_XPATH_NODETYPE:
                // The relative location path is set to 0; it will be updated in
                // the call to RelativeLocationPath below.
                dwRet = Make1Node(STREE_ROOT_SLASH, (DWORD)0);
                dwRet = RelativeLocationPath(lookahead, dwRet);
                break;
            default:
                // Create a ROOT_SLASH node and attach child node.
                dwRet = Make1Node(STREE_ROOT_SLASH, (DWORD)0); // No relative location path.
                break;
        }
    }
    else
        dwRet = AbbreviatedAbsoluteLocationPath(lookahead);

    return dwRet;
}

// Rules [3] and [11]:
// dwLeft (left context) is non-zero in 4 situations:
//                 '/'  RelativeLocationPath    -> dwLeft = /
//                 '//' RelativeLocationPath    -> dwLeft = /descendant-or-self::node()
//      FilterExpr '/'  RelativeLocationPath    -> dwLeft = FilterExpr
//      FilterExpr '//' RelativeLocationPath    -> dwLeft = FilterExpr/descendant-or-self::node()
template<typename U>
DWORD XPathExpression<U>::RelativeLocationPath(Token& lookahead, DWORD dwLeft)
{
    DWORD dwRet = Step(lookahead);

    if(dwLeft)
    {
        // There is a left context.
        if(m_syntaxTree[dwLeft] == (DWORD)STREE_ROOT_SLASH && m_syntaxTree[dwLeft + 1] == 0)
        {
            // Update the child of the ROOT_SLASH node, and set the ROOT_SLASH node
            // to be the left context:
            m_syntaxTree[dwLeft + 1] = dwRet;
            dwRet = dwLeft;
        }
        else
        {
            // Compose the left context with the step:
            dwRet = Make2Node(STREE_SLASH, dwLeft, dwRet);
        }
    }

    while(  lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
           (lookahead.m_eSubtype == OPERATOR_SLASH ||
            lookahead.m_eSubtype == OPERATOR_DOUBLE_SLASH)
         )
    {
        // Save the operator subtype.
        XPATH_TOKEN_SUBTYPE eSubtype = lookahead.m_eSubtype;

        GetNextXPathToken(lookahead);
        DWORD dwRet2 = Step(lookahead);

        if(eSubtype == OPERATOR_DOUBLE_SLASH)
        {
            // DOUBLE_SLASH is equivalent to: /descendant-or-self::node()/
            // Append /descendant-or-self::node() to the left context.
            DWORD dwAxis     = Make1Node(STREE_AXIS, (DWORD)descendant_or_self);
            DWORD dwNodeTest = Make1Node(STREE_NODETEST, (DWORD)nt_node);
            DWORD dwTmp      = Make3Node(STREE_STEP, dwAxis, dwNodeTest, 0);
            dwRet            = Make2Node(STREE_SLASH, dwRet, dwTmp);
        }

        // Create a SLASH node, and attach child nodes.
        dwRet = Make2Node(STREE_SLASH, dwRet, dwRet2);
    }

    return dwRet;
}

// Rule [4]:
template<typename U>
DWORD XPathExpression<U>::Step(Token& lookahead)
{
    DWORD dwRet = 0, dwRet2, dwRet3 = 0;    // By default, no predicate.
    switch(lookahead.m_eToken)
    {
        case TOKEN_XPATH_AXISNAME:
        case TOKEN_XPATH_AT:
        case TOKEN_XPATH_NAMETEST:
        case TOKEN_XPATH_NODETYPE:
            dwRet  = AxisSpecifier(lookahead);
            dwRet2 = NodeTest(lookahead);
            {
                DWORD dwPreviousPred = 0;
                while(lookahead.m_eToken == TOKEN_XPATH_LEFT_SQUAREBRACKET)
                {
                    DWORD dwPred = Predicate(lookahead);

                    if(dwRet3 == 0)         // First predicate.
                        dwPreviousPred = dwRet3 = dwPred;
                    else                    // Link to previous predicate.
                        SetNodeParam(dwPreviousPred, 2, dwPred);
                    dwPreviousPred = dwPred;
                }
            }

            // Create a STEP node, and attach child nodes.
            dwRet = Make3Node(STREE_STEP, dwRet, dwRet2, dwRet3);
            break;

        case TOKEN_XPATH_DOT:
        case TOKEN_XPATH_DOTDOT:
            dwRet = AbbreviatedStep(lookahead);
            break;

        default:
            throw SyntaxError(lookahead);
    }

    return dwRet;
}

// Rule [5]:
template<typename U>
DWORD XPathExpression<U>::AxisSpecifier(Token& lookahead)
{
    DWORD dwRet;
    if(lookahead.m_eToken == TOKEN_XPATH_AXISNAME)
    {
        // Extract the axis name:
        Axis eAxis = child;
        switch(lookahead.m_pcBeg[0])
        {
            case 'a':
                if(lookahead.m_pcBeg[1] == 'n')
                {
                    eAxis = lookahead.m_pcBeg[8] == '-' ?
                        ancestor_or_self :
                        ancestor;
                }
                else
                    eAxis = attribute;
                break;
            case 'c':
                eAxis = child;
                break;
            case 'd':
                eAxis = lookahead.m_pcBeg[10] == '-' ?
                    descendant_or_self :
                    descendant;
                break;
            case 'f':
                eAxis = lookahead.m_pcBeg[9] == '-' ?
                    following_sibing :
                    following;
                break;
            case 'n':
                eAxis = namespace_axis;
                break;
            case 'p':
                if(lookahead.m_pcBeg[1] == 'r')
                {
                    eAxis = lookahead.m_pcBeg[9] == '-' ?
                        preceding_sibling :
                        preceding;
                }
                else
                    eAxis = parent;
                break;
            case 's':
                eAxis = self;
                break;
        }

        GetNextXPathToken(lookahead);
        Match(TOKEN_XPATH_AXIS_DELIMITER, lookahead);

        // Create an AXIS node, and attach child node.
        dwRet = Make1Node(STREE_AXIS, (DWORD)eAxis);
    }
    else
        dwRet = AbbreviatedAxisSpecifier(lookahead);

    return dwRet;
}

// Rule [6] is implemented by the lexical analyzer.

// Rule [7]:
template<typename U>
DWORD XPathExpression<U>::NodeTest(Token& lookahead)
{
    DWORD dwRet = 0;
    NodeTestKind eSubtype = nt_star;
    U* pString = NULL;
    switch(lookahead.m_eToken)
    {
        case TOKEN_XPATH_NAMETEST:
            {
                // Extract the nametest type:
                if(*lookahead.m_pcBeg == '*')
                    eSubtype = nt_star;
                else if(lookahead.m_pcCur[-1] == '*')
                {
                    eSubtype = nt_ncname_star;
                    pString = CopyLexeme(lookahead, 0, lookahead.LexemeLength() - 2);
                }
                else
                {
                    eSubtype = nt_qname;
                    pString = CopyLexeme(lookahead);
                }

                GetNextXPathToken(lookahead);

                // Create a NODETEST node, attach child nodes.
                if(eSubtype == nt_star)
                    dwRet = Make1Node(STREE_NODETEST, (DWORD)eSubtype);
                else
                    dwRet = Make2Node(STREE_NODETEST, eSubtype, (DWORD)pString);
            }
            break;

        case TOKEN_XPATH_NODETYPE:
            {
                // Extract the nodetype type:
                switch(*lookahead.m_pcBeg)
                {
                    case 'c': eSubtype = nt_comment;                break;
                    case 'p': eSubtype = nt_processing_instruction; break;
                    case 'n': eSubtype = nt_node;                   break;
                    case 't': eSubtype = nt_text;                   break;
                }

                BOOL bIsProcessingInstruction =
                    lookahead.m_eSubtype == NODETYPE_PROCESSING_INSTRUCTION;
                GetNextXPathToken(lookahead);
                Match(TOKEN_XPATH_LEFT_PARENTHESIS, lookahead);
                if(bIsProcessingInstruction && lookahead.m_eToken == TOKEN_XPATH_LITERAL)
                {
                    // Extract the literal (remove the surrounding quotes):
                    pString = CopyLexeme(lookahead, 1, lookahead.LexemeLength() - 2);

                    // Update the nodetype type:
                    eSubtype = nt_processing_instruction_literal;

                    GetNextXPathToken(lookahead);

                }

                Match(TOKEN_XPATH_RIGHT_PARENTHESIS, lookahead);

                // Create a NODETEST node, attach child nodes.
                if(eSubtype == nt_processing_instruction_literal)
                    dwRet = Make2Node(STREE_NODETEST, eSubtype, (DWORD)pString);
                else
                    dwRet = Make1Node(STREE_NODETEST, (DWORD)eSubtype);
            }
            break;
    }

    return dwRet;
}

// Rule [8]:
template<typename U>
DWORD XPathExpression<U>::Predicate(Token& lookahead)
{
    Match(TOKEN_XPATH_LEFT_SQUAREBRACKET, lookahead);
    DWORD dwRet = PredicateExpr(lookahead);
    Match(TOKEN_XPATH_RIGHT_SQUAREBRACKET, lookahead);

    // Create a new PREDICATE node, and attach its child nodes.
    // Set the second parameter to "no next predicate"; it will
    // possibly be updated by the calling function.
    dwRet = Make2Node(STREE_PREDICATE, dwRet, 0);

    return dwRet;
}

// Rule [9]:
template<typename U>
DWORD XPathExpression<U>::PredicateExpr(Token& lookahead)
{
    return Expr(lookahead);
}

// Rule [10]:
template<typename U>
DWORD XPathExpression<U>::AbbreviatedAbsoluteLocationPath(Token& lookahead)
{
    DWORD dwRet;
    if(lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
       lookahead.m_eSubtype == OPERATOR_DOUBLE_SLASH)
    {
        // DOUBLE_SLASH is equivalent to: /descendant-or-self::node()/
        // Pass /descendant-or-self::node() as left context to RelativeLocationPath.
        DWORD dwAxis     = Make1Node(STREE_AXIS, (DWORD)descendant_or_self);
        DWORD dwNodeTest = Make1Node(STREE_NODETEST, (DWORD)nt_node);
        dwRet            = Make3Node(STREE_STEP, dwAxis, dwNodeTest, 0);
        dwRet            = Make1Node(STREE_ROOT_SLASH, dwRet);

        GetNextXPathToken(lookahead);
        dwRet = RelativeLocationPath(lookahead, dwRet);
    }
    else
        throw SyntaxError(lookahead);

    return dwRet;
}

// Rule [11]: see rule [3].

// Rule [12]:
template<typename U>
DWORD XPathExpression<U>::AbbreviatedStep(Token& lookahead)
{
    DWORD dwRet = 0, dwRet2;
    if(lookahead.m_eToken == TOKEN_XPATH_DOT ||
       lookahead.m_eToken == TOKEN_XPATH_DOTDOT)
    {
        // Save the token.
        XPATH_TOKEN token = lookahead.m_eToken;

        GetNextXPathToken(lookahead);

        if(token == TOKEN_XPATH_DOT)
        {
            // DOT is equivalent to self::node()
            // Generate the equivalent nodes.
            dwRet = Make1Node(STREE_AXIS, (DWORD)self);
        }
        else
        {
            // DOTDOT is equivalent to parent::node()
            // Generate the equivalent nodes.
            dwRet = Make1Node(STREE_AXIS, (DWORD)parent);
        }
        dwRet2 = Make1Node(STREE_NODETEST, (DWORD)nt_node);

        // Create a STEP node, and attach child nodes.
        dwRet = Make3Node(STREE_STEP, dwRet, dwRet2, 0);    // No predicate.
    }

    return dwRet;
}

// Rule [13]:
template<typename U>
DWORD XPathExpression<U>::AbbreviatedAxisSpecifier(Token& lookahead)
{
    Axis eAxis;
    if(lookahead.m_eToken == TOKEN_XPATH_AT)
    {
        GetNextXPathToken(lookahead);
        eAxis = attribute;
    }
    else
        eAxis = child;

    // Create an AXIS node, and attach child node.
    return Make1Node(STREE_AXIS, (DWORD)eAxis);
}

// Rule [14]:
template<typename U>
DWORD XPathExpression<U>::Expr(Token& lookahead)
{
    return OrExpr(lookahead);
}

// Rule [15]:
template<typename U>
DWORD XPathExpression<U>::PrimaryExpr(Token& lookahead)
{
    DWORD dwRet;
    U* pString;
    double d;
    switch(lookahead.m_eToken)
    {
        case TOKEN_XPATH_VARIABLE_REFERENCE:
            // Extract the variable ref.
            pString = CopyLexeme(lookahead, 1, lookahead.LexemeLength() - 1);

            GetNextXPathToken(lookahead);

            // Create a VARIABLE_REF node.
            dwRet = Make1Node(STREE_VARIABLE_REF, (DWORD)pString);
            break;

        case TOKEN_XPATH_LITERAL:
            // Extract the literal (remove the surrounding quotes).
            pString = CopyLexeme(lookahead, 1, lookahead.LexemeLength() - 2);

            GetNextXPathToken(lookahead);

            // Create a LITERAL node.
            dwRet = Make1Node(STREE_LITERAL, (DWORD)pString);
            break;

        case TOKEN_XPATH_NUMBER:
            // Extract the number.
            d = lookahead.ToNumber();

            GetNextXPathToken(lookahead);

            // Create a NUMBER node. To store the double, 2 parameters will be needed!
            dwRet = Make1Node(STREE_NUMBER, d);
            break;

        case TOKEN_XPATH_LEFT_PARENTHESIS:
            GetNextXPathToken(lookahead);
            dwRet = Expr(lookahead);
            Match(TOKEN_XPATH_RIGHT_PARENTHESIS, lookahead);
            break;

        default:
            dwRet = FunctionCall(lookahead);
            break;
    }

    return dwRet;
}

// Rule [16]:
template<typename U>
DWORD XPathExpression<U>::FunctionCall(Token& lookahead)
{
    DWORD dwRet = 0;    // By default, no argument.
    U* pString = NULL;

    if(lookahead.m_eToken == TOKEN_XPATH_FUNCTIONNAME)
        pString = CopyLexeme(lookahead);

    Match(TOKEN_XPATH_FUNCTIONNAME, lookahead);
    Match(TOKEN_XPATH_LEFT_PARENTHESIS, lookahead);
    if(lookahead.m_eToken != TOKEN_XPATH_RIGHT_PARENTHESIS)
    {
        DWORD dwRet2 = dwRet = Argument(lookahead);
        while(lookahead.m_eToken == TOKEN_XPATH_COMMA)
        {
            GetNextXPathToken(lookahead);
            DWORD dwArg = Argument(lookahead);

            // Link to previous argument.
            SetNodeParam(dwRet2, 2, dwArg);
            dwRet2 = dwArg;
        }
    }
    Match(TOKEN_XPATH_RIGHT_PARENTHESIS, lookahead);

    // Create a FUNCTION_CALL node, and attach child nodes.
    dwRet = Make2Node(STREE_FUNCTION_CALL, (DWORD)pString, dwRet);

    return dwRet;
}

// Rule [17]:
template<typename U>
DWORD XPathExpression<U>::Argument(Token& lookahead)
{
    DWORD dwRet = Expr(lookahead);

    // Create a FUNCTION_ARGUMENT node, and attach child nodes.
    // Set the second parameter to "no next argument"; it will
    // possibly be updated by the calling function.
    dwRet = Make2Node(STREE_FUNCTION_ARGUMENT, dwRet, 0);

    return dwRet;
}

// Rule [18]:
template<typename U>
DWORD XPathExpression<U>::UnionExpr(Token& lookahead)
{
    DWORD dwRet = PathExpr(lookahead);
    while(lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
          lookahead.m_eSubtype == OPERATOR_PIPE)
    {
        GetNextXPathToken(lookahead);
        DWORD dwRet2 = PathExpr(lookahead);

        // Create a new UNION node, and attach its child nodes.
        dwRet = Make2Node(STREE_UNION, dwRet, dwRet2);
    }

    return dwRet;
}

// Rule [19]:
template<typename U>
DWORD XPathExpression<U>::PathExpr(Token& lookahead)
{
    DWORD dwRet;
    switch(lookahead.m_eToken)
    {
        case TOKEN_XPATH_VARIABLE_REFERENCE:
        case TOKEN_XPATH_LEFT_PARENTHESIS:
        case TOKEN_XPATH_LITERAL:
        case TOKEN_XPATH_NUMBER:
        case TOKEN_XPATH_FUNCTIONNAME:
            dwRet = FilterExpr(lookahead);
            if(lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
               (lookahead.m_eSubtype == OPERATOR_SLASH ||
                lookahead.m_eSubtype == OPERATOR_DOUBLE_SLASH)
              )
            {
                if(lookahead.m_eSubtype == OPERATOR_DOUBLE_SLASH)
                {
                    // DOUBLE_SLASH is equivalent to: /descendant-or-self::node()/
                    // Pass FILTER_EXPRESSION/descendant-or-self::node() as left
                    // context to RelativeLocationPath.
                    DWORD dwAxis     = Make1Node(STREE_AXIS, (DWORD)descendant_or_self);
                    DWORD dwNodeTest = Make1Node(STREE_NODETEST, (DWORD)nt_node);
                    DWORD dwTmp      = Make3Node(STREE_STEP, dwAxis, dwNodeTest, 0);
                    dwRet            = Make2Node(STREE_SLASH, dwRet, dwTmp);
                }

                GetNextXPathToken(lookahead);
                dwRet = RelativeLocationPath(lookahead, dwRet);
            }
            break;

        default:
            dwRet = LocationPath(lookahead);
            break;
    }

    return dwRet;
}

// Rule [20]:
template<typename U>
DWORD XPathExpression<U>::FilterExpr(Token& lookahead)
{
    DWORD dwRet = PrimaryExpr(lookahead);
    DWORD dwRet2 = 0;   // By default, no predicate.

    DWORD dwPreviousPred = 0;
    while(lookahead.m_eToken == TOKEN_XPATH_LEFT_SQUAREBRACKET)
    {
        DWORD dwPred = Predicate(lookahead);

        if(dwRet2 == 0)                             // First predicate.
            dwPreviousPred = dwRet2 = dwPred;
        else                                        // Link to previous predicate.
            SetNodeParam(dwPreviousPred, 2, dwPred);
        dwPreviousPred = dwPred;
    }

    // Create a new FILTER_EXPRESSION node, and attach its child nodes.
    // dwRet2 == 0 means "no predicate".
    dwRet = Make2Node(STREE_FILTER_EXPRESSION, dwRet, dwRet2);

    return dwRet;
}

// Rule [21]:
template<typename U>
DWORD XPathExpression<U>::OrExpr(Token& lookahead)
{
    DWORD dwRet = AndExpr(lookahead);
    while(lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
        lookahead.m_eSubtype == OPERATOR_NAME_OR)
    {
        GetNextXPathToken(lookahead);
        DWORD dwRet2 = AndExpr(lookahead);

        // Create a new OR node, and attach its child nodes.
        dwRet = Make2Node(STREE_OR, dwRet, dwRet2);
    }

    return dwRet;
}

// Rule [22]:
template<typename U>
DWORD XPathExpression<U>::AndExpr(Token& lookahead)
{
    DWORD dwRet = EqualityExpr(lookahead);
    while(lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
        lookahead.m_eSubtype == OPERATOR_NAME_AND)
    {
        GetNextXPathToken(lookahead);
        DWORD dwRet2 = EqualityExpr(lookahead);

        // Create a new AND node, and attach its child nodes.
        dwRet = Make2Node(STREE_AND, dwRet, dwRet2);
    }

    return dwRet;
}

// Rule [23]:
template<typename U>
DWORD XPathExpression<U>::EqualityExpr(Token& lookahead)
{
    DWORD dwRet = RelationalExpr(lookahead);
    while(lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
        (lookahead.m_eSubtype == OPERATOR_EQUALS ||
         lookahead.m_eSubtype == OPERATOR_NOT_EQUALS))
    {
        // Save the subtype.
        DWORD dwSubtype = lookahead.m_eSubtype == OPERATOR_EQUALS ?
            EQUALITY_EQUALS : EQUALITY_NOT_EQUALS;

        GetNextXPathToken(lookahead);
        DWORD dwRet2 = RelationalExpr(lookahead);

        // Create a new EQUALITY node, and attach its child nodes.
        dwRet = Make3Node(STREE_EQUALITY, dwSubtype, dwRet, dwRet2);
    }

    return dwRet;
}

// Rule [24]:
template<typename U>
DWORD XPathExpression<U>::RelationalExpr(Token& lookahead)
{
    DWORD dwRet = AdditiveExpr(lookahead);
    while(   lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
            (lookahead.m_eSubtype == OPERATOR_GREATER ||
             lookahead.m_eSubtype == OPERATOR_LESS ||
             lookahead.m_eSubtype == OPERATOR_GREATER_OR_EQUAL ||
             lookahead.m_eSubtype == OPERATOR_LESS_OR_EQUAL
             )
         )
    {
        // Save the subtype.
        DWORD dwSubtype;
        switch(lookahead.m_eSubtype)
        {
            case OPERATOR_GREATER:          dwSubtype = RELATIONAL_GREATER;          break;
            case OPERATOR_GREATER_OR_EQUAL: dwSubtype = RELATIONAL_GREATER_OR_EQUAL; break;
            case OPERATOR_LESS:             dwSubtype = RELATIONAL_LESS;             break;
            case OPERATOR_LESS_OR_EQUAL:
            default:                        dwSubtype = RELATIONAL_LESS_OR_EQUAL;    break;
        }

        GetNextXPathToken(lookahead);
        DWORD dwRet2 = AdditiveExpr(lookahead);

        // Create a new RELATIONAL node, and attach its child nodes.
        dwRet = Make3Node(STREE_RELATIONAL, dwSubtype, dwRet, dwRet2);
    }

    return dwRet;
}

// Rule [25]:
template<typename U>
DWORD XPathExpression<U>::AdditiveExpr(Token& lookahead)
{
    DWORD dwRet = MultiplicativeExpr(lookahead);
    while(   lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
            (lookahead.m_eSubtype == OPERATOR_PLUS ||
             lookahead.m_eSubtype == OPERATOR_MINUS
             )
         )
    {
        // Save the subtype.
        DWORD dwSubtype = lookahead.m_eSubtype == OPERATOR_PLUS ?
            ADDITIVE_PLUS : ADDITIVE_MINUS;

        GetNextXPathToken(lookahead);
        DWORD dwRet2 = MultiplicativeExpr(lookahead);

        // Create a new ADDITIVE node, and attach its child nodes.
        dwRet = Make3Node(STREE_ADDITIVE, dwSubtype, dwRet, dwRet2);
    }

    return dwRet;
}

// Rule [26]:
template<typename U>
DWORD XPathExpression<U>::MultiplicativeExpr(Token& lookahead)
{
    DWORD dwRet = UnaryExpr(lookahead);
    while(   lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
            (lookahead.m_eSubtype == OPERATOR_MULTIPLY ||
             lookahead.m_eSubtype == OPERATOR_NAME_MOD ||
             lookahead.m_eSubtype == OPERATOR_NAME_DIV
             )
         )
    {
        // Save the subtype.
        DWORD dwSubtype;
        switch(lookahead.m_eSubtype)
        {
            case OPERATOR_MULTIPLY: dwSubtype = MULTIPLICATIVE_MUTIPLY; break;
            case OPERATOR_NAME_MOD: dwSubtype = MULTIPLICATIVE_MOD;     break;
            case OPERATOR_NAME_DIV:
            default:                dwSubtype = MULTIPLICATIVE_DIV;     break;
        }

        GetNextXPathToken(lookahead);
        DWORD dwRet2 = UnaryExpr(lookahead);

        // Create a new MULTIPLICATIVE node, and attach its child nodes.
        dwRet = Make3Node(STREE_MULTIPLICATIVE, dwSubtype, dwRet, dwRet2);
    }

    return dwRet;
}

// Rule [27]:
template<typename U>
DWORD XPathExpression<U>::UnaryExpr(Token& lookahead)
{
    BOOL bHasMinus = FALSE;
    while(lookahead.m_eToken == TOKEN_XPATH_OPERATOR &&
          lookahead.m_eSubtype == OPERATOR_MINUS)
    {
        GetNextXPathToken(lookahead);
        bHasMinus = !bHasMinus;
    }

    DWORD dwRet = UnionExpr(lookahead);

    if(bHasMinus)
    {
        // Create a new UNARY_MINUS node, and attach its child node.
        dwRet = Make1Node(STREE_UNARY_MINUS, dwRet);
    }

    return dwRet;
}

// Rules [28] to [39] are lexical rules, and are directly implemented in the
// lexical analyzer.



                /////////// Dumping /////////////
#ifdef XML_DUMP

template<typename U>
void XPathExpression<U>::DumpSyntaxTree(std::basic_ostream<U>& os) const
{
    DumpNode(os, m_dwRootNode);
    os << std::endl;
}

template<typename U>
void XPathExpression<U>::DumpSyntaxTreeAsArray(std::basic_ostream<U>& os) const
{
    os << STR("Root: entry ") << m_dwRootNode << std::endl;
    DWORD dwSize = m_syntaxTree.size();
    for(DWORD dwIndex = 0; dwIndex < dwSize; ++dwIndex)
    {
        DWORD dwSkip = DumpNode(os, dwIndex, 0, FALSE);
        for(DWORD dw = 1; dw <= dwSkip; ++dw)
            os << dwIndex + dw << STR("\t") << m_syntaxTree[dwIndex + dw] << STR("\n");
        dwIndex += dwSkip;
    }
    os << std::endl;
}

template<typename U>
DWORD XPathExpression<U>::DumpNode(std::basic_ostream<U>& os, DWORD dwIndex, DWORD dwIndent, BOOL bRecurse) const
{
    DWORD dwSkip = 0;
    Margin(os, dwIndent);
    switch(m_syntaxTree[dwIndex])
    {
        case STREE_OR:
            os << dwIndex << STR(" or\n");
            if(bRecurse)
            {
                DumpNodeIndirect(os, dwIndex + 1, dwIndent + 2);
                DumpNodeIndirect(os, dwIndex + 2, dwIndent + 2);
            }
            else
                dwSkip = 2;
            break;
        case STREE_AND:
            os << dwIndex << STR(" and\n");
            if(bRecurse)
            {
                DumpNodeIndirect(os, dwIndex + 1, dwIndent + 2);
                DumpNodeIndirect(os, dwIndex + 2, dwIndent + 2);
            }
            else
                dwSkip = 2;
            break;
        case STREE_EQUALITY:
            os << dwIndex << (m_syntaxTree[dwIndex+1] == EQUALITY_EQUALS ? STR(" =") : STR(" !="))
               << STR("\n");
            if(bRecurse)
            {
                DumpNodeIndirect(os, dwIndex + 2, dwIndent + 2);
                DumpNodeIndirect(os, dwIndex + 3, dwIndent + 2);
            }
            else
                dwSkip = 3;
            break;
        case STREE_RELATIONAL:
            os << dwIndex;
            switch(m_syntaxTree[dwIndex+1])
            {
                case RELATIONAL_GREATER:            os << STR(" >");  break;
                case RELATIONAL_GREATER_OR_EQUAL:   os << STR(" >="); break;
                case RELATIONAL_LESS:               os << STR(" <");  break;
                case RELATIONAL_LESS_OR_EQUAL:      os << STR(" <="); break;
            }
            os << STR("\n");
            if(bRecurse)
            {
                DumpNodeIndirect(os, dwIndex + 2, dwIndent + 2);
                DumpNodeIndirect(os, dwIndex + 3, dwIndent + 2);
            }
            else
                dwSkip = 3;
            break;
        case STREE_ADDITIVE:
            os << dwIndex << (m_syntaxTree[dwIndex+1] == ADDITIVE_PLUS ? STR(" +") : STR(" -"))
               << STR("\n");
            if(bRecurse)
            {
                DumpNodeIndirect(os, dwIndex + 2, dwIndent + 2);
                DumpNodeIndirect(os, dwIndex + 3, dwIndent + 2);
            }
            else
                dwSkip = 3;
            break;
        case STREE_MULTIPLICATIVE:
            os << dwIndex;
            switch(m_syntaxTree[dwIndex+1])
            {
                case MULTIPLICATIVE_MUTIPLY:    os << STR(" *");    break;
                case MULTIPLICATIVE_DIV:        os << STR(" div");  break;
                case MULTIPLICATIVE_MOD:        os << STR(" mod");  break;
            }
            os << STR("\n");
            if(bRecurse)
            {
                DumpNodeIndirect(os, dwIndex + 2, dwIndent + 2);
                DumpNodeIndirect(os, dwIndex + 3, dwIndent + 2);
            }
            else
                dwSkip = 3;
            break;
        case STREE_UNARY_MINUS:
            os << dwIndex << STR(" - (unary)\n");
            if(bRecurse)
                DumpNodeIndirect(os, dwIndex + 1, dwIndent + 2);
            else
                dwSkip = 1;
            break;
        case STREE_UNION:
            os << dwIndex << STR(" |\n");
            if(bRecurse)
            {
                DumpNodeIndirect(os, dwIndex + 1, dwIndent + 2);
                DumpNodeIndirect(os, dwIndex + 2, dwIndent + 2);
            }
            else
                dwSkip = 2;
            break;
        case STREE_ROOT_SLASH:
            os << dwIndex << STR(" / (root)\n");
            if(bRecurse)
            {
                if(m_syntaxTree[dwIndex+1])
                    DumpNodeIndirect(os, dwIndex + 1, dwIndent + 2);
            }
            else
                dwSkip = 1;
            break;
        case STREE_SLASH:
            os << dwIndex << STR(" /\n");
            if(bRecurse)
            {
                DumpNodeIndirect(os, dwIndex + 1, dwIndent + 2);
                DumpNodeIndirect(os, dwIndex + 2, dwIndent + 2);
            }
            else
                dwSkip = 2;
            break;
        case STREE_FILTER_EXPRESSION:
            os << dwIndex << STR(" FILTER_EXPRESSION\n");
            if(bRecurse)
            {
                DumpNodeIndirect(os, dwIndex + 1, dwIndent + 2);
                {
                    // Predicate list.
                    for(DWORD dwPred = m_syntaxTree[dwIndex + 2]; dwPred; dwPred = m_syntaxTree[dwPred + 2])
                        DumpNode(os, dwPred, dwIndent + 2);
                }
            }
            else
                dwSkip = 2;
            break;
        case STREE_STEP:
            os << dwIndex << STR(" STEP\n");
            if(bRecurse)
            {
                DumpNodeIndirect(os, dwIndex + 1, dwIndent + 2);
                DumpNodeIndirect(os, dwIndex + 2, dwIndent + 2);
                if(m_syntaxTree[dwIndex + 3] == 0)
                {
                    Margin(os, dwIndent + 2);
                    os << STR("No Predicate.") << std::endl;
                }
                {
                    // Predicate list.
                    for(DWORD dwPred = m_syntaxTree[dwIndex + 3]; dwPred; dwPred = m_syntaxTree[dwPred + 2])
                        DumpNode(os, dwPred, dwIndent + 2);
                }
            }
            else
                dwSkip = 3;
            break;
        case STREE_AXIS:
            os << dwIndex << STR(" AXIS: ");
            switch(m_syntaxTree[dwIndex + 1])
            {
                case ancestor:            os << STR("ancestor\n");           break;
                case ancestor_or_self:    os << STR("ancestor_or_self\n");   break;
                case attribute:           os << STR("attribute\n");          break;
                case child:               os << STR("child\n");              break;
                case descendant:          os << STR("descendant\n");         break;
                case descendant_or_self:  os << STR("descendant_or_self\n"); break;
                case following:           os << STR("following\n");          break;
                case following_sibing:    os << STR("following_sibing\n");   break;
                case namespace_axis:      os << STR("namespace_axis\n");     break;
                case parent:              os << STR("parent\n");             break;
                case preceding:           os << STR("preceding\n");          break;
                case preceding_sibling:   os << STR("preceding_sibling\n");  break;
                case self:                os << STR("self\n");               break;
            }
            dwSkip = 1;
            break;
        case STREE_NODETEST:
            os << dwIndex << STR(" NODETEST: ");
            dwSkip = 1;
            switch(m_syntaxTree[dwIndex + 1])
            {
                case nt_star:        os << STR("*\n"); break;
                case nt_ncname_star: os << (CUPTR)m_syntaxTree[dwIndex + 2] << STR("::*\n"); dwSkip = 2; break;
                case nt_qname:       os << (CUPTR)m_syntaxTree[dwIndex + 2] << STR("\n"); dwSkip = 2; break;
                case nt_comment:     os << STR("comment()\n"); break;
                case nt_text:        os << STR("text()\n"); break;
                case nt_processing_instruction: os << STR("processing-instruction()\n"); break;
                case nt_processing_instruction_literal: os << STR("processing-instruction(") << (CUPTR)m_syntaxTree[dwIndex + 2] << STR(")\n"); dwSkip = 2; break;
                case nt_node:        os << STR("node()\n"); break;
            }
            break;
        case STREE_PREDICATE:
            os << dwIndex << STR(" PREDICATE\n");
            if(bRecurse)
                DumpNodeIndirect(os, dwIndex + 1, dwIndent + 2);
            else
                dwSkip = 2;
            break;
        case STREE_LITERAL:
            os << dwIndex << STR(" LITERAL: '") << (CUPTR)m_syntaxTree[dwIndex + 1] << STR("'\n");
            dwSkip = 1;
            break;
        case STREE_VARIABLE_REF:
            os << dwIndex << STR(" VARIABLE_REF: ") << (CUPTR)m_syntaxTree[dwIndex + 1] << STR("\n");
            dwSkip = 1;
            break;
        case STREE_NUMBER:
            os << dwIndex << STR(" NUMBER: ") << GetDoubleParam(dwIndex + 1) << STR("\n");
            dwSkip = 2;
            break;
        case STREE_FUNCTION_CALL:
            os << dwIndex << STR(" FUNCTION_CALL: ") << (CUPTR)m_syntaxTree[dwIndex + 1] << STR("\n");
            if(bRecurse)
            {
                if(m_syntaxTree[dwIndex + 2] == 0)
                {
                    Margin(os, dwIndent + 2);
                    os << STR("No function argument.") << std::endl;
                }

                // Argument list.
                for(DWORD dwArg = m_syntaxTree[dwIndex + 2]; dwArg; dwArg = m_syntaxTree[dwArg + 2])
                    DumpNode(os, dwArg, dwIndent + 2);
            }
            else
                dwSkip = 2;
            break;
        case STREE_FUNCTION_ARGUMENT:
            os << dwIndex << STR(" FUNCTION_ARGUMENT\n");
            if(bRecurse)
                DumpNodeIndirect(os, dwIndex + 1, dwIndent + 2);
            else
                dwSkip = 2;
            break;
        default:
            os << dwIndex << STR(" UNKNOWN NODE TYPE\n");
            break;
    }

    return dwSkip;
}

#endif


            ////////////////// Expression evaluation //////////////////

template<typename U>
void XPathExpression<U>::ComposeWithStep(
    const OrderedNodeSet& initialNodeSet, DWORD dwIndex, OrderedNodeSet& outNodeSet)
{
    // Initialize the output node set.
    outNodeSet.clear();

    // Get the axis of the step.
    Axis eAxis = (Axis)m_syntaxTree[m_syntaxTree[dwIndex + 1] + 1];

    // Update the sorting direction using the axis.
    {
        DocumentOrder order(IsForwardAxis(eAxis) ? DocumentOrder::forward : DocumentOrder::reverse);
        OrderedNodeSet tmpSet(order);
        outNodeSet = tmpSet;    // The predicate is copied!
    }

    // If the initial node set is empty, the output node set is also empty.
    if(!initialNodeSet.empty())
    {
        // Use each node in the initial node set as the context node.
        // All subsets produced by the following loop are directly unioned
        // together into outNodeSet.
        for(OrderedNodeSet::const_iterator pCtxtNode = initialNodeSet.begin(); pCtxtNode != initialNodeSet.end(); ++pCtxtNode)
            EvaluateStep(*pCtxtNode, dwIndex, eAxis, outNodeSet);

        // Filter the resulting node set through the list of predicates.
        for(DWORD dwPred = m_syntaxTree[dwIndex + 3]; dwPred != 0; dwPred = m_syntaxTree[dwPred + 2])
            Filter(dwPred, eAxis, outNodeSet);
    }
}

template<typename U>
void XPathExpression<U>::EvaluateStep(
    NodePTR pContextNode, DWORD dwIndex, Axis eAxis, OrderedNodeSet& outNodeSet)
{
    // Get the node test.
    DWORD dwNodeTestIndex = m_syntaxTree[dwIndex + 2];
    NodeTestKind nt   = (NodeTestKind)m_syntaxTree[dwNodeTestIndex + 1];
    CUPTR pString     = HasString(nt) ? (CUPTR)m_syntaxTree[dwNodeTestIndex + 2] : NULL;

    switch(eAxis)
    {
        case self:
            EvaluateNodeTestForNode(pContextNode, nt, pString, outNodeSet);
            break;

        case ancestor_or_self:
            EvaluateNodeTestForNode(pContextNode, nt, pString, outNodeSet);
            // Fall through:

        case ancestor:
            {
                const XmlParse<U>::XMLTree::NodeParent* pParent;
                const XmlParse<U>::XMLTree::NodeChild* pChild = dynamic_cast<const XmlParse<U>::XMLTree::NodeChild*>(pContextNode);
                if(!pChild)
                {
                    const XmlParse<U>::XMLTree::NodeAttribute* pAtt = dynamic_cast<const XmlParse<U>::XMLTree::NodeAttribute*>(pContextNode);
                    if(pAtt)
                        pParent = &pAtt->m_parent;
                    else
                        pParent = NULL;
                }
                else
                    pParent = &pChild->m_parent;

                while(pParent)
                {
                    EvaluateNodeTestForNode(pParent, nt, pString, outNodeSet);

                    // Next ancestor:
                    pChild = dynamic_cast<const XmlParse<U>::XMLTree::NodeChild*>(pParent);
                    pParent = pChild ? &pChild->m_parent : NULL;
                }
            }
            break;

        case child:
            EvaluateNodeTestForChildren(pContextNode, nt, pString, outNodeSet);
            break;

        case attribute:
            {
                const XmlParse<U>::XMLTree::NodeElement* pElement = dynamic_cast<const XmlParse<U>::XMLTree::NodeElement*>(pContextNode);
                if(pElement)
                {
                    for(XmlParse<U>::XMLTree::Attributes::const_iterator pAtt = pElement->m_attributes.begin(); pAtt != pElement->m_attributes.end(); ++pAtt)
                        EvaluateNodeTestForNode(*pAtt, nt, pString, outNodeSet, TRUE);
                }
            }
            break;

        case descendant_or_self:
            EvaluateNodeTestForNode(pContextNode, nt, pString, outNodeSet);
            // Fall through:

        case descendant:
            EvaluateNodeTestForDescendants(pContextNode, nt, pString, outNodeSet);
            break;

        case parent:
            {
                const XmlParse<U>::XMLTree::NodeParent* pParent;
                const XmlParse<U>::XMLTree::NodeChild* pChild = dynamic_cast<const XmlParse<U>::XMLTree::NodeChild*>(pContextNode);
                if(!pChild)
                {
                    const XmlParse<U>::XMLTree::NodeAttribute* pAtt = dynamic_cast<const XmlParse<U>::XMLTree::NodeAttribute*>(pContextNode);
                    if(pAtt)
                        pParent = &pAtt->m_parent;
                    else
                        pParent = NULL;
                }
                else
                    pParent = &pChild->m_parent;

                if(pParent)
                    EvaluateNodeTestForNode(pParent, nt, pString, outNodeSet);
            }
            break;

        case preceding_sibling:
            {
                const XmlParse<U>::XMLTree::NodeChild* pChild = dynamic_cast<const XmlParse<U>::XMLTree::NodeChild*>(pContextNode);
                if(pChild)
                {
                    for(pChild = pChild->m_pPrecedingSibling; pChild; pChild = pChild->m_pPrecedingSibling)
                        EvaluateNodeTestForNode(pChild, nt, pString, outNodeSet);
                }
            }
            break;

        case following_sibing:
            {
                const XmlParse<U>::XMLTree::NodeChild* pChild = dynamic_cast<const XmlParse<U>::XMLTree::NodeChild*>(pContextNode);
                if(pChild)
                {
                    for(pChild = pChild->m_pFollowingSibling; pChild; pChild = pChild->m_pFollowingSibling)
                        EvaluateNodeTestForNode(pChild, nt, pString, outNodeSet);
                }
            }
            break;

        case preceding:
            {
                const XmlParse<U>::XMLTree::Node* pNode = pContextNode;
                const XmlParse<U>::XMLTree::NodeChild* pChild;
                while(pNode && (pChild = dynamic_cast<const XmlParse<U>::XMLTree::NodeChild*>(pNode)) != NULL)
                {
                    for(XmlParse<U>::XMLTree::NodeChild* pSibling = pChild->m_pPrecedingSibling; pSibling; pSibling = pSibling->m_pPrecedingSibling)
                    {
                        EvaluateNodeTestForNode(pSibling, nt, pString, outNodeSet);
                        EvaluateNodeTestForDescendants(pSibling, nt, pString, outNodeSet);
                    }

                    // Go to parent.
                    pNode = &pChild->m_parent;
                }
            }
            break;

        case following:
            {
                const XmlParse<U>::XMLTree::Node* pNode = pContextNode;
                const XmlParse<U>::XMLTree::NodeChild* pChild;
                while(pNode && (pChild = dynamic_cast<const XmlParse<U>::XMLTree::NodeChild*>(pNode)) != NULL)
                {
                    for(XmlParse<U>::XMLTree::NodeChild* pSibling = pChild->m_pFollowingSibling; pSibling; pSibling = pSibling->m_pFollowingSibling)
                    {
                        EvaluateNodeTestForNode(pSibling, nt, pString, outNodeSet);
                        EvaluateNodeTestForDescendants(pSibling, nt, pString, outNodeSet);
                    }

                    // Go to parent.
                    pNode = &pChild->m_parent;
                }
            }
            break;

        case namespace_axis:
            break;
    }
}

template<typename U>
void XPathExpression<U>::Filter(DWORD dwIndexPredicate, Axis eAxis, OrderedNodeSet& outNodeSet)
{
    // Iterate over the nodes.
    DWORD dwContextPosition = 1;
    DWORD dwContextSize     = outNodeSet.size();
    for(OrderedNodeSet::iterator pNode = outNodeSet.begin(); pNode != outNodeSet.end(); )
    {
        // The sequence below correctly erases a node without invalidating pNode:
        if(FilterNode(dwIndexPredicate, eAxis, *pNode, dwContextSize, dwContextPosition))
            outNodeSet.erase(pNode++);
        else
            ++pNode;

        ++dwContextPosition;
    }
}

template<typename U>
BOOL XPathExpression<U>::FilterNode(
    DWORD dwIndexPredicate, Axis eAxis, NodePTR pContextNode,
    DWORD dwContextSize, DWORD dwContextPosition)
{
    BOOL bRet;

    // Get the predicate's expression.
    DWORD dwIndexExpression = m_syntaxTree[dwIndexPredicate + 1];

    // Evaluate the expression.
    XPathExprValue val;
    EvaluateExpression(dwIndexExpression, eAxis, pContextNode, dwContextSize, dwContextPosition, val);

    // If the result is a number, then is must match the context position.
    if(val.m_eType == TYPE_NUMBER)
        bRet = ((DWORD)val.m_dNumber) != dwContextPosition;
    else
    {
        // Convert the result to a boolean.
        ToBOOL(val);

        // Return TRUE if the expression if FALSE or unknown.
        bRet = val.m_eType != TYPE_BOOLEAN || !val.m_bBoolean;
    }

    return bRet;
}

template<typename U>
void XPathExpression<U>::EvaluateNodeTestForNode(
    NodePTR pContextNode, NodeTestKind nt, CUPTR pString, OrderedNodeSet& outNodeSet, BOOL bAttributeAxis)
{
    BOOL bEval = FALSE;
    switch(nt)
    {
        case nt_star:
            if(bAttributeAxis)
            {
                // TRUE if node is an attribute node.
                bEval = dynamic_cast<const XmlParse<U>::XMLTree::NodeAttribute*>(pContextNode) != NULL;
            }
            else
            {
                // TRUE if node is an element node.
                bEval = dynamic_cast<const XmlParse<U>::XMLTree::NodeElement*>(pContextNode) != NULL;
            }
            break;
        case nt_node:
            bEval = TRUE;
            break;
        case nt_ncname_star:    // Not implemented.
            break;
        case nt_qname:
            {
                const XmlParse<U>::String* pName = XmlParse<U>::XMLTree::GetNodeName(pContextNode);
                if(pName)
                    bEval = *pName == pString;
            }
            break;
        case nt_comment:
            bEval = dynamic_cast<const XmlParse<U>::XMLTree::NodeComment*>(pContextNode) != NULL;
            break;
        case nt_text:
            bEval = dynamic_cast<const XmlParse<U>::XMLTree::NodeText*>(pContextNode) != NULL;
            break;
        case nt_processing_instruction:
            bEval = dynamic_cast<const XmlParse<U>::XMLTree::NodeProcessingInstruction*>(pContextNode) != NULL;
            break;
        case nt_processing_instruction_literal:
            {
                const XmlParse<U>::XMLTree::NodeProcessingInstruction* pPI =
                    dynamic_cast<const XmlParse<U>::XMLTree::NodeProcessingInstruction*>(pContextNode);
                bEval = pPI && pPI->m_strTarget == pString;
            }
            break;
    }

    if(bEval)
    {
        // Add the node to the output node set.
        outNodeSet.insert(pContextNode);
    }
}

template<typename U>
void XPathExpression<U>::EvaluateNodeTestForChildren(
    NodePTR pNode, NodeTestKind nt, CUPTR pString, OrderedNodeSet& outNodeSet)
{
    const XmlParse<U>::XMLTree::NodeParent* pParent = dynamic_cast<const XmlParse<U>::XMLTree::NodeParent*>(pNode);
    if(pParent)
    {
        for(XmlParse<U>::XMLTree::Children::const_iterator pChild = pParent->m_children.begin(); pChild != pParent->m_children.end(); ++pChild)
            EvaluateNodeTestForNode(*pChild, nt, pString, outNodeSet);
    }
}

template<typename U>
void XPathExpression<U>::EvaluateNodeTestForDescendants(
    NodePTR pNode, NodeTestKind nt, CUPTR pString, OrderedNodeSet& outNodeSet)
{
    const XmlParse<U>::XMLTree::NodeParent* pParent = dynamic_cast<const XmlParse<U>::XMLTree::NodeParent*>(pNode);
    if(pParent)
    {
        for(XmlParse<U>::XMLTree::Children::const_iterator pChild = pParent->m_children.begin(); pChild != pParent->m_children.end(); ++pChild)
        {
            EvaluateNodeTestForNode(*pChild, nt, pString, outNodeSet);
            EvaluateNodeTestForDescendants(*pChild, nt, pString, outNodeSet);
        }
    }
}

template<typename U>
void XPathExpression<U>::EvaluateExpression(
    DWORD dwIndexExpression, Axis eAxis, NodePTR pContextNode,
    DWORD dwContextSize, DWORD dwContextPosition, XPathExprValue& val)
{
    SYNTAX_TREE_NODE eNodeType = (SYNTAX_TREE_NODE)m_syntaxTree[dwIndexExpression];
    switch(eNodeType)
    {
        case STREE_OR:
            EvaluateExpression(m_syntaxTree[dwIndexExpression + 1], eAxis, pContextNode,
                dwContextSize, dwContextPosition, val);
            if(val.IsValid())
            {
                ToBOOL(val);
                if(!val.m_bBoolean) // Short-circuit evaluation.
                {
                    EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                        dwContextSize, dwContextPosition, val);
                    if(val.IsValid())
                        ToBOOL(val);
                }
            }
            break;

        case STREE_AND:
            EvaluateExpression(m_syntaxTree[dwIndexExpression + 1], eAxis, pContextNode,
                dwContextSize, dwContextPosition, val);
            if(val.IsValid())
            {
                ToBOOL(val);
                if(val.m_bBoolean)  // Short-circuit evaluation.
                {
                    EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                        dwContextSize, dwContextPosition, val);
                    if(val.IsValid())
                        ToBOOL(val);
                }
            }
            break;

        case STREE_EQUALITY:
            EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                dwContextSize, dwContextPosition, val);
            if(val.IsValid())
            {
                XPathExprValue val2;
                EvaluateExpression(m_syntaxTree[dwIndexExpression + 3], eAxis, pContextNode,
                    dwContextSize, dwContextPosition, val2);
                if(val2.IsValid())
                {
                    // val and val2 are valid.
                    SYNTAX_TREE_NODE_SUBTYPE eOperatorType = (SYNTAX_TREE_NODE_SUBTYPE)m_syntaxTree[dwIndexExpression + 1];
                    if(val.m_eType == TYPE_NODESET || val2.m_eType == TYPE_NODESET)
                    {
                        if(val.m_eType == TYPE_NODESET)
                        {
                            if(val2.m_eType == TYPE_NODESET)
                                val = NodeSetCompare(val.m_nodeSet, val2.m_nodeSet, eOperatorType);
                            else
                                val = NodeSetCompare(val.m_nodeSet, val2, eOperatorType);
                        }
                        else
                            val = NodeSetCompare(val, val2.m_nodeSet, eOperatorType);
                    }
                    else
                    {
                        // Neither val nor val2 are node sets.
                        if(val.m_eType == TYPE_BOOLEAN)
                        {
                            if(val2.m_eType != TYPE_BOOLEAN)
                                ToBOOL(val2);
                            if(eOperatorType == EQUALITY_EQUALS)
                                val = val.m_bBoolean == val2.m_bBoolean;
                            else
                                val = val.m_bBoolean != val2.m_bBoolean;
                        }
                        else if(val2.m_eType == TYPE_BOOLEAN)
                        {
                            if(val.m_eType != TYPE_BOOLEAN)
                                ToBOOL(val);
                            if(eOperatorType == EQUALITY_EQUALS)
                                val = val.m_bBoolean == val2.m_bBoolean;
                            else
                                val = val.m_bBoolean != val2.m_bBoolean;
                        }
                        else if(val.m_eType == TYPE_NUMBER)
                        {
                            if(val2.m_eType != TYPE_NUMBER)
                                ToNumber(val2);
                            if(val2.m_eType == TYPE_UNKNOWN)
                                val.Clear();
                            else
                            {
                                if(eOperatorType == EQUALITY_EQUALS)
                                    val = val.m_dNumber == val2.m_dNumber;
                                else
                                    val = val.m_dNumber != val2.m_dNumber;
                            }
                        }
                        else if(val2.m_eType == TYPE_NUMBER)
                        {
                            if(val.m_eType != TYPE_NUMBER)
                                ToNumber(val);
                            if(val.m_eType == TYPE_NUMBER)
                            {
                                if(eOperatorType == EQUALITY_EQUALS)
                                    val = val.m_dNumber == val2.m_dNumber;
                                else
                                    val = val.m_dNumber != val2.m_dNumber;
                            }
                        }
                        else
                        {
                            // val and val2 are strings.
                            if(eOperatorType == EQUALITY_EQUALS)
                                val = compare(val.m_string, val2.m_string) == 0;
                            else
                                val = compare(val.m_string, val2.m_string) != 0;
                        }
                    }
                }
                else
                    val.Clear();
            }
            break;

        case STREE_RELATIONAL:
            EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                dwContextSize, dwContextPosition, val);
            if(val.IsValid())
            {
                XPathExprValue val2;
                EvaluateExpression(m_syntaxTree[dwIndexExpression + 3], eAxis, pContextNode,
                    dwContextSize, dwContextPosition, val2);
                if(val2.IsValid())
                {
                    // val and val2 are valid.
                    SYNTAX_TREE_NODE_SUBTYPE eOperatorType = (SYNTAX_TREE_NODE_SUBTYPE)m_syntaxTree[dwIndexExpression + 1];
                    if(val.m_eType == TYPE_NODESET || val2.m_eType == TYPE_NODESET)
                    {
                        if(val.m_eType == TYPE_NODESET)
                        {
                            if(val2.m_eType == TYPE_NODESET)
                                val = NodeSetCompare(val.m_nodeSet, val2.m_nodeSet, eOperatorType);
                            else
                                val = NodeSetCompare(val.m_nodeSet, val2, eOperatorType);
                        }
                        else
                            val = NodeSetCompare(val, val2.m_nodeSet, eOperatorType);
                    }
                    else
                    {
                        // Neither val nor val2 are node sets.
                        if(val.m_eType != TYPE_NUMBER)
                            ToNumber(val);
                        if(val.m_eType == TYPE_NUMBER)
                        {
                            if(val2.m_eType != TYPE_NUMBER)
                                ToNumber(val2);
                            if(val2.m_eType == TYPE_NUMBER)
                            {
                                SYNTAX_TREE_NODE_SUBTYPE eSubtype = (SYNTAX_TREE_NODE_SUBTYPE)m_syntaxTree[dwIndexExpression + 1];
                                switch(eSubtype)
                                {
                                    case RELATIONAL_GREATER:
                                        val = val.m_dNumber >  val2.m_dNumber;
                                        break;
                                    case RELATIONAL_GREATER_OR_EQUAL:
                                        val = val.m_dNumber >= val2.m_dNumber;
                                        break;
                                    case RELATIONAL_LESS:
                                        val = val.m_dNumber <  val2.m_dNumber;
                                        break;
                                    case RELATIONAL_LESS_OR_EQUAL:
                                        val = val.m_dNumber <= val2.m_dNumber;
                                        break;
                                    default:    // Error: should not arise.
                                        val.Clear();
                                        break;
                                }
                            }
                            else
                                val.Clear();
                        }
                    }
                }
                else
                    val.Clear();
            }
            break;

        case STREE_ADDITIVE:
            EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                dwContextSize, dwContextPosition, val);
            if(val.IsValid())
            {
                XPathExprValue val2;
                EvaluateExpression(m_syntaxTree[dwIndexExpression + 3], eAxis, pContextNode,
                    dwContextSize, dwContextPosition, val2);
                if(val2.IsValid())
                {
                    // val and val2 are valid.
                    if(val.m_eType != TYPE_NUMBER)
                        ToNumber(val);
                    if(val.m_eType == TYPE_NUMBER)
                    {
                        if(val2.m_eType != TYPE_NUMBER)
                            ToNumber(val2);
                        if(val2.m_eType == TYPE_NUMBER)
                        {
                            SYNTAX_TREE_NODE_SUBTYPE eSubtype = (SYNTAX_TREE_NODE_SUBTYPE)m_syntaxTree[dwIndexExpression + 1];
                            if(eSubtype == ADDITIVE_PLUS)
                                val = val.m_dNumber + val2.m_dNumber;
                            else
                                val = val.m_dNumber - val2.m_dNumber;
                        }
                    }
                }
            }
            break;

        case STREE_MULTIPLICATIVE:
            EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                dwContextSize, dwContextPosition, val);
            if(val.IsValid())
            {
                XPathExprValue val2;
                EvaluateExpression(m_syntaxTree[dwIndexExpression + 3], eAxis, pContextNode,
                    dwContextSize, dwContextPosition, val2);
                if(val2.IsValid())
                {
                    // val and val2 are valid.
                    if(val.m_eType != TYPE_NUMBER)
                        ToNumber(val);
                    if(val.m_eType == TYPE_NUMBER)
                    {
                        if(val2.m_eType != TYPE_NUMBER)
                            ToNumber(val2);
                        if(val2.m_eType == TYPE_NUMBER)
                        {
                            SYNTAX_TREE_NODE_SUBTYPE eSubtype = (SYNTAX_TREE_NODE_SUBTYPE)m_syntaxTree[dwIndexExpression + 1];
                            if(eSubtype == MULTIPLICATIVE_MUTIPLY)
                                val = val.m_dNumber * val2.m_dNumber;
                            else if(eSubtype == MULTIPLICATIVE_DIV)
                                val = val.m_dNumber / val2.m_dNumber;
                            else // MULTIPLICATIVE_MOD
                            {
                                // First, convert operands to integers;
                                int i1 = (int)val.m_dNumber;
                                int i2 = (int)val2.m_dNumber;
                                BOOL bNegative = FALSE;
                                if(i1 < 0)
                                {
                                    bNegative = TRUE;
                                    i1 = -i1;
                                }
                                if(i2 < 0)
                                    i2 = -i2;
                                i1 %= i2;
                                if(bNegative)
                                    i1 = -i1;
                                val = i1;
                            }
                        }
                    }
                }
            }
            break;

        case STREE_UNARY_MINUS:
            EvaluateExpression(m_syntaxTree[dwIndexExpression + 1], eAxis, pContextNode,
                dwContextSize, dwContextPosition, val);
            if(val.IsValid())
            {
                if(val.m_eType != TYPE_NUMBER)
                    ToNumber(val);
                if(val.m_eType == TYPE_NUMBER)
                    val = -val.m_dNumber;
            }
            break;

        case STREE_UNION:
            // The operands must evaluate to node sets.
            EvaluateExpression(m_syntaxTree[dwIndexExpression + 1], eAxis, pContextNode,
                dwContextSize, dwContextPosition, val);
            if(val.IsValid())
            {
                if(val.m_eType == TYPE_NODESET)
                {
                    XPathExprValue val2;
                    EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                        dwContextSize, dwContextPosition, val2);
                    if(val2.IsValid())
                    {
                        if(val2.m_eType == TYPE_NODESET)
                        {
                            // val and val2 are valid. Perform the union.
                            while(!val2.m_nodeSet.empty()) {
                                val.m_nodeSet.insert(*(val2.m_nodeSet.begin()));
                                val2.m_nodeSet.erase(val2.m_nodeSet.begin());
                            }
                        }
                        else
                            val.Clear();
                    }
                }
                else
                    val.Clear();
            }
            break;

        case STREE_ROOT_SLASH:
            // Compute the location step with the root node as the unique context node.
            {
                DocumentOrder order(DocumentOrder::forward);
                OrderedNodeSet inputNodeSet(order);
                inputNodeSet.insert(m_pXMLTreeRoot);
                val.Clear();
                val.m_eType = TYPE_NODESET;
                if(m_syntaxTree[dwIndexExpression + 1])
                    ComposeWithStep(inputNodeSet, m_syntaxTree[dwIndexExpression + 1], val.m_nodeSet);
                else
                    val.m_nodeSet.insert(m_pXMLTreeRoot);   // The resulting set is
                                                            // reduced to the root node.
            }
            break;

        case STREE_SLASH:
            {
                val.Clear();
                XPathExprValue val2;
                EvaluateExpression(m_syntaxTree[dwIndexExpression + 1], eAxis, pContextNode,
                    dwContextSize, dwContextPosition, val2);
                if(val2.IsValid())
                {
                    // The first expression must yield a node set.
                    if(val2.m_eType == TYPE_NODESET)
                    {
                        val.m_eType = TYPE_NODESET;
                        ComposeWithStep(val2.m_nodeSet, m_syntaxTree[dwIndexExpression + 2], val.m_nodeSet);
                    }
                }
            }
            break;

        case STREE_FILTER_EXPRESSION:
            EvaluateExpression(m_syntaxTree[dwIndexExpression + 1], eAxis, pContextNode,
                dwContextSize, dwContextPosition, val);
            if(val.IsValid())
            {
                // If at least one predicate follows, then the expression must
                // evaluate to a node set.
                if(m_syntaxTree[dwIndexExpression + 2] != 0)
                {
                    if(val.m_eType == TYPE_NODESET)
                    {
                        // The default axis we wish to use is child, which is a
                        // forward axis.
                        // Update the sorting direction using the axis if needed.
                        if(!IsDocumentOrder(val.m_nodeSet))
                        {
                            DocumentOrder order(DocumentOrder::forward);
                            OrderedNodeSet tmpSet(order);
                            while(!val.m_nodeSet.empty()) {
                                tmpSet.insert(*(val.m_nodeSet.begin()));
                                val.m_nodeSet.erase(val.m_nodeSet.begin());
                            }
                            tmpSet.swap(val.m_nodeSet);
                        }

                        // Filter the node set through the list of predicates.
                        for(DWORD dwPred = m_syntaxTree[dwIndexExpression + 2]; dwPred != 0; dwPred = m_syntaxTree[dwPred + 2])
                        {
                            // Filter. Use child as the default axis.
                            Filter(dwPred, child, val.m_nodeSet);
                        }
                    }
                    else
                        val.Clear();
                }
            }
            break;

        case STREE_STEP:
            // If the context node is NULL, it means that this step begins a relative
            // location path, so assume the root of the XML document as the unique
            // context node, as with an absolute location path.
            if(pContextNode == NULL)
            {
                // Compute the location step with the root node as the unique context node.
                DocumentOrder order(DocumentOrder::forward);
                OrderedNodeSet inputNodeSet(order);
                inputNodeSet.insert(m_pXMLTreeRoot);
                val.Clear();
                val.m_eType = TYPE_NODESET;
                ComposeWithStep(inputNodeSet, dwIndexExpression, val.m_nodeSet);
            }
            else
            {
                // Compute the location step with the context node as the unique context node.
                DocumentOrder order(IsForwardAxis(eAxis) ? DocumentOrder::forward : DocumentOrder::reverse);
                OrderedNodeSet inputNodeSet(order);
                inputNodeSet.insert(pContextNode);
                val.Clear();
                val.m_eType = TYPE_NODESET;
                ComposeWithStep(inputNodeSet, dwIndexExpression, val.m_nodeSet);
            }
            break;

        case STREE_AXIS:                // Nothing to do.
            break;

        case STREE_NODETEST:            // Nothing to do.
            break;

        case STREE_PREDICATE:           // Nothing to do.
            break;

        case STREE_LITERAL:
            val = (CUPTR)m_syntaxTree[dwIndexExpression + 1];
            break;

        case STREE_VARIABLE_REF:
            val.Clear();    // TODO: not implemented yet.
            break;

        case STREE_NUMBER:
            val = GetDoubleParam(dwIndexExpression + 1);
            break;

        case STREE_FUNCTION_CALL:
            {
                CUPTR pFuncName = (CUPTR)m_syntaxTree[dwIndexExpression + 1];
                if(compare(pFuncName, STR("position")) == 0)    // position()
                    val = (double)dwContextPosition;
                else if(compare(pFuncName, STR("string")) == 0) // string(object?)
                {
                    if(m_syntaxTree[dwIndexExpression + 2] == 0)
                    {
                        // Call to string function with no argument.
                        // Defaults to a node set with the context node as only member.
                        val.Clear();
                        NodePTR pNode = pContextNode ? pContextNode : m_pXMLTreeRoot;
                        val.m_eType = TYPE_NODESET;
                        val.m_nodeSet.insert(pNode);
                    }
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                    }
                    if(val.IsValid() && val.m_eType != TYPE_STRING)
                        ToString(val);
                }
                else if(compare(pFuncName, STR("number")) == 0) // number(object?)
                {
                    if(m_syntaxTree[dwIndexExpression + 2] == 0)
                    {
                        // Call to string function with no argument.
                        // Defaults to a node set with the context node as only member.
                        val.Clear();
                        NodePTR pNode = pContextNode ? pContextNode : m_pXMLTreeRoot;
                        val.m_eType = TYPE_NODESET;
                        val.m_nodeSet.insert(pNode);
                    }
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                    }
                    if(val.IsValid() && val.m_eType != TYPE_NUMBER)
                        ToNumber(val);
                }
                else if(compare(pFuncName, STR("string-length")) == 0) // string-length(string?)
                {
                    if(m_syntaxTree[dwIndexExpression + 2] == 0)
                    {
                        // Call to string-length function with no argument.
                        // Defaults to a node set with the context node as only member.
                        val.Clear();
                        NodePTR pNode = pContextNode ? pContextNode : m_pXMLTreeRoot;
                        val.m_eType = TYPE_NODESET;
                        val.m_nodeSet.insert(pNode);
                    }
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                    }
                    if(val.IsValid() && val.m_eType != TYPE_STRING)
                        ToString(val);
                    if(val.m_eType == TYPE_STRING)
                        val = (double)XmlParse<U>::stringlen(val.m_string);
                }
                else if(compare(pFuncName, STR("boolean")) == 0) // boolean(object)
                {
                    if(m_syntaxTree[dwIndexExpression + 2] == 0)
                        val.Clear();    // Must have one argument.
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.IsValid() && val.m_eType != TYPE_BOOLEAN)
                            ToBOOL(val);
                    }
                }
                else if(compare(pFuncName, STR("not")) == 0) // not(boolean)
                {
                    if(m_syntaxTree[dwIndexExpression + 2] == 0)
                        val.Clear();    // Must have one argument.
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.IsValid() && val.m_eType != TYPE_BOOLEAN)
                            ToBOOL(val);
                        if(val.m_eType == TYPE_BOOLEAN)
                            val.m_bBoolean = !val.m_bBoolean;
                    }
                }
                else if(compare(pFuncName, STR("true")) == 0)   // true()
                    val = (BOOL)TRUE;
                else if(compare(pFuncName, STR("false")) == 0)  // false()
                    val = (BOOL)FALSE;
                else if(compare(pFuncName, STR("last")) == 0)   // last()
                    val = (double)dwContextSize;
                else if(compare(pFuncName, STR("count")) == 0)  // count(node-set)
                {
                    if(m_syntaxTree[dwIndexExpression + 2] != 0)
                    {
                        EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_NODESET)
                            val.Clear();
                        else
                            val = (double)val.m_nodeSet.size();
                    }
                    else
                        val.Clear();
                }
                else if(compare(pFuncName, STR("sum")) == 0)    // sum(node-set)
                {
                    if(m_syntaxTree[dwIndexExpression + 2] != 0)
                    {
                        EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_NODESET)
                            val.Clear();
                        else
                        {
                            double d = 0.0;
                            for(OrderedNodeSet::const_iterator ppNode = val.m_nodeSet.begin(); ppNode != val.m_nodeSet.end(); ++ppNode)
                            {
                                double d1;
                                if(ToNumber((*ppNode)->StringValue().c_str(), d1))
                                    d += d1;
                            }
                            val = d;
                        }
                    }
                    else
                        val.Clear();
                }
                else if(compare(pFuncName, STR("avg")) == 0)    // avg(node-set)
                {
                    if(m_syntaxTree[dwIndexExpression + 2] != 0)
                    {
                        EvaluateExpression(m_syntaxTree[dwIndexExpression + 2], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_NODESET)
                            val.Clear();
                        else
                        {
                            double d = 0.0;
                            for(OrderedNodeSet::const_iterator ppNode = val.m_nodeSet.begin(); ppNode != val.m_nodeSet.end(); ++ppNode)
                            {
                                double d1;
                                if(ToNumber((*ppNode)->StringValue().c_str(), d1))
                                    d += d1;
                            }
                            val = d / val.m_nodeSet.size();
                        }
                    }
                    else
                        val.Clear();
                }
                else if(compare(pFuncName, STR("concat")) == 0) // concat(string,string,string*)
                {
                    DWORD dwArg = m_syntaxTree[dwIndexExpression + 2];
                    DWORD dwArg2 = dwArg ? m_syntaxTree[dwArg + 2] : 0;
                    if(dwArg == 0 || dwArg2 == 0)   // At least 2 arguments needed.
                        val.Clear();
                    else
                    {
                        // Concatenate strings.
                        EvaluateExpression(m_syntaxTree[dwArg + 1], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_UNKNOWN && val.m_eType != TYPE_STRING)
                            ToString(val);
                        if(val.m_eType == TYPE_STRING)
                        {
                            while(dwArg2)
                            {
                                XPathExprValue val2;
                                EvaluateExpression(m_syntaxTree[dwArg2 + 1], eAxis, pContextNode,
                                    dwContextSize, dwContextPosition, val2);
                                if(val2.m_eType != TYPE_UNKNOWN && val2.m_eType != TYPE_STRING)
                                    ToString(val2);
                                if(val2.m_eType == TYPE_STRING)
                                    val += val2.m_string;
                                else
                                {
                                    val.Clear();
                                    break;
                                }

                                // Next argument.
                                dwArg2 = m_syntaxTree[dwArg2 + 2];
                            }
                        }
                    }
                }
                else if(compare(pFuncName, STR("starts-with")) == 0) // starts-with(string,string)
                {
                    DWORD dwArg = m_syntaxTree[dwIndexExpression + 2];
                    DWORD dwArg2 = dwArg ? m_syntaxTree[dwArg + 2] : 0;
                    if(dwArg == 0 || dwArg2 == 0)   // 2 arguments needed.
                        val.Clear();
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwArg + 1], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_UNKNOWN && val.m_eType != TYPE_STRING)
                            ToString(val);
                        if(val.m_eType == TYPE_STRING)
                        {
                            XPathExprValue val2;
                            EvaluateExpression(m_syntaxTree[dwArg2 + 1], eAxis, pContextNode,
                                dwContextSize, dwContextPosition, val2);
                            if(val2.m_eType != TYPE_UNKNOWN && val2.m_eType != TYPE_STRING)
                                ToString(val2);
                            if(val2.m_eType == TYPE_STRING)
                                val = StartsWith(val.m_string, val2.m_string);
                            else
                                val.Clear();
                        }
                    }
                }
                else if(compare(pFuncName, STR("substring-before")) == 0) // substring-before(string,string)
                {
                    DWORD dwArg = m_syntaxTree[dwIndexExpression + 2];
                    DWORD dwArg2 = dwArg ? m_syntaxTree[dwArg + 2] : 0;
                    if(dwArg == 0 || dwArg2 == 0)   // 2 arguments needed.
                        val.Clear();
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwArg + 1], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_UNKNOWN && val.m_eType != TYPE_STRING)
                            ToString(val);
                        if(val.m_eType == TYPE_STRING)
                        {
                            XPathExprValue val2;
                            EvaluateExpression(m_syntaxTree[dwArg2 + 1], eAxis, pContextNode,
                                dwContextSize, dwContextPosition, val2);
                            if(val2.m_eType != TYPE_UNKNOWN && val2.m_eType != TYPE_STRING)
                                ToString(val2);
                            if(val2.m_eType == TYPE_STRING)
                                val = SubStringBefore(val.m_string, val2.m_string);
                            else
                                val.Clear();
                        }
                    }
                }
                else if(compare(pFuncName, STR("substring-after")) == 0) // substring-after(string,string)
                {
                    DWORD dwArg = m_syntaxTree[dwIndexExpression + 2];
                    DWORD dwArg2 = dwArg ? m_syntaxTree[dwArg + 2] : 0;
                    if(dwArg == 0 || dwArg2 == 0)   // 2 arguments needed.
                        val.Clear();
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwArg + 1], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_UNKNOWN && val.m_eType != TYPE_STRING)
                            ToString(val);
                        if(val.m_eType == TYPE_STRING)
                        {
                            XPathExprValue val2;
                            EvaluateExpression(m_syntaxTree[dwArg2 + 1], eAxis, pContextNode,
                                dwContextSize, dwContextPosition, val2);
                            if(val2.m_eType != TYPE_UNKNOWN && val2.m_eType != TYPE_STRING)
                                ToString(val2);
                            if(val2.m_eType == TYPE_STRING)
                                val = SubStringAfter(val.m_string, val2.m_string);
                            else
                                val.Clear();
                        }
                    }
                }
                else if(compare(pFuncName, STR("contains")) == 0) // contains(string,string)
                {
                    DWORD dwArg = m_syntaxTree[dwIndexExpression + 2];
                    DWORD dwArg2 = dwArg ? m_syntaxTree[dwArg + 2] : 0;
                    if(dwArg == 0 || dwArg2 == 0)   // 2 arguments needed.
                        val.Clear();
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwArg + 1], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_UNKNOWN && val.m_eType != TYPE_STRING)
                            ToString(val);
                        if(val.m_eType == TYPE_STRING)
                        {
                            XPathExprValue val2;
                            EvaluateExpression(m_syntaxTree[dwArg2 + 1], eAxis, pContextNode,
                                dwContextSize, dwContextPosition, val2);
                            if(val2.m_eType != TYPE_UNKNOWN && val2.m_eType != TYPE_STRING)
                                ToString(val2);
                            if(val2.m_eType == TYPE_STRING)
                                val = Contains(val.m_string, val2.m_string);
                            else
                                val.Clear();
                        }
                    }
                }
                else if(compare(pFuncName, STR("round")) == 0)  // round(number)
                {
                    DWORD dwArg = m_syntaxTree[dwIndexExpression + 2];
                    if(dwArg == 0)   // 1 argument needed.
                        val.Clear();
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwArg + 1], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_UNKNOWN && val.m_eType != TYPE_NUMBER)
                            ToNumber(val);
                        if(val.m_eType == TYPE_NUMBER)
                            val = round(val.m_dNumber);
                    }
                }
                else if(compare(pFuncName, STR("floor")) == 0)  // floor(number)
                {
                    DWORD dwArg = m_syntaxTree[dwIndexExpression + 2];
                    if(dwArg == 0)   // 1 argument needed.
                        val.Clear();
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwArg + 1], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_UNKNOWN && val.m_eType != TYPE_NUMBER)
                            ToNumber(val);
                        if(val.m_eType == TYPE_NUMBER)
                            val = floor(val.m_dNumber);
                    }
                }
                else if(compare(pFuncName, STR("ceiling")) == 0)  // ceiling(number)
                {
                    DWORD dwArg = m_syntaxTree[dwIndexExpression + 2];
                    if(dwArg == 0)   // 1 argument needed.
                        val.Clear();
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwArg + 1], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_UNKNOWN && val.m_eType != TYPE_NUMBER)
                            ToNumber(val);
                        if(val.m_eType == TYPE_NUMBER)
                            val = ceil(val.m_dNumber);
                    }
                }
                else if(compare(pFuncName, STR("substring")) == 0) // substring(string,number,number?)
                {
                    DWORD dwArg = m_syntaxTree[dwIndexExpression + 2];
                    DWORD dwArg2 = dwArg ? m_syntaxTree[dwArg + 2] : 0;
                    if(dwArg == 0 || dwArg2 == 0)   // 2 arguments needed.
                        val.Clear();
                    else
                    {
                        EvaluateExpression(m_syntaxTree[dwArg + 1], eAxis, pContextNode,
                            dwContextSize, dwContextPosition, val);
                        if(val.m_eType != TYPE_UNKNOWN && val.m_eType != TYPE_STRING)
                            ToString(val);
                        if(val.m_eType == TYPE_STRING)
                        {
                            XPathExprValue val2;
                            EvaluateExpression(m_syntaxTree[dwArg2 + 1], eAxis, pContextNode,
                                dwContextSize, dwContextPosition, val2);
                            if(val2.m_eType != TYPE_UNKNOWN && val2.m_eType != TYPE_NUMBER)
                                ToNumber(val2);
                            if(val2.m_eType == TYPE_NUMBER)
                            {
                                double dStart = val2.m_dNumber;
                                double dLength = -1;

                                // Look for an optional third argument.
                                dwArg2 = m_syntaxTree[dwArg2 + 2];
                                if(dwArg2)
                                {
                                    EvaluateExpression(m_syntaxTree[dwArg2 + 1], eAxis, pContextNode,
                                        dwContextSize, dwContextPosition, val2);
                                    if(val2.m_eType != TYPE_UNKNOWN && val2.m_eType != TYPE_NUMBER)
                                        ToNumber(val2);
                                    if(val2.m_eType == TYPE_NUMBER)
                                        dLength = val2.m_dNumber;
                                }

                                // Extract the substring.
                                val = SubString(val.m_string, (DWORD)round(dStart), (DWORD)round(dLength));
                            }
                            else
                                val.Clear();
                        }
                    }
                }
                else
                    val.Clear();    // Other functions not implemented.
            }
            break;

        case STREE_FUNCTION_ARGUMENT:
            EvaluateExpression(m_syntaxTree[dwIndexExpression + 1], eAxis, pContextNode,
                dwContextSize, dwContextPosition, val);
            break;
    }
}

template<typename U>
void XPathExpression<U>::Evaluate(const XmlParse<U>::XMLTree::NodeRoot* pRootNode, XPathExprValue& outVal)
{
    m_pXMLTreeRoot = pRootNode;

    EvaluateExpression(
        m_dwRootNode,   // Root of the XPath expression tree.
        child,          // Default axis (child).
        NULL,           // No default context node.
        0,              // Context size.
        0,              // Context position.
        outVal);        // Ouptut value.
}

template<typename U>
BOOL XPathExpression<U>::NodeSetCompare(
    const OrderedNodeSet& nodeSet, const XPathExprValue& val,
    SYNTAX_TREE_NODE_SUBTYPE eOperatorType)
{
    BOOL bRet = FALSE;

    switch(val.m_eType)
    {
        case TYPE_BOOLEAN:
            {
                // Convert the node set to a boolean:
                BOOL b = !nodeSet.empty();

                switch(eOperatorType)
                {
                    case EQUALITY_EQUALS:
                        bRet = b == val.m_bBoolean;
                        break;
                    case EQUALITY_NOT_EQUALS:
                        bRet = b != val.m_bBoolean;
                        break;
                    case RELATIONAL_GREATER:
                        // Compare is if the booleans were converted to numbers with
                        // double d = b ? 1.0 : 0.0
                        bRet = (int)b > (int)val.m_bBoolean;
                        break;
                    case RELATIONAL_GREATER_OR_EQUAL:
                        // Compare is if the booleans were converted to numbers with
                        // double d = b ? 1.0 : 0.0
                        bRet = (int)b >= (int)val.m_bBoolean;
                        break;
                    case RELATIONAL_LESS:
                        // Compare is if the booleans were converted to numbers with
                        // double d = b ? 1.0 : 0.0
                        bRet = (int)b < (int)val.m_bBoolean;
                        break;
                    case RELATIONAL_LESS_OR_EQUAL:
                        // Compare is if the booleans were converted to numbers with
                        // double d = b ? 1.0 : 0.0
                        bRet = (int)b <= (int)val.m_bBoolean;
                        break;
                }
            }
            break;
        case TYPE_NUMBER:
            {
                for(OrderedNodeSet::const_iterator ppNode = nodeSet.begin(); !bRet && ppNode != nodeSet.end(); ++ppNode)
                {
                    // Get the node's string-value:
                    double d;
                    if(ToNumber((*ppNode)->StringValue().c_str(), d))
                    {
                        switch(eOperatorType)
                        {
                            case EQUALITY_EQUALS:
                                bRet = d == val.m_dNumber;
                                break;
                            case EQUALITY_NOT_EQUALS:
                                bRet = d != val.m_dNumber;
                                break;
                            case RELATIONAL_GREATER:
                                bRet = d > val.m_dNumber;
                                break;
                            case RELATIONAL_GREATER_OR_EQUAL:
                                bRet = d >= val.m_dNumber;
                                break;
                            case RELATIONAL_LESS:
                                bRet = d < val.m_dNumber;
                                break;
                            case RELATIONAL_LESS_OR_EQUAL:
                                bRet = d <= val.m_dNumber;
                                break;
                        }
                    }
                }
            }
            break;
        case TYPE_STRING:
            {
                for(OrderedNodeSet::const_iterator ppNode = nodeSet.begin(); !bRet && ppNode != nodeSet.end(); ++ppNode)
                {
                    // Get the node's string-value:
                    int iRet = compare((*ppNode)->StringValue().c_str(), val.m_string);
                    switch(eOperatorType)
                    {
                        case EQUALITY_EQUALS:
                            bRet = iRet == 0;
                            break;
                        case EQUALITY_NOT_EQUALS:
                            bRet = iRet != 0;
                            break;
                        case RELATIONAL_GREATER:
                            bRet = iRet > 0;
                            break;
                        case RELATIONAL_GREATER_OR_EQUAL:
                            bRet = iRet >= 0;
                            break;
                        case RELATIONAL_LESS:
                            bRet = iRet < 0;
                            break;
                        case RELATIONAL_LESS_OR_EQUAL:
                            bRet = iRet <= 0;
                            break;
                    }
                }
            }
            break;
    }

    return bRet;
}

template<typename U>
BOOL XPathExpression<U>::NodeSetCompare(
    const XPathExprValue& val, const OrderedNodeSet& nodeSet,
    SYNTAX_TREE_NODE_SUBTYPE eOperatorType)
{
    // This function is evaluated by calling the other overload.
    // Change the operator to reflect the change in argument order:
    switch(eOperatorType)
    {
        case RELATIONAL_GREATER:
            eOperatorType = RELATIONAL_LESS;
            break;
        case RELATIONAL_GREATER_OR_EQUAL:
            eOperatorType = RELATIONAL_LESS_OR_EQUAL;
            break;
        case RELATIONAL_LESS:
            eOperatorType = RELATIONAL_GREATER;
            break;
        case RELATIONAL_LESS_OR_EQUAL:
            eOperatorType = RELATIONAL_GREATER_OR_EQUAL;
            break;
    }

    return NodeSetCompare(nodeSet, val, eOperatorType);
}

template<typename U>
BOOL XPathExpression<U>::NodeSetCompare(
    const OrderedNodeSet& nodeSet1, const OrderedNodeSet& nodeSet2,
    SYNTAX_TREE_NODE_SUBTYPE eOperatorType)
{
    for(OrderedNodeSet::const_iterator ppNode1 = nodeSet1.begin(); ppNode1 != nodeSet1.end(); ++ppNode1)
    {
        // Take the string-value of the first node:
        XmlParse<U>::XmlString str1 = (*ppNode1)->StringValue();
        for(OrderedNodeSet::const_iterator ppNode2 = nodeSet2.begin(); ppNode2 != nodeSet2.end(); ++ppNode2)
        {
            int iRet = compare(str1.c_str(), (*ppNode2)->StringValue().c_str());
            switch(eOperatorType)
            {
                case EQUALITY_EQUALS:
                    if(iRet == 0)
                        return TRUE;
                    break;
                case EQUALITY_NOT_EQUALS:
                    if(iRet != 0)
                        return TRUE;
                    break;
                case RELATIONAL_GREATER:
                    if(iRet > 0)
                        return TRUE;
                    break;
                case RELATIONAL_GREATER_OR_EQUAL:
                    if(iRet >= 0)
                        return TRUE;
                    break;
                case RELATIONAL_LESS:
                    if(iRet < 0)
                        return TRUE;
                    break;
                case RELATIONAL_LESS_OR_EQUAL:
                    if(iRet <= 0)
                        return TRUE;
                    break;
            }
        }
    }

    return FALSE;
}

template<typename U>
XPathExpression<U>::~XPathExpression()
{
    if(m_pExpressionString)
        delete m_pExpressionString;

    // The syntax tree needs to be properly deallocated.
    // Walk the tree, and deallocate each string.
    DWORD dwSize = m_syntaxTree.size();
    DWORD dwSkip;
    for(DWORD dwIndex = 0; dwIndex < dwSize; ++dwIndex)
    {
        switch(m_syntaxTree[dwIndex])
        {
            case STREE_OR:
            case STREE_AND:
            case STREE_UNION:
            case STREE_SLASH:
            case STREE_FILTER_EXPRESSION:
            case STREE_PREDICATE:
            case STREE_NUMBER:
            case STREE_FUNCTION_ARGUMENT:
                dwSkip = 2;
                break;

            case STREE_EQUALITY:
            case STREE_RELATIONAL:
            case STREE_ADDITIVE:
            case STREE_MULTIPLICATIVE:
            case STREE_STEP:
                dwSkip = 3;
                break;

            case STREE_UNARY_MINUS:
            case STREE_ROOT_SLASH:
            case STREE_AXIS:
                dwSkip = 1;
                break;

            case STREE_NODETEST:
                switch(m_syntaxTree[dwIndex + 1])
                {
                    case nt_processing_instruction_literal:
                    case nt_ncname_star:
                    case nt_qname:
                        delete [] (U*)m_syntaxTree[dwIndex + 2];
                        dwSkip = 2;
                        break;
                    default:
                        dwSkip = 1;
                        break;
                }
                break;

            case STREE_LITERAL:
            case STREE_VARIABLE_REF:
                delete [] (U*)m_syntaxTree[dwIndex + 1];
                dwSkip = 1;
                break;

            case STREE_FUNCTION_CALL:
                delete [] (U*)m_syntaxTree[dwIndex + 1];
                dwSkip = 2;
                break;

            default:
                dwSkip = 0;
                break;
        }

        dwIndex += dwSkip;
    }
}
