/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _XML_PARSER_HPP
#define _XML_PARSER_HPP

///////////////////////////////////////////////////////////////////////////////
// This file defines an XML parser class.
// It is based on the document:
//      "Extensible Markup Language (XML) 1.0 (Second Edition)
//       W3C Recommendation 6 October 2000"
// available at http://www.w3.org/TR/2000/REC-xml-20001006.
//
// The class is defined as a template, in order to be able to process
// single-byte or double-byte character sets:
// - If a single-byte character set is used, the template parameter U is char.
// - If the character set needs more than a byte per character, as UTF-8 or
//   UTF-16, then the template parameter U is wchar_t.
// The idea is to avoid a call to MultiByteToWideChar to convert the input
// buffer.
///////////////////////////////////////////////////////////////////////////////

#include "charclasses.hpp"
#pragma warning (disable: 4250 4290 4702 4786 4710 4711)
#include <vector>
#include <string>
#include <iosfwd>
#include <sstream>
#include <typeinfo>


// Comment out the following to allow text nodes as pure whitespace
// and to avoid removing leading and trailing whitespace:
#define XML_IGNORE_WHITESPACE

template <typename U>
class XmlParse
{
public:
    typedef std::basic_ostream<U, std::char_traits<U> > XmlParseStream;
    typedef std::basic_string <U, std::char_traits<U>, std::allocator<U> > XmlString;
    typedef std::basic_ostringstream<U, std::char_traits<U>, std::allocator<U> > XmlStringStream;

    // Exceptions thrown when encoutering a parsing error.
    struct ParsingError
    {
        DWORD m_dwLine, m_dwCol;
    protected:
        ParsingError(DWORD dwLine, DWORD dwCol) : m_dwLine(dwLine), m_dwCol(dwCol) {}
    };
    struct LexicalError : public ParsingError
    {
        LexicalError(DWORD dwLine, DWORD dwCol) : ParsingError(dwLine, dwCol) {}
    };
    struct SyntaxError  : public ParsingError
    {
        SyntaxError(DWORD dwLine, DWORD dwCol) : ParsingError(dwLine, dwCol) {}
    };

    // String type.
    // This type assumes that the lifetime of an instance is always enclosed
    // within the lifetime of the actual string of characters pointed to.
    class String
    {
    public:
        String() : m_pcStr(NULL), m_dwLen(0) {}
        String(const String& str) : m_pcStr(str.m_pcStr), m_dwLen(str.m_dwLen) {}
        String(const U* pcStr, DWORD dwLen) : m_pcStr(pcStr), m_dwLen(dwLen) {}

        String& operator= (const String& str) {m_pcStr = str.m_pcStr; m_dwLen = str.m_dwLen; return *this;}
        BOOL    operator==(const String& str) const {return m_dwLen == str.m_dwLen && compare(m_pcStr, str.m_pcStr, m_dwLen) == 0;}
        BOOL    operator==(const U* zStr)     const {return m_dwLen == (DWORD)stringlen(zStr) && compare(m_pcStr, zStr, m_dwLen) == 0;}
        BOOL    operator!=(const String& str) const {return !(str == *this);}
        BOOL    operator!=(const U* zStr)     const {return !(*this == zStr);}
        U       operator[](int ix)            const {return m_pcStr[ix];}
        operator XmlString() const {return XmlString(m_pcStr, (XmlString::size_type)m_dwLen);}

        // Remove leading and trailing whitespace.
        String& Trim()
        {
            const U* pcLast  = m_pcStr + m_dwLen - 1;
            while(m_pcStr < pcLast && (*m_pcStr == ' ' || *m_pcStr == '\t' || *m_pcStr == '\r' || *m_pcStr == '\n'))
                ++m_pcStr;
            while(pcLast >= m_pcStr && (*pcLast == ' ' || *pcLast  == '\t' || *pcLast  == '\r' || *pcLast  == '\n'))
                --pcLast;
            m_dwLen = pcLast + 1 - m_pcStr;
            return *this;
        }

        String& Assign(const U* pcStr, DWORD dwLen) {return *this;} // Requires specialization.
        DWORD   GetSize() const {return m_dwLen;}
        void    Copy(U* zBuf) const // CAUTION: no length check!
        {
            CopyMemory(zBuf, m_pcStr, m_dwLen * sizeof U);
            zBuf[m_dwLen] = 0;
        }

        // Returns a substring of the string starting at position dwStart and ending (but
        // NOT including) position dwEnd. Positions are zero-based.
        String Substring(DWORD dwStart, DWORD dwEnd)
        {
            if(dwEnd > m_dwLen)
                dwEnd = m_dwLen;
            if(dwStart >= m_dwLen || dwEnd <= dwStart)
                return String();
            return String(m_pcStr + dwStart, dwEnd - dwStart);
        }

        // Returns a suffix of the string starting at position dwPos (0 based).
        String AfterPosition(DWORD dwPos)
        {
            return dwPos >= m_dwLen ? String() : String(m_pcStr + dwPos, m_dwLen - dwPos);
        }

        // Assumes the 2 strings are in the same buffer, and that str comes after.
        // If the string pointed to by "this" was initially NULL, it is set to str.
        String& operator+=(const String& str)
        {
            if(!m_pcStr)
                *this = str;
            else
                m_dwLen = str.m_pcStr + str.m_dwLen - m_pcStr;
            return *this;
        }

        String operator+(const String& str)
        {
            return m_pcStr ? String(m_pcStr, str.m_pcStr + str.m_dwLen - m_pcStr) : str;
        }

        void Clear() {m_pcStr = NULL; m_dwLen = 0;}
        BOOL IsEmpty() const {return m_pcStr == NULL || m_dwLen == 0;}
        BOOL is_pubidliteral() const
        {
            BOOL bRet = TRUE;
            for(DWORD dw = 0; bRet && dw < m_dwLen; ++dw)
            {
                if(!IS_PUBIDCHAR(m_pcStr[dw]))
                    bRet = FALSE;
            }

            return bRet;
        }
        BOOL IsWhitespace() const
        {
            BOOL bRet = TRUE;
            for(DWORD dw = 0; bRet && dw < m_dwLen; ++dw)
            {
                U c = m_pcStr[dw];
                if(c != ' ' && c != '\t' && c != '\n' && c != '\r')
                    bRet = FALSE;
            }

            return bRet;
        }

    private:
        const U*    m_pcStr;    // Pointer to the actual string.
        DWORD       m_dwLen;    // Number of characters in string.
    };

    // XML document tree class.
    // An instance of this class can be produced by the XML parser.
    struct XMLTree
    {
        XMLTree() {}

        struct Node // The base type of all nodes.
        {
            DWORD m_dwID;       // Node identifier, unique in the tree,
                                // reflecting the position of the node in document order.
            Node(DWORD dwID = 0) : m_dwID(dwID) {}
            virtual ~Node(){}

            // Return the "string-value" defined by XPath.
            virtual XmlString StringValue() const = 0;

            // Return a simple absolute XPath location path to the node.
            virtual XmlString GenerateAddress() const = 0;

        protected:
            virtual XmlString GenerateNodeString() const = 0;
            static const char*    MAKE_obracket     (const char*)   {return  "[";}
            static const wchar_t* MAKE_obracket     (const wchar_t*){return L"[";}
            static const char*    MAKE_cbracket     (const char*)   {return  "]";}
            static const wchar_t* MAKE_cbracket     (const wchar_t*){return L"]";}
        };

        struct NodeParent;
        struct NodeChild : public virtual Node
        {
            NodeParent&     m_parent;
            String          m_strValue;
            NodeChild*      m_pPrecedingSibling;
            NodeChild*      m_pFollowingSibling;

            NodeChild(NodeParent& parent, const String& strValue, DWORD dwID = 0)
                : Node(dwID), m_parent(parent), m_strValue(strValue),
                  m_pPrecedingSibling(NULL), m_pFollowingSibling(NULL) {}
            virtual XmlString GenerateAddress() const
            {
                XmlStringStream os;
                os <<  m_parent.GenerateAddress() << GenerateNodeString()
                   << MAKE_obracket((const U*)0)  << GetSiblingNumber()
                   << MAKE_cbracket((const U*)0);
                return os.str();
            }

            DWORD GetSiblingNumber() const
            {
                DWORD dwNum = 1;
                for(const NodeChild* pNodeChild = m_pPrecedingSibling;
                    pNodeChild;
                    pNodeChild = pNodeChild->m_pPrecedingSibling)
                {
                    if(typeid(*pNodeChild) == typeid(*this))
                        ++dwNum;
                }
                return dwNum;
            }
#ifdef XML_DUMP
            virtual void Dump(XmlParseStream& /*os*/, int /*iLevel*/) {} // Requires specialization.
#endif
        private:
            //// Forbidden:
            NodeChild();
            NodeChild(const NodeChild&);
            operator=(const NodeChild&);
        };
        typedef std::vector<NodeChild*> Children;

        struct NodeParent : public virtual Node
        {
            Children        m_children;

            NodeParent(DWORD dwID = 0) : Node(dwID) {}
            virtual ~NodeParent()
            {
                for(Children::iterator pChild = m_children.begin(); pChild != m_children.end(); ++pChild)
                    delete *pChild;
            }

            // Add a new child and link with previous sibling.
            void AddChild(NodeChild* pNewChild)
            {
                NodeChild* pPrevious = m_children.empty() ? NULL : m_children.back();
                m_children.push_back(pNewChild);
                if(pPrevious)
                {
                    pPrevious->m_pFollowingSibling = pNewChild;
                    pNewChild->m_pPrecedingSibling = pPrevious;
                }
            }

            // Retrieve the Nth child. Returns NULL if no such child.
            const NodeChild* operator[](DWORD dwPos) const {return dwPos < (DWORD)m_children.size() ? m_children[dwPos] : NULL;}

            // Retrieve the child with value zStr. Returns NULL if no such child.
            const NodeChild* GetChildByValue(const U* zStr) const
            {
                const NodeChild* pChildRet = NULL;
                for(Children::const_iterator pChild = m_children.begin(); pChild != m_children.end(); ++pChild)
                {
                    if((*pChild)->m_strValue == zStr)
                    {
                        pChildRet = *pChild;
                        break;
                    }
                }
                return pChildRet;
            }

            virtual XmlString StringValue() const {return GetStringValue(this);}

        protected:
            static const char*    MAKE_slash        (const char*)   {return  "/";}
            static const wchar_t* MAKE_slash        (const wchar_t*){return L"/";}

        private:
            static XmlString GetStringValue(const Node* pNode)
            {
                const NodeText* pNodeText = dynamic_cast<const NodeText*>(pNode);
                if(pNodeText)
                    return pNodeText->StringValue();
                else
                {
                    const NodeParent* pParent = dynamic_cast<const NodeParent*>(pNode);
                    if(pParent)
                    {
                        // Concatenate the string values (recursively).
                        XmlString str;
                        for(Children::const_iterator pChild = pParent->m_children.begin(); pChild != pParent->m_children.end(); ++pChild)
                            str += GetStringValue(*pChild);
                        return str;
                    }
                }

                return XmlString(); // For non-parent nodes that are also not text nodes.
            }
        };

        struct NodeComment : public NodeChild
        {
            NodeComment(NodeParent& parent, const String& strValue, DWORD dwID)
                : Node(dwID), NodeChild(parent, strValue, dwID) {}
            virtual XmlString StringValue() const {return m_strValue;}
#ifdef XML_DUMP
            virtual void Dump(XmlParseStream& os, int iLevel) {} // Requires specialization.
#endif
        private:
            //// Forbidden:
            NodeComment();
            NodeComment(const NodeComment&);
            operator=(const NodeComment&);

            static const char*    MAKE_comment      (const char*)   {return  "/comment()";}
            static const wchar_t* MAKE_comment      (const wchar_t*){return L"/comment()";}

            virtual XmlString GenerateNodeString() const
            {
                return MAKE_comment((const U*)0);
            }
        };
        struct NodeProcessingInstruction : public NodeChild
        {
            String          m_strTarget;

            NodeProcessingInstruction(NodeParent& parent, const String& strValue,
                                      const String& strTarget, DWORD dwID)
                : Node(dwID),
                  NodeChild(parent, strValue, dwID),
                  m_strTarget(strTarget)
            {}
            virtual XmlString StringValue() const {return m_strValue;}
#ifdef XML_DUMP
            virtual void Dump(XmlParseStream& os, int iLevel) {} // Requires specialization.
#endif
        private:
            //// Forbidden:
            NodeProcessingInstruction();
            NodeProcessingInstruction(const NodeProcessingInstruction&);
            operator=(const NodeProcessingInstruction&);

            static const char*    MAKE_proc_instr   (const char*)   {return  "/processing-instruction()";}
            static const wchar_t* MAKE_proc_instr   (const wchar_t*){return L"/processing-instruction()";}

            virtual XmlString GenerateNodeString() const
            {
                return MAKE_proc_instr((const U*)0);
            }
        };
        struct NodeElement;
        struct NodeText : public NodeChild  // Only NodeElement can be a parent.
        {
            NodeText(NodeParent& parent, const String& strValue, DWORD dwID)
                : Node(dwID), NodeChild(parent, strValue, dwID) {}
            virtual XmlString StringValue() const {return m_strValue;}
#ifdef XML_DUMP
            virtual void Dump(XmlParseStream& os, int iLevel) {} // Requires specialization.
#endif
        private:
            //// Forbidden:
            NodeText();
            NodeText(const NodeText&);
            operator=(const NodeText&);

            static const char*    MAKE_text         (const char*)   {return  "/text()";}
            static const wchar_t* MAKE_text         (const wchar_t*){return L"/text()";}

            virtual XmlString GenerateNodeString() const
            {
                return MAKE_text((const U*)0);
            }
        };
        struct NodeAttribute : public Node
        {
            NodeElement&    m_parent;
            String          m_strName;
            String          m_strValue;
            NodeAttribute*  m_pPrecedingSibling;
            NodeAttribute*  m_pFollowingSibling;

            NodeAttribute(NodeElement& parent, const String& strName,
                const String& strValue, DWORD dwID)
                : Node(dwID), m_parent(parent), m_strName(strName), m_strValue(strValue),
                  m_pPrecedingSibling(NULL), m_pFollowingSibling(NULL) {}
            virtual XmlString StringValue() const {return m_strValue;}
            virtual XmlString GenerateAddress() const
            {
                XmlStringStream os;
                os <<  m_parent.GenerateAddress() << GenerateNodeString()
                   << MAKE_obracket((const U*)0)  << GetSiblingNumber()
                   << MAKE_cbracket((const U*)0);
                return os.str();
            }

            DWORD GetSiblingNumber() const
            {
                DWORD dwNum = 1;
                for(const NodeAttribute* pNodeAttribute = m_pPrecedingSibling;
                    pNodeAttribute;
                    pNodeAttribute = pNodeAttribute->m_pPrecedingSibling)
                        ++dwNum;
                return dwNum;
            }

        private:
            //// Forbidden:
            NodeAttribute();
            NodeAttribute(const NodeAttribute&);
            operator=(const NodeAttribute&);

            static const char*    MAKE_attribute    (const char*)   {return  "/@*";}
            static const wchar_t* MAKE_attribute    (const wchar_t*){return L"/@*";}

            virtual XmlString GenerateNodeString() const
            {
                return MAKE_attribute((const U*)0);
            }
        };
        typedef std::vector<NodeAttribute*> Attributes;
        struct NodeElement : public NodeParent, public NodeChild
        {
            Attributes      m_attributes;

            NodeElement(NodeParent& parent, const String& strName, DWORD dwID)
                : Node(dwID), NodeChild(parent, strName) {}
            virtual ~NodeElement()
            {
                for(Attributes::iterator pAtt = m_attributes.begin(); pAtt != m_attributes.end(); ++pAtt)
                    delete *pAtt;
            }

            // Is this element the root element?
            BOOL IsRoot() const {return dynamic_cast<NodeRoot*>(&m_parent) != NULL;}

            // Retrieve the first child element with name zStr. Returns NULL if no such child.
            const NodeElement* GetChildElementByName(const U* zStr) const
            {
                const NodeElement* pChildRet = NULL;
                for(Children::const_iterator pChild = m_children.begin(); pChild != m_children.end(); ++pChild)
                {
                    if((pChildRet = dynamic_cast<NodeElement*>(*pChild)) != NULL
                        && pChildRet->m_strValue == zStr)
                        break;
                }
                return pChildRet;
            }

            // Iterate through child element nodes.
            typedef Children::const_iterator ChildElementIterator;
            ChildElementIterator FirstChildElement() const
            {
                ChildElementIterator it =  m_children.begin();
                while(it != m_children.end() && dynamic_cast<NodeElement*>(*it) == NULL)
                    ++it;
                return it;
            }
            ChildElementIterator ChildElementEnd() const {return m_children.end();}
            void NextChildElement(ChildElementIterator& it) const
            {
                do
                    ++it;
                while(it != m_children.end() && dynamic_cast<NodeElement*>(*it) == NULL);
            }

            // Retrieve first text node, if it exists (otherwise NULL).
            const NodeText* GetFirstChildText() const
            {
                NodeText* pText = NULL;
                Children::const_iterator it =  m_children.begin();
                while(it != m_children.end() && (pText = dynamic_cast<NodeText*>(*it)) == NULL)
                    ++it;
                return pText;
            }
#ifdef XML_DUMP
            virtual void Dump(XmlParseStream& os, int iLevel) {} // Requires specialization.
#endif
        private:
            //// No copy:
            NodeElement(const NodeElement&);
            operator=(const NodeElement&);

            static const char*    MAKE_star         (const char*)   {return  "*";}
            static const wchar_t* MAKE_star         (const wchar_t*){return L"*";}
            static const char*    MAKE_slashstar    (const char*)   {return  "/*";}
            static const wchar_t* MAKE_slashstar    (const wchar_t*){return L"/*";}

            virtual XmlString GenerateNodeString() const
            {
                return IsRoot() ? MAKE_star((const U*)0) : MAKE_slashstar((const U*)0);
            }
        };
        struct NodeRoot : public NodeParent
        {
            NodeElement*    m_pRootElement;

            NodeRoot(DWORD dwID = 0) : Node(dwID), NodeParent(dwID) {}
            virtual XmlString GenerateAddress() const
            {
                return MAKE_slash((const U*)0);
            }

        private:
            //// Forbidden:
            NodeRoot(const NodeRoot&);
            operator=(const NodeRoot&);

            virtual XmlString GenerateNodeString() const {return XmlString();}
        };

        // Returns the name of a node. Here, the definition of "name" depends
        // on the node type, but is suitable for the "QName" of a test node in an
        // XPath location step.
        // Returns NULL if no name is associated with the node.
        static const String* GetNodeName(const Node* pNode)
        {
            const XmlParse<U>::String* pStr = NULL;

            // Child node derivative? This includes NodeComment, NodeText,
            // NodeProcessingInstruction and NodeElement.
            const XMLTree::NodeChild* pChild = dynamic_cast<const XMLTree::NodeChild*>(pNode);
            if(pChild)
                pStr = &pChild->m_strValue;
            else
            {
                const XMLTree::NodeAttribute* pAtt = dynamic_cast<const XMLTree::NodeAttribute*>(pNode);
                if(pAtt)
                    pStr = &pAtt->m_strName;
            }

            return pStr;
        }

        //// Data:
        NodeRoot m_root;

#ifdef XML_DUMP
        // The following functions require specialization.
        void Dump(XmlParseStream& os) {}
        static void DumpAttribute(const NodeAttribute& attribute, XmlParseStream& os, int iLevel) {}
        static void DumpChildren(const Children& children, XmlParseStream& os, int iLevel) {}
        static void Margin(XmlParseStream& os, int n)
        {
            U* pcMargin = (U*)_alloca((n + 1) * sizeof(U));
            for(int i = 0; i < n; ++i)
                pcMargin[i] = 32;
            pcMargin[n] = 0;
            os << pcMargin;
        }
        static XmlParseStream& operator<<(XmlParseStream& os, const String& str)
        {
            return os << (const XmlString&)str;
        }
#endif

    private:

        //// No copy:
        XMLTree(const XMLTree&);
        operator=(const XMLTree&);
    };

    // The following method examines the input buffer, and determines which
    // character set / character encoding is used. If necessary, the buffer
    // can be byte swapped or transformed.
    // The output buffer is either an 8-bit character buffer, or a 16-bit
    // wide character UTF-16 little endian buffer.
    // For the moment, UTF-32 is not supported.
    // The original buffer is input via pInputBuf. A transformation can
    // occur in place directly in this original buffer, or a new buffer
    // can be allocated if space is insufficient, in which case bNewBuffer
    // is set to TRUE.
    // After parsing, an allocated buffer should be freed using the FreeBuf
    // helper.
    // The dwOffset output argument is the byte offset at which parsing should
    // start.
    // The "charset" directive of an HTTP "Content-type" response header has
    // priority over the possible "encoding" declaration in the XML document.
    // If such a directive exists, it should be input via the zCharset string;
    // or NULL should be used otherwise.
    // The eBufType output argument gives the character size in the transformed
    // buffer.
    // The return value indicates success or failure.
    enum XML_BUFFER_TYPE {OCTET_BUFFER, WIDECHAR_BUFFER};
    static BOOL NormalizeBuffer(BYTE*& pInputBuf, const char* zCharset,
                                XML_BUFFER_TYPE& eBufType, BOOL& bNewBuffer,
                                DWORD& dwOffset);
    static void FreeBuf(BYTE* pBuf);

    // Constructor.
    XmlParse(const U* pcInputBuf, BOOL bFetchEncodingOnly = FALSE);

    // Parsing an XML document. May throw a ParsingError derivative.
    void Parse(XMLTree* pOutTree = NULL) throw(ParsingError);
    BOOL GetEncoding(U* zOutBuf, DWORD dwBufSizeInChars)
    {
        DWORD dwEncodingSize = m_lexeme.GetSize();
        BOOL bRet = m_bFetchEncodingOnly && dwEncodingSize && dwBufSizeInChars > dwEncodingSize;
        if(bRet)
            m_lexeme.Copy(zOutBuf);
        return bRet;
    }

    //// Public helpers:
    static int compare(const char *string1, const char *string2, size_t count)
    {
        return strncmp(string1, string2, count);
    }
    static int compare(const wchar_t *string1, const wchar_t *string2, size_t count)
    {
        return wcsncmp(string1, string2, count);
    }
    static int stringlen(const char* lpString)
    {
        return lstrlenA(lpString);
    }
    static int stringlen(const wchar_t* lpString)
    {
        return lstrlenW(lpString);
    }

private:
    //// Static helpers:

    // String selector:
    static const char*    string_selector(const char*, const char* s1, const wchar_t*)
    {
        return s1;
    }
    static const wchar_t* string_selector(const wchar_t*, const char*, const wchar_t* s2)
    {
        return s2;
    }
    #define STR(s)        string_selector((const U*)0, s, L##s)

    static UINT ParseCharset(const char* zCharset); // Returns 0xFFFFFFFF in case of failure.

    // Assumptions based on byte-order mark.
    enum XML_CHARACTERS
    {
        UTF8,
        UTF16_LITTLE_ENDIAN,
        UTF16_BIG_ENDIAN,
        UTF32_LITTLE_ENDIAN,
        UTF32_BIG_ENDIAN,
    };
    static XML_CHARACTERS GetXMLBufferLayout(const BYTE* pBuffer);
    static BOOL IsBasicLatin(const BYTE* pBuffer);
    static void ConvertByteOrder(wchar_t* pBuffer);

    // Parsing context.
    enum Context
    {
        CONTEXT_MARKUP,                     // Inside markup.
        CONTEXT_ELEMENT_CONTENT_CHARDATA,   // Inside an element content's character data.
        CONTEXT_CDATA_SECTION,              // Inside a CDATA section.
        CONTEXT_ATTRIBUTE_VALUE,            // Inside an attribute value.
        CONTEXT_ENTITY_VALUE                // Inside an entity value.
    };

    // Lexical token:
    enum XML_TOKEN
    {
        TOKEN_XML_ERROR,                        // Lexical error.
        TOKEN_XML_EQUAL,                        // =
        TOKEN_XML_LEFT_SQUAREBRACKET,           // [
        TOKEN_XML_RIGHT_SQUAREBRACKET,          // ]
        TOKEN_XML_LEFT_PARENTHESIS,             // (
        TOKEN_XML_RIGHT_PARENTHESIS,            // )
        TOKEN_XML_SHARP,                        // #
        TOKEN_XML_PIPE,                         // |
        TOKEN_XML_STAR,                         // *
        TOKEN_XML_QUESTIONMARK,                 // ?
        TOKEN_XML_PLUS,                         // +
        TOKEN_XML_COMMA,                        // ,
        TOKEN_XML_PERCENT,                      // %
        TOKEN_XML_SEMICOLON,                    // ;
        TOKEN_XML_BEGIN_START_TAG,              // <
        TOKEN_XML_BEGIN_END_TAG,                // </
        TOKEN_XML_BEGIN_DOCTYPE,                // <!DOCTYPE
        TOKEN_XML_BEGIN_ELEMENT,                // <!ELEMENT
        TOKEN_XML_BEGIN_ATTLIST,                // <!ATTLIST
        TOKEN_XML_BEGIN_ENTITY,                 // <!ENTITY
        TOKEN_XML_BEGIN_NOTATION,               // <!NOTATION
        TOKEN_XML_BEGIN_CDATA,                  // <![CDATA[
        TOKEN_XML_BEGIN_XMLDECL,                // <?xml
        TOKEN_XML_END_TAG,                      // >
        TOKEN_XML_END_EMPTY_ELEMENT_TAG,        // />
        TOKEN_XML_PROCESSING_INSTRUCTION,       // <? PITarget xxx ?> as in rule [16].
        TOKEN_XML_COMMENT,                      // <!-- xxx --> as in rule [15].
        TOKEN_XML_ENTITY_REF,                   // Entity reference: &xxx;
        TOKEN_XML_CHAR_REF,                     // Character reference: &#nnn; or &#xhhh;
        TOKEN_XML_WHITESPACE,                   // space, tab, carriage return, line feed, as defined in rule [3].
        TOKEN_XML_END_CDATA,                    // Matches the actual characters of the CDATA section and the trailling ]]>.
        TOKEN_XML_CHARDATA,                     // Character data, as defined in rule [14].
        TOKEN_XML_NAME,                         // Name, as defined in rule [5].
        TOKEN_XML_ATTVALUE,                     // Attribute value, as defined in rule [10].
        TOKEN_XML_SYSTEMLITERAL,                // System literal, as defined in rule [11].
        TOKEN_XML_END                           // End of input buffer.
    };

    //// Lexical helpers:
    XML_TOKEN GetNextXMLToken();
    void Match(XML_TOKEN eToken);
    void StringMatch(const U* zString);
    void MatchNoAdvance(XML_TOKEN eToken) const;
    static BOOL is_digit(U c)    {return c >= '0' && c <= '9';}
    static BOOL is_hexdigit(U c) {return c >= '0' && c <= '9' || c >= 'A' && c <= 'Z' || c >= 'a' && c <= 'z';}

    //// Syntactic helpers:
    void document(XMLTree::NodeRoot* pRootNode = NULL);
    void prolog(XMLTree::NodeRoot* pRootNode = NULL);
    XMLTree::NodeElement* element(XMLTree::NodeParent* pParent = NULL);
    void content(XMLTree::NodeElement* pParent = NULL);
    void Misc();
    void Attribute(XMLTree::NodeElement* pParent = NULL);
    void ETag(const String& strName);
    void XMLDecl();
    void VersionInfo();
    void doctypedecl();
    void DeclSep();
    void markupdecl();
    void EntityDecl();
    void EntityDef();
    void PEDef();
    void elementdecl();
    void contentspec();
    void choice_or_seq_continuation();
    void cp();
    void AttlistDecl();
    void DefaultDecl();
    void AttType();
    void EnumeratedType();
    void NotationDecl();
    void ExternalID();

    //// Data:
    BOOL        m_bGoodInputBuffer; // TRUE if the input buffer is valid.
    BOOL        m_bFetchEncodingOnly; // If TRUE, parsing is started only to fetch
                                      // the encoding scheme, and stops as soon as
                                      // it was found or it was determined that no
                                      // encoding scheme is specified. Use only with
                                      // a monobyte U type.
    const U*    m_pcInputBuf;       // Pointer to the beginning of the input buffer.
    const U*    m_pcCur;            // Pointer to the current character in the input buffer.
    DWORD       m_dwLine, m_dwCol;  // Current line and column in the input buffer. An end
                                    // of line is defined by the character '\n'.
    Context     m_eContext;         // The parsing context.
    BOOL        m_bInStartTag;      // TRUE if we are in a start tag.
    BOOL        m_bInDTD;           // TRUE if we are in the DTD.
    BOOL        m_bEntityValue;     // TRUE if we are in an entity value.
    DWORD       m_dwDepth;          // Depth level in element content.
    XML_TOKEN   m_eLookahead;       // The current lookahead.
    DWORD       m_dwLookaheadLine, m_dwLookaheadCol; // The line and column of the first
                                                     // character of the current lookahead.
    String      m_lexeme;           // Lexeme for current token.
    int         m_iCounter;         // Counter to generate XML tree node IDs in document order.

    // Relevant only when the token is a TOKEN_XML_PROCESSING_INSTRUCTION:
    DWORD       m_dwPITargetBegin;  // Position of the first character of a processing
                                    // instruction target.
    DWORD       m_dwPITargetEnd;    // Position of the character just after a processing
                                    // instruction target.
    DWORD       m_dwPIBodyBegin;    // Position of the first character of a processing
                                    // instruction body (what is after the target).
};

#include "xmlparser.inl"

#endif  // _XML_PARSER_HPP
