// TestDebugCommsProxy.cpp: implementation of the CTestDebugCommsProxy class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TestDebugCommsProxy.h"
#include "TestDebug.h"

extern void Trace( const char *format ,... );
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTestDebugCommsProxy::CTestDebugCommsProxy(CTestDebug *pTestDebug):
	m_pTestDebug(pTestDebug),
	m_pCommandThread(NULL),
	m_hPipe(INVALID_HANDLE_VALUE)
{
	m_pCommandThread = AfxBeginThread(&CommandThread,this);
	
	m_hTerminateEvent = CreateEvent(NULL,TRUE,FALSE,NULL);
}

CTestDebugCommsProxy::~CTestDebugCommsProxy()
{
	CloseHandle(m_hTerminateEvent);
}

BOOL CTestDebugCommsProxy::NotifyTestComplete()
{
	//Notify that we are about to shutdown
	SetEvent(m_hTerminateEvent);

	//Wait for everything to terminate
	const DWORD dwNumHandles = 1;
	HANDLE pHandles[dwNumHandles] = {m_pCommandThread->m_hThread};
	if(WaitForMultipleObjects(dwNumHandles,pHandles,TRUE,INFINITE) != WAIT_OBJECT_0)
		return false;

	return true;
}

BOOL CTestDebugCommsProxy::WaitForPipeConnection()
{
	//Wait for a client to connect to pipe or for us to terminate
	OVERLAPPED ol;
	memset(&ol,0,sizeof(ol));
	ol.hEvent = CreateEvent(NULL,TRUE,FALSE,NULL);

	if(!ConnectNamedPipe(m_hPipe, &ol) )
	{
		if(GetLastError() != ERROR_IO_PENDING)
		{
			Trace("Error (%d): Connect pipe failed \n",GetLastError());
			CloseHandle(ol.hEvent);
			return false;
		}

		const int nNumHandles = 2;
		HANDLE pHandles[nNumHandles] = {m_hTerminateEvent,ol.hEvent};

		//Wait for client to connect or us to terminate
		while(1)
		{
			Trace("Waiting for connect\n");
			
			DWORD dwResult = WaitForMultipleObjectsEx(nNumHandles,pHandles,FALSE,1000,FALSE);
			if(dwResult == WAIT_TIMEOUT)
				continue;

			CloseHandle(ol.hEvent);

			if(dwResult-WAIT_OBJECT_0 == 1)	
			{
				Trace("Client Connected\n");
				return true;	//Client connected
			}
			else	
				return false;		//Test complete or error
		}
	}

	Trace("Client Connected\n");

	CloseHandle(ol.hEvent);
	return true;
}

BOOL CTestDebugCommsProxy::CreateCommsPipe()
{
	if(m_hPipe != INVALID_HANDLE_VALUE)
		return true;

	const CString strPipename = "\\\\.\\pipe\\" + m_pTestDebug->GetTestName();
	PSECURITY_DESCRIPTOR pSD = new SECURITY_DESCRIPTOR;
	if (!InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION))
	{
		Trace("Error(%d): Could not initialise security descriptor\n",GetLastError());
		delete pSD;
		return FALSE;
	}

	if (!SetSecurityDescriptorDacl(pSD, TRUE, (PACL) NULL, FALSE))
	{	
		Trace("Error(%d): Could not set security dacl\n",GetLastError());
		delete pSD;
		return false;
	}

	SECURITY_ATTRIBUTES     sa;
	sa.nLength = sizeof(sa);
	sa.lpSecurityDescriptor = pSD;
	sa.bInheritHandle = TRUE;

	// Create a named pipe
	m_hPipe = CreateNamedPipe (strPipename,
				PIPE_ACCESS_DUPLEX                // 2 way pipe.
				| FILE_FLAG_OVERLAPPED,           // Use overlapped structure.
				PIPE_WAIT                         // Wait on messages.
				| PIPE_READMODE_BYTE
				|PIPE_TYPE_BYTE,
				100,				              // Maximum instance limit.
				0,			                  // Buffer sizes.
				0,
				0,		                          // Specify time out.
				&sa);                             // Security attributes.

	delete pSD;
	
	if (m_hPipe == INVALID_HANDLE_VALUE)
	{
		Trace("Error(%d): Could create comms pipe (%s)\n",GetLastError(),strPipename);
		return false;
	}

	return true;
}

BOOL CTestDebugCommsProxy::WriteToPipe(LPBYTE pBuffer, DWORD dwBufSize)
{
	CSingleLock lock(&m_CritSect);
	lock.Lock();

	OVERLAPPED ol;
	memset(&ol,0,sizeof(ol));
	ol.hEvent = CreateEvent(NULL,TRUE,FALSE,NULL);
	
	DWORD dwBytesWritten=0;
	BOOL bResult = WriteFile(m_hPipe,pBuffer,dwBufSize,&dwBytesWritten,&ol);

	if(!bResult)
	{
		if(GetLastError() == ERROR_IO_PENDING)
		{
			const DWORD dwNumHandles = 2;
			HANDLE pHandles[dwNumHandles] = {ol.hEvent,m_hTerminateEvent};
			DWORD dwResult = WaitForMultipleObjects(dwNumHandles,pHandles,FALSE,INFINITE);
	
			bResult = ((dwResult == WAIT_OBJECT_0) && 
						GetOverlappedResult(m_hPipe,&ol,&dwBytesWritten,FALSE) &&
						(dwBytesWritten == dwBufSize));
		}

		if(!bResult)
		{
			Trace("Writefile failed Error %d\n",GetLastError());
			CancelIo(m_hPipe);
		}
	}
	
	CloseHandle(ol.hEvent);
	return bResult;
}

BOOL CTestDebugCommsProxy::ReadFromPipe(LPBYTE pBuffer,DWORD dwBufSize)
{
	OVERLAPPED ol;
	memset(&ol,0,sizeof(ol));
	ol.hEvent = CreateEvent(NULL,TRUE,FALSE,NULL);

	BOOL bResult = true;
	DWORD dwBytesRead=0;
	if(!ReadFile(m_hPipe,pBuffer,dwBufSize,&dwBytesRead,&ol))
	{
		if(GetLastError() == ERROR_IO_PENDING)
		{
			const DWORD dwNumHandles = 2;
			HANDLE pHandles[dwNumHandles] = {ol.hEvent,m_hTerminateEvent};
			DWORD dwResult = WaitForMultipleObjects(dwNumHandles,pHandles,FALSE,INFINITE);
			if(dwResult == WAIT_OBJECT_0)
			{
				if(GetOverlappedResult(m_hPipe,&ol,&dwBytesRead,FALSE) && (dwBytesRead == dwBufSize))
					bResult = true;
				else
					bResult = false;
			}
			else
			if(dwResult == WAIT_OBJECT_0+1)
			{
				Trace("Test Complete in ReadFromPipe\n");
				CancelIo(m_hPipe);
				bResult = false;	//Terminated
			}
			else
				bResult = false;
		}
		else
			bResult = false;

		Trace("Readfile failed Error %d\n",GetLastError());
	}
	CloseHandle(ol.hEvent);
	return bResult;
}

UINT CTestDebugCommsProxy::CommandThread(LPVOID pParam)
{
	CTestDebugCommsProxy *pComms = (CTestDebugCommsProxy*)pParam;

	//Create the named pipe
	if(!pComms->CreateCommsPipe())
		return 0;

	while(1)
	{
		//Wait for a client to connect
		if(!pComms->WaitForPipeConnection())
			return 0;

		int nCommand = 0;
		while(1)
		{
			//Start a read operation
			if(!pComms->ReadFromPipe((LPBYTE)&nCommand,sizeof(int)))
				break;
			
			//Received command
			switch(nCommand)
			{
			case enStepSingleUser:
			{
				Trace("Got Step User Command\n");
				
				int nVUId;
				if(!pComms->ReadFromPipe((LPBYTE)&nVUId,sizeof(int)))
					break;
				
				pComms->m_pTestDebug->StepSingleUser(nVUId);
			}
			break;
			case enStepSingleEvent:
				Trace("Got Step Single Event Command\n");
				pComms->m_pTestDebug->StepSingleEvent();
			break;
			case enRun:
				Trace("Got Run Command\n");
				pComms->m_pTestDebug->Run();
			break;
			case enBreak:
				Trace("Got Break command\n");
				pComms->m_pTestDebug->Break();
			break;
			case enSetBreakPoint:
			{
				Trace("Got Set Breakpoint\n");

				int nScriptNameLen=0;
				CString strScriptName;
				int nLineNumber =0;
				BOOL bSetBreakPoint=false;

				//Read additional data
				if(!pComms->ReadFromPipe((LPBYTE)&nScriptNameLen,sizeof(int)) ||
					!pComms->ReadFromPipe((LPBYTE)strScriptName.GetBufferSetLength(nScriptNameLen),nScriptNameLen) ||
					!pComms->ReadFromPipe((LPBYTE)&nLineNumber,sizeof(int)) ||
					!pComms->ReadFromPipe((LPBYTE)&bSetBreakPoint,sizeof(BOOL)))
				{
					break;
				}
				strScriptName.ReleaseBuffer();
				pComms->m_pTestDebug->SetBreakPoint(strScriptName,nLineNumber,bSetBreakPoint);
			
				Trace("%s Breakpoint %s(%d)\n",(bSetBreakPoint?"Set":"UnSet"),strScriptName,nLineNumber);
			}
			break;
			case enGetState:
				Trace("Got Get State Command\n");
				pComms->m_pTestDebug->SendState();
			break;
			case enAbort:
				Trace("Got Abort Command\n");
				pComms->m_pTestDebug->Abort();
			break;
			default:
				Trace("Got unknown server command %d\n",nCommand);
			break;
			}
		}
		DisconnectNamedPipe(pComms->m_hPipe);
	}

	CloseHandle (pComms->m_hPipe);
	return 0;
}

BOOL CTestDebugCommsProxy::NotifyNextRequest(CNextRequestData* pData)
{
	Trace("Notify Next Request\n");

	DWORD dwBufLen;
	LPBYTE pBuffer = pData->SaveToBuffer(dwBufLen);

	//Write command
	int nCommand = enNextRequest;
	BOOL bResult = WriteToPipe((LPBYTE)&nCommand,sizeof(int));
	
	//Write Num bytes to follow
	bResult &= WriteToPipe((LPBYTE)&dwBufLen,sizeof(DWORD));

	//Write data
	bResult &= WriteToPipe(pBuffer,dwBufLen);
	delete []pBuffer;
	
	Trace("Notify Next Request Done\n");

	return bResult;
}

BOOL CTestDebugCommsProxy::NotifyRequestComplete(CRequestCompleteData* pData)
{
	Trace("Notify Request Complete\n");

	DWORD dwBufLen;
	LPBYTE pBuffer = pData->SaveToBuffer(dwBufLen);
	
	//Write command
	int nCommand = enRequestComplete;
	BOOL bResult = WriteToPipe((LPBYTE)&nCommand,sizeof(int));

	//Write num bytes to follow
	bResult &= WriteToPipe((LPBYTE)&dwBufLen,sizeof(DWORD));
	
	//Write Data
	bResult &= WriteToPipe(pBuffer,dwBufLen);
	delete []pBuffer;
	
	Trace("Notify Request Complete Done\n");

	return bResult;
}

BOOL CTestDebugCommsProxy::NotifyReadyToStep()
{
	Trace("Notify Ready To Step\n");

	int nCommand = enReadyToStep;
	BOOL bResult = WriteToPipe((LPBYTE)&nCommand,sizeof(int));

	Trace("Notify Ready To Step Done\n");
	
	return bResult;
}

BOOL CTestDebugCommsProxy::NotifyRunning()
{
	Trace("Notify Running\n");

	int nCommand = enRunning;
	BOOL bResult = WriteToPipe((LPBYTE)&nCommand,sizeof(int));

	Trace("Notify Running Done\n");
	
	return bResult;
}

BOOL CTestDebugCommsProxy::NotifyStepped(int nVUId,int nConId)
{
	Trace("Notify Stepped\n");

	int nBufSize = sizeof(int)*3;
	
	LPBYTE pBuf = new BYTE[nBufSize];
	LPBYTE pBufPos = pBuf;

	int nCommand = enRequestStepped;
	memcpy(pBufPos,&nCommand,sizeof(int));		pBufPos += sizeof(int);
	memcpy(pBufPos,&nVUId,sizeof(int));			pBufPos += sizeof(int);
	memcpy(pBufPos,&nConId,sizeof(int));		pBufPos += sizeof(int);

	BOOL bResult = WriteToPipe(pBuf,nBufSize);
	delete []pBuf;

	Trace("Notify Stepped Done\n");

	return bResult;
}
