/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

////////////////////////////////////////////////////////////////////////////////
// Fast lock implementation file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include <windows.h>
#include "Fastlock.hpp"

// -------
// Statics
// -------

BOOL CFastLock::s_bSingleProcessorMachine = (BOOL)-1;


// -------
// Methods
// -------

void CFastLock::Lock()
{
    BOOL bTookOwnership;
    DWORD dwSpinCount = m_dwSpinCount;
    DWORD dwThreadId = GetCurrentThreadId();

    // Spin and try to acquire the lock.
    do
    {
        bTookOwnership = (InterlockedCompareExchange(&m_lLockRequests, 1, 0) == 0);
        if(bTookOwnership)
        {
            m_dwOwningThreadId = dwThreadId;
            m_dwLockDepth      = 1;
        }
        else if(m_dwOwningThreadId == dwThreadId)
        {
            // The current thread owns the lock; increment the lock depth
            // and lock request count.
            ++m_dwLockDepth;
            InterlockedIncrement(&m_lLockRequests);
            bTookOwnership = TRUE;  // Mark ownership.
        }
    } while(!bTookOwnership && dwSpinCount-- != 0);

    if(!bTookOwnership)
    {
        // Failed to acquire the lock while spinning.
        // Issue a normal lock request.
        if(InterlockedIncrement(&m_lLockRequests) == 1)
        {
            // Initial ownership.
            m_dwOwningThreadId = dwThreadId;
            m_dwLockDepth      = 1;
        }
        else
        {
            if(m_dwOwningThreadId == dwThreadId)
            {
                // The current thread owns the lock; increment the lock depth.
                ++m_dwLockDepth;
            }
            else
            {
                // Sorry, the lock is owned by another thread. Wait.
                if(!m_hEvent)
                {
                    // Just-in-time event creation.
                    // Create the event kernel object as auto-reset.
                    HANDLE hNewEvent = CreateEventW(NULL, FALSE, FALSE, NULL);
                    if(InterlockedCompareExchange((LPLONG)&m_hEvent, (LONG)hNewEvent, 0) != 0)
                    {
                        // Another thread has already created the event: delete it!
                        CloseHandle(hNewEvent);
                    }
                }
                WaitForSingleObject(m_hEvent, INFINITE);

                // OK, may own the lock now.
                m_dwOwningThreadId = dwThreadId;
                m_dwLockDepth      = 1;
            }
        }
    }
}

void CFastLock::Unlock()
{
    // Decrement the lock depth.
    --m_dwLockDepth;

    if(m_dwLockDepth)
        InterlockedDecrement(&m_lLockRequests);
    else
    {
        // The current thread loses ownership of the lock.
        m_dwOwningThreadId = 0;

        if(InterlockedDecrement(&m_lLockRequests))
        {
            // Some threads are waiting for the lock.
            // Set the event, which releases one of the waiting threads.
            // The event is then automatically reset.
            if(!m_hEvent)
            {
                // Just-in-time event creation.
                // Create the event kernel object as auto-reset.
                HANDLE hNewEvent = CreateEventW(NULL, FALSE, FALSE, NULL);
                if(InterlockedCompareExchange((LPLONG)&m_hEvent, (LONG)hNewEvent, 0) != 0)
                {
                    // Another thread has already created the event: delete it!
                    CloseHandle(hNewEvent);
                }
            }
            SetEvent(m_hEvent);
        }
    }
}
