/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

////////////////////////////////////////////////////////////////////////////////
// OpenSTA
//
// Performance counters implementation file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <malloc.h>
#ifdef _DEBUG
#   include <iostream>
#endif
#include "PerfCounters.hpp"

namespace OpenSTA_Performance {

CPerfObject::~CPerfObject()
{
    // Delete the counters.
    for(Counters::iterator pCnt = m_counters.begin(); pCnt != m_counters.end(); ++pCnt)
        delete *pCnt;

    // Delete the instances.
    for(Instances::const_iterator pInst = m_id.begin(); pInst != m_id.end(); ++pInst)
    {
        if(*pInst)
            delete *pInst;
    }
}

DWORD CPerfObject::GetStructSize(DWORD dwNumInstances) const
{
    // The PERF_OBJECT_TYPE structure.
    DWORD dwSize = sizeof(PERF_OBJECT_TYPE);

    // The number of counters.
    DWORD dwNumCounters = m_counters.size();

    // Add the size of the PERF_COUNTER_DEFINITION structures.
    dwSize += sizeof(PERF_COUNTER_DEFINITION) * dwNumCounters;

    // Add the size of the instances and counter blocks.
    if(m_bSupportsInstances)
        dwSize += dwNumInstances *
            CPerfInstance::GetSizeWithCounterBlock(dwNumCounters, m_dwMaxInstanceNameSize);
    else
        dwSize += CPerfCounterBlock::GetSize(dwNumCounters);

    // Add padding to reach an 8-byte boundary.
    if((dwSize & 0x00000007) != 0)  // If at least one of the 3 low-order bits is set.
    {
        dwSize &= 0xFFFFFFF8;       // Clear 3 low-order bits.
        dwSize += 8;
    }

    return dwSize;
}

PERF_OBJECT_TYPE* CPerfObject::InitializeStruct(PERF_OBJECT_TYPE* pPot)
{
    DWORD dwNumCounters = m_counters.size();
    DWORD dwDefinitionLen = sizeof(PERF_OBJECT_TYPE) + sizeof(PERF_COUNTER_DEFINITION) * dwNumCounters;

    ZeroMemory(pPot, sizeof(PERF_OBJECT_TYPE));

    pPot->TotalByteLength      = GetMaxStructSize(); // The value will be different when collecting
                                                     // counters for a client app.
    pPot->DefinitionLength     = dwDefinitionLen;
    pPot->HeaderLength         = sizeof(PERF_OBJECT_TYPE);
    pPot->ObjectNameTitleIndex = m_dwNameTitleIndex;
    pPot->ObjectHelpTitleIndex = m_dwHelpTitleIndex;
    pPot->DetailLevel          = m_dwDetailLevel;
    pPot->NumCounters          = dwNumCounters;
    pPot->DefaultCounter       = m_pDefaultCounter->m_dwIndex;
    pPot->NumInstances         = m_bSupportsInstances ? 0 : PERF_NO_INSTANCES;
    pPot->CodePage             = 0;

    // Initialize counter definitions.
    PERF_COUNTER_DEFINITION* pPcd = (PERF_COUNTER_DEFINITION*)((PBYTE)pPot + sizeof(PERF_OBJECT_TYPE));
    for(Counters::iterator pCnt = m_counters.begin(); pCnt != m_counters.end(); ++pCnt)
        (*pCnt)->InitializeStruct(pPcd++);

    if(!m_bSupportsInstances)
    {
        // Initialize the counter block.
        m_cb.m_dwNumCounters = dwNumCounters;
        m_cb.InitializeStruct((PERF_COUNTER_BLOCK*)((PBYTE)pPot + dwDefinitionLen));
    }

    m_pPot = pPot;

    return (PERF_OBJECT_TYPE*)((PBYTE)pPot + GetMaxStructSize());
}

CPerfCore::CPerfCore(PCWSTR zSourceName, PCWSTR zOpenFuncName, PCWSTR zCollectFuncName,PCWSTR zCloseFuncName)
    : m_pbView(NULL), m_dwFirstCounter(0), m_dwFirstHelp(0), m_dwLastCounter(0),
      m_dwLastHelp(0), m_zSourceName(zSourceName), m_zOpenFuncName(zOpenFuncName),
      m_zCollectFuncName(zCollectFuncName), m_zCloseFuncName(zCloseFuncName),
      m_hMutex(NULL), m_hMap(NULL)
{
    // Create security attributes granting GENERIC_ALL access to everyone.
    SECURITY_DESCRIPTOR sd;
    InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION);
    PSID psidEveryone;
    SID_IDENTIFIER_AUTHORITY siauth = SECURITY_WORLD_SID_AUTHORITY;
    AllocateAndInitializeSid(&siauth, 1, SECURITY_WORLD_RID, 0, 0, 0, 0, 0, 0, 0, &psidEveryone);
    DWORD dwAclSize = sizeof(ACL) + offsetof(ACCESS_ALLOWED_ACE, SidStart) + GetLengthSid(psidEveryone);
    PACL pAcl = (PACL) _alloca(dwAclSize);
    InitializeAcl(pAcl, dwAclSize, ACL_REVISION);
    SetSecurityDescriptorDacl(&sd, TRUE, pAcl, FALSE);
    AddAccessAllowedAce(pAcl, ACL_REVISION, GENERIC_ALL, psidEveryone);
    SECURITY_ATTRIBUTES sa = { sizeof(sa), &sd, FALSE };

    // Create the mutex.
    WCHAR zMutexName[128];
    lstrcpyW(zMutexName, zSourceName);
    lstrcatW(zMutexName, L"_MTX");
    m_hMutex = CreateMutexW(&sa, FALSE, zMutexName);
    FreeSid(psidEveryone);
}

CPerfCore::~CPerfCore()
{
    // Delete the objects.
    for(PerfObjects::iterator it = m_po.begin(); it != m_po.end(); ++it)
        delete *it;

    // Close the shared memory.
    if(m_hMap)
        CloseHandle(m_hMap);
    if(m_pbView)
        UnmapViewOfFile(m_pbView);

    // Close the mutex object.
    CloseHandle(m_hMutex);
}

// Parsing the object / counter static map.
DWORD CPerfCore::ParseMap(const CPerfMapType* pMap, DWORD dwMapSize)
{
    DWORD dwRet = ERROR_SUCCESS;

    DWORD dwCounter = m_dwFirstCounter;
    DWORD dwHelp    = m_dwFirstHelp;
    CPerfObject* pObject = NULL;            // Current object.
    const CPerfMapType* pMapObject = NULL;  // Current object in the map.

    for(DWORD dw = 0; dw != dwMapSize; ++dw, ++pMap)
    {
        const CPerfMapType& entry = *pMap;

        if(entry.m_bIsCounter)
        {
            if(pObject == NULL)
                return ERROR_INVALID_DATA;
            CPerfCounter* pNewCounter = new CPerfCounter();
            if(pNewCounter)
            {
                // Common.
                pNewCounter->m_dwKey                = entry.m_dwKey;
                pNewCounter->m_zName                = entry.m_zName;
                pNewCounter->m_zHelp                = entry.m_zHelp;
                pNewCounter->m_dwDetailLevel        = entry.m_dwDetailLevel;
                pNewCounter->m_dwNameTitleIndex     = dwCounter;
                pNewCounter->m_dwHelpTitleIndex     = dwHelp;

                // Counter.
                pNewCounter->m_dwType               = entry.m_dwType;
                pNewCounter->m_lDefaultScale        = entry.m_lDefaultScale;
                pNewCounter->m_dwIndex              = pObject->m_counters.size();
                pNewCounter->m_pPcd                 = NULL;

                pObject->m_counters.push_back(pNewCounter);
                if(pNewCounter->m_dwKey == pMapObject->m_dwDefaultCounterKey)
                    pObject->m_pDefaultCounter = pNewCounter;
            }
            else
                dwRet = ERROR_OUTOFMEMORY;
        }
        else
        {
            CPerfObject* pNewObject = new CPerfObject();
            if(pNewObject)
            {
                // Common.
                pNewObject->m_dwKey                 = entry.m_dwKey;
                pNewObject->m_zName                 = entry.m_zName;
                pNewObject->m_zHelp                 = entry.m_zHelp;
                pNewObject->m_dwDetailLevel         = entry.m_dwDetailLevel;
                pNewObject->m_dwNameTitleIndex      = dwCounter;
                pNewObject->m_dwHelpTitleIndex      = dwHelp;

                // Object.
                pNewObject->m_bSupportsInstances    = entry.m_bSupportsInstances;
                pNewObject->m_dwMaxInstances        = entry.m_dwMaxInstances;
                pNewObject->m_dwMaxInstanceNameSize = entry.m_dwMaxInstanceNameSize;
                pNewObject->m_pDefaultCounter       = NULL;     // Until the default counter is found.
                pNewObject->m_pPot                  = NULL;

                m_po.push_back(pNewObject);
                pMapObject = &entry;
                pObject = pNewObject;
            }
            else
                dwRet = ERROR_OUTOFMEMORY;
        }

        dwCounter += 2;
        dwHelp    += 2;
    }

    return dwRet;
}

DWORD CPerfCore::AdjustTitleIndexes()
{
    WCHAR zSvcSubkey[512];
    wsprintfW(zSvcSubkey, L"SYSTEM\\CurrentControlSet\\Services\\%s\\Performance", m_zSourceName);
    CRegistryKey regSvc;
    DWORD dwRet = regSvc.Open(zSvcSubkey);
    if(dwRet == ERROR_SUCCESS)
    {
        regSvc.GetValue(L"First Counter",    m_dwFirstCounter);
        regSvc.GetValue(L"Last Counter",     m_dwLastCounter);
        regSvc.GetValue(L"First Help",       m_dwFirstHelp);
        regSvc.GetValue(L"Last Help",        m_dwLastHelp);
        regSvc.Close();

        DWORD dwIndex     = m_dwFirstCounter;
        DWORD dwHelpIndex = m_dwFirstHelp;

        for(PerfObjects::const_iterator it = m_po.begin(); it != m_po.end(); ++it)
        {
            // Additional bytes for an object.
            CPerfObject* pObj = *it;
            pObj->m_dwNameTitleIndex = dwIndex;
            pObj->m_dwHelpTitleIndex = dwHelpIndex;
            dwIndex     += 2;
            dwHelpIndex += 2;

            // Additional bytes for a counter.
            for(CPerfObject::Counters::const_iterator cnt = pObj->m_counters.begin(); cnt != pObj->m_counters.end(); ++cnt)
            {
                CPerfCounter* pCnt = *cnt;
                pCnt->m_dwNameTitleIndex = dwIndex;
                pCnt->m_dwHelpTitleIndex = dwHelpIndex;
                dwIndex     += 2;
                dwHelpIndex += 2;
            }
        }
    }

    return dwRet;
}

DWORD CPerfCore::CalculateMaxSize() const
{
    DWORD dwSize = 0;
    for(PerfObjects::const_iterator pObj = m_po.begin(); pObj != m_po.end(); ++pObj)
        dwSize += (*pObj)->GetMaxStructSize();
    return dwSize;
}

DWORD CPerfCore::BuildSharedMemory(DWORD dwSize, BOOL bIsPerfSource)
{
    // Create security attributes granting GENERIC_ALL access to everyone.
    SECURITY_DESCRIPTOR sd;
    InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION);
    PSID psidEveryone;
    SID_IDENTIFIER_AUTHORITY siauth = SECURITY_WORLD_SID_AUTHORITY;
    AllocateAndInitializeSid(&siauth, 1, SECURITY_WORLD_RID, 0, 0, 0, 0, 0, 0, 0, &psidEveryone);
    DWORD dwAclSize = sizeof(ACL) + offsetof(ACCESS_ALLOWED_ACE, SidStart) + GetLengthSid(psidEveryone);
    PACL pAcl = (PACL) _alloca(dwAclSize);
    InitializeAcl(pAcl, dwAclSize, ACL_REVISION);
    SetSecurityDescriptorDacl(&sd, TRUE, pAcl, FALSE);
    AddAccessAllowedAce(pAcl, ACL_REVISION, GENERIC_ALL, psidEveryone);
    SECURITY_ATTRIBUTES sa = { sizeof(sa), &sd, FALSE };

    // Create a file mapping kernel object.
    WCHAR zMappingName[128];
    lstrcpyW(zMappingName, m_zSourceName);
    lstrcatW(zMappingName, L"_MPG");
    m_hMap = CreateFileMappingW(INVALID_HANDLE_VALUE, &sa, PAGE_READWRITE, 0, dwSize, zMappingName);
    DWORD dwRet = GetLastError();
    FreeSid(psidEveryone);
    if(m_hMap)
    {
        // dwRet can be ERROR_ALREADY_EXISTS if the file mapping object already exists.
        m_pbView = (PBYTE)MapViewOfFile(m_hMap, FILE_MAP_READ|FILE_MAP_WRITE, 0, 0, 0);
        if(m_pbView == NULL)
            dwRet = GetLastError();

        if(dwRet == ERROR_SUCCESS || dwRet == ERROR_ALREADY_EXISTS && bIsPerfSource)
        {
            CCrossProcessLock lock(m_hMutex);       // Lock the access.

            // Initialize the shared memory.
            PERF_OBJECT_TYPE* pPot = (PERF_OBJECT_TYPE*)m_pbView;
            for(PerfObjects::iterator pObj = m_po.begin(); pObj != m_po.end(); ++pObj)
                pPot = (*pObj)->InitializeStruct(pPot);
        }
    }

    return dwRet;
}

DWORD CPerfSource::AddCounterHelpInfo(CRegistryKey& reg, DWORD& dwTitleIndex, BOOL bIsHelp)
{
    DWORD dwRet = ERROR_SUCCESS;

    // Determine the number of additional bytes to add to the multistring.
    DWORD dwIndex = dwTitleIndex;
    DWORD dwExtraBytes = 0;
    for(PerfObjects::const_iterator it = m_po.begin(); it != m_po.end(); ++it)
    {
        // Additional bytes for an object.
        CPerfObject* pObj = *it;
        dwExtraBytes += (GetDigitCount(dwIndex) + 1) * sizeof(WCHAR);   // Title index size.
        dwExtraBytes += (lstrlenW(bIsHelp ? pObj->m_zHelp : pObj->m_zName) + 1)  * sizeof(WCHAR);   // Title size.
        dwIndex += 2;

        // Additional bytes for a counter.
        for(CPerfObject::Counters::const_iterator cnt = pObj->m_counters.begin(); cnt != pObj->m_counters.end(); ++cnt)
        {
            CPerfCounter* pCnt = *cnt;
            dwExtraBytes += (GetDigitCount(dwIndex) + 1) * sizeof(WCHAR);   // Title index size.
            dwExtraBytes += (lstrlenW(bIsHelp ? pCnt->m_zHelp : pCnt->m_zName) + 1)  * sizeof(WCHAR);   // Title size.
            dwIndex += 2;
        }
    }
    dwIndex = dwTitleIndex;

    WCHAR zLang[8];
    wsprintfW(zLang, L"%03x", GetSystemDefaultLangID() & 0x00FF);
    CRegistryKey regStrings;
    dwRet = regStrings.Open(zLang, reg, FALSE);
    if(dwRet == ERROR_SUCCESS)
    {
        // Get the number of bytes of the current multistring.
        DWORD dwCurSize = 0;
        PWSTR zValue = bIsHelp ? L"Help" : L"Counter";
        regStrings.GetValueSize(zValue, dwCurSize);

        // Get the multistring and append our strings.
        PWSTR zNewString = (PWSTR)_alloca(dwCurSize + dwExtraBytes);
        regStrings.GetValue(zValue, zNewString, dwCurSize + dwExtraBytes);
        PWSTR zNewData = (PWSTR)((PBYTE)zNewString + dwCurSize) - 1;
        if(dwCurSize == 0)  // The subkey was empty.
            --zNewData;
        for(PerfObjects::const_iterator it = m_po.begin(); it != m_po.end(); ++it)
        {
            // Write object title index and title.
            CPerfObject* pObj = *it;
            zNewData += wsprintfW(zNewData, L"%u", dwTitleIndex) + 1;
            lstrcpyW(zNewData, bIsHelp ? pObj->m_zHelp : pObj->m_zName);
            zNewData += lstrlenW(zNewData) + 1;
            dwTitleIndex += 2;

            // Write counter title index and title for the object's counters.
            for(CPerfObject::Counters::const_iterator cnt = pObj->m_counters.begin(); cnt != pObj->m_counters.end(); ++cnt)
            {
                CPerfCounter* pCnt = *cnt;
                zNewData += wsprintfW(zNewData, L"%u", dwTitleIndex) + 1;
                lstrcpyW(zNewData, bIsHelp ? pCnt->m_zHelp : pCnt->m_zName);
                zNewData += lstrlenW(zNewData) + 1;
                dwTitleIndex += 2;
            }
        }
        if(dwTitleIndex > dwIndex)
            dwTitleIndex -= 2;
        *zNewData = 0;  // Trailing zero character.
        regStrings.SetValue(zValue, zNewString, TRUE);
    }

    return dwRet;
}

DWORD CPerfSource::DeleteCounterHelpInfo(CRegistryKey& reg, DWORD dwFirstTitleIndex,
    DWORD& dwLastTitleIndex, BOOL bIsHelp)
{
    DWORD dwRet = ERROR_SUCCESS;

    if(dwFirstTitleIndex > dwLastTitleIndex)
        return dwRet;

    WCHAR zLang[8];
    wsprintfW(zLang, L"%03x", GetSystemDefaultLangID() & 0x00FF);
    CRegistryKey regStrings;
    dwRet = regStrings.Open(zLang, reg, FALSE);
    if(dwRet == ERROR_SUCCESS)
    {
        // Get the number of bytes of the current multistring.
        DWORD dwCurSize = 0;
        PWSTR zValue = bIsHelp ? L"Help" : L"Counter";
        regStrings.GetValueSize(zValue, dwCurSize);

        // Read the current multistring.
        PWSTR zString = (PWSTR)_alloca(dwCurSize);
        regStrings.GetValue(zValue, zString, dwCurSize);

        // Find the beginning of the first counter/help.
        WCHAR zNum[8];
        wsprintfW(zNum, L"%u", dwFirstTitleIndex);
        PWSTR zFirst;
        for(zFirst = zString;  *zFirst != 0 && lstrcmpW(zNum, zFirst); zFirst += lstrlenW(zFirst) + 1);
        if(*zFirst)
        {
            // Find the beginning of the last counter/help.
            wsprintfW(zNum, L"%u", dwLastTitleIndex);
            PWSTR zLast;
            for(zLast = zFirst;  *zLast != 0 && lstrcmpW(zNum, zLast); zLast += lstrlenW(zLast) + 1);
            if(*zLast)
            {
                // Skip the last title index and title.
                zLast += lstrlenW(zLast) + 1;
                zLast += lstrlenW(zLast) + 1;

                // Delete the strings.
                SIZE_T size = ((PBYTE)zString + dwCurSize) - (PBYTE)zLast;
                MoveMemory(zFirst, zLast, size);
                regStrings.SetValue(zValue, zString, TRUE);

                // Retrieve the new last title index.
                if(*zFirst == 0)
                {
                    // The set of strings we removed was the last.
                    // Search the last title index from the beginning.
                    zFirst = zString;
                }

                // Move on to the last title index.
                for(zLast = zFirst;  *zLast != 0; zLast += lstrlenW(zLast) + 1)
                {
                    // Is the string a title index?.
                    PWSTR p = zLast;
                    while(*p)
                    {
                        if(!iswdigit(*p))
                            break;
                        ++p;
                    }

                    if(*p == 0)
                    {
                        // It's a title index: read it.
                        dwLastTitleIndex = _wtol(zLast);
                    }
                }
            }
            else
                dwRet = ERROR_NOT_FOUND;
        }
        else
            dwRet = ERROR_NOT_FOUND;
    }

    return dwRet;
}

DWORD CPerfSource::GetDigitCount(DWORD dw)
{
    WCHAR z[32];
    z[0] = 0;
    wsprintfW(z, L"%u", dw);
    return lstrlenW(z);
}

DWORD CPerfSource::Install(PCWSTR zPerfDllPath)
{
    DWORD dwRet = ERROR_SUCCESS;

    if(zPerfDllPath == NULL)
        return ERROR_BAD_PATHNAME;

    // Check the existence of the DLL.
    HANDLE hDLL = CreateFileW(zPerfDllPath, GENERIC_READ, FILE_SHARE_READ, NULL,
        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if(hDLL == INVALID_HANDLE_VALUE)
        return GetLastError();
    else
        CloseHandle(hDLL);

    // Remove any previous installation.
    dwRet = Uninstall();

    // Install the source.
    if(dwRet == ERROR_SUCCESS)
    {
        WCHAR zSvcSubkey[512];
        wsprintfW(zSvcSubkey, L"SYSTEM\\CurrentControlSet\\Services\\%s\\Performance", m_zSourceName);
        CRegistryKey regSvc;
        dwRet = regSvc.Open(zSvcSubkey, HKEY_LOCAL_MACHINE, FALSE);
        if(dwRet == ERROR_SUCCESS)
        {
            CRegistryKey regPerflib;
            dwRet = regPerflib.Open(L"SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Perflib", HKEY_LOCAL_MACHINE, FALSE);
            if(dwRet == ERROR_SUCCESS)
            {
                DWORD dwLastCounter, dwLastHelp;
                regPerflib.GetValue(L"Last Counter", dwLastCounter);
                regPerflib.GetValue(L"Last Help", dwLastHelp);

                // Ensure that dwLastCounter is even and that dwLastHelp = dwLastCounter + 1.
                if((dwLastCounter & 1) != 0 || dwLastHelp != dwLastCounter + 1)
                {
                    if(dwLastCounter < dwLastHelp)
                        dwLastCounter = dwLastHelp;
                    if((dwLastCounter & 1) != 0)
                        ++dwLastCounter;
                    dwLastHelp = dwLastCounter + 1;
                }

                m_dwFirstCounter = m_dwLastCounter = dwLastCounter + 2;
                m_dwFirstHelp = m_dwLastHelp = dwLastHelp + 2;
                AddCounterHelpInfo(regPerflib, m_dwLastCounter, FALSE);
                AddCounterHelpInfo(regPerflib, m_dwLastHelp, TRUE);
                regPerflib.SetValue(L"Last Counter", m_dwLastCounter);
                regPerflib.SetValue(L"Last Help",    m_dwLastHelp);
                regSvc.SetValue(L"First Counter",    m_dwFirstCounter);
                regSvc.SetValue(L"Last Counter",     m_dwLastCounter);
                regSvc.SetValue(L"First Help",       m_dwFirstHelp);
                regSvc.SetValue(L"Last Help",        m_dwLastHelp);
                regSvc.SetValue(L"Open",             m_zOpenFuncName);
                regSvc.SetValue(L"Collect",          m_zCollectFuncName);
                regSvc.SetValue(L"Close",            m_zCloseFuncName);
                regSvc.SetValue(L"Library",          zPerfDllPath);
            }
        }
    }

    return dwRet;
}

DWORD CPerfSource::Uninstall()
{
    DWORD dwRet = ERROR_SUCCESS;

    WCHAR zSvcSubkey[512];
    wsprintfW(zSvcSubkey, L"SYSTEM\\CurrentControlSet\\Services\\%s\\Performance", m_zSourceName);
    CRegistryKey regSvc;
    if(regSvc.Open(zSvcSubkey) == ERROR_SUCCESS)
    {
        CRegistryKey regPerflib;
        dwRet = regPerflib.Open(L"SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Perflib", HKEY_LOCAL_MACHINE, FALSE);
        if(dwRet == ERROR_SUCCESS)
        {
            DWORD dwLastCounter, dwLastHelp;
            regPerflib.GetValue(L"Last Counter", dwLastCounter);
            regPerflib.GetValue(L"Last Help", dwLastHelp);

            regSvc.GetValue(L"First Counter",    m_dwFirstCounter);
            regSvc.GetValue(L"Last Counter",     m_dwLastCounter);
            regSvc.GetValue(L"First Help",       m_dwFirstHelp);
            regSvc.GetValue(L"Last Help",        m_dwLastHelp);
            if(m_dwLastCounter <= dwLastCounter && m_dwLastHelp <= dwLastHelp)
            {
                DWORD dwNewLastCounter = m_dwLastCounter;
                DeleteCounterHelpInfo(regPerflib, m_dwFirstCounter, dwNewLastCounter, FALSE);
                DWORD dwNewLastHelp = m_dwLastHelp;
                DeleteCounterHelpInfo(regPerflib, m_dwFirstHelp, dwNewLastHelp, TRUE);

                if(dwNewLastCounter != dwLastCounter)
                    regPerflib.SetValue(L"Last Counter", dwNewLastCounter);
                if(dwNewLastHelp != dwLastHelp)
                    regPerflib.SetValue(L"Last Help", dwNewLastHelp);

                regSvc.Close();
                RegDeleteKeyW(HKEY_LOCAL_MACHINE, zSvcSubkey);
                wsprintfW(zSvcSubkey, L"SYSTEM\\CurrentControlSet\\Services\\%s", m_zSourceName);
                RegDeleteKeyW(HKEY_LOCAL_MACHINE, zSvcSubkey);
            }
            else
                dwRet = ERROR_INVALID_PARAMETER;
        }
    }

    return dwRet;
}

void CPerfInstance::InitializeStruct(PERF_INSTANCE_DEFINITION* pPid)
{
    DWORD dwInstanceNameSizeInBytes = m_zInstanceName ? (lstrlenW(m_zInstanceName) + 1) * sizeof(WCHAR) : 0;
    DWORD dwSize = InternalGetSize(dwInstanceNameSizeInBytes);
    ZeroMemory(pPid, dwSize);

    pPid->ByteLength             = dwSize;
    pPid->ParentObjectTitleIndex = m_dwParentObjectTitleIndex;
    pPid->ParentObjectInstance   = m_pParentInstance == NULL ? 0 : m_pParentInstance->m_dwInstanceIndex;
    pPid->UniqueID               = m_zInstanceName ? PERF_NO_UNIQUE_ID : m_lUniqueId;
    pPid->NameOffset             = m_zInstanceName ? sizeof(PERF_INSTANCE_DEFINITION) : 0;
    pPid->NameLength             = dwInstanceNameSizeInBytes;

    if(m_zInstanceName)
    {
        // Copy the instance name.
        PWSTR zName = (PWSTR)((PBYTE)pPid + pPid->NameOffset);
        lstrcpyW(zName, m_zInstanceName);
    }

    // Initialize the counter block.
    m_cb.InitializeStruct((PERF_COUNTER_BLOCK*)((PBYTE)pPid + dwSize));

    m_pPid = pPid;
}

CPerfInstance* CPerfObject::AddInstance(LONG lUniqueID,
    DWORD dwParentObjectTitleIndex, CPerfInstance* pParentInstance)
{
    // Append the new instance to the end of the vector.
    CPerfInstance* pInstance = NULL;
    if(m_id.size() < (Instances::size_type)m_dwMaxInstances)
    {
        pInstance = new CPerfInstance(m_id.size(), lUniqueID,
            m_dwMaxInstanceNameSize, m_counters.size(), dwParentObjectTitleIndex, pParentInstance);
        if(pInstance)
        {
            PERF_INSTANCE_DEFINITION* pid;
            if(m_id.size() == 0)
                pid = (PERF_INSTANCE_DEFINITION*)(((PBYTE)m_pPot) + m_pPot->DefinitionLength);
            else
                pid = (PERF_INSTANCE_DEFINITION*)(((PBYTE)m_id.back()->m_pPid) + m_id.back()->GetSizeWithCounterBlock());
            m_id.push_back(pInstance);
            pInstance->InitializeStruct(pid);
            if(pParentInstance)
                pParentInstance->m_sons.push_back(pInstance);
            ++m_pPot->NumInstances;
        }
    }
    return pInstance;
}

CPerfInstance* CPerfObject::AddInstance(PCWSTR zInstanceName,
    DWORD dwParentObjectTitleIndex, CPerfInstance* pParentInstance)
{
    // Append the new instance to the end of the vector.
    CPerfInstance* pInstance = NULL;
    if(m_id.size() < (Instances::size_type)m_dwMaxInstances)
    {
        pInstance = new CPerfInstance(m_id.size(), zInstanceName,
            m_dwMaxInstanceNameSize, m_counters.size(), dwParentObjectTitleIndex, pParentInstance);
        if(pInstance)
        {
            PERF_INSTANCE_DEFINITION* pid;
            if(m_id.size() == 0)
                pid = (PERF_INSTANCE_DEFINITION*)((PBYTE)m_pPot + m_pPot->DefinitionLength);
            else
                pid = (PERF_INSTANCE_DEFINITION*)(((PBYTE)m_id.back()->m_pPid) + m_id.back()->GetSizeWithCounterBlock());
            m_id.push_back(pInstance);
            pInstance->InitializeStruct(pid);
            if(pParentInstance)
                pParentInstance->m_sons.push_back(pInstance);
            ++m_pPot->NumInstances;
        }
    }
    return pInstance;
}

void CPerfObject::DeleteInstance(CPerfInstance* pInstance)
{
    // Remove the reference in the parent.
    if(pInstance->m_pParentInstance)
    {
        pInstance->m_pParentInstance->m_sons.erase(
            std::find(pInstance->m_pParentInstance->m_sons.begin(),
                pInstance->m_pParentInstance->m_sons.end(), pInstance)
        );
    }

    // If the instance is the last used instance in index order, then simply delete it.
    --m_pPot->NumInstances;
    if(pInstance->m_dwInstanceIndex == m_id.size() - 1)
    {
        delete pInstance;
        m_id.pop_back();
    }
    else
    {
        // Delete the instance, move the last instance to fill the hole, and fix the references.
        m_id[pInstance->m_dwInstanceIndex] = m_id.back();
        m_id[pInstance->m_dwInstanceIndex]->m_dwInstanceIndex = pInstance->m_dwInstanceIndex;
        CPerfInstance* pMovedInstance = m_id[pInstance->m_dwInstanceIndex];
        for(CPerfInstance::InstanceSons::iterator it = pMovedInstance->m_sons.begin();
            it != pMovedInstance->m_sons.end(); ++it)
        {
            (*it)->m_pPid->ParentObjectInstance = pMovedInstance->m_dwInstanceIndex;
        }
        delete pInstance;
    }
}

PLONG CPerfObject::GetCounterValue(DWORD dwCounterKey, CPerfInstance* pInstance)
{
    PLONG plCounterVal = NULL;

    // Locate the counter definition.
    for(Counters::const_iterator pCtDef = m_counters.begin(); pCtDef != m_counters.end(); ++pCtDef)
    {
        if((*pCtDef)->m_dwKey == dwCounterKey)
        {
            CPerfCounterBlock* pCb = NULL;
            if(pInstance)
            {
                if(m_bSupportsInstances)
                    pCb = &pInstance->m_cb;
            }
            else
            {
                if(!m_bSupportsInstances)
                    pCb = &m_cb;
            }
            if(pCb)
                plCounterVal = (PLONG)(*pCb)[CPerfCounterBlock::IndexToOffset((*pCtDef)->m_dwIndex)];
            break;
        }
    }

    return plCounterVal;
}

DWORD CPerfSource::Initialize()
{
    return BuildSharedMemory(CalculateMaxSize(), TRUE);
}

PLONG CPerfSource::GetCounterValue(DWORD dwCounterKey, CPerfInstance* pInstance)
{
    PLONG plCounterValue = NULL;
    for(PerfObjects::iterator pObj = m_po.begin(); pObj != m_po.end(); ++pObj)
    {
        plCounterValue = (*pObj)->GetCounterValue(dwCounterKey, pInstance);
        if(plCounterValue)
            break;
    }

    return plCounterValue;
}

CPerfInstance* CPerfSource::AddInstance(DWORD dwObjectKey, LONG lUniqueID, DWORD dwParentObjectTitleIndex,
    CPerfInstance* pParentInstance)
{
    CPerfInstance* pNewInstance = NULL;
    CPerfObject* pObj = GetObject(dwObjectKey);
    if(pObj)
    {
        CCrossProcessLock lock(m_hMutex);       // Lock the access.
        pNewInstance = pObj->AddInstance(lUniqueID, dwParentObjectTitleIndex, pParentInstance);
    }

    return pNewInstance;
}

CPerfInstance* CPerfSource::AddInstance(DWORD dwObjectKey, PCWSTR zInstanceName, DWORD dwParentObjectTitleIndex,
    CPerfInstance* pParentInstance)
{
    CPerfInstance* pNewInstance = NULL;
    CPerfObject* pObj = GetObject(dwObjectKey);
    if(pObj)
    {
        CCrossProcessLock lock(m_hMutex);       // Lock the access.
        pNewInstance = pObj->AddInstance(zInstanceName, dwParentObjectTitleIndex, pParentInstance);
    }

    return pNewInstance;
}

void CPerfSource::DeleteInstance(DWORD dwObjectKey, CPerfInstance* pInstance)
{
    CPerfObject* pObj = GetObject(dwObjectKey);
    if(pObj)
    {
        CCrossProcessLock lock(m_hMutex);       // Lock the access.
        pObj->DeleteInstance(pInstance);
    }
}

DWORD CPerfClient::Initialize()
{
    DWORD dwRet = BuildSharedMemory(CalculateMaxSize(), FALSE);

    return dwRet;
}

DWORD CPerfClient::Collect(PCWSTR zObjectNumbers, PVOID* ppData, PDWORD pdwBytes, PDWORD pdwObjectTypes)
{
    DWORD dwRet = ERROR_SUCCESS;
    PVOID pData = *ppData;

    // Protect against exceptions.
    __try
    {
        dwRet = DoCollect(zObjectNumbers, ppData, pdwBytes, pdwObjectTypes);
    }
    __except(EXCEPTION_EXECUTE_HANDLER)
    {
        *ppData = pData;
        *pdwBytes = 0;
        *pdwObjectTypes = 0;
    }

    return dwRet;
}

// Caution: this method must be thread safe!
DWORD CPerfClient::DoCollect(PCWSTR zObjectNumbers, PVOID* ppData, PDWORD pdwBytes, PDWORD pdwObjectTypes)
{
    DWORD dwRet = ERROR_SUCCESS;

    CCrossProcessLock lock(m_hMutex);       // Lock the access.

    // Collect the objects.
    if(m_pbView)
    {
        BOOL bCollectAll = lstrcmpiW(zObjectNumbers, L"Global") == 0;

        // Array mapping object title indexes to BOOLs indicating
        // whether the object should be collected.
        // The last entry is set to zero, to mark the end.
        ObjToCollect* pOc = (ObjToCollect*)_alloca(sizeof(ObjToCollect) * (m_po.size() + 1));
        ObjToCollect* pOcTmp = pOc;
        PerfObjects::const_iterator pObjDef;
        for(pObjDef = m_po.begin(); pObjDef != m_po.end(); ++pObjDef)
        {
            pOcTmp->m_dwObjTitleIndex = (*pObjDef)->m_dwNameTitleIndex;
            pOcTmp->m_bToCollect      = bCollectAll;
            ++pOcTmp;
        }
        pOcTmp->m_dwObjTitleIndex = 0; // Array's end.

        if(!bCollectAll)
        {
            // Determine which objects to collect.
            PWSTR zON = (PWSTR)_alloca((lstrlenW(zObjectNumbers) + 1) * sizeof(WCHAR));
            lstrcpyW(zON, zObjectNumbers);
            for(WCHAR* pTok = wcstok(zON, L" "); pTok != NULL; pTok = wcstok(NULL, L" "))
            {
                long lObj = _wtol(pTok);
                if(lObj > 0)
                    MarkToCollect(pOc, lObj);
            }
        }

        // Compute the size of the buffer passed on input.
        DWORD dwTotalSize = 0;
        PERF_OBJECT_TYPE* pObj = (PERF_OBJECT_TYPE*)m_pbView;
        pOcTmp = pOc;
        for(pObjDef = m_po.begin(); pObjDef != m_po.end(); ++pObjDef)
        {
            CPerfObject* pPo = *pObjDef;

            if(pOcTmp->m_bToCollect)
            {
                // Get the total object's size.
                dwTotalSize += pPo->GetStructSize(pObj->NumInstances);
            }

            // Move on to the next object.
            pObj = (PERF_OBJECT_TYPE*)((PBYTE)pObj + pPo->GetMaxStructSize());
            ++pOcTmp;
        }

        // Determine if the buffer passed on input is large enough.
        *pdwObjectTypes = 0;
        if(dwTotalSize <= *pdwBytes)
        {
            // Ok, the buffer is large enough.
            *pdwBytes = 0;
            PBYTE* ppPerfData = (PBYTE*)ppData;
            pObj = (PERF_OBJECT_TYPE*)m_pbView;
            pOcTmp = pOc;
            for(pObjDef = m_po.begin(); pObjDef != m_po.end(); ++pObjDef)
            {
                CPerfObject* pPo = *pObjDef;

                // Get the total object's size.
                dwTotalSize = pPo->GetStructSize(pObj->NumInstances);

                if(pOcTmp->m_bToCollect)
                {
                    // Collect this object.
                    CopyMemory(*ppPerfData, pObj, dwTotalSize);

                    // Fix the Total byte length.
                    PERF_OBJECT_TYPE* pCopiedObj = (PERF_OBJECT_TYPE*)*ppPerfData;
                    pCopiedObj->TotalByteLength = dwTotalSize;

                    ++pOcTmp;
                    *ppPerfData += dwTotalSize;
                    ++*pdwObjectTypes;
                    *pdwBytes += dwTotalSize;
                }

                // Move on to the next object.
                pObj = (PERF_OBJECT_TYPE*)((PBYTE)pObj + pPo->GetMaxStructSize());
            }
        }
        else
        {
            // The buffer passed on input is too small: ask for a larger one.
            *pdwBytes = 0;
            dwRet = ERROR_MORE_DATA;
        }
    }
    else
    {
        // Not ready: return nothing.
        *pdwBytes = 0;
        *pdwObjectTypes = 0;
    }

    return dwRet;
}

void CPerfClient::MarkToCollect(ObjToCollect* paOc, DWORD dwObjTitleIndex)
{
    ObjToCollect* paOcSave = paOc;
    while(paOc->m_dwObjTitleIndex)
    {
        if(paOc->m_dwObjTitleIndex == dwObjTitleIndex)
        {
            if(!paOc->m_bToCollect)
            {
                paOc->m_bToCollect = TRUE;

                // Get the object by title index.
                PERF_OBJECT_TYPE* pObj = GetObject(dwObjTitleIndex);
                if(pObj && pObj->NumInstances != PERF_NO_INSTANCES)
                {
                    // This object exists and supports instances.
                    // Loop through all instances, looking for object references.
                    DWORD dwNumInstances = pObj->NumInstances;
                    PERF_INSTANCE_DEFINITION* pInst = (PERF_INSTANCE_DEFINITION*)((PBYTE)pObj + pObj->DefinitionLength);
                    while(dwNumInstances)
                    {
                        if(pInst->ParentObjectTitleIndex)
                            MarkToCollect(paOcSave, pInst->ParentObjectTitleIndex);
                        PERF_COUNTER_BLOCK* pCb = (PERF_COUNTER_BLOCK*)((PBYTE)pInst + pInst->ByteLength);
                        pInst = (PERF_INSTANCE_DEFINITION*)((PBYTE)pCb + pCb->ByteLength);
                        --dwNumInstances;
                    }
                }
            }
            break;
        }
        ++paOc;
    }
}

PERF_OBJECT_TYPE* CPerfClient::GetObject(DWORD dwObjTitleIndex)
{
    DWORD dwObjects = m_po.size();
    PERF_OBJECT_TYPE* pObj = (PERF_OBJECT_TYPE*)m_pbView;
    while(dwObjects && pObj->ObjectNameTitleIndex != dwObjTitleIndex)
    {
        pObj = (PERF_OBJECT_TYPE*)((PBYTE)pObj + pObj->TotalByteLength);
        --dwObjects;
    }
    if(dwObjects == 0)
        pObj = NULL;

    return pObj;
}

////////////////////////////////////// Dump routines ////////////////////////////////////
#ifdef _DEBUG

using std::endl;
using std::flush;

std::ostream& ToAnsi(std::ostream& os, LPCWSTR z)
{
    if(z && *z)
    {
        int iSize = WideCharToMultiByte(CP_ACP, 0, z, -1, NULL, 0, NULL, NULL);
        PSTR zA = (PSTR)_alloca((size_t)iSize);
        WideCharToMultiByte(CP_ACP, 0, z, -1, zA, iSize, NULL, NULL);
        os << (const char*)zA;
    }
    return os;
}

PCSTR DetailLevel(DWORD dwDetailLevel)
{
    PCSTR zDl = NULL;
    switch(dwDetailLevel)
    {
        case PERF_DETAIL_NOVICE:    zDl = "PERF_DETAIL_NOVICE";
            break;
        case PERF_DETAIL_ADVANCED:  zDl = "PERF_DETAIL_ADVANCED";
            break;
        case PERF_DETAIL_EXPERT:    zDl = "PERF_DETAIL_EXPERT";
            break;
        case PERF_DETAIL_WIZARD:    zDl = "PERF_DETAIL_WIZARD";
            break;
    }
    return zDl;
}

PCSTR CounterType(DWORD dwCounterType)
{
    PCSTR zCt;
    switch(dwCounterType)
    {
        case PERF_COUNTER_COUNTER:    zCt = "PERF_COUNTER_COUNTER";
            break;
        case PERF_COUNTER_TIMER:  zCt = "PERF_COUNTER_TIMER";
            break;
        case PERF_COUNTER_QUEUELEN_TYPE:    zCt = "PERF_COUNTER_QUEUELEN_TYPE";
            break;
        case PERF_COUNTER_LARGE_QUEUELEN_TYPE:    zCt = "PERF_COUNTER_LARGE_QUEUELEN_TYPE";
            break;
        case PERF_COUNTER_100NS_QUEUELEN_TYPE:    zCt = "PERF_COUNTER_100NS_QUEUELEN_TYPE";
            break;
        case PERF_COUNTER_OBJ_TIME_QUEUELEN_TYPE:    zCt = "PERF_COUNTER_OBJ_TIME_QUEUELEN_TYPE";
            break;
        case PERF_COUNTER_BULK_COUNT:    zCt = "PERF_COUNTER_BULK_COUNT";
            break;
        case PERF_COUNTER_TEXT:    zCt = "PERF_COUNTER_TEXT";
            break;
        case PERF_COUNTER_RAWCOUNT:    zCt = "PERF_COUNTER_RAWCOUNT";
            break;
        case PERF_COUNTER_LARGE_RAWCOUNT:    zCt = "PERF_COUNTER_LARGE_RAWCOUNT";
            break;
        case PERF_COUNTER_RAWCOUNT_HEX:    zCt = "PERF_COUNTER_RAWCOUNT_HEX";
            break;
        case PERF_COUNTER_LARGE_RAWCOUNT_HEX:    zCt = "PERF_COUNTER_LARGE_RAWCOUNT_HEX";
            break;
        case PERF_SAMPLE_FRACTION:    zCt = "PERF_SAMPLE_FRACTION";
            break;
        case PERF_SAMPLE_COUNTER:    zCt = "PERF_SAMPLE_COUNTER";
            break;
        case PERF_COUNTER_NODATA:    zCt = "PERF_COUNTER_NODATA";
            break;
        case PERF_COUNTER_TIMER_INV:    zCt = "PERF_COUNTER_TIMER_INV";
            break;
        case PERF_SAMPLE_BASE:    zCt = "PERF_SAMPLE_BASE";
            break;
        case PERF_AVERAGE_TIMER:    zCt = "PERF_AVERAGE_TIMER";
            break;
        case PERF_AVERAGE_BASE:    zCt = "PERF_AVERAGE_BASE";
            break;
        case PERF_AVERAGE_BULK:    zCt = "PERF_AVERAGE_BULK";
            break;
        case PERF_OBJ_TIME_TIMER:    zCt = "PERF_OBJ_TIME_TIMER";
            break;
        case PERF_100NSEC_TIMER:    zCt = "PERF_100NSEC_TIMER";
            break;
        case PERF_100NSEC_TIMER_INV:    zCt = "PERF_100NSEC_TIMER_INV";
            break;
        case PERF_COUNTER_MULTI_TIMER:    zCt = "PERF_COUNTER_MULTI_TIMER";
            break;
        case PERF_COUNTER_MULTI_TIMER_INV:    zCt = "PERF_COUNTER_MULTI_TIMER_INV";
            break;
        case PERF_COUNTER_MULTI_BASE:    zCt = "PERF_COUNTER_MULTI_BASE";
            break;
        case PERF_100NSEC_MULTI_TIMER:    zCt = "PERF_100NSEC_MULTI_TIMER";
            break;
        case PERF_100NSEC_MULTI_TIMER_INV:    zCt = "PERF_100NSEC_MULTI_TIMER_INV";
            break;
        case PERF_RAW_FRACTION:    zCt = "PERF_RAW_FRACTION";
            break;
        case PERF_LARGE_RAW_FRACTION:    zCt = "PERF_LARGE_RAW_FRACTION";
            break;
        case PERF_RAW_BASE:    zCt = "PERF_RAW_BASE";
            break;
        case PERF_LARGE_RAW_BASE:    zCt = "PERF_LARGE_RAW_BASE";
            break;
        case PERF_ELAPSED_TIME:    zCt = "PERF_ELAPSED_TIME";
            break;
        case PERF_COUNTER_HISTOGRAM_TYPE:    zCt = "PERF_COUNTER_HISTOGRAM_TYPE";
            break;
        case PERF_COUNTER_DELTA:    zCt = "PERF_COUNTER_DELTA";
            break;
        case PERF_COUNTER_LARGE_DELTA:    zCt = "PERF_COUNTER_LARGE_DELTA";
            break;
        case PERF_PRECISION_SYSTEM_TIMER:    zCt = "PERF_PRECISION_SYSTEM_TIMER";
            break;
        case PERF_PRECISION_100NS_TIMER:    zCt = "PERF_PRECISION_100NS_TIMER";
            break;
        case PERF_PRECISION_OBJECT_TIMER:    zCt = "PERF_PRECISION_OBJECT_TIMER";
            break;
        default:    zCt = "<other>";
            break;
    }
    return zCt;
}

void CPerfCounterBlock::Dump(std::ostream& os)
{
    os << "0x" << std::hex << (DWORD)m_pPcb << std::dec
       << ": PERF_COUNTER_BLOCK, " << m_dwNumCounters << " counters:\n"
       << "\tByteLength:            " << m_pPcb->ByteLength << endl;
    for(DWORD dw = 0; dw < m_dwNumCounters; ++dw)
        os <<
          "\tCnt:                   " << *(PDWORD)operator[](IndexToOffset(dw)) << "\n";
    os << flush;
}

void CPerfCounter::Dump(std::ostream& os)
{
    os << "0x" << std::hex << (DWORD)m_pPcd << std::dec
       << ": PERF_COUNTER_DEFINITION, index " << m_dwIndex << ", key " << m_dwKey << "\n"
       << "\tByteLength:            " << m_pPcd->ByteLength << "\n"
       << "\tCounterNameTitleIndex: " << m_pPcd->CounterNameTitleIndex << "\n"
       << "\tCounterNameTitle:      ";
    ToAnsi(os, m_pPcd->CounterNameTitle) << "\n"
       << "\tCounterHelpTitleIndex: " << m_pPcd->CounterHelpTitleIndex << "\n"
       << "\tCounterHelpTitle:      ";
    ToAnsi(os, m_pPcd->CounterHelpTitle) << "\n"
       << "\tDefaultScale:          " << m_pPcd->DefaultScale << "\n"
       << "\tDetailLevel:           " << DetailLevel(m_pPcd->DetailLevel) << "\n"
       << "\tCounterType:           " << CounterType(m_pPcd->CounterType) << "\n"
       << "\tCounterSize:           " << m_pPcd->CounterSize << "\n"
       << "\tCounterOffset:         " << m_pPcd->CounterOffset
       << endl;
}

void CPerfInstance::Dump(std::ostream& os)
{
    os << "0x" << std::hex << (DWORD)m_pPid << std::dec
       << ": PERF_INSTANCE_DEFINITION, index " << m_dwInstanceIndex << "\n"
       << "\tByteLength:            " << m_pPid->ByteLength << "\n"
       << "\tParentObjectTitleIndex:" << m_pPid->ParentObjectTitleIndex << "\n"
       << "\tParentObjectInstance:  " << m_pPid->ParentObjectInstance << "\n"
       << "\tUniqueID:              " << m_pPid->UniqueID << "\n"
       << "\tNameOffset:            " << m_pPid->NameOffset << "\n"
       << "\tNameLength:            " << m_pPid->NameLength << "\n";
    if(m_pPid->UniqueID == -1)
    {
        os <<
          "\tName:                  ";
        ToAnsi(os, (PCWSTR)(((PBYTE)m_pPid) + m_pPid->NameOffset)) << "\n";
    }
    os << flush;

    // Dump the associated counter block.
    m_cb.Dump(os);
}

void CPerfObject::Dump(std::ostream& os)
{
    os << "0x" << std::hex << (DWORD)m_pPot << std::dec
       << ": PERF_OBJECT_TYPE, key "  << m_dwKey << "\n"
       << "\tTotalByteLength:       " << m_pPot->TotalByteLength << "\n"
       << "\tDefinitionLength:      " << m_pPot->DefinitionLength << "\n"
       << "\tHeaderLength:          " << m_pPot->HeaderLength << "\n"
       << "\tObjectNameTitleIndex:  " << m_pPot->ObjectNameTitleIndex << "\n"
       << "\tObjectNameTitle:       ";
    ToAnsi(os, m_pPot->ObjectNameTitle) << "\n"
       << "\tObjectHelpTitleIndex:  " << m_pPot->ObjectHelpTitleIndex << "\n"
       << "\tObjectHelpTitle:       " ;
    ToAnsi(os, m_pPot->ObjectHelpTitle) << "\n"
       << "\tDetailLevel:           " << DetailLevel(m_pPot->DetailLevel) << "\n"
       << "\tNumCounters:           " << m_pPot->NumCounters << "\n"
       << "\tDefaultCounter:        " << m_pPot->DefaultCounter << "\n"
       << "\tNumInstances:          " << m_pPot->NumInstances << "\n"
       << "\tCodePage:              " << m_pPot->CodePage
       << endl;

    // Dump the counter definitions.
    for(Counters::const_iterator pCnt = m_counters.begin(); pCnt != m_counters.end(); ++pCnt)
        (*pCnt)->Dump(os);

    if(m_bSupportsInstances)
    {
        // Dump the instances.
        for(Instances::const_iterator pInst = m_id.begin(); pInst != m_id.end(); ++pInst)
            (*pInst)->Dump(os);
    }
    else
    {
        // Dump the counter block.
        m_cb.Dump(os);
    }

    os << endl;
}

void CPerfSource::Dump(std::ostream& os)
{
    for(PerfObjects::const_iterator pOt = m_po.begin(); pOt != m_po.end(); ++pOt)
        (*pOt)->Dump(os);
}

#endif
////////////////////////////////////// End of dump routines ////////////////////////////////////

}   // OpenSTA_Performance
