/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// SCLWriter.cpp: implementation of the SCLWriter class.
//
//////////////////////////////////////////////////////////////////////
#include "SCLWriter.h"
#include "TofPrmDef.h"
#include "MinMax.h"

extern char szSCLFilename[MAX_PATH];
extern char szHeadersIniPath[MAX_PATH];
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
SCLWriter::SCLWriter()
{
    m_nBufferLen=0;
    m_nBufferPos=0;
    m_pszBuffer = NULL;

    m_bHeadersInitialised = false;
    m_bOnPrimaryGet = true;
    m_strDefIndent = _T("\t");
    m_bLastWriteWasConnect = true;
	m_nPageTimer=1;
	m_bMatchedTimers = true;

	//Generate the Page timer name format
    char *pszTofName = strrchr(szSCLFilename,'\\');
    if(pszTofName)
        pszTofName++;

    m_strPageTimerName = pszTofName;
	m_strPageTimerName = m_strPageTimerName.SpanExcluding(".");
	m_strPageTimerName.MakeUpper();
	m_strPageTimerName = "T_" + m_strPageTimerName.Left(57);
	m_strPageTimerName.TrimRight("_-");
	m_strPageTimerName += "_%d";
}

SCLWriter::~SCLWriter()
{
    if(m_pszBuffer != NULL)
    {
		WriteExitString();

		//Flush to buffer
		WriteString("");

		//Insert header at start of file
		CString strBuffer,strDefs,strCodeSection;

		GetEnvironmentString(strBuffer);
		GetDefinitionsString(strDefs);
		GetBeginCodeString(strCodeSection);

		strBuffer += strDefs + strCodeSection + m_pszBuffer;

		free(m_pszBuffer);

		try
		{
			CStdioFile SclFile;
			SclFile.Open(szSCLFilename,CFile::modeCreate|CFile::modeWrite);
			SclFile.WriteString(strBuffer);
			SclFile.Close();
		}
		catch(CFileException *e)
		{
			e->Delete();
		}
	}

	m_arVariables.RemoveAll();
    m_arConstants.RemoveAll();
	m_mapDefaultHeaders.RemoveAll();
}

//Get environment section of SCL File
void SCLWriter::GetEnvironmentString(CString& strEnvironment)
{
    char *pszTofName = strrchr(szSCLFilename,'\\');
    if(pszTofName)
        pszTofName++;

    CString strTofName(pszTofName);

    strEnvironment = "Environment\n";
    strEnvironment += "\tDescription \"\"\n";// + strTofName + "\"\n";
    strEnvironment += "\tMode\t\tHTTP\n";
    strEnvironment += "\tWait\t\tUNIT MILLISECONDS\n\n";
}

// Get the definitions section.
void SCLWriter::GetDefinitionsString(CString& strDefinitions)
{
    // Create the Definitions section:
    strDefinitions = "Definitions\n";
    strDefinitions += "\t! Standard Defines\n";
    strDefinitions += "\tInclude\t\t\t\"RESPONSE_CODES.INC\"\n";
    strDefinitions += "\tInclude\t\t\t\"GLOBAL_VARIABLES.INC\"\n\n";
    strDefinitions += "\tCHARACTER*512\tUSER_AGENT\n";
	strDefinitions += "\tInteger\t\t\tUSE_PAGE_TIMERS\n";

    // And add in definitions for the request timeout
    strDefinitions += "\tCHARACTER*256\tMESSAGE\n\n";

    // Add the variable definitions:
    for(int nVariable=0;nVariable<m_arVariables.GetSize();nVariable++)
        strDefinitions += m_arVariables[nVariable] + "\n";

    // Add the DEFAULT_HEADERS constant:
    strDefinitions += "\n\tCONSTANT\t\tDEFAULT_HEADERS = ";
    CString strHeader,strValue;
    CString strTemp;
    POSITION pos = m_mapDefaultHeaders.GetStartPosition();
    while(pos!=NULL)
    {
        m_mapDefaultHeaders.GetNextAssoc(pos,strHeader,strValue);

        strTemp = "\"" + strHeader + ": " + strValue + "^J\"";
        ValidateString(strTemp,2);

        strDefinitions += strTemp;
        strDefinitions += "\t&\n\t\t";
    }
    strDefinitions += "\"User-Agent: " + m_strUserAgent + "\"\n";

    // Add the other constant definitions:
    if(m_arConstants.GetSize())
    {
        for(int nConstants=0;nConstants<m_arConstants.GetSize();nConstants++)
		{
			CString strVarTmp = m_arConstants[nConstants] ;
			ValidateString(strVarTmp, 0);
            strDefinitions += strVarTmp + "\n";
		}
    }

    strDefinitions += "\n";
}

// Get the beginning of the code section.
void SCLWriter::GetBeginCodeString(CString& strBeginCodeString)
{
    strBeginCodeString = "Code\n";

    strBeginCodeString += "\t!Read in the default browser user agent field\n";
    strBeginCodeString += "\tEntry[USER_AGENT,USE_PAGE_TIMERS]\n\n";
}

//Get Any scl which should go at the end of the script
void SCLWriter::WriteExitString()
{
	if(!m_bMatchedTimers)
		WritePageTimer();
	
    CString strExit;
    strExit += "Exit\n\n";
    //
    // Add in error label for the HTTP Request Timeout to jump to - which
    // appears after the main exit statement so that only http timeouts go through
    // the code.
    strExit += "ERR_LABEL:\n\n";
    strExit += "\tIf (MESSAGE <> \"\") Then\n";
    strExit += "\t\tReport MESSAGE\n";
    strExit += "\tEndif\n\n";
    strExit += "Exit\n";

    WriteString(strExit);
}

void SCLWriter::WriteConnectString(const char* szHostName,int nConId)
{
    m_bLastWriteWasConnect = true;

    CString strConnect;

    CString strHost(szHostName);
    ValidateString(strHost,1);

    strConnect.Format("%sCONNECT TO \"%s\" ON %d\n\n",m_strDefIndent,strHost,nConId);

    WriteString(strConnect);
}

void SCLWriter::WriteWaitString(int nDelay)
{
    m_bLastWriteWasConnect = false;

    CString strWait;
    strWait.Format("%sWAIT %d\n\n",m_strDefIndent,nDelay);

    WriteString(strWait);
}

void SCLWriter::WriteDisconnectString(int nConId,int nOptions)
{
    m_bLastWriteWasConnect = false;

    CString strDisconnect;
    strDisconnect.Format("%sDISCONNECT FROM %d",m_strDefIndent,nConId);

    if(nOptions == 1)
        strDisconnect += ",WITH CANCEL";

    strDisconnect += "\n\n";

    WriteString(strDisconnect);
}

void SCLWriter::WriteRequestString(int nConId,
                                   RequestTypes RequestType,
                                   const char* szHttpVersion,
                                   const char* szHeaders,
                                   const char* szBody,
								   const char* szCertificate,
								   int nOption)
{
    CString strType;
	
	if(nOption & TOF_REQOPT_PRIMARY)
	{
//		m_bOnPrimaryGet = true;
		strType += "PRIMARY ";
	}

    switch(RequestType)
    {
    case enGet:		strType += "GET URI";	break;
    case enHead:	strType += "HEAD URI";	break;
    case enPost:	strType += "POST URI";	break;
    default:
        ASSERT(FALSE);
        break;
    }

    BOOL bLastGetWasSecondary = !m_bOnPrimaryGet;

    m_bOnPrimaryGet = false;

    CString strHttpVersion(szHttpVersion);
    strHttpVersion.TrimRight();
    int nPos = strHttpVersion.ReverseFind(' '); //This will find the gap between the URL and HTTP/X.X
    if(nPos!=-1)
    {
        CString strURL = strHttpVersion.Left(nPos);
        strURL.TrimRight();
        m_bOnPrimaryGet = (strURL.Right(1) == "/");
    }

    CString strHeader;
    ProcessHeaders(szHeaders,strHeader);

	if(m_bOnPrimaryGet && !m_bMatchedTimers)
	{
		WriteSynchronize();
		WritePageTimer();
	}

	if(m_bOnPrimaryGet)
		WritePageTimer();

    if(m_bLastWriteWasConnect && bLastGetWasSecondary && m_bOnPrimaryGet)
    {
        //Remove indent from connect on previous line
        m_strPrevLine.Replace("\t\t","\t");
    }

    m_strDefIndent = (m_bOnPrimaryGet ? "\t" : "\t\t");

    CString strRequest;
    strRequest.Format("%s%s \"%s\" ON %d",
                      m_strDefIndent, strType, strHttpVersion, nConId);
    ValidateString(strRequest, (m_bOnPrimaryGet?2:3));
    strRequest += " &\n" + m_strDefIndent + "\tHEADER DEFAULT_HEADERS";
    strRequest += strHeader;

    if(szBody && RequestType == enPost)
    {
        strRequest += m_strDefIndent + " &\n\t\t,";

        if(strncmp(szBody,"0x",2)==0)
            strRequest += "BINARY ";

        CString strBody("\"");
		strBody += szBody;
		strBody += "\"";
        ValidateString(strBody,2);
        strRequest += "BODY " + strBody;
    }

	if(szCertificate)
	{
        strRequest += m_strDefIndent + " &\n\t\t,CERTIFICATE ";
		
		if(nOption & TOF_REQOPT_CERTFIL)
			strRequest += "FILE ";
		else
			strRequest += "NAME ";

		CString strCertificate("\"");
		strCertificate += szCertificate;
		strCertificate += "\"";
        ValidateString(strCertificate,2);
		strRequest += strCertificate;
	}

    strRequest += "\n\n";

    WriteString(strRequest);
}

void SCLWriter::WriteLoadResponseInfoString(int nConId,int nData,const char* szIdentifier,const char *szName)
{
    m_bLastWriteWasConnect = false;

    CString strLoadRespInfo(m_strDefIndent + "Load Response_Info ");

    CString strType;
    switch(nData)
    {
    case 0:
        strLoadRespInfo += "Header ";
        break;
    case 1:
        strLoadRespInfo += "Body ";
        break;
    default:
        ASSERT(FALSE);
        break;
    }

    CString strConId;
    strConId.Format("on %d\t\t&\n",nConId);
    strLoadRespInfo += strConId;

    strLoadRespInfo += m_strDefIndent + "\tInto ";

    CString strTemp(szName);
    ValidateString(strTemp,2);

    strLoadRespInfo += strTemp;

    if(szIdentifier)
    {
        strLoadRespInfo += "\t&\n";
        strLoadRespInfo += m_strDefIndent + "\t,WITH \"";
        strLoadRespInfo += szIdentifier;
        strLoadRespInfo += "\"";
    }

    strLoadRespInfo += "\n\n";

    WriteString(strLoadRespInfo);
}

void SCLWriter::WriteString(const CString& str)
{
    if(!m_strPrevLine.IsEmpty())
    {
        int nStrLen = m_strPrevLine.GetLength();
        if(nStrLen+1 > (m_nBufferLen-m_nBufferPos))
        {
            m_nBufferLen += nStrLen + 10000;
            char *pszNewBuffer = (char*)malloc(m_nBufferLen);

            memcpy(pszNewBuffer,m_pszBuffer,m_nBufferPos);

			free(m_pszBuffer);
            m_pszBuffer = pszNewBuffer;
        }

        strcpy(&m_pszBuffer[m_nBufferPos],m_strPrevLine);
        m_nBufferPos+=nStrLen;
    }
    m_strPrevLine = str;
}

void SCLWriter::ProcessHeaders(const char* szHeaders,CString& strHeader)
{
    CString strCopy(szHeaders);
    if(strCopy.IsEmpty())
    {
        m_bHeadersInitialised = true;
        return;
    }

	strCopy.TrimLeft("\"");
	strCopy.TrimRight("\"");

    CString strKey,strValue,strDefValue;
    CMapStringToString mapNewHeaders;

    BOOL bFirstWith=true;
    //Go through each header in the list (separated by newline)

    BOOL bFoundAcceptAll=m_bOnPrimaryGet;
    char *pszToken = strCopy.GetBuffer(0);
    char *pszTemp = pszToken;
    while(pszTemp!=NULL)
    {
        pszTemp = strstr(pszToken,"^J");

        if(pszTemp)
            pszTemp[0] = NULL;
#ifdef _DEBUG_ON_
        CString str(pszToken);
        OutputDebugString(str + "\n");
#endif

        //Separate each header into a key and a value
        //The key always has a : following it
        char *pValue = strchr(pszToken,':');

        if(pValue++)
        {
            int nColonPos = pValue - pszToken-1;

            //Obtain the key and value values
            strKey = CString(pszToken).Left(nColonPos);
            if(pValue[0] == '\0')
                strValue.Empty();
			else
				strValue = pValue+1;
			strValue.TrimLeft("\"");
			strValue.TrimRight("\" +");

            if(!bFoundAcceptAll && strKey.CompareNoCase("Accept")==0)
            {
                if(strValue == "*/*")
                    bFoundAcceptAll = true;
            }

            if(strKey.CompareNoCase("user-agent")!=0)
            {
                //The headers for the first request are assumed to be the default for
                //the remainder, so we need a copy of these to compare.
                if(!m_bHeadersInitialised)
                    m_mapDefaultHeaders.SetAt(strKey,strValue);
                else
                {
                    //Decide whether this key is the same as the default
                    //If not it must be added in the WITH
                    mapNewHeaders.SetAt(strKey,strValue);

                    BOOL bFound = m_mapDefaultHeaders.Lookup(strKey,strDefValue);
                    if(!bFound || (strDefValue != strValue))
                    {
                        CString strTemp("\t\t");
                        if(bFirstWith)
                        {
                            bFirstWith = false;
                            strHeader += " &\n";
                            strTemp += ",WITH {";
                        }
                        else
                        {
                            strHeader += ", &\n";
                            strTemp += "\t";
                        }

                        strTemp += "\"" + strKey + ": ";
						if(strValue.Left(1) == "+")
						{
							strTemp += "\"" + strValue;
						}
						else
						{
							strTemp += strValue + "\"";
						}

                        ValidateString(strTemp,3);
                        strHeader += strTemp;
                    }
                }
            }
            else
                if(!m_bHeadersInitialised)
                    m_strUserAgent = strValue;
        }
        pszToken = pszTemp+2;
    }

    if(!bFirstWith)
        strHeader += "}";

    if(m_bHeadersInitialised)
    {
        BOOL bFirstWithout=true;

        //Go through all the default headers ...
        POSITION pos = m_mapDefaultHeaders.GetStartPosition();
        while(pos != NULL)
        {
            m_mapDefaultHeaders.GetNextAssoc(pos,strKey,strDefValue);

            //..if this request doesn't contain the default then
            //we will need a WITHOUT
            if(!mapNewHeaders.Lookup(strKey,strValue))
            {
                if(bFirstWithout)
                {
                    bFirstWithout = false;
                    strHeader += "\t\t\t&\n\t\t,WITHOUT {";
                }
                else
                    strHeader += ",\t\t\t&\n\t\t\t";

                ValidateString(strKey,3);

                strHeader += "\"" + strKey + "\"";
            }
        }
        mapNewHeaders.RemoveAll();

        if(!bFirstWithout)
            strHeader += "}";
    }
    else
    {
        CString strIniFile;
        if(szHeadersIniPath)
        {
            strIniFile = szHeadersIniPath;
            strIniFile.TrimRight();

            if(!strIniFile.IsEmpty() && strIniFile.Right(1) != "\\")
                strIniFile += "\\";
        }
        strIniFile += "headers.ini";

        ProcessIniFile(strIniFile);
        m_bHeadersInitialised = true;
        ProcessHeaders(szHeaders,strHeader);
    }

    m_bOnPrimaryGet |= !bFoundAcceptAll;

    if(!m_bOnPrimaryGet)
    {
        strHeader = "\t" + strHeader;
        strHeader.Replace("\n\t","\n\t\t");
    }
}

void SCLWriter::WriteCommentString(const char* szComment)
{
    CString strComment;

    CString strCopy(szComment);

    CString strTemp;
    char *pszToken = strtok(strCopy.GetBuffer(0),"\n");
    while(pszToken!=NULL)
    {
        strComment += m_strDefIndent + "!";

        strTemp = pszToken;
        int nLength = strTemp.GetLength();
        int nIndex=80;
        while(nLength>nIndex)
        {
            strTemp.Insert(nIndex,"\n"+m_strDefIndent + "!");
            nIndex+=83;
            nLength++;
        }
        //  ValidateString(strTemp,1);

        strComment += strTemp;
        strComment += "\n";

        pszToken = strtok(NULL,"\n");
    }

    WriteString(strComment);
}

void SCLWriter::ProcessIniFile(const CString& strIniFile)
{
    try
    {
        CStdioFile f(strIniFile,CFile::modeRead);

        CString strKey,strValue,strHeader;
        CMapStringToString mapIniHeaders;
        while(f.ReadString(strHeader))
        {
            int nEqualsPos = strHeader.Find('=');
            if(nEqualsPos != -1)
            {
                strKey = strHeader.Left(nEqualsPos);
                strValue = strHeader.Right(strHeader.GetLength()-nEqualsPos-1);
                if(strValue == "1")
                {
                    strKey.MakeUpper();
                    mapIniHeaders.SetAt(strKey,strValue);
                }
            }
        }

        CString strKeyCopy;
        POSITION pos = m_mapDefaultHeaders.GetStartPosition();
        while(pos != NULL)
        {
            m_mapDefaultHeaders.GetNextAssoc(pos,strKey,strValue);
            strKeyCopy = strKey;
            strKey.MakeUpper();
            if(!mapIniHeaders.Lookup(strKey,strValue))
            {
                m_mapDefaultHeaders.RemoveKey(strKeyCopy);
            }
        }

        f.Close();

    }
    catch(CFileException* e)
    {
        m_mapDefaultHeaders.RemoveAll();
        e->Delete();
    }
}

//Check that a quoted line in SCL is not too long
//if it is, split it into 2 or more lines
void SCLWriter::ValidateString(CString& strLine,int nNumTabs)
{
    const int nMaxUsable = 95; // Less than max of 132, just for style ...
    // Quickly find the case where we have nothing to do and return.
    if(strLine.GetLength() <= nMaxUsable) return;
    const int nMaxStartNew = 65; // Don't start strings after, for style ...
    CString strIndent('\t', nNumTabs); // Define indentation for split lines
    CString strLineBreak(" &\n"); // Define characters to break line with
    CString strExtra(strLine); // Create string of all chars yet to deal with
    // Empty original string to later put modified contents back in
    strLine.Empty();
    // Work string contains max possible in a line to start with
    CString strWork = strExtra.Left(nMaxUsable);
    int nWorkPos; // Position in the work string we are looking at
    int nWorkAdded = 0; // Number of chars added to work string not from extra 
    int nSclStringSize = -1; // Keep track of total SCL string length
    int nSclStringStart = -1; // Where an SCL string begins in work
    // Loop around until no more work to do
    while(!strExtra.IsEmpty()) {
        nSclStringStart = (nSclStringSize >= 0) ? 0 : -1;
        // Handle string breaks - loop through quote marks in work string
        nWorkPos = 0;
        while((nWorkPos < strWork.GetLength()) &&
              ((nWorkPos = strWork.Find('\"', nWorkPos)) >= 0)) {
            if(nSclStringStart < 0) { // no current SCL string, so starting quote
                // If not started string note string has started (don't count size yet)
                if(nSclStringSize < 0) nSclStringSize = 0;
                nSclStringStart = nWorkPos+1; // and where it starts
            } else { // found ending quote
                nSclStringSize = -1; // Not in SCL string
                nSclStringStart = -1;
            }
            nWorkPos++;
        } // while quote characters in work string

        if(nSclStringSize < 0) { // not partway through an SCL string
            nWorkPos = strWork.GetLength();  // Start at end
            // Only do anything if more left is extra buffer 
            if(strExtra.GetLength() > nWorkPos) {
                bool bFound = false;
                while(!bFound && (nWorkPos > 0)) {
                    switch(strWork.GetAt(nWorkPos-1)) {
                      case '\"':
                      case ',':
                      case ')':
                      case '>':
                      case ' ':
                      case '\t':
                      case '+':
                        bFound = true;
                        break;
                      default:
                        nWorkPos--;
                        break;
                    }
                }
                // If didn't find anything good, strip off any partial ~ commands
                if(!bFound && ((nWorkPos = strWork.ReverseFind('~')) <= 0)) {
                    nWorkPos = strWork.GetLength();
                }
                if(nWorkPos < strWork.GetLength()) { // Delete if some to delete
                    strWork.Delete(nWorkPos, (strWork.GetLength()-nWorkPos));
                }
            }
        } else { // partway through a big SCL string
            if((nSclStringSize == 0) && // SCL string started this work buffer
               (nSclStringStart > nMaxStartNew) && // and was after given point
               ((nWorkPos = strExtra.Find('\"')) > 0) && // and the end quote 
               ((nWorkPos + (strWork.GetLength()-nSclStringStart)) < nMaxUsable)) {
                // means the quote could fit on its own line, so put it there
                nWorkPos = nSclStringStart-1;
            // Find a good place to break the string - we prefer encoded LF's
            } else if((nWorkPos = strWork.Find("~<LF>", nSclStringStart)) > 0) {
                nWorkPos += 5;
                int nPos;
                // Look for more later in string - get as much on a line
                while((nPos = strWork.Find("~<LF>", nWorkPos)) > 0) {
                    nWorkPos = nPos+5;
                }
            } else if((nWorkPos = strWork.Find("^J", nSclStringStart)) > 0) {
                nWorkPos += 2;
            } else if(((nWorkPos = strWork.ReverseFind('>')) > 0) || // ~<XX>
                      ((nWorkPos = strWork.ReverseFind(' ')) > 0) || // space
                      ((nWorkPos = strWork.ReverseFind(',')) > 0) || // comma
                      ((nWorkPos = strWork.ReverseFind('/')) > 0)) { // slash
                // Above checks from end of string giving preference
                nWorkPos += 1;
            } else { // Nothing good just make sure not breaking control or cmd
                nWorkPos = strWork.GetLength();
                // This loop could produce odd results for strings full of
                // these chars - but it shouldn't break anything.
                while((nWorkPos > 4) &&
                      ((strWork[nWorkPos-1] == '~') ||
                       (strWork[nWorkPos-1] == '<') ||
                       (strWork[nWorkPos-2] == '<') ||
                       (strWork[nWorkPos-3] == '<') ||
                       (strWork[nWorkPos-4] == '<') ||
                       (strWork[nWorkPos-1] == '^'))) nWorkPos--;
            }
            // If above puts us before beginning of string
            if((nWorkPos > 0) && (nWorkPos < nSclStringStart)) {
                nWorkPos = nSclStringStart-1; // Put string on next line
                nSclStringSize = -1; // 
                nSclStringStart = -1;
            } 
            // Delete end of Work string if needed
            if((nWorkPos > 0) && (nWorkPos < strWork.GetLength())) {
                strWork.Delete(nWorkPos, (strWork.GetLength()-nWorkPos));
            }
            if(nSclStringSize >= 0) { // If we're within a broken SCL string
                // Increment total Scl string size
                nSclStringSize += (strWork.GetLength()-nSclStringStart);
                ASSERT(nSclStringSize!=0); //this would just be stupid
            }
        } 

        // Remove the work string used from the extra
        strExtra.Delete(0, strWork.GetLength()-nWorkAdded);
        nWorkAdded = 0;
        // Add the work string to end of the modified original
        strLine += strWork;

        if(!strExtra.IsEmpty()) {
            // Grab another section of the extra into the work string.
            strWork = strExtra.Left(nMaxUsable);
            // if we are mid SCL string.
            if(nSclStringSize > 0) {
                strLine += '\"'; // Need to add quote for end of line
                // If next line could possibly go beyond half SCL max var size
                if((nSclStringSize + strWork.GetLength()) >= 32768) {
                    strLine += " +"; // Chop string with SCL string concatenation    
                    strWork.Insert(0, '\"'); // make work strings start with quote
                    nWorkAdded++; // Keep track of quote added to work string
                    nSclStringSize = -1; // no longer really in broken SCL string
                }
            }
            strLine.TrimRight(); // Remove any trailing whitespace
#ifdef _DEBUG_ON_
            strLineBreak.Format(" & !%d,%d\n", nSclStringStart, nSclStringSize);
#endif
            strLine += strLineBreak; // Add line break
            strLine += strIndent; // Indent for the next line
            // And if mid string add more indent and quote
            if(nSclStringSize > 0) strLine += "\t\"";
        }
    } // Loop while there's still more extra string to be consumed

}

void SCLWriter::WriteConstantDef(const char *szName,const char *szValue)
{
    CString strVarDef("\tCONSTANT");

    strVarDef += "\t";
    strVarDef += szName;
    strVarDef += "\t= \"";
    strVarDef += szValue;
    strVarDef += "\"";

    m_arConstants.Add(strVarDef);
}

void SCLWriter::WriteVariableDef(const char *szName,int nOption,int nSize)
{
    CString strVarDef("\t");

    //Get the variable type
    if(nOption & VTCHAR)
        strVarDef += "CHARACTER";
    else
        if(nOption & VTINT)
            strVarDef += "Integer";
        else
        {
            ASSERT(FALSE);
        }

    if(nSize != 1)
    {
        CString strSize;
        strSize.Format("*%d",nSize);
        strVarDef += strSize;
    }

    strVarDef += "\t";
    strVarDef += szName;

    //Get scope
    if(nOption & VCLOCAL);
    else if(nOption & VCSCRIPT)
        strVarDef += ",SCRIPT";
    else if(nOption & VCGLOBAL)
        strVarDef += ",GLOBAL";
    else if(nOption & VCTHREAD)
        strVarDef += ",THREAD";
    else
    {
        ASSERT(FALSE);
    }

    m_arVariables.Add(strVarDef);
}

void SCLWriter::WriteTimer(const char *szName,bool bStart)
{
    CString strTimer;
    CString szTimerName(szName);

	int nLength = min(szTimerName.GetLength(),64);
    if(bStart)
    {
        strTimer.Format("\tStart Timer %s\n",szTimerName.Left(nLength));
    }
    else
    {
        strTimer.Format("\tEnd Timer %s\n",szTimerName.Left(nLength));
    }

    WriteString(strTimer);
}

void SCLWriter::WritePageTimer()
{
	if (!m_bAddPageTimer)
		return ;
	CString strName;
	strName.Format(m_strPageTimerName,m_nPageTimer);

	CString str;
	str = "\n\tIf (USE_PAGE_TIMERS=1) then\n\t";
	
	//If previous line was a comment insert the timer before it
	CString strPrevLine;
	if(m_strPrevLine.Find("\t!")!=-1)
	{
		strPrevLine = m_strPrevLine;
		m_strPrevLine.Empty();
	}
	
	WriteString(str);
	WriteTimer(strName,m_bMatchedTimers);
	str = "\tEndIf\n\n";
	WriteString(str);
	
	if(m_bMatchedTimers)
	{
		WriteTimerDef(strName);
		m_bMatchedTimers = false;
	}
	else
	{
		m_nPageTimer++;
		m_bMatchedTimers = true;
	}
	if(!strPrevLine.IsEmpty())
		WriteString(strPrevLine);
}

void SCLWriter::WriteTimerDef(const char *szName)
{
    CString strTimDef;

    // Trim the timer variable if over 64 characters
    CString szTimerName = szName;

    strTimDef.Format("\tTimer\t\t\t%s",szTimerName.Left(64));
    m_arVariables.Add(strTimDef);
}

void SCLWriter::WriteSynchronize()
{
    CString strSynchronize;
    strSynchronize = m_strDefIndent + "SYNCHRONIZE REQUESTS\n\n";

	//If previous line was a comment insert the timer before it
	CString strPrevLine;
	if(m_strPrevLine.Find("\t!")!=-1)
	{
		strPrevLine = m_strPrevLine;
		m_strPrevLine.Empty();
	}

    WriteString(strSynchronize);

	if(!strPrevLine.IsEmpty())
		WriteString(strPrevLine);
}

void SCLWriter::WriteSetVar(const char *szName,const char *szValue)
{
	CString strSetVar;
	strSetVar.Format("%sSet %s = \"%s\"\n",m_strDefIndent,szName,szValue);
	ValidateString(strSetVar,m_strDefIndent.GetLength());
	WriteString(strSetVar);
}

void SCLWriter::WriteBAuthentication(int nType,int nTPArg,const char* szUserName,const char* szPassword,
									  const char* szDomain,const char* szBlob,const char* szInto)
{
	CString strBAuth;

	strBAuth = m_strDefIndent + "BUILD AUTHENTICATION BLOB\t&\n";
	strBAuth += m_strDefIndent + "\tFOR ";

	switch(nType)
	{
	case TOF_AUTTYP_BASIC:
		strBAuth += "BASIC\t&\n";
	break;
	case TOF_AUTTYP_NTLM:
		strBAuth += "NTLM\t&\n";
	break;
	case TOF_AUTTYP_NEGOTIATE:
		strBAuth += "NEGOTIATE\t&\n";
	break;
	case TOF_AUTTYP_KERBEROS:
	default:
		return;
	}

	CString strTPArg;
	switch(nTPArg)
	{
	case TOF_AUTIN_CURRENT:
		strTPArg = "FROM CURRENT USER";
	break;
	case TOF_AUTIN_CREDENTIALS:
		if(szDomain && strlen(szDomain))
			strTPArg.Format("FROM USER %s PASSWORD %s DOMAIN %s",szUserName,szPassword,szDomain);
		else
			strTPArg.Format("FROM USER %s PASSWORD %s",szUserName,szPassword);
	break;
	case TOF_AUTIN_BLOB:
		strTPArg.Format("FROM BLOB %s",szBlob);
	break;
	}
	strBAuth += m_strDefIndent + "\t" + strTPArg + "\t&\n";

	strBAuth += m_strDefIndent + "\tINTO " + CString(szInto) + "\n\n";
	WriteString(strBAuth);
}
