/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//    CyrExecuteLocalCommand.cpp                                            //
//    =====================                                                 //
//                                                                          //
//    This is the implementation file for the CyrExecuteLocalCommand Class. //
//                                                                          //
//                                                                          //
//    This class supports the execution of an image or DOSCommand           //
//////////////////////////////////////////////////////////////////////////////
//    Class definition files.
//
#include "stdafx.h"
#include "CyrExecuteLocalCommand.h"
#include <winbase.h>
#include <malloc.h>
#include <iostream>
#define BUFSIZE 4096

/////////////////////////////////////////////////////////////////////////////
// CyrExecuteLocalCommand

//
//    Default constructor.
//    @mfunc Constructor
//
CyrExecuteLocalCommand::CyrExecuteLocalCommand  (CString _strCurrentDir)
{
    m_strCurrentDir  = _strCurrentDir;
    m_hChildStdoutWr = NULL;
    m_hChildStdoutRd = NULL;

}

//
// @mfunc Destructor
//
//    Destructor
//
CyrExecuteLocalCommand::~CyrExecuteLocalCommand()
{}

//
// @mfunc ExecuteCommand
//        ==============
//
//    This routine allows a DOS Command to be executed, optionally sending the output from the
//    command to a CWnd derived object.
//
//    The parameter bAppendOutput is used to indicate whether the output window will be
//    appended (TRUE) or its previous contents erased (FALSE).
//
//    If a remote node name is specified then the command is executed on the specified remote node.
//    Note however that in this case we shall also have to pass across to this routine a
//    username and password which will be used to log onto the remote node.
//
//    Calling Format:
//      status = ExecuteCommand (Command
//                              ,OutputString
// @parm    char* | pCommandLine | Pointer to buffer for the command to execute
// @parm    Pointer to string to which output from the command is sent
// @rdesc   0 if success, non-zero otherwise
//

int CyrExecuteLocalCommand::ExecuteCommand  (char *pCommandLine, CString &OutputStr)
{
    m_strOutput=OutputStr;
    int stat=GoMultiThread (pCommandLine);
    OutputStr=m_strOutput;
    return stat;
}


//
// @mfunc GoMultiThread
//        =============
//
//    This routine is called as a new thread to execute the specified DOSCommand and route its
//    output to the output window.
//
// @parm char* | pCommandLine | Pointer to command line to execute
// @rdesc   0 if success, non-zero otherwise
//
int     CyrExecuteLocalCommand::GoMultiThread   (char*    pCommandLine)

{

    //
    //    The command is being executed locally.
    //
    SECURITY_ATTRIBUTES saAttr;

    //
    // Set the bInheritHandle flag so pipe handles are inherited.
    //
    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = NULL;

    //
    // The steps for redirecting child's STDOUT:
    //     2.  Create anonymous pipe to be STDOUT for child.
    //     3.  Set STDOUT of parent to be write handle of pipe, so
    //         it is inherited by child.
    //

    //
    // Create a pipe for the child's STDOUT.
    //
    if (!CreatePipe (&m_hChildStdoutRd, &m_hChildStdoutWr, &saAttr, 0))
        ErrorExit("Stdout pipe creation failed\n");

    //
    // Now create the child process to execute the command
    //
    InvokeImage (pCommandLine);

    //
    //  Read from pipe that is the standard output for child process.
    //
    ReadFromChild();

    return 0;
}

//
//    ReadFromChild
//    =============
//
//    This routine is called from a multi-threaded operation to read from the pipe
//    created to a child process which is executing a DOS command for us. We send
//    the output recovered to the output window
//
void    CyrExecuteLocalCommand::ReadFromChild   ()
{
    CString TextBuffer;
    CString s;

    DWORD dwRead;
    CHAR chBuf[BUFSIZE];

    //
    // Close the write end of the pipe before reading from the
    // read end of the pipe.
    //
    if (!CloseHandle(m_hChildStdoutWr))
        ErrorExit("Closing handle failed");

    //
    // Read output from child, and write it to the output window edit box.
    //
    TextBuffer="";
    for (;;)
    {
        if ((!ReadFile(m_hChildStdoutRd, chBuf, BUFSIZE, &dwRead, NULL)) || (dwRead == 0)) break;
        chBuf[dwRead] = '\0';
        TextBuffer += chBuf;
    }

    m_strOutput = TextBuffer;

}

void CyrExecuteLocalCommand::ErrorExit (CString lpszMessage)
{
    std::cerr<<(const char*)lpszMessage<<std::endl;
    ExitProcess(0);
}

//
//    CyrExecuteLocalCommand::InvokeImage
//    ==============================
//
//    This routine is used to invoke the specified image. or execute the specified DOS Command
//
int CyrExecuteLocalCommand::InvokeImage (char*  pImageName)
{
    PROCESS_INFORMATION ProcInfo;
    CString     ErrorBuffer;
    STARTUPINFO StartInfo = {sizeof(StartInfo)};
    BOOL        InheritHandles = FALSE;
    int         CreateFlags = 0;


    //
    //    If we have requested that the window should be hidden then set up the STARTUPINFO
    //
    StartInfo.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
    StartInfo.hStdOutput = m_hChildStdoutWr;
    StartInfo.hStdError = m_hChildStdoutWr;
    StartInfo.wShowWindow = SW_HIDE;
    InheritHandles = TRUE;
    CreateFlags = CREATE_NEW_CONSOLE;

    // Copy the command line, because CreateProcess may modify it directly!
    char* zCmdLine = (char*)_alloca(lstrlen(pImageName) + 1);
    lstrcpy(zCmdLine, pImageName);

    //
    //    Create a process to execute the specified image.
    //
    if (!CreateProcess (NULL
                        ,zCmdLine                        // Command Line
                        ,NULL                            // Process Security attributes
                        ,NULL                            // Primary thread security attributes
                        ,InheritHandles                  // Handles are {NOT} inherited
                        ,CreateFlags                     // Creation flags
                        ,NULL                            // Use Parents environment
                        ,m_strCurrentDir                 // Use Parents current directory
                        ,&StartInfo                      // STARTUPINFO pointer
                        ,&ProcInfo))                     // Receives Process Information
    {
        LPTSTR lpMsgBuf;

        FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                       NULL,
                       GetLastError(),
                       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                       (LPTSTR) &lpMsgBuf,
                       0,
                       NULL);

        //
        //    Output the error message via a message box if we have requested the child process output
        //    be directed to a window, If we have requested that child process output be sent to a string
        //    then the error should be logged into that string
        //
        m_strOutput = lpMsgBuf;


        //
        //    Free the buffer.
        //
        LocalFree( lpMsgBuf );
        return -1;

    }

    // Close unused handles:
    CloseHandle(ProcInfo.hThread);
    CloseHandle(ProcInfo.hProcess);

    return 0;
}



//
//    DidCommandSucceed
//    =================
//
//    This is an over-ridable routine which can be called to determine if the output from the
//    last command executed indicated that the command was successful. The success criteria string
//    is passed through to us and is a standard text string, delimited by newline characters of words
//    or phrases which, if they appear in the last output, indicate that the command was sucessful.
//
//    Return Status:
//      true if any of the keywords or phrases appear in the last output string recovered
//      false otherwise
//
//    Parameters:
//      SuccessText - CString passsed by reference. This should be a list of the
//                  phrases or strings to search for separated by newline ("\n") characters.
//
bool CyrExecuteLocalCommand::DidCommandSucceed(CString & SuccessText) const
{
    //std::cout<<"Text: "<<(const char*)SuccessText<<std::endl;
    int SubStrIdx = SuccessText.Find('\n');

    while (SubStrIdx != -1)
    {
        CString SubString;

        //
        //    We have detected a delimiter of a substring so extract the substring itself.
        //
        SubString = SuccessText.Left(SubStrIdx);
        //std::cout<<"SubString: "<<(const char*)SubString<<std::endl;
        //
        //    Scan the output buffer to determine if this sub-string is present in it.
        //
        if (m_strOutput.Find(SubString) != -1)
        {
            return true;
        }

        //
        //    Substring not found so move on to the next chunk.
        //
        SuccessText = SuccessText.Mid(SubStrIdx + 1,SuccessText.GetLength());
        //std::cout<<"Text: "<<(const char*)SuccessText<<std::endl;
        SubStrIdx = SuccessText.Find('\n');
    }

    //
    //    At this point we are left with (hopefully) the final chunk of the string to check.
    //

    if ((SuccessText!="") && (m_strOutput.Find (SuccessText) == TRUE))
    {
        //std::cout<<"Return True!"<<std::endl;
        //std::cout<<"Text: "<<(const char*)SuccessText<<std::endl;
        //std::cout<<"Output: "<<(const char*)m_strOutput<<std::endl;
        return true;
    }

    return false;
}
