/* Copyright (C) 2000-2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     NT_Performance
//
// @module  NT Performance Class | This module contains the implementation of
//          the NT Performance class.  This class allows the retrieval of NT
//          Performance Statistics. Obviously this class will only work on a
//          Windows platform.
//
///////////////////////////////////////////////////////////////////////////////

// OpenSTA includes.
#include <afx.h>
#include "NTPerfMain.h"
#include "Ntp.h"
#include "OstaTraceDef.h"
#include <OSTAORB/OSTAORB.h>
#include <pdhmsg.h>
#include <winbase.h>
#include <CECSimpleParse.h>
#include <malloc.h>
///////////////////////////////////////////////////////////////////////////////
// Constants.
//const char* OtaNTPerform::CLASS_NAME = "OtaNTPerform";


///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor initialises the member variables and then opens a
//          performance query.  This query will be used throughout the lifetime
//          of the object.  
//
OtaNTPerform::OtaNTPerform()
:m_hQuery(NULL)
{
    // Error information
    const char* where = "OtaNTPerform::OtaNTPerform";

    // Open query if we don't already have one
    if (m_hQuery == NULL)
    {
        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Opening NT Performance query");

        // Open performance query - assume success
        PDH_STATUS pdhStatus = ERROR_SUCCESS;

        try
        {
            // Make the query
            pdhStatus = PdhOpenQuery( NULL, 0, &m_hQuery );
            if ( pdhStatus != ERROR_SUCCESS )
			{
		        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE,"PdhOpenQuery failed");
			}
			if(m_hQuery == NULL)
			{
		        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE,"m_hQuery == NULL");
			}
        }
        catch (...)
        {
            // And throw an exception if not possible - as we may be finishing
		    LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE,"Exception: PdhOpenQuery failed");
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor removes the counters from the query.
OtaNTPerform::~OtaNTPerform()
{
	if (m_hQuery)
		PdhCloseQuery(m_hQuery);
}

bool OtaNTPerform::InitializeQuery(const string &query) 
{
    PDH_STATUS pdhStatus;
    HCOUNTER hCounter = NULL;
    try
    {
		unsigned short * w = (unsigned short *)_alloca(query.size() * 2 + 10);
		memset(w, 0, query.size() * 2 + 10);
		MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED,query.c_str(), query.size(), w, query.size() * 2 + 9) ;
        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Adding " + query + "...");
        pdhStatus = PdhAddCounterW( m_hQuery, w, NULL, &hCounter );
        if ( pdhStatus != ERROR_SUCCESS )
        {
			m_hCounters.push_back(NULL);
			m_lastVal.push_back(-1);
			m_queries.push_back(query);
            LogPDHError("InitializeQuery : PdhAddCounter", query, pdhStatus);
            return false;
        }
	    m_hCounters.push_back(hCounter);
		m_lastVal.push_back(-1);
		m_queries.push_back(query);
    }
    catch (...)
    {
        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE,"InitializeQuery : Exception in PdhAddCounter");
        return false;
    }
    LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,query + " added.");
	return true ;
}

bool OtaNTPerform::ExecQueries ()
{
    PDH_STATUS pdhStatus;

    // Collect data
    try
    {
        Sleep(1);   // milliseconds
        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Executing...");

		pdhStatus = PdhCollectQueryData( m_hQuery );
		if ( pdhStatus != ERROR_SUCCESS )
		{
			LogPDHError("PdhCollectQueryData", "", pdhStatus);
			return false;
		}
    }
    catch (...)
    {
        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE,"Exception in PdhCollectQueryData");
        return false;
    }
	return true ;
}

bool OtaNTPerform::GetValue(int iIndex, long & lValue)
{
    // Get formatted data
    PDH_STATUS pdhStatus;
    PDH_FMT_COUNTERVALUE value;
    try
    {
        pdhStatus = PdhGetFormattedCounterValue( m_hCounters[iIndex],
                    (PDH_FMT_LONG|PDH_FMT_NOCAP100), NULL, &value );
        if ( pdhStatus != ERROR_SUCCESS )
        {
            LogPDHError("PdhGetFormattedCounterValue", m_queries[iIndex], pdhStatus);
            return false;
        }
    }
    catch (...)
    {
        LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE,"Exception in PdhGetFormattedCounterValue");
        return false;
    }
	lValue = value.longValue;
	m_lastVal[iIndex] = lValue ;
    return true;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method logs a PDH error
// @parm    The PDH function name
// @parm    The query
// @parm    The error status
void OtaNTPerform::LogPDHError (const string func, const string query, PDH_STATUS error)
{
    string errtxt = "ERROR: error from " + func + " for " + query + ": ";
    char    pdherr[256];
    if (error == PDH_CSTATUS_BAD_COUNTERNAME)
    {
        strcpy (pdherr, "The counter name path string could not be parsed or interpreted.");
    }
    else if (error == PDH_CSTATUS_NO_COUNTER)
    {
        strcpy (pdherr, "The specified counter was not found.");
    }
    else if (error == PDH_CSTATUS_NO_COUNTERNAME)
    {
        strcpy (pdherr, "An empty counter name path string was passed in.");
    }
    else if (error == PDH_CSTATUS_NO_MACHINE)
    {
        strcpy (pdherr, "A machine entry could not be created.");
    }
    else if (error == PDH_CSTATUS_NO_OBJECT)
    {
        strcpy (pdherr, "The specified object could not be found.");
    }
    else if (error == PDH_FUNCTION_NOT_FOUND)
    {
        strcpy (pdherr, "The calculation function for this counter could not be determined.");
    }
    else if (error == PDH_INVALID_ARGUMENT)
    {
        strcpy (pdherr, "One or more arguments are invalid.");
    }
    else if (error == PDH_INVALID_HANDLE)
    {
        strcpy (pdherr, "The query handle is not valid.");
    }
    else if (error == PDH_INVALID_DATA)
    {
        strcpy (pdherr, "The specified counter does not contain valid data or a successful status code.");
    }
    else if (error == PDH_MEMORY_ALLOCATION_FAILURE)
    {
        strcpy (pdherr, "A memory buffer could not be allocated.");
    }
    else if (error == PDH_NO_DATA)
    {
        strcpy (pdherr, "The query does not currently have any counters.");
    }
    else
    {
        sprintf (pdherr, "Unknown error %d", error);
    }
    errtxt += pdherr;
    LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_NONE,errtxt);
}

long OtaNTPerform::GetLastVal(int iIndex)
{
	return m_lastVal[iIndex];
}
